//	TorusGamesApples.c
//
//	Code specific to the Apples game.
//	This code knows about OpenGL, but does not
//	know about Windows or any other operating system.
//
//	Apples coordinate conventions
//
//	Integer coordinates (h,v) label the cells, for h,v ∈ {0, 1, … , n-1}.
//	The first coordinate runs left-to-right, while the second runs bottom-to-top.
//	Cell centers align with the natural n×n grid, so for example
//	cells straddle the edges of the fundamental domain
//	and a single cell sits centered at the fundamental domain's corners.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <math.h>


//	The body of the apple isn't quite centered vertically in the cell,
//	so the smiley face marker shouldn't be either.
#define APPLE_MARK_OFFSET	0.06


//	A queue of CellCoordinates will keep track of which
//	apples have been recently uncovered.
typedef struct
{
	unsigned int	itsH,
					itsV;
} CellCoordinates;


//	Public functions
static void	ApplesShutDown(ModelData *md);
static void	ApplesSetUpTextures(ModelData *md, ErrorText *aFirstError);
static void	ApplesShutDownTextures(ModelData *md);
static void	ApplesReset(ModelData *md);
static void	ApplesHandMoved(ModelData *md);
static bool	ApplesDragBegin(ModelData *md, bool aShiftOrControlKeyIsDown);
static void	ApplesDraw(ModelData *md);

//	Private functions
static Byte	CountNeighbors(ModelData *md, unsigned int h, unsigned int v);
static void	FindCurrentApple(ModelData *md, unsigned int *h, unsigned int *v, bool *aFlip);
static void	DeclareLoss(ModelData *md);
static bool	PlayerHasWon(ModelData *md);
static void	DeclareWin(ModelData *md);
static void	ExposeHitCells(ModelData *md, unsigned int h, unsigned int v);
static void	DrawSquares(ModelData *md);


void ApplesSetUp(ModelData *md)
{
	unsigned int	i;

	//	Initialize function pointers.
	md->itsGameShutDown			= &ApplesShutDown;
	md->itsGameSetUpTextures	= &ApplesSetUpTextures;
	md->itsGameShutDownTextures	= &ApplesShutDownTextures;
	md->itsGameReset			= &ApplesReset;
	md->itsGameHandMoved		= &ApplesHandMoved;
	md->itsGameDragBegin		= &ApplesDragBegin;
	md->itsGameDragObject		= NULL;
	md->itsGameDragEnd			= NULL;
	md->itsGameCharacterInput	= NULL;
	md->itsGameSpecialCommand	= NULL;
	md->itsGameSimulationUpdate	= NULL;
	md->itsGameDraw				= &ApplesDraw;
	
	//	Initialize game-specific texture names.
	md->itsGameOf.Apples.itsTextureNameBackground	= 0;
	md->itsGameOf.Apples.itsTextureNameApple		= 0;
	md->itsGameOf.Apples.itsTextureNameWorm			= 0;
	md->itsGameOf.Apples.itsTextureNameFaceYucky	= 0;
	md->itsGameOf.Apples.itsTextureNameFaceHappy	= 0;
	for (i = 0; i < 9; i++)
		md->itsGameOf.Apples.itsTextureNameNumeral[i] = 0;

	//	Set up the board.
	ApplesReset(md);
}


static void ApplesShutDown(ModelData *md)
{
	//	The caller must ensure that the OpenGL context is set
	//	so textures may be released.
	ApplesShutDownTextures(md);
}


static void ApplesSetUpTextures(
	ModelData	*md,
	ErrorText	*aFirstError)
{
	unsigned int	i;

	SetUpOneTexture(md, &md->itsGameOf.Apples.itsTextureNameBackground,	"Apples/Background.rgba",	TextureWrap,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Apples.itsTextureNameApple,		"Apples/Apple.rgba",		TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Apples.itsTextureNameWorm,		"Apples/Worm.rgba",			TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Apples.itsTextureNameFaceYucky,	"Apples/FaceYucky.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Apples.itsTextureNameFaceHappy,	"Apples/FaceHappy.rgba",	TextureClamp,	TextureRGBA,	aFirstError);

	for (i = 0; i < 9; i++)
	{
		//	Initialize number textures.
		//	The texture for '0' gets created, but never gets used.
		SetUpOneTexture(md, &md->itsGameOf.Apples.itsTextureNameNumeral[i],	NULL,					TextureClamp,	TextureAlpha,	aFirstError);
		SetAlphaTextureFromCharacter(
			md->itsGameOf.Apples.itsTextureNameNumeral[i],
			'0' + i,
			md->itsFonts[FontApplesDigit],
			APPLES_CHARACTER_TEXTURE_SIZE,
			md->itsShadersAvailableFlag && md->itsShadersEnabledFlag,
			aFirstError);
	}
}


static void ApplesShutDownTextures(ModelData *md)
{
	unsigned int	i;

	ShutDownOneTexture(&md->itsGameOf.Apples.itsTextureNameBackground	);
	ShutDownOneTexture(&md->itsGameOf.Apples.itsTextureNameApple		);
	ShutDownOneTexture(&md->itsGameOf.Apples.itsTextureNameWorm			);
	ShutDownOneTexture(&md->itsGameOf.Apples.itsTextureNameFaceYucky	);
	ShutDownOneTexture(&md->itsGameOf.Apples.itsTextureNameFaceHappy	);

	for (i = 0; i < 9; i++)
		ShutDownOneTexture(&md->itsGameOf.Apples.itsTextureNameNumeral[i]);
}


static void ApplesReset(ModelData *md)
{
	unsigned int	h,
					v;

	//	Set the board size.
	switch (md->itsDifficultyLevel)
	{
		//	Board sizes had been {2, 4, 8, 16, 32}.
		//	Changed to odd values to better show Klein bottle axes.
		case 0:  md->itsGameOf.Apples.itsBoardSize =  3;  break;
		case 1:  md->itsGameOf.Apples.itsBoardSize =  5;  break;
		case 2:  md->itsGameOf.Apples.itsBoardSize =  7;  break;
		case 3:  md->itsGameOf.Apples.itsBoardSize = 15;  break;
		case 4:  md->itsGameOf.Apples.itsBoardSize = 31;  break;
		default: md->itsGameOf.Apples.itsBoardSize =  2;  break;	//	should never occur
	}

	//	Clear the board.
	for (h = 0; h < MAX_APPLE_BOARD_SIZE; h++)
		for (v = 0; v < MAX_APPLE_BOARD_SIZE; v++)
			md->itsGameOf.Apples.itsBoard[h][v] = 0x00;

	//	Just to be safe...
	if (md->itsGameOf.Apples.itsBoardSize > MAX_APPLE_BOARD_SIZE)
		return;

	//	Assign a worm to each apple with probability 1-in-5.
	for (h = 0; h < md->itsGameOf.Apples.itsBoardSize; h++)
		for (v = 0; v < md->itsGameOf.Apples.itsBoardSize; v++)
			if (rand() % 5 == 0)
				md->itsGameOf.Apples.itsBoard[h][v] |= APPLE_HAS_WORM_MASK;

	//	Count each square's neighbors.
	//	(In a 2×2 or smaller board, some neighbors will get counted more than once.)
	for (h = 0; h < md->itsGameOf.Apples.itsBoardSize; h++)
		for (v = 0; v < md->itsGameOf.Apples.itsBoardSize; v++)
			md->itsGameOf.Apples.itsBoard[h][v] |= CountNeighbors(md, h, v);

	//	In a Klein bottle, assign flips randomly.
	if (md->itsTopology == TopologyKlein)
		for (h = 0; h < md->itsGameOf.Apples.itsBoardSize; h++)
			for (v = 0; v < md->itsGameOf.Apples.itsBoardSize; v++)
				if (rand() % 2)
					md->itsGameOf.Apples.itsBoard[h][v] |=  APPLE_FLIPPED_MASK;

	//	Abort any pending simulation.
	SimulationEnd(md);

	//	Ready to go.
	md->itsGameIsOver = false;
}


static Byte CountNeighbors(
	ModelData		*md,
	unsigned int	h,
	unsigned int	v)
{
	Byte		theCount;
	signed int	dh,
				dv,
				hh,
				vv,
				hhh,
				vvv;

	theCount = 0;

	for (dv = -1; dv <= +1; dv++)
	{
		hh = h;
		vv = v + dv;
		if (vv < 0)
		{
			vv = md->itsGameOf.Apples.itsBoardSize - 1;
			if (md->itsTopology == TopologyKlein)
				hh = (hh > 0) ? md->itsGameOf.Apples.itsBoardSize - hh : 0;
		}
		if (vv > (signed int)(md->itsGameOf.Apples.itsBoardSize - 1))
		{
			vv = 0;
			if (md->itsTopology == TopologyKlein)
				hh = (hh > 0) ? md->itsGameOf.Apples.itsBoardSize - hh : 0;
		}

		for (dh = -1; dh <= +1; dh++)
		{
			if (dh == 0 && dv == 0)
				continue;

			hhh = hh + dh;
			vvv = vv;
			if (hhh < 0)
				hhh = md->itsGameOf.Apples.itsBoardSize - 1;
			if (hhh > (signed int)(md->itsGameOf.Apples.itsBoardSize - 1))
				hhh = 0;

			if (md->itsGameOf.Apples.itsBoard[hhh][vvv] & APPLE_HAS_WORM_MASK)
				theCount++;
		}
	}

	return theCount;
}


static void ApplesHandMoved(ModelData *md)
{
	unsigned int	h,
					v;
	bool			theFlip;

	//	Which apple is the hand cursor over?
	FindCurrentApple(md, &h, &v, &theFlip);

	//	The orientation of the symbol relative to the fundamental domain
	//	should match the current orientation of the hand relative 
	//	to the fundamental domain.
	if (theFlip)
		md->itsGameOf.Apples.itsBoard[h][v] |=  APPLE_FLIPPED_MASK;
	else
		md->itsGameOf.Apples.itsBoard[h][v] &= ~APPLE_FLIPPED_MASK;
}


static bool ApplesDragBegin(
	ModelData	*md,
	bool		aShiftOrControlKeyIsDown)
{
	unsigned int	h,
					v;

	//	Keep the cell flip consistent with the cursor flip.
	ApplesHandMoved(md);

	//	If the game is already over, ignore all hits.
	if (md->itsGameIsOver)
		return false;

	//	Which cell got hit?
	FindCurrentApple(md, &h, &v, NULL);

	//	Ignore already exposed cells.
	if (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_EXPOSED_MASK)
		return false;

	//	Is the shift or control key down?
	//
	//	Adam suggested an illogical yet user-friendly special case:
	//	if an apple is marked, clicking it should merely unmark it, not expose it.
	if (aShiftOrControlKeyIsDown									//	typical case
	 || (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_MARKED_MASK))	//	special case
	{
		//	Toggle the mark.
		md->itsGameOf.Apples.itsBoard[h][v] ^= APPLE_MARKED_MASK;
		return false;
	}

	//	Expose the cell.
	//	Did the user hit a worm?
	if (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_HAS_WORM_MASK)
		DeclareLoss(md);
	else
	{
		ExposeHitCells(md, h, v);

		if (PlayerHasWon(md))
			DeclareWin(md);
	}

	return false;
}


static void FindCurrentApple(
	ModelData		*md,	//	input
	unsigned int	*h,		//	output
	unsigned int	*v,		//	output
	bool			*aFlip)	//	output, may be NULL
{
	double	theHandH,
			theHandV;

	//	Convert hand coordinates from [-0.5, +0.5] to [0.0, itsGameOf.Apples.itsBoardSize].
	theHandH = md->itsGameOf.Apples.itsBoardSize*(md->itsHandPlacement.itsH + 0.5);
	theHandV = md->itsGameOf.Apples.itsBoardSize*(md->itsHandPlacement.itsV + 0.5);

	//	Find the integer coordinates (h,v) of the hit cell.
	//	Results will lie in the range [0, n],
	//	but n will wrap to 0 immediately below.
	*h = (unsigned int) floor(theHandH + 0.5);
	*v = (unsigned int) floor(theHandV + 0.5);
	
	//	If the chirality of the hit cell is to agree with
	//	the chirality of the hand, should the hit cell be flipped?
	if (aFlip != NULL)
		*aFlip = md->itsHandPlacement.itsFlip;

	//	Wrap values on the boundary and
	//	protect against unexpected errors.
	if (*v >= md->itsGameOf.Apples.itsBoardSize)
	{
		*v = 0;
		if (md->itsTopology == TopologyKlein)
		{
			*h = md->itsGameOf.Apples.itsBoardSize - *h;
			if (aFlip != NULL)
				*aFlip = ! *aFlip;
		}
	}
	if (*h >= md->itsGameOf.Apples.itsBoardSize)
		*h = 0;
}


static void DeclareLoss(ModelData *md)
{
	unsigned int	h,
					v;

	md->itsGameIsOver = true;

	for (h = 0; h < md->itsGameOf.Apples.itsBoardSize; h++)
		for (v = 0; v < md->itsGameOf.Apples.itsBoardSize; v++)
			md->itsGameOf.Apples.itsBoard[h][v] |= APPLE_EXPOSED_MASK;

	PlayTheSound(SoundApplesLoss);
}


static bool PlayerHasWon(ModelData *md)
{
	unsigned int	h,
					v;

	for (h = 0; h < md->itsGameOf.Apples.itsBoardSize; h++)
		for (v = 0; v < md->itsGameOf.Apples.itsBoardSize; v++)
		{
			if (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_HAS_WORM_MASK)
			{
				if ((md->itsGameOf.Apples.itsBoard[h][v] & APPLE_EXPOSED_MASK) != 0)
					return false;
			}
			else
			{
				if ((md->itsGameOf.Apples.itsBoard[h][v] & APPLE_EXPOSED_MASK) == 0)
					return false;
			}
		}

	return true;
}


static void DeclareWin(ModelData *md)
{
	unsigned int	h,
					v;

	md->itsGameIsOver = true;

	for (h = 0; h < md->itsGameOf.Apples.itsBoardSize; h++)
		for (v = 0; v < md->itsGameOf.Apples.itsBoardSize; v++)
			if (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_HAS_WORM_MASK)
				md->itsGameOf.Apples.itsBoard[h][v] |= APPLE_MARKED_MASK;

	PlayTheSound(SoundApplesWin);
}


static void ExposeHitCells(
	ModelData		*md,
	unsigned int	h,
	unsigned int	v)
{
	//	Keep a queue of all cells that get uncovered.
	//	If a newly uncovered cell has zero wormy neighbors,
	//	uncover any still-covered neighbors and add them
	//	to the queue as well.

	CellCoordinates	*theQueue		= NULL,
					*theQueueFirst	= NULL,
					*theQueueLast	= NULL;
	signed int		dh,
					dv,
					hh,
					vv,
					hhh,
					vvv;

	//	Uncover the original hit cell.
	md->itsGameOf.Apples.itsBoard[h][v] |= APPLE_EXPOSED_MASK;

	//	Allocate space for the queue.
	//	If we can't get it, fail silently.
	theQueue = GET_MEMORY(md->itsGameOf.Apples.itsBoardSize * md->itsGameOf.Apples.itsBoardSize * sizeof(CellCoordinates));
	if (theQueue == NULL)
		return;

	//	Put the original hit cell on the queue.
	theQueueFirst		= theQueue;
	theQueueLast		= theQueue;
	theQueueFirst->itsH	= h;
	theQueueFirst->itsV	= v;

	//	Process each newly uncovered cell.  If the cell has no wormy neighbors,
	//	uncover all previously covered neighbors and add them to the queue.
	while (theQueueFirst <= theQueueLast)
	{
		//	Pull the first apple off the queue and check its neighbors,
		//	uncovering them and adding them to the queue as necessary.
		if ((md->itsGameOf.Apples.itsBoard[theQueueFirst->itsH][theQueueFirst->itsV] & APPLE_NBR_COUNT_MASK) == 0)
		{
			for (dv = -1; dv <= +1; dv++)
			{
				hh = theQueueFirst->itsH;
				vv = theQueueFirst->itsV + dv;
				if (vv < 0)
				{
					vv = md->itsGameOf.Apples.itsBoardSize - 1;
					if (md->itsTopology == TopologyKlein)
						hh = (hh > 0) ? md->itsGameOf.Apples.itsBoardSize - hh : 0;
				}
				if (vv > (signed int)(md->itsGameOf.Apples.itsBoardSize - 1))
				{
					vv = 0;
					if (md->itsTopology == TopologyKlein)
						hh = (hh > 0) ? md->itsGameOf.Apples.itsBoardSize - hh : 0;
				}

				for (dh = -1; dh <= +1; dh++)
				{
					if (dh == 0 && dv == 0)
						continue;

					hhh = hh + dh;
					vvv = vv;
					if (hhh < 0)
						hhh = md->itsGameOf.Apples.itsBoardSize - 1;
					if (hhh > (signed int)(md->itsGameOf.Apples.itsBoardSize - 1))
						hhh = 0;

					if ((md->itsGameOf.Apples.itsBoard[hhh][vvv] & APPLE_EXPOSED_MASK) == 0)
					{
						md->itsGameOf.Apples.itsBoard[hhh][vvv] |= APPLE_EXPOSED_MASK;
						theQueueLast++;
						theQueueLast->itsH = hhh;
						theQueueLast->itsV = vvv;
					}
				}
			}
		}

		//	Move on.
		theQueueFirst++;
	}

	//	Free the queue.
	FREE_MEMORY_SAFELY(theQueue);
}


static void ApplesDraw(ModelData *md)
{
	//	Draw the background.
	DrawFundamentalDomain(md, md->itsGameOf.Apples.itsTextureNameBackground, 11);

	//	Draw glide reflection axes if appropriate.
	DrawKleinAxes(	md,
					PREMULTIPLY_RGBA(0.0, 0.0, 0.0, 1.0),
					PREMULTIPLY_RGBA(1.0, 1.0, 1.0, 1.0) );

	//	Draw the squares.
	DrawSquares(md);
}


static void DrawSquares(ModelData *md)
{
	double			theScaleFactor;
	Placement		thePlacement;
	unsigned int	h,
					v,
					theNumNbrWorms;

	static float	theNumeralColor[4] = PREMULTIPLY_RGBA(0.0, 0.0, 1.0, 1.0);

	//	For convenience we'll think in terms of a board with coordinates
	//	running from 0.0 to itsBoardSize, but then we'll rescale to [-0.5, +0.5].
	theScaleFactor = 1.0 / md->itsGameOf.Apples.itsBoardSize;

	//	Draw the apples.
	thePlacement.itsAngle	= 0.0;
	thePlacement.itsSizeH	= theScaleFactor;
	thePlacement.itsSizeV	= theScaleFactor;
	for (h = 0; h < md->itsGameOf.Apples.itsBoardSize; h++)
		for (v = 0; v < md->itsGameOf.Apples.itsBoardSize; v++)
			if ( ! (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_EXPOSED_MASK) )
			{
				thePlacement.itsH		=  -0.5  +  h * theScaleFactor;
				thePlacement.itsV		=  -0.5  +  v * theScaleFactor;
				thePlacement.itsFlip	= (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_FLIPPED_MASK);
				
				DrawSprite(	md,
							md->itsGameOf.Apples.itsTextureNameApple,
							&thePlacement);
			}

	//	Draw the marks.
	thePlacement.itsAngle	= 0.0;
	thePlacement.itsSizeH	= 0.5 * theScaleFactor;
	thePlacement.itsSizeV	= 0.5 * theScaleFactor;
	for (h = 0; h < md->itsGameOf.Apples.itsBoardSize; h++)
		for (v = 0; v < md->itsGameOf.Apples.itsBoardSize; v++)
			if ( ! (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_EXPOSED_MASK)
			 &&    (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_MARKED_MASK ) )
			{
				thePlacement.itsH		=  -0.5  +  h * theScaleFactor;
				thePlacement.itsV		=  -0.5  +  (v - APPLE_MARK_OFFSET) * theScaleFactor;
				thePlacement.itsFlip	= (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_FLIPPED_MASK);
				
				DrawSprite(	md,
							md->itsGameIsOver ?
								md->itsGameOf.Apples.itsTextureNameFaceHappy :
								md->itsGameOf.Apples.itsTextureNameFaceYucky,
							&thePlacement);
			}

	//	Draw the worms.
	thePlacement.itsAngle	= 0.0;
	thePlacement.itsSizeH	= theScaleFactor;
	thePlacement.itsSizeV	= theScaleFactor;
	for (h = 0; h < md->itsGameOf.Apples.itsBoardSize; h++)
		for (v = 0; v < md->itsGameOf.Apples.itsBoardSize; v++)
			if ((md->itsGameOf.Apples.itsBoard[h][v] & APPLE_EXPOSED_MASK )
			 && (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_HAS_WORM_MASK))
			{
				thePlacement.itsH		=  -0.5  +  h * theScaleFactor;
				thePlacement.itsV		=  -0.5  +  v * theScaleFactor;
				thePlacement.itsFlip	= (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_FLIPPED_MASK);
				
				DrawSprite(	md,
							md->itsGameOf.Apples.itsTextureNameWorm,
							&thePlacement);
			}

	//	Draw the numbers.
	//	To minimize texture changes, first draw all the 1's, then all the 2's, etc.
	thePlacement.itsAngle	= 0.0;
	thePlacement.itsSizeH	= theScaleFactor;
	thePlacement.itsSizeV	= theScaleFactor;
	for (theNumNbrWorms = 1; theNumNbrWorms <= 8; theNumNbrWorms++)	//	skip 0
		for (h = 0; h < md->itsGameOf.Apples.itsBoardSize; h++)
			for (v = 0; v < md->itsGameOf.Apples.itsBoardSize; v++)
				if (   (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_EXPOSED_MASK )
				 &&  ! (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_HAS_WORM_MASK)
				 &&    (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_NBR_COUNT_MASK) == theNumNbrWorms)
				{
					thePlacement.itsH		=  -0.5  +  h * theScaleFactor;
					thePlacement.itsV		=  -0.5  +  v * theScaleFactor;
					thePlacement.itsFlip	= (md->itsGameOf.Apples.itsBoard[h][v] & APPLE_FLIPPED_MASK);
					
					DrawColorWithAlphaMask(	md,
											md->itsGameOf.Apples.itsTextureNameNumeral[theNumNbrWorms],
											&thePlacement,
											theNumeralColor);
				}
}
