//	TorusGamesGraphics.c
//
//	Drawing code for Torus Games.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <math.h>
#include <string.h>	//	for memcpy()


//	Stacking Note:  When we render the scene in the universal cover
//	we draw all images of one object, then all images of the next object,
//	and so on.  This is efficient from the GPU's perspective
//	(we get the OpenGL state set the way we like it and then draw
//	all images of the object, thus minimizing state changes) 
//	and it also ensures that overlapping objects are drawn consistently.


//	The near clipping distance is 1/INVERSE_NEAR_CLIP.
#define INVERSE_NEAR_CLIP	128.0


static void	MatrixIdentity(float m[4][4]);
static void	MatrixCopy(float dst[4][4], float src[4][4]);
static void	DrawHandCursor(ModelData *md);
static void	DrawRepeatingImages(ModelData *md, Placement *aPlacement,
				unsigned int aShaderIndex, unsigned int aTextureName, unsigned int aVertexBufferIndex,
				float aColor[4], TexCoordTransform *aTexCoordTransform);


char *GetRendererName(void)
{
	return (char *)glGetString(GL_RENDERER);
}


void SetTextureImageFancy(
	unsigned int	aTextureName,
	unsigned int	aWidth,
	unsigned int	aHeight,
	unsigned int	aDepth,			//	Pass 4 for RGBA or 1 for alpha-only.
	Byte			*aPixelArray,	//	4 bytes per pixel or 1 byte per pixel
	bool			aShadersFlag)
{
	unsigned int	thePrevTextureName;
	unsigned int	thePixelFormat;
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	unsigned int	theNumTexels;
	Byte			*theLuminanceAlphaBuffer	= NULL,
					*theSrcByte,
					*theDstByte;
	unsigned int	theCount;
#endif

	if (aTextureName == 0
	 || ! IsPowerOfTwo(aWidth )
	 || ! IsPowerOfTwo(aHeight)
	 || (aDepth != 1 && aDepth != 4))
		return;

	//	Select an appropriate pixel format.
	if (aDepth == 4)	//	RGBA texture
	{
		thePixelFormat = GL_RGBA;
	}
	else				//	alpha-only texture
	{
		//	Different versions of OpenGL support
		//	alpha-only textures in different ways.
		//
		//	OpenGL 1.1 expects a texture with internal format GL_ALPHA,
		//	and the fixed-function pipeline handles it correctly.
		//
		//	OpenGL 3.0 doesn't support GL_ALPHA in a forward-compatible context.
		//	Instead we may create a 1-component texture with internal format GL_RED,
		//	and let the shader interpret the sample's red component as an opacity.
		//
		//	OpenGL 2.0 offers a choice of GL_ALPHA as in OpenGL 1.1
		//	or a 1-component texture as in OpenGL 3.0, but with the catch that
		//	we must use internal format GL_LUMINANCE instead of GL_RED.
		//	We opt for GL_LUMINANCE over GL_ALPHA, so the fragment shader
		//	code may remain the same for OpenGL 2.0 and OpenGL 3.0.
		//
#ifdef TORUS_GAMES_FOR_OPENGL_ES
		if (aShadersFlag)
		{
			thePixelFormat = GL_LUMINANCE;
		}
		else
		{
			//	With non-premultiplied (R,G,B,α), the texture mask
			//	needs to modulate the current color's alpha channel only,
			//	which is exactly what GL_ALPHA does.
			//
			//	With premultiplied (αR, αG, αB, α), the texture mask
			//	needs to modulate all four components, which is exactly
			//	what GL_INTENSITY does.
			//
			//	OpenGL ES 1.1 does not support GL_INTENSITY, so we'll 
			//	use GL_LUMINANCE_ALPHA instead.  The price we pay is that 
			//	this is a 2 bytes/texel format instead of GL_INTENSITY's 1 byte/texel.
			//
			//	I'm not pleased with this ugly workaround, but take some comfort
			//	in the fact that within a few years current iPhones and iPod Touches 
			//	will have been superceded by newer models that provide programmable shaders.
			//	(Adam tells me that iPhones and iPod Touches last only 2-3 years, 
			//	not the 5-10 years that a desktop computer might last).
			//	Once shader-based devices are the norm, this fixed-function code
			//	will disappear entirely.
			theNumTexels			= aWidth * aHeight;
			theLuminanceAlphaBuffer	= GET_MEMORY(2 * theNumTexels);
			if (theLuminanceAlphaBuffer != NULL)
			{
				theSrcByte = aPixelArray;
				theDstByte = theLuminanceAlphaBuffer;
				for (theCount = theNumTexels; theCount-- > 0; )
				{
					*theDstByte++ = *theSrcByte;
					*theDstByte++ = *theSrcByte++;
				}
				
				//	Hijack aPixelArray and let it point
				//	to theLuminanceAlphaBuffer instead.
				//	(Yes, I know this is ugly -- see comments above!)
				aPixelArray		= theLuminanceAlphaBuffer;
				thePixelFormat	= GL_LUMINANCE_ALPHA;
			}
			else
				return;
		}
#else	//	desktop OpenGL
		if (aShadersFlag)
		{
			if (RUNNING_OPENGL_3)
				thePixelFormat = GL_RED;
			else
				thePixelFormat = GL_LUMINANCE;
		}
		else
		{
			thePixelFormat = GL_INTENSITY;
		}
#endif
	}

	//	SetTextureImageFancy() typically gets called within SetUpGraphicsAsNeeded(),
	//	but JigsawDraw() calls it from with Render().  The Render() algorithm keeps
	//	track of the currently bound texture, so if we change it here, we must be sure
	//	to change it back again when we're done.
	glGetIntegerv(GL_TEXTURE_BINDING_2D, (GLint *)&thePrevTextureName);
	if (aTextureName != thePrevTextureName)
		glBindTexture(GL_TEXTURE_2D, aTextureName);
	
	//	glTexImage2D()'s default 4-byte row alignment is fine for all RGBA textures
	//	as well as for alpha-only textures with power-of-two width at least 4,
	//	but we must relax the row alignment for alpha-only textures of width 1 or 2.
	if (aDepth == 1 && aWidth < 4)
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	//	When completing the migration to OpenGL 3.0 and OpenGL ES 2.0,
	//	eliminate this call to glTexParameteri(…GL_GENERATE_MIPMAP…)
	//	in favor of the glGenerateMipmap() call below.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	if ( ! aShadersFlag )	//	(aShadersFlag == false) ⇔ (running OpenGL 1.1)
	{
#ifdef SUPPORT_OPENGL_ES_1
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
#endif // SUPPORT_OPENGL_ES_1
	}
#else // desktop OpenGL
	if ( ! RUNNING_OPENGL_3)
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
#endif // TORUS_GAMES_FOR_OPENGL_ES

	glTexImage2D(	GL_TEXTURE_2D,
					0,
					thePixelFormat,
					aWidth,
					aHeight,
					0,
//	Once the desktop version moves to shaders,
//	this #ifdef/#else/#endif constuction will go away,
//	because just plain "thePixelFormat" will be OK
//	on both the desktop (where it'll be GL_LUMINANCE or GL_RED)
//	and the iPhone (where it'll be GL_LUMINANCE_ALPHA or GL_LUMINANCE).
#ifdef TORUS_GAMES_FOR_OPENGL_ES
					thePixelFormat,
#else
					(thePixelFormat == GL_INTENSITY ? GL_LUMINANCE : thePixelFormat),
#endif
					GL_UNSIGNED_BYTE,
					aPixelArray);

	//	To average two pixels correctly when generating mipmaps,
	//	we need to weight them according to their alpha values.
	//	With traditional (R,G,B,α) the formula is a bit messy
	//
	//							 α0*(R0,G0,B0) + α1*(R1,G1,B1)
	//		(Ravg, Gavg, Bavg) = -----------------------------
	//							            α0 + α1
	//
	//							 α0 + α1
	//		              αavg = -------
	//							    2
	//
	//	With premultiplied (αR,αG,αB,α) the formula becomes a simple average
	//
	//		(αavg*Ravg, αavg*Gavg, αavg*Bavg, αavg)
	//
	//			  (α0*R0, α0*G0, α0*B0, α0) + (α1*R1, α1*G1, α1*B1, α1)
	//			= -----------------------------------------------------
	//			                            2
	//
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	if (aShadersFlag)	//	(aShadersFlag == true) ⇔ (running OpenGL 2.0 or newer)
	{
#ifdef SUPPORT_OPENGL_ES_2
		glGenerateMipmap(GL_TEXTURE_2D);
#endif // SUPPORT_OPENGL_ES_2
	}
#else // desktop OpenGL
//UNCOMMENT THESE LINES WHEN ADDING OPENGL 3.0 SUPPORT
//	if (RUNNING_OPENGL_3)
//		glGenerateMipmap(GL_TEXTURE_2D);
#endif // TORUS_GAMES_FOR_OPENGL_ES

	//	Restore the default row alignment, if necessary.
	if (aDepth == 1 && aWidth < 4)
		glPixelStorei(GL_UNPACK_ALIGNMENT, 4);

	//	Restore the previous texture binding, if necessary.
	if (aTextureName != thePrevTextureName)
		glBindTexture(GL_TEXTURE_2D, thePrevTextureName);
	
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	Free the luminance-alpha buffer.
	FREE_MEMORY_SAFELY(theLuminanceAlphaBuffer);
#endif
}


bool IsPowerOfTwo(unsigned int n)
{
	if (n == 0)
		return false;

	while ((n & 0x00000001) == 0)
		n >>= 1;

	return (n == 0x00000001);
}


void ResizeGLView(
	unsigned int	aNewViewWidth,	//	in pixels
	unsigned int	aNewViewHeight)	//	in pixels
{
	//	Let OpenGL know about the new view size.
	glViewport(0, 0, aNewViewWidth, aNewViewHeight);
}


ErrorText ReadRGBAPixels(
	unsigned int	aWidth,
	unsigned int	aHeight,
	Byte			*aBuffer)
{
#ifndef TORUS_GAMES_FOR_OPENGL_ES
	//	Mac OS seems to literally swap buffers, rather than copying
	//	the contents of a fixed back buffer to a fixed front buffer.
	//	As a result, reading the back buffer gives the contents
	//	of the frame previous to the one currently visible.
	//	Solution:  Read the front buffer instead.
	glReadBuffer(GL_FRONT);
#endif

	glReadPixels(	0, 0,
					(GLsizei) aWidth, (GLsizei) aHeight,
					GL_RGBA, GL_UNSIGNED_BYTE,
					(GLvoid *) aBuffer);
	
	return GetErrorString();
}


ErrorText Render(ModelData *md)
{
	GLint			theFullViewport[4],
					theSmallWidth,
					theSmallHeight;
	float			f,
					dh,
					dv,
					s,
					theViewMatrix[4][4];
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
	unsigned int	i;
#endif

	//	If no game is active, clear the color buffer to medium grey and return.
	if (md->itsGame == GameNone)
	{
		//	In the iPhone version, be sure to give TorusGamesEAGLView
		//	this same color, to avoid an unsightly flash at launch.
		glClearColor(0.5, 0.5, 0.5, 1.0);		//	pre-multiplied alpha
		glClear(GL_COLOR_BUFFER_BIT);
		return NULL;
	}

	//	In ViewSmall mode...
	if (md->itsView == ViewSmall)
	{
		//	Clear the color buffer.
		glClearColor(0.75, 0.75, 0.75, 1.00);	//	pre-multiplied alpha
		glClear(GL_COLOR_BUFFER_BIT);
		
		//	Restrict the viewport to the middle 1/9.
		//	The platform-dependent code restricts the content rectangle
		//	to be a square with multiple-of-three width and height.
		//	Even if it didn't, the following code would still work,
		//	but with a 1/3 or 2/3 pixel shift between the ViewSmall
		//	and ViewTiling images.
		glGetIntegerv(GL_VIEWPORT, theFullViewport);
		theSmallWidth	= theFullViewport[2] / 3;	//	integer division, remainder gets discarded
		theSmallHeight	= theFullViewport[3] / 3;	//	integer division, remainder gets discarded
		glViewport(theSmallWidth, theSmallHeight, theSmallWidth, theSmallHeight);
	}

	//	Disable depth testing.
	glDisable(GL_DEPTH_TEST);

	//	Enable blending, so partially transparent sprites
	//	overlay the background correctly.
	//
	//	Rigorously correct blending requires
	//
	//				   αs*(Rs,Gs,Bs) + (1 - αs)*αd*(Rd,Gd,Bd)
	//		(R,G,B) = ----------------------------------------
	//				         αs      +      (1 - αs)*αd
	//
	//		      α = αs + (1 - αs)*αd
	//
	//	Replacing the traditional (R,G,B,α) with the premultiplied (αR,αG,αB,α)
	//	simplifies the formula to 
	//
	//		(αR, αG, αB) = (αs*Rs, αs*Gs, αs*Bs) + (1 - αs)*(αd*Rd, αd*Gd, αd*Bd)
	//
	//		           α = αs + (1 - αs)*αd
	//
	//	Because they share the same coefficients, we may merge the RGB and α parts
	//	into a single formula
	//
	//		(αR, αG, αB, α) = (αs*Rs, αs*Gs, αs*Bs, αs) + (1 - αs)*(αd*Rd, αd*Gd, αd*Bd, αd)
	//
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

	//	Use shaders if possible, otherwise fall back to the fixed-function pipeline.
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
		//	Render all primitives using position and texture coordinate arrays.
		//	The default shader (see below) doesn't need a color.
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		glEnableVertexAttribArray(ATTRIBUTE_POSITION	);
		glEnableVertexAttribArray(ATTRIBUTE_TEX_COORD	);
#endif
	}
	else
	{
#if defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		//	Render all primitives using position and texture coordinate arrays.
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		//	Enable texturing.
		glEnable(GL_TEXTURE_2D);

		//	No lighting.
		glDisable(GL_LIGHTING);
#endif
	}

	//	Disable fixed-function arrays when running shaders,
	//	and disable generic attribute arrays when running in fixed-function mode.
	//	This guards against the possibility that the wrong kinds of arrays
	//	were left enabled from a previous frame, when itsShadersEnabledFlag
	//	may have been different.
	//
	//	Note #1.  A forward-compatible OpenGL 3.0 context
	//	will not support the fixed-function calls.
	//
	//	Note #2.  This whole paragraph of code may be deleted 
	//	when fixed-function support gets removed from the desktop version.
	//	Even if fixed-function support remains in the iPhone version,
	//	this paragraph isn't necessary, because the iPhone version
	//	never switches between GLES1 and GLES2.
	//
#ifndef TORUS_GAMES_FOR_OPENGL_ES
	if (md->itsShadersAvailableFlag && ! RUNNING_OPENGL_3)
	{
		if (md->itsShadersEnabledFlag)
		{
			//	We'll use only the forward-compatible glVertexAttribPointer(),
			//	not the backward-compatible glVertexPointer(), glTexCoordPointer(), etc.
			glDisableClientState(GL_VERTEX_ARRAY		);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY	);
		}
		else
		{
			//	We'll use only the backward-compatible glVertexPointer(), glTexCoordPointer(), etc.,
			//	not the forward-compatible glVertexAttribPointer().
			glDisableVertexAttribArray(ATTRIBUTE_POSITION	);
			glDisableVertexAttribArray(ATTRIBUTE_TEX_COORD	);
		}
	}
#endif
	
	//	According to the formal specification, one may call
	//
	//			glEnable(GL_MULTISAMPLE)
	//	or
	//			glDisable(GL_MULTISAMPLE)
	//
	//	but in practice the pixel format (with or without a multisample buffer)
	//	wholly determines whether multisampling takes place.
	//	This latter approach makes sense, given that 
	//	multisampling into a single-sample buffer would be impossible, and
	//	single-sampling into a multisample buffer would be pointless.
	//	Nevertheless, let's call glEnable/glDisable(GL_MULTISAMPLE),
	//	just in case some future OpenGL graphics driver wants the call.
	//
	//	Note #1:  OpenGL ES 1.1 and desktop OpenGL both use glEnable(GL_MULTISAMPLE),
	//	but OpenGL ES 2.0 does not.  OpenGL ES 2.0 enables multisampling
	//	iff a multisample buffer is present.
	//
	//	Note #2:  Torus Games doesn't need multisampling,
	//	and so md->itsMultisamplingEnabledFlag is always false.
	//
#if defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	Call glEnable(GL_MULTISAMPLE) or glDisable(GL_MULTISAMPLE)
	//	in OpenGL 1.1 but not in OpenGL 2.0 or later.
	if ( ! (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag) )	//	running GLES 1.1 ?
#else
	//	Always call glEnable(GL_MULTISAMPLE) or glDisable(GL_MULTISAMPLE) in desktop OpenGL.
	if (true)
#endif	//	TORUS_GAMES_FOR_OPENGL_ES
	{
		if (md->itsMultisamplingAvailableFlag && md->itsMultisamplingEnabledFlag)
			glEnable(GL_MULTISAMPLE);
		else
			glDisable(GL_MULTISAMPLE);
	}
#endif	//	defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)


	//	The view matrix is the product 
	//	of a possible reflection, a translation and a dilation:
	//
	//		(f 0 0 0) ( 1  0  0  0) (s 0 0 0)   ( s∙r   0    0    0 )
	//		(0 1 0 0) ( 0  1  0  0) (0 s 0 0) = (  0    s    0    0 )
	//		(0 0 1 0) ( 0  0  1  0) (0 0 1 0)   (  0    0    1    0 )
	//		(0 0 0 1) (∆h ∆v  0  1) (0 0 0 1)   ( s∙∆h s∙∆v  0    1 )
	//
	
	//	The reflection factor f realizes a possible flip.
	f = (md->itsOffset.itsFlip ? -1.0 : +1.0);
	
	//	The translation (∆h, ∆v) offsets the fundamental domain.
	dh = md->itsOffset.itsH;
	dv = md->itsOffset.itsV;
	
	//	The dilation factor s effectively converts OpenGL's default [-1, +1] coordinates
	//	to either [-1/2, +1/2] coordinates in ViewDomain or ViewSmall,
	//	or [-3/2, +3/2] coordinates in ViewTiling.
	s = (md->itsView != ViewTiling ? 2.0 : 2.0/3.0);
	
	//	Initialize theViewMatrix to the identity,
	//	then fill in the four non-identity values.
	MatrixIdentity(theViewMatrix);
	theViewMatrix[0][0] = s*f;
	theViewMatrix[1][1] = s;
	theViewMatrix[3][0] = s*dh;
	theViewMatrix[3][1] = s*dv;

	//	Set theViewMatrix in all shaders.
	//	It will remain constant for the whole frame.
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		for (i = 0; i < NumShaders; i++)
		{
			glUseProgram(md->itsShaderPrograms[i]);
			glUniformMatrix4fv(	glGetUniformLocation(md->itsShaderPrograms[i], "uniViewMatrix"),
								1,
								GL_FALSE,
								(float *)theViewMatrix);
			glUniform1i(glGetUniformLocation(md->itsShaderPrograms[i], "uniTexture"), 0);	//	redundant to set to 0
		}
		glUseProgram(0);
#endif
	}
	else
	{
		//	When I someday eliminate the fixed-function code,
		//	theViewMatrix may become 3×3 instead of 4×4.
#if defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		glMatrixMode(GL_PROJECTION);
		glLoadMatrixf((float *)theViewMatrix);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
#endif
	}


	//	Remember the current OpenGL state, to avoid redundant state changes.
	//	Rumor has it that even redundant state changes can be expensive.

	//		shader
	md->itsCurrentShaderIndex = ShaderTexture;
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		glUseProgram(md->itsShaderPrograms[ShaderTexture]);
#endif
	}
	else
	{
#if defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
#endif
	}
	
	//		texture
	md->itsCurrentTextureName = 0;
	glBindTexture(GL_TEXTURE_2D, 0);
	
	//		vertex buffers
	md->itsCurrentVertexBufferIndex = VertexBufferRectangle;
	BindRectangleVAO(	md->itsVertexArrayNames	[VertexBufferRectangle],
						md->itsVertexBufferNames[VertexBufferRectangle],
						md->itsIndexBufferNames [VertexBufferRectangle],
//						md->itsOpenGLVersion);	RUNNING_OPENGL_3
						0,	//	temporarily pass OpenGL version number 0.0
						md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);

	//	Draw the game.
	if (md->itsGameDraw != NULL)
		(*md->itsGameDraw)(md);

	//	Draw the hand if we have the cursor,
	//	or the progress meter if the computer
	//	is choosing a chess move.
	DrawHandCursor(md);

	//	In ViewSmall mode, restore the original viewport.
	if (md->itsView == ViewSmall)
		glViewport(theFullViewport[0], theFullViewport[1], theFullViewport[2], theFullViewport[3]);

	//	Return a string describing the first of any OpenGL errors
	//	that may have occurred.  If no errors occurred, return NULL.
	return GetErrorString();
}


static void MatrixIdentity(float m[4][4])
{
	unsigned int	i,
					j;

	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++)
			m[i][j] = (i == j ? 1.0 : 0.0);
}


static void MatrixCopy(
	float	dst[4][4],
	float	src[4][4])
{
	unsigned int	i,
					j;

	for (i = 0; i < 4; i++)
		for (j = 0; j < 4; j++)
			dst[i][j] = src[i][j];
}


static void DrawHandCursor(ModelData *md)
{
	Placement	theHandPlacement;
	
	static float	theClockBaseColor[4]	= PREMULTIPLY_RGBA(0.6250, 0.8750, 0.9375, 1.0000),
					theClockSweepColor[4]	= PREMULTIPLY_RGBA(1.0000, 0.0000, 0.0000, 1.0000),
					theClockRimColor[4]		= PREMULTIPLY_RGBA(0.0000, 0.0000, 0.0000, 1.0000),
					theClockArrowColor[4]	= PREMULTIPLY_RGBA(0.0000, 0.0000, 0.0000, 1.0000);

	//	Copy the current hand placement.
	theHandPlacement = md->itsHandPlacement;

	//	On the one hand, it would be nice to keep the cursor
	//	a single fixed size at all times.  On the other hand,
	//	it really is too big for the more complicated mazes,
	//	so let's make an exception there and shrink the cursor.
	switch (md->itsGame)
	{
		case GameMaze:
			switch (md->itsDifficultyLevel)
			{
				case 0:
				case 1:
				case 2:
					break;
		
				case 3:
					theHandPlacement.itsSizeH *= 0.50;
					theHandPlacement.itsSizeV *= 0.50;
					break;

				case 4:
					theHandPlacement.itsSizeH *= 0.25;
					theHandPlacement.itsSizeV *= 0.25;
					break;
			}
			break;

		case GameApples:
			switch (md->itsDifficultyLevel)
			{
				case 0:
				case 1:
				case 2:
					break;
				
				case 3:
					theHandPlacement.itsSizeH *= 0.6;
					theHandPlacement.itsSizeV *= 0.6;
					break;
					
				case 4:
					theHandPlacement.itsSizeH *= 0.3;
					theHandPlacement.itsSizeV *= 0.3;
					break;
			}
			break;

		default:
			break;
	}

	//	Draw the cursor according to itsHandStatus.
	switch (md->itsHandStatus)
	{
		case HandNone:
#ifdef TORUS_GAMES_FOR_OPENGL_ES
			//	On the iPhone, once the user touches the board,
			//	HandNone should never come back again.
			//	Nevertheless, in the interest of robust design,
			//	let HandNone fall through to HandFree on iPhone OS.
#else
			break;
#endif

		case HandScroll:
		case HandDrag:
#ifdef TORUS_GAMES_FOR_OPENGL_ES
			//	On the iPhone, HandScroll and HandDrag fall through to HandFree,
			//	so when the program is displaying a progress meter, it will
			//	get shown even while the user scrolls the board.
#else
			DrawSprite(md, md->itsTextureNamesCommon[TextureCommonHandGrab], &theHandPlacement);
			break;
#endif

		case HandFree:

			if (md->itsSimulationStatus == SimulationNone)
			{
#ifndef TORUS_GAMES_FOR_OPENGL_ES
				//	Draw an opaque flat-hand cursor.
				DrawSprite(md, md->itsTextureNamesCommon[TextureCommonHandFlat], &theHandPlacement);
#endif
			}
			else
			if ( ! md->itsProgressMeterIsActive )
			{
#ifndef TORUS_GAMES_FOR_OPENGL_ES
				//	Draw a semi-transparent flat-hand cursor to let
				//	the user know the cursor is inactive, while not 
				//	distracting the user's attention from the simulation,
				//	for example from the rolling pool balls.
				DrawSpriteTransparent(md, md->itsTextureNamesCommon[TextureCommonHandFlat], &theHandPlacement);
#endif
			}
			else
			{
#ifdef TORUS_GAMES_FOR_OPENGL_ES
				//	On the iPhone version, we want to draw the progress indicator
				//	at the center of the display.  To do this, subtract off
				//	the fundamental domain's translation and possible flip.
				theHandPlacement.itsH		= md->itsOffset.itsFlip ? +md->itsOffset.itsH : -md->itsOffset.itsH;
				theHandPlacement.itsV		= -md->itsOffset.itsV;
				theHandPlacement.itsFlip	= md->itsOffset.itsFlip;
				theHandPlacement.itsAngle	= 0.0;
				theHandPlacement.itsSizeH	= 0.25;
				theHandPlacement.itsSizeV	= 0.25;
#endif
				//	A long computation is in progress (namely the computer
				//	is deciding its chess move).  Display a watch face to let
				//	the user know what fraction of the required time has elapsed.
				DrawColorWithAlphaMask(md, md->itsTextureNamesCommon[TextureCommonClockBase], &theHandPlacement, theClockBaseColor);
				if (md->itsProgressMeter <= 0.5)
				{
					theHandPlacement.itsAngle	= -TWOPI * md->itsProgressMeter;
					DrawColorWithAlphaMask(md, md->itsTextureNamesCommon[TextureCommonClockSweep], &theHandPlacement, theClockSweepColor);
					theHandPlacement.itsAngle	= 0.0;
					DrawColorWithAlphaMask(md, md->itsTextureNamesCommon[TextureCommonClockSweep], &theHandPlacement, theClockBaseColor);
				}
				else
				{
					theHandPlacement.itsAngle	= -TWOPI * 0.5;
					DrawColorWithAlphaMask(md, md->itsTextureNamesCommon[TextureCommonClockSweep], &theHandPlacement, theClockSweepColor);
					theHandPlacement.itsAngle	= -TWOPI * md->itsProgressMeter;
					DrawColorWithAlphaMask(md, md->itsTextureNamesCommon[TextureCommonClockSweep], &theHandPlacement, theClockSweepColor);
				}
				theHandPlacement.itsAngle	= 0.0;
				DrawColorWithAlphaMask(md, md->itsTextureNamesCommon[TextureCommonClockRim], &theHandPlacement, theClockRimColor);
				theHandPlacement.itsAngle	= -TWOPI * md->itsProgressMeter;
				theHandPlacement.itsSizeH	*= 0.75;
				theHandPlacement.itsSizeV	*= 0.75;
				DrawColorWithAlphaMask(md, md->itsTextureNamesCommon[TextureCommonClockArrow], &theHandPlacement, theClockArrowColor);
			}

			break;
	}
}


void DrawFundamentalDomain(
	ModelData		*md,
	unsigned int	aTextureName,
	unsigned int	aNumReps)	//	How many times does the texture repeat in each direction?
{
	DrawRepeatingImages(md,
						&(Placement){0.0, 0.0, false, 0.0, 1.0, 1.0},
						ShaderTexture,
						aTextureName,
						VertexBufferRectangle,
						NULL,
						&(TexCoordTransform){{aNumReps, aNumReps}, {0.0, 0.0}});
}


void DrawSprite(
	ModelData		*md,
	unsigned int	aTextureName,
	Placement		*aPlacement)
{
	DrawRepeatingImages(md,
						aPlacement,
						ShaderTexture,
						aTextureName,
						VertexBufferRectangle,
						NULL,
						NULL);
}


void DrawColor(
	ModelData		*md,
	Placement		*aPlacement,
	float			aColor[4])	//	premultiplied alpha
{
	DrawRepeatingImages(md,
						aPlacement,
						ShaderColor,
						0,
						VertexBufferRectangle,
						aColor,
						NULL);
}


void DrawColorWithAlphaMask(
	ModelData		*md,
	unsigned int	aTextureName,
	Placement		*aPlacement,
	float			aColor[4])	//	premultiplied alpha
{
	DrawRepeatingImages(md,
						aPlacement,
						ShaderColorWithAlphaMask,
						aTextureName,
						VertexBufferRectangle,
						aColor,
						NULL);
}


void DrawSpriteTransparent(
	ModelData		*md,
	unsigned int	aTextureName,
	Placement		*aPlacement)
{
	DrawRepeatingImages(md,
						aPlacement,
						ShaderColorWithTexture,
						aTextureName,
						VertexBufferRectangle,
						PREMULTIPLY_RGBA(1.0, 1.0, 1.0, 0.25),
						NULL);
}


void DrawSpriteWithTextureTransform(
	ModelData			*md,
	Placement			*aPlacement,
	unsigned int		aTextureName,
	TexCoordTransform	*aTexCoordTransform)	//	may be NULL
{
	DrawRepeatingImages(md,
						aPlacement,
						ShaderTexture,
						aTextureName,
						VertexBufferRectangle,
						NULL,
						aTexCoordTransform);
}


void DrawKleinAxes(	
	ModelData	*md,
	float		aColorA[4],	//	premultiplied alpha
	float		aColorB[4])	//	premultiplied alpha
{
	if (md->itsTopology == TopologyKlein
	 && md->itsShowGlideAxes)
	{
		DrawSegment(md,
					aColorA,
					md->itsTextureNamesCommon[TextureCommonLineDashed],
					0.0, 0.0, 1.0, 0.01, 0.5*PI, 8);

		DrawSegment(md,
					aColorB,
					md->itsTextureNamesCommon[TextureCommonLineDashed],
					0.5, 0.0, 1.0, 0.01, 0.5*PI, 8);
	}
}


void DrawSegment(
	ModelData		*md,
	float			aColor[4],			//	premultiplied alpha;  may be NULL
	unsigned int	aLineTextureName,	//	RGBA texture if aColor = NULL;  alpha mask if aColor ≠ NULL
	double			aCenterH,
	double			aCenterV,
	double			aLength,
	double			aThickness,
	double			anAngle,			//	in radians, measured ccw from horizontal
	unsigned int	aNumReps)			//	how many times the texture repeats in the long direction
{
	Placement	thePlacement =
				{
					aCenterH,
					aCenterV,
					false,
					anAngle,
					aLength,
					aThickness
				};
	
	//	Normalize the segment's center to lie within the fundamental domain.
	//	This normalization lets the caller ensure that no object extend 
	//	more than 0.5 units beyond the fundamental domain in any direction,
	//	thus guaranteeing that it will be drawn correctly.
	NormalizePlacement(&thePlacement, md->itsTopology);
		
	//	Drag the segment.
	DrawRepeatingImages(md,
						&thePlacement,
						aColor != NULL ?
							ShaderColorWithAlphaMask :
							ShaderTexture,
						aLineTextureName,
						VertexBufferRectangle,
						aColor,
						&(TexCoordTransform){{aNumReps, 1.0}, {0.0, 0.0}});
}


void DrawSegmentWithEnds(
	ModelData		*md,
	float			aColor[4],			//	premultiplied alpha;  may be NULL
	unsigned int	aLineTextureName,	//	always RGBA texture; will be modulated by aColor if present
	double			aCenterH,
	double			aCenterV,
	double			aLength,
	double			aThickness,
	double			anAngle)			//	in radians, measured ccw from horizontal
{
	Placement	thePlacement =
				{
					aCenterH,
					aCenterV,
					false,
					anAngle,
					aLength,
					aThickness
				};

	//	Normalize the segment's center to lie within the fundamental domain.
	//	This normalization lets the caller ensure that no object extend 
	//	more than 0.5 units beyond the fundamental domain in any direction,
	//	thus guaranteeing that it will be drawn correctly.
	NormalizePlacement(&thePlacement, md->itsTopology);
		
	//	Drag the segment.
	DrawRepeatingImages(md,
						&thePlacement,
						aColor != NULL ?
							ShaderColorWithTexture :
							ShaderTexture,
						aLineTextureName,
						VertexBufferRectangleWithEndcaps,
						aColor,
						NULL);
}


void DrawTriangle(
	ModelData		*md,
	Placement		*aPlacement,
	float			aColor[4])	//	premultiplied alpha
{
	DrawRepeatingImages(md,
						aPlacement,
						ShaderColor,
						0,
						VertexBufferTriangle,
						aColor,
						NULL);
}


static void DrawRepeatingImages(
	ModelData			*md,
	Placement			*aPlacement,
	unsigned int		aShaderIndex,
	unsigned int		aTextureName,
	unsigned int		aVertexBufferIndex,
	float				aColor[4],				//	premultiplied alpha;  may be NULL
	TexCoordTransform	*aTexCoordTransform)	//	may be NULL
{
	void			(*theBindVAOFunction)(
						unsigned int aVertexArrayName,
						unsigned int aVertexBufferName,
						unsigned int anIndexBufferName,
						unsigned int anOpenGLVersion,
						bool aShadersFlag) = NULL;
	void			(*theDrawVAOFunction)(void) = NULL;
	double			c,
					s,
					f;
	float			theTextureRepetitionMatrix[4][4],
					theModelMatrix[4][4],
					theModelMatrix2[4][4];
	signed int		theGridSize;
	signed int		dh,
					dv;
	double			theBaseOffsetH;
	unsigned int	i;

	//	Change shader iff necessary.
	if (aShaderIndex != md->itsCurrentShaderIndex)
	{
		md->itsCurrentShaderIndex = aShaderIndex;

		if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
		{
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
			glUseProgram(md->itsShaderPrograms[aShaderIndex]);
#endif
		}
		else
		{
#if defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, aShaderIndex == ShaderTexture ? GL_REPLACE : GL_MODULATE);
#endif
		}
	}

	//	Change texture iff necessary.
	if (aTextureName != md->itsCurrentTextureName)
	{
		md->itsCurrentTextureName = aTextureName;
		glBindTexture(GL_TEXTURE_2D, aTextureName);
	}
	
	//	Change vertex buffer if necessary.
	if (aVertexBufferIndex != md->itsCurrentVertexBufferIndex)
	{
		md->itsCurrentVertexBufferIndex = aVertexBufferIndex;
		switch (aVertexBufferIndex)
		{
			case VertexBufferRectangle:
				theBindVAOFunction = BindRectangleVAO;
				break;

			case VertexBufferRectangleWithEndcaps:
				theBindVAOFunction = BindRectangleWithEndcapsVAO;
				break;

			case VertexBufferTriangle:
				theBindVAOFunction = BindTriangleVAO;
				break;
			
			default:	//	should never occur
				return;
		}
		theBindVAOFunction(	md->itsVertexArrayNames	[aVertexBufferIndex],
							md->itsVertexBufferNames[aVertexBufferIndex],
							md->itsIndexBufferNames [aVertexBufferIndex],
//							md->itsOpenGLVersion,	RUNNING_OPENGL_3
							0,	//	temporarily pass OpenGL version number 0.0
							md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
	}
	
	//	Set the color, if present.
	if (aColor != NULL)
	{
		if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
		{
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
			glVertexAttrib4fv(ATTRIBUTE_COLOR, aColor);
#endif
		}
		else
		{
#if defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
			glColor4f(aColor[0], aColor[1], aColor[2], aColor[3]);
#endif
		}
	}
	
	//	Set the texture coordinate transformation.
	//	If none are given, use the default coordinates.
	if (aTexCoordTransform == NULL)
		aTexCoordTransform = &(TexCoordTransform){{1.0, 1.0}, {0.0, 0.0}};
	if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
	{
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		glVertexAttrib4f(	ATTRIBUTE_TEX_COORD_TRANSFORM, 
							aTexCoordTransform->itsScaleFactor[0],
							aTexCoordTransform->itsScaleFactor[1], 
							aTexCoordTransform->itsOffset[0],
							aTexCoordTransform->itsOffset[1]);
#endif
	}
	else
	{
#if defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		MatrixIdentity(theTextureRepetitionMatrix);
		theTextureRepetitionMatrix[0][0] = aTexCoordTransform->itsScaleFactor[0];
		theTextureRepetitionMatrix[1][1] = aTexCoordTransform->itsScaleFactor[1];
		theTextureRepetitionMatrix[3][0] = aTexCoordTransform->itsOffset[0];
		theTextureRepetitionMatrix[3][1] = aTexCoordTransform->itsOffset[1];
		glMatrixMode(GL_TEXTURE);
		glLoadMatrixf((float *)theTextureRepetitionMatrix);
#endif
	}

	//	Select the appropriate drawing function.
	switch (aVertexBufferIndex)
	{
		case VertexBufferRectangle:
			theDrawVAOFunction = DrawRectangleVAO;
			break;

		case VertexBufferRectangleWithEndcaps:
			theDrawVAOFunction = DrawRectangleWithEndcapsVAO;
			break;

		case VertexBufferTriangle:
			theDrawVAOFunction = DrawTriangleVAO;
			break;
		
		default:	//	should never occur
			return;
	}
	
	//	Represent aPlacement as a 4×4 matrix (which could become 
	//	a 4×2 matrix when we abandon OpenGL ES 1.1 compatibility).
	//	The matrix should allow for ordinary vertices
	//
	//		         (h 0 0 0)( c s 0 0)(f 0 0 0)(1 0 0 0)
	//		(x y 0 1)(0 v 0 0)(-s c 0 0)(0 1 0 0)(0 1 0 0)
	//		         (* * * *)( 0 0 0 0)(0 0 0 0)(0 0 0 0)
	//		         (0 0 0 1)( 0 0 0 1)(0 0 0 1)(a b 0 1)
	//					 ↑		   ↑		↑		 ↑
	//				 dilation  rotation   flip  translation
	//				  factor    factor   factor   factor
	//
	//	as well as the special endcap vertices (which typically
	//	have e = 0.5)
	//
	//		         (h 0 0 0)( c s 0 0)(f 0 0 0)(1 0 0 0)
	//		(x y e 1)(0 v 0 0)(-s c 0 0)(0 1 0 0)(0 1 0 0)
	//		         (v 0 0 0)( 0 0 0 0)(0 0 0 0)(0 0 0 0)
	//		         (0 0 0 1)( 0 0 0 1)(0 0 0 1)(a b 0 1)
	//
	if (aPlacement->itsAngle != 0.0)
	{
		c = cos(aPlacement->itsAngle);
		s = sin(aPlacement->itsAngle);
	}
	else
	{
		c = 1.0;
		s = 0.0;
	}
	f = (aPlacement->itsFlip ? -1.0 : 1.0);
	MatrixIdentity(theModelMatrix);
	theModelMatrix[0][0] = aPlacement->itsSizeH * c    * f;
	theModelMatrix[0][1] = aPlacement->itsSizeH * s;
	theModelMatrix[1][0] = aPlacement->itsSizeV * (-s) * f;
	theModelMatrix[1][1] = aPlacement->itsSizeV * c;
	theModelMatrix[2][0] = aPlacement->itsSizeV * c    * f;
	theModelMatrix[2][1] = aPlacement->itsSizeV * s;
	theModelMatrix[2][2] = 0.0;
	theModelMatrix[3][0] = aPlacement->itsH;
	theModelMatrix[3][1] = aPlacement->itsV;
	
	//	Make a working copy of theModelMatrix,
	//	which will store the product of the ModelMatrix
	//	with each element of the tiling group.
	MatrixCopy(theModelMatrix2, theModelMatrix);

	//	Draw translated images of the sprite.
	theGridSize = (md->itsView == ViewTiling ? 2 : 1);
	for (dv = -theGridSize; dv <= +theGridSize; dv++)
	{
		theModelMatrix2[3][1] = (float)(theModelMatrix[3][1] + (double)dv);
		if (md->itsTopology == TopologyKlein)
		{
			if (dv & 0x00000001)	//	flip odd rows
			{
				for (i = 0; i < 4; i++)
					theModelMatrix2[i][0] = - theModelMatrix[i][0];
			}
			else
			{
				for (i = 0; i < 4; i++)
					theModelMatrix2[i][0] =   theModelMatrix[i][0];
			}
		}
		else
		{
			theModelMatrix2[3][0] = theModelMatrix[3][0];
		}

		theBaseOffsetH = theModelMatrix2[3][0];

		for (dh = -theGridSize; dh <= +theGridSize; dh++)
		{
			theModelMatrix2[3][0] = (float)(theBaseOffsetH + (double)dh);

			if (md->itsShadersAvailableFlag && md->itsShadersEnabledFlag)
			{
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
				glVertexAttrib4f(	ATTRIBUTE_TRANSFORMATION_X,
									theModelMatrix2[0][0],
									theModelMatrix2[1][0],
									theModelMatrix2[2][0],
									theModelMatrix2[3][0]);
				glVertexAttrib4f(	ATTRIBUTE_TRANSFORMATION_Y,
									theModelMatrix2[0][1],
									theModelMatrix2[1][1],
									theModelMatrix2[2][1],
									theModelMatrix2[3][1]);
#endif
			}
			else
			{
#if defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
				glMatrixMode(GL_MODELVIEW);
				glLoadMatrixf((float *)theModelMatrix2);
#endif
			}
			
			theDrawVAOFunction();
		}
	}
}


ErrorText GetErrorString(void)
{
	switch (glGetError())
	{
		case GL_NO_ERROR:
			return NULL;
		case GL_INVALID_ENUM:
			return "GL_INVALID_ENUM:  GLenum argument out of range.";
		case GL_INVALID_VALUE:
			return "GL_INVALID_VALUE:  Numeric argument out of range.";
		case GL_INVALID_OPERATION:
			return "GL_INVALID_OPERATION:  Operation illegal in current state.";
#if defined(SUPPORT_OPENGL_ES_1) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
		case GL_STACK_OVERFLOW:
			return "GL_STACK_OVERFLOW:  Command would cause a stack overflow.";
		case GL_STACK_UNDERFLOW:
			return "GL_STACK_UNDERFLOW:  Command would cause a stack underflow.";
#endif
		case GL_OUT_OF_MEMORY:
			return "GL_OUT_OF_MEMORY:  Not enough memory left to execute command.";
		default:
			return "Unknown OpenGL error.";
	}
}
