//	TorusGamesInit.c
//
//	Initializes the ModelData and prepares OpenGL extensions and textures.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <time.h>	//	for seeding srand()
#include <math.h>	//	for floor()
#include <stdio.h>	//	for sscanf()


//	How big is the hand cursor?
#define HAND_SIZE	0.1


static ErrorText	SetUpExtensions(ModelData *md);
static bool			WholeWordSubstring(const char *aString, const char *aPotentialSubstring);
static void			ShutDownExtensions(ModelData *md);
static ErrorText	SetUpShaders(ModelData *md);
static ErrorText	SetUpOneShaderProgram(unsigned int *aShaderProgram, ShaderIndex aShaderIndex);
static void			ShutDownShaders(ModelData *md);
static ErrorText	SetUpTextures(ModelData *md);
static void			ShutDownTextures(ModelData *md);
static ErrorText	SetUpVAOs(ModelData *md);
static void			ShutDownVAOs(ModelData *md);


void SetUpModelData(ModelData *md)
{
	unsigned int	i;

	//	All OpenGL elements require initialization.
	md->itsPreparedExtensions	= false;
	md->itsPreparedShaders		= false;
	md->itsPreparedTextures		= false;
	md->itsPreparedVAOs			= false;

	//	We'll test for OpenGL extensions at the last minute,
	//	just before rendering the first frame.
	//	For now set safe defaults.
	md->itsMultisamplingAvailableFlag	= false;
	md->itsAnisotropicAvailableFlag		= false;
	md->itsShadersAvailableFlag			= false;
	
	//	No shaders, textures or vertex/index buffers are present yet.

	for (i = 0; i < NumShaders; i++)
		md->itsShaderPrograms[i] = 0;

	for (i = 0; i < NumTexturesCommon; i++)
		md->itsTextureNamesCommon[i] = 0;

	for (i = 0; i < NumVertexBuffers; i++)
	{
		md->itsVertexArrayNames [i] = 0;
		md->itsVertexBufferNames[i] = 0;
		md->itsIndexBufferNames [i] = 0;
	}

	//	Enable synchronization by default.
	md->itsSyncFlag	= true;

	//	Disable multisampling by default.  In a 2D program like Torus Games,
	//	all the antialiasing takes place in the textures.  The pool cue
	//	is almost an exception, but with anisotropic texturing enabled
	//	its edges are effectively defined within the pool cue texture.
	//	(Multisampling takes place iff itsMultisamplingAvailableFlag and 
	//	itsMultisamplingEnabledFlag are both true.)
	md->itsMultisamplingEnabledFlag	= false;
	
	//	Enable shaders by default.  This setting may be changed only 
	//	via a "secret" keystroke 's' on the Mac, and not at all on Windows.
//	md->itsShadersEnabledFlag		= true;
//DISABLE SHADERS,
//FOR MAXIMAL ROBUSTNESS IN THIS ARCHIVED VERSION OF THE PROGRAM.
md->itsShadersEnabledFlag = false;

	//	No redraw request is pending.
	md->itsRedrawRequestFlag		= false;


	//	Set up application-specific data.

	//	Scrolling position
	md->itsOffset.itsH		= 0.0;
	md->itsOffset.itsV		= 0.0;
	md->itsOffset.itsFlip	= false;

	//	Options
	md->itsGame				= GameNone;
	md->itsTopology			= TopologyTorus;
	md->itsView				= ViewDomain;
	md->itsHumanVsComputer	= true;
	md->itsDifficultyLevel	= 2;
	md->itsShowGlideAxes	= true;

	//	Win tracking
	md->itsGameIsOver	= false;
	md->itsFlashFlag	= false;

	//	Hand status
	md->itsHandStatus				= HandNone;
	md->itsHandPlacement.itsH		= 0.0;
	md->itsHandPlacement.itsV		= 0.0;
	md->itsHandPlacement.itsFlip	= false;
	md->itsHandPlacement.itsAngle	= 0.0;
	md->itsHandPlacement.itsSizeH	= HAND_SIZE;
	md->itsHandPlacement.itsSizeV	= HAND_SIZE;
	
	//	Double-click testing
	md->itsPrevClickPlacement		= md->itsHandPlacement;
	
	//	Touch velocity
	md->itsTouchIsCoasting			= false;
	md->itsTouchVelocity[0]			= 0.0;
	md->itsTouchVelocity[1]			= 0.0;
	
	//	Simulation status
	md->itsSimulationStatus			= SimulationNone;
	md->itsSimulationElapsedTime	= 0.0;
	
	//	Progress meter
	md->itsProgressMeterIsActive	= false;
	md->itsProgressMeter			= 0.0;
	
	//	Function pointers
	ClearFunctionPointers(md);
	
	//	Note:  The platform-dependent code will have set
	//	the opaque pointers md->itsFonts[] and md->itsSideWindow.
	//	In particular, we do *not* want to set them to NULL!

	//	Initialize the random number generator.
	//	If one runs the program several times within the course
	//	of, say, a half-hour, time(NULL) will of course vary from
	//	one run to the next.  However, the first random number one
	//	subsequently gets from rand() is decidedly non-random
	//	(at least in the Win32 implementation of rand()).
	//	For example the horizontal position of the first jigsaw
	//	puzzle piece is always the same.  To avoid this problem,
	//	get a random number and throw it away before proceeding.
	srand(time(NULL));
	(void) rand();	//	Discard first not-so-random number.
}


void ShutDownModelData(ModelData *md)
{
	if (md->itsGame != GameNone)
		ErrorMessage("ShutDownModelData() called when itsGame ≠ GameNone", "Internal Error", false);

	//	If the null game held any allocated memory we'd free it here, but it doesn't.
}


void ClearFunctionPointers(ModelData *md)
{
	md->itsGameShutDown			= NULL;
	md->itsGameSetUpTextures	= NULL;
	md->itsGameShutDownTextures	= NULL;
	md->itsGameReset			= NULL;
	md->itsGameHandMoved		= NULL;
	md->itsGameDragBegin		= NULL;
	md->itsGameDragObject		= NULL;
	md->itsGameDragEnd			= NULL;
	md->itsGameCharacterInput	= NULL;
	md->itsGameSpecialCommand	= NULL;
	md->itsGameSimulationUpdate	= NULL;
	md->itsGameDraw				= NULL;
}


ErrorText SetUpGraphicsAsNeeded(ModelData *md)
{
	//	These are one-time initializations,
	//	so usually no work is required here.
	//	Only in exceptional circumstances, for example
	//	when the user has changed the contents or display settings,
	//	will something need an update.

	//	Assume the OpenGL context has already been set.
	
	ErrorText	theError;

	if ( ! md->itsPreparedExtensions )
	{
		if ((theError = SetUpExtensions(md)) != NULL)
			return theError;
		md->itsPreparedExtensions = true;

		//	Availability of extensions may influence
		//	the construction of textures, etc.,
		//	so rebuild everything from scratch.
		md->itsPreparedShaders	= false;
		md->itsPreparedTextures	= false;
		md->itsPreparedVAOs		= false;
	}

	if (md->itsPreparedShaders == false)
	{
		if ((theError = SetUpShaders(md)) != NULL)
			return theError;
		md->itsPreparedShaders	= true;
		
		//	Alpha-only textures get handled differently
		//	in the OpenGL 1.1 fixed-function pipeline,
		//	in OpenGL 2.0 shaders, and in OpenGL 3.0 shaders.
		//	Please see the comments in SetTextureImageFancy()
		//	for details.
		md->itsPreparedTextures = false;
	}

	if (md->itsPreparedTextures == false)
	{
		if ((theError = SetUpTextures(md)) != NULL)
			return theError;
		md->itsPreparedTextures	= true;
	}

	if ( ! md->itsPreparedVAOs )
	{
		if ((theError = SetUpVAOs(md)) != NULL)
			return theError;
		md->itsPreparedVAOs		= true;
	}
	
	return NULL;
}


void ShutDownGraphicsAsNeeded(ModelData *md)
{
	//	Assume the OpenGL context has already been set.

	//	Shut down the various elements in the opposite order
	//	from that in which they were created.  In particular,
	//	be sure to shut down the extensions last, so the other
	//	shutdown routines will know what extensions have been
	//	available.  For example, ShutDownShaders() will need
	//	to know whether shaders have been available or not.
	ShutDownVAOs(md);
	ShutDownTextures(md);
	ShutDownShaders(md);
	ShutDownExtensions(md);

	md->itsPreparedExtensions	= false;
	md->itsPreparedShaders		= false;
	md->itsPreparedTextures		= false;
	md->itsPreparedVAOs			= false;
}


static ErrorText SetUpExtensions(ModelData *md)
{
	const char		*theExtensionsString;
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	const char		*theVersionString;
	unsigned int	theMajorVersionNumber,
					theMinorVersionNumber;
#endif

	//	Desired OpenGL extensions are assumed absent until found.
	md->itsMultisamplingAvailableFlag	= false;
	md->itsAnisotropicAvailableFlag		= false;
	md->itsShadersAvailableFlag			= false;

	//	Get the renderer's extensions string.
	theExtensionsString = (const char *) glGetString(GL_EXTENSIONS);

	if (theExtensionsString != NULL)
	{
		//	Is multisampling available?
		if (WholeWordSubstring(theExtensionsString, "GL_ARB_multisample"))
			md->itsMultisamplingAvailableFlag = true;

		//	Is anisotropic texture filtering available?
		if (WholeWordSubstring(theExtensionsString, "GL_EXT_texture_filter_anisotropic"))
			md->itsAnisotropicAvailableFlag = true;

		//	Are shaders available?
#ifdef TORUS_GAMES_FOR_OPENGL_ES
		//	The OpenGL 1.1 spec says that the version string should have the format
		//
		//		"OpenGL ES-XX 1.1"
		//
		//	where XX tells the profile (CM or CL).
		//
		//	The OpenGL 2.0 spec says that the version string should have the format
		//
		//		"OpenGL ES N.M vendor-specific information"
		//
		//	where N and M are the major and minor version numbers.
		theVersionString = (const char *) glGetString(GL_VERSION);
		if (sscanf(theVersionString,
			"OpenGL %*s %d.%d",
			&theMajorVersionNumber,
			&theMinorVersionNumber) == 2)
		{
			md->itsShadersAvailableFlag = (theMajorVersionNumber >= 2);
		}
		else
		{
			md->itsShadersAvailableFlag = false;
		}
#else
		if (WholeWordSubstring(theExtensionsString, "GL_ARB_shader_objects" )
		 && WholeWordSubstring(theExtensionsString, "GL_ARB_vertex_shader"  )
		 && WholeWordSubstring(theExtensionsString, "GL_ARB_fragment_shader"))
			md->itsShadersAvailableFlag = true;
#endif
	}
	
	//	Did any OpenGL errors occur?
	return GetErrorString();
}


static bool WholeWordSubstring(
	const char	*aString,				//	null-terminated
	const char	*aPotentialSubstring)	//	null-terminated
{
	const char	*thePotentialMatch,
				*s,
				*p;
	bool		theMatchFlag;

	//	Does aPotentialSubstring occur as a complete space-delimited word in aString?

	//	Irrelevant technical detail:  an empty aPotentialSubstring
	//	(consisting of a terminating zero alone) counts as a substring
	//	of any string, even an empty one.

	//	Consider the first character in each space-delimited word of aString
	//	as the starting point for a potential match.
	thePotentialMatch = aString;
	while (*thePotentialMatch == ' ')
		thePotentialMatch++;
	do
	{
		//	Compare each character in aPotentialSubstring
		//	(excluding the terminating zero) to thePotentialMatch.
		s = thePotentialMatch;
		p = aPotentialSubstring;
		theMatchFlag = true;
		while (theMatchFlag && *p != 0)
		{
			if (*p++ != *s++)
				theMatchFlag = false;
		}

		//	If a space or terminating 0 follows a match, return true.
		if (theMatchFlag
		 && (*s == ' ' || *s == 0))
			return true;

		//	Skip to the beginning of the next whole word in thePotentialMatch.
		//		Skip by non-spaces.
		while (*thePotentialMatch != ' ' && *thePotentialMatch != 0)
			thePotentialMatch++;
		//		Skip by intervening spaces.
		while (*thePotentialMatch == ' ')
			thePotentialMatch++;

		//	Continue with the loop to examine the next whole word (if any).

	} while (*thePotentialMatch != 0);

	return false;
}


static void ShutDownExtensions(ModelData *md)
{
	//	Just for good form...
	md->itsMultisamplingAvailableFlag	= false;
	md->itsAnisotropicAvailableFlag		= false;
	md->itsShadersAvailableFlag			= false;
}


static ErrorText SetUpShaders(ModelData *md)
{
	ErrorText		theError;
	unsigned int	i;

	if (md->itsShadersAvailableFlag)
	{
		//	Delete any pre-existing shaders.
		ShutDownShaders(md);

		//	If shaders are desired, then...
		if (md->itsShadersEnabledFlag)
		{
			//	...create a shader of each type.
			//
			//	At run time, swapping simple shaders in and out as needed turns out
			//	to be faster than running a single all-purpose shader with options,
			//	because (much to my surprise) an if/then block containing
			//	texture sampling slows the fragment shader to a crawl
			//	(at least on my Radeon X1600).
			for (i = 0; i < NumShaders; i++)
			{
				theError = SetUpOneShaderProgram(&md->itsShaderPrograms[i], (ShaderIndex) i );
				if (theError != NULL)
					return theError;
			}
		}
	}
	
	//	Did any OpenGL errors occur?
	return GetErrorString();
}


static ErrorText SetUpOneShaderProgram(
	unsigned int	*aShaderProgram,	//	output
	ShaderIndex		aShaderIndex)		//	input
{
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
	ErrorText		theErrorMessage					= NULL;
	char			*theVertexShaderFileName		= NULL,
					*theFragmentShaderFileName		= NULL,
					theFilePath[2048];
	unsigned int	theVertexShaderSourceLength		= 0,
					theFragmentShaderSourceLength	= 0;
	Byte			*theVertexShaderSource			= NULL,
					*theFragmentShaderSource		= NULL;
	const char		*theVertexShaderPrefix			= NULL,
					*theFragmentShaderPrefix		= NULL;
	unsigned int	theVertexShader					= 0,
					theFragmentShader				= 0;
	GLint			theSuccessFlag;

	static GLchar	theInfoLog[1024] = {0};	//	static so it persists after this function returns
	
	//	For safe error handling.
	*aShaderProgram = 0;	//	Should already be 0, but reset it anyway.
	
	//	Select an appropriate pair of shaders for the given ShaderIndex.
	switch (aShaderIndex)
	{
		case ShaderTexture:
			theVertexShaderFileName		= "TorusGames-Texture.vs";
			theFragmentShaderFileName	= "TorusGames-Texture.fs";
			break;

		case ShaderColor:
			theVertexShaderFileName		= "TorusGames-Color.vs";
			theFragmentShaderFileName	= "TorusGames-Color.fs";
			break;

		case ShaderColorWithAlphaMask:
			theVertexShaderFileName		= "TorusGames-ColorWithAlphaMask.vs";
			theFragmentShaderFileName	= "TorusGames-ColorWithAlphaMask.fs";
			break;

		case ShaderColorWithTexture:
			theVertexShaderFileName		= "TorusGames-ColorWithTexture.vs";
			theFragmentShaderFileName	= "TorusGames-ColorWithTexture.fs";
			break;

		default:
			theErrorMessage = "Invalid shader index";
			goto CleanUpSetUpOneShaderProgram;
	}

	//	Load the vertex shader source code.
	if ((theErrorMessage = GetAbsolutePath(	"Shaders", 
											theVertexShaderFileName,
											theFilePath,
											BUFFER_LENGTH(theFilePath))) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	if ((theErrorMessage = GetFileContents(theFilePath, &theVertexShaderSourceLength, &theVertexShaderSource)) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	
	//	Load the fragment shader source code.
	if ((theErrorMessage = GetAbsolutePath(	"Shaders",
											theFragmentShaderFileName,
											theFilePath,
											BUFFER_LENGTH(theFilePath))) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	if ((theErrorMessage = GetFileContents(theFilePath, &theFragmentShaderSourceLength, &theFragmentShaderSource)) != NULL)
		goto CleanUpSetUpOneShaderProgram;
	
	//	Select appropriate prefixes.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	Convert GLSL 1.3 storage qualifiers "in" and "out"
	//	to their GLSL ES 1.0 equivalents "attribute" and "varying".
	theVertexShaderPrefix	= "#version 100\n#define in attribute\n#define out varying\n";
	theFragmentShaderPrefix	= "#version 100\n#define in varying\nprecision highp float;\n";
#else
	if (RUNNING_OPENGL_3)
	{
		theVertexShaderPrefix	= "#version 130\n";
		theFragmentShaderPrefix	= "#version 130\n";
	}
	else
	{
		//	Convert GLSL 1.3 storage qualifiers "in" and "out"
		//	to their GLSL 1.1 equivalents "attribute" and "varying".
		theVertexShaderPrefix	= "#version 110\n#define in attribute\n#define out varying\n";
		theFragmentShaderPrefix	= "#version 110\n#define in varying\n";
	}
#endif

	//	Create the vertex shader.
	theVertexShader = glCreateShader(GL_VERTEX_SHADER);
	glShaderSource(	theVertexShader,
					2,
					(const char *[2])	{	theVertexShaderPrefix,	(const char *)theVertexShaderSource	},
					(GLint [2])			{	-1,						theVertexShaderSourceLength			});
	glCompileShader(theVertexShader);
	glGetShaderiv(theVertexShader, GL_COMPILE_STATUS, &theSuccessFlag);
	if ( ! theSuccessFlag )
	{
		glGetShaderInfoLog(theVertexShader, BUFFER_LENGTH(theInfoLog), NULL, theInfoLog);
		theErrorMessage = theInfoLog;
		goto CleanUpSetUpOneShaderProgram;
	}

	//	Create the fragment shader.
	theFragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(	theFragmentShader,
					2,
					(const char *[2])	{	theFragmentShaderPrefix,	(const char *)theFragmentShaderSource	},
					(GLint [2])			{	-1,							theFragmentShaderSourceLength			});
	glCompileShader(theFragmentShader);
	glGetShaderiv(theFragmentShader, GL_COMPILE_STATUS, &theSuccessFlag);
	if ( ! theSuccessFlag )
	{
		glGetShaderInfoLog(theFragmentShader, BUFFER_LENGTH(theInfoLog), NULL, theInfoLog);
		theErrorMessage = theInfoLog;
		goto CleanUpSetUpOneShaderProgram;
	}

	//	Create the program and attach the vertex and fragment shaders.
	*aShaderProgram = glCreateProgram();
	if (*aShaderProgram == 0)
	{
		theErrorMessage = "Couldn't create *aShaderProgram";
		goto CleanUpSetUpOneShaderProgram;
	}
	glAttachShader(*aShaderProgram, theVertexShader  );
	glAttachShader(*aShaderProgram, theFragmentShader);
	
	//	Assign locations to vertex shader attributes.
	//	It's always safe to bind an attribute name, whether or not the given shader uses it.
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_POSITION,				"atrPosition"				);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_TEX_COORD,				"atrTextureCoordinates"		);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_COLOR,					"atrColor"					);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_TRANSFORMATION_X,		"atrTransformationX"		);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_TRANSFORMATION_Y,		"atrTransformationY"		);
	glBindAttribLocation(*aShaderProgram, ATTRIBUTE_TEX_COORD_TRANSFORM,	"atrTextureCoordTransform"	);
	
	//	Link the program.
	glLinkProgram(*aShaderProgram);
	glGetProgramiv(*aShaderProgram, GL_LINK_STATUS, &theSuccessFlag);
	if ( ! theSuccessFlag )
	{
		glGetProgramInfoLog(*aShaderProgram, BUFFER_LENGTH(theInfoLog), NULL, theInfoLog);
		theErrorMessage = theInfoLog;
		goto CleanUpSetUpOneShaderProgram;
	}

CleanUpSetUpOneShaderProgram:

	//	Free our private copy of the source code.
	FreeFileContents(&theVertexShaderSourceLength,   &theVertexShaderSource  );
	FreeFileContents(&theFragmentShaderSourceLength, &theFragmentShaderSource);

	//	Release our temporary references (if any) to the vertex and fragments shaders.
	//	The shader program will keep its own references to the vertex and fragments shaders,
	//	and will automatically delete them when it itself gets deleted.  
	//	Passing 0 is harmless.
	glDeleteShader(theVertexShader);
	glDeleteShader(theFragmentShader);

	//	If an error occured, delete the shader program (if any).
	if (theErrorMessage != NULL)
	{
		glDeleteProgram(*aShaderProgram);	//	OK to pass 0
		*aShaderProgram = 0;
	}

	return theErrorMessage;
#else
	return NULL;
#endif
}


static void ShutDownShaders(ModelData *md)
{
#if defined(SUPPORT_OPENGL_ES_2) || ! defined(TORUS_GAMES_FOR_OPENGL_ES)
	unsigned int	i;

	if (md->itsShadersAvailableFlag)
	{
		glUseProgram(0);

		for (i = 0; i < NumShaders; i++)
		{
			glDeleteProgram(md->itsShaderPrograms[i]);
			md->itsShaderPrograms[i] = 0;
		}
	}
#endif
}


static ErrorText SetUpTextures(ModelData *md)
{
	ErrorText	theFirstError	= NULL;

	//	Set up the common textures.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	The iPhone OS Torus Games has no hand cursor,
	//	but when the computer is choosing its chess move,
	//	it will display a wait cursor.
#else
	SetUpOneTexture(md, &md->itsTextureNamesCommon[TextureCommonHandFlat],		"Common/HandFlat.rgba",		TextureClamp,	TextureRGBA,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNamesCommon[TextureCommonHandGrab],		"Common/HandGrab.rgba",		TextureClamp,	TextureRGBA,	&theFirstError);
#endif
	SetUpOneTexture(md, &md->itsTextureNamesCommon[TextureCommonClockBase],		"Common/ClockBase.rgba",	TextureClamp,	TextureAlpha,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNamesCommon[TextureCommonClockSweep],	"Common/ClockSweep.rgba",	TextureClamp,	TextureAlpha,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNamesCommon[TextureCommonClockRim],		"Common/ClockRim.rgba",		TextureClamp,	TextureAlpha,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNamesCommon[TextureCommonClockArrow],	"Common/ClockArrow.rgba",	TextureClamp,	TextureAlpha,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNamesCommon[TextureCommonLineDashed],	"Common/LineDashed.rgba",	TextureWrap,	TextureAlpha,	&theFirstError);
	SetUpOneTexture(md, &md->itsTextureNamesCommon[TextureCommonHalo],			"Common/Halo.rgba",			TextureClamp,	TextureAlpha,	&theFirstError);

	//	Set up the textures for the current game.
	if (md->itsGameSetUpTextures != NULL)
		(*md->itsGameSetUpTextures)(md, &theFirstError);

	return theFirstError;

	//	Did any OpenGL errors occur?
	return GetErrorString();
}


void SetUpOneTexture(
	ModelData		*md,
	unsigned int	*aTextureName,		//	output
	char			*aTextureFileName,	//	pass NULL to set up an empty texture
	TextureBoundary	aTextureBoundary,
	TextureFormat	aTextureFormat,
	ErrorText		*aFirstError)		//	may be NULL
{
	ErrorText		theErrorMessage		= NULL;
	float			theMaxAnisotropy	= 1.0;
	ImageRGBA		*theImageRGBA		= NULL;
	PixelRGBA		*theSrcPixel;
	Byte			*theDstByte;
	unsigned int	theCount;
	
	//	Release any pre-existing texture.
	glDeleteTextures(1, (GLuint *) aTextureName);	//	silently ignores *aTextureName == 0
	*aTextureName = 0;
	
	//	Generate a new texture name.
	glGenTextures(1, (GLuint *) aTextureName);

	//	Bind the texture.
	glBindTexture(GL_TEXTURE_2D, *aTextureName);

	//	Set minification and magnification filters.
	//	Mipmapping is almost always a good idea.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	//	On the iPhone and iPod Touch (PowerVR MBX Lite) GL_LINEAR_MIPMAP_NEAREST is much faster
	//	than GL_LINEAR_MIPMAP_LINEAR (40-42 fps instead of 25-27 fps in torus Jigsaw).
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
#else
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
#endif
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR               );

	//	Set anisotropic texture filtering.
	if (md->itsAnisotropicAvailableFlag)
	{
		//	Only the pool cue needs anisotropic filtering.
		//	Otherwise leave it off -- on desktop computers it makes little difference,
		//	but on the iPhone and iPod Touch (PowerVR MBX Lite) the frame rate
		//	increases tremendously without it.  For example, disabling anisotropic
		//	texture filtering increased the Jigsaw's frame rate from 40-41 fps 
		//	to the maximum 60 fps.
		if (md->itsGame == GamePool && *aTextureName == md->itsGameOf.Pool.itsTextureNameStick)
			glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &theMaxAnisotropy);	//	Allow the maximum.
		else
			theMaxAnisotropy = 1.0;	//	Suppress anisotropic texture filtering.
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, theMaxAnisotropy);
	}

	//	Set wrapping or clamping.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, aTextureBoundary == TextureClamp ? GL_CLAMP_TO_EDGE : GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, aTextureBoundary == TextureClamp ? GL_CLAMP_TO_EDGE : GL_REPEAT);
	
	//	If the caller didn't want to load a file, we're done.
	if (aTextureFileName == NULL)
		return;

	//	Read the texture file.
	if ((theErrorMessage = ReadImageRGBA(aTextureFileName, &theImageRGBA)) != NULL)
		goto CleanUpSetUpOneTexture;
	
	//	What sort of texture does the caller want?
	switch (aTextureFormat)
	{
		case TextureRGBA:

			//	Pass the unmodified image to OpenGL.
			SetTextureImageFancy(	*aTextureName,
									theImageRGBA->itsWidth,
									theImageRGBA->itsHeight,
									4,
									(Byte *) theImageRGBA->itsPixels,
									md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);

			break;

		case TextureAlpha:

			//	Ignore the red, green and blue bytes, keeping only the alpha bytes.
			//	Note that we condense the array in place -- no new memory gets allocated.
			theSrcPixel =          theImageRGBA->itsPixels;
			theDstByte  = (Byte *) theImageRGBA->itsPixels;
			for (theCount = theImageRGBA->itsWidth * theImageRGBA->itsHeight; theCount-- > 0; )
				*theDstByte++ = (theSrcPixel++)->a;

			//	Pass the alpha values to OpenGL.
			SetTextureImageFancy(	*aTextureName,
									theImageRGBA->itsWidth,
									theImageRGBA->itsHeight,
									1,
									(Byte *) theImageRGBA->itsPixels,
									md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);

			break;
	}

CleanUpSetUpOneTexture:

	//	Free the image (if necessary).
	FreeImageRGBA(&theImageRGBA);

	if (theErrorMessage != NULL)
	{
		//	Substitute a pure red texture for the missing one.
		SetTextureImageFancy(	*aTextureName,
								1,
								1,
								4,
								(Byte [4]){0xFF, 0x00, 0x00, 0xFF},	//	premultiplied alpha
								md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);
		
		//	Report this error iff an error report is desired
		//	and no earlier error has occurred.
		if (aFirstError != NULL
		 && *aFirstError == NULL)
		{
			*aFirstError = theErrorMessage;
		}

		//	Let the caller push on without the desired texture...
	}
}

ErrorText ReadImageRGBA(
	char		*aTextureFileName,	//	input, zero-terminated UTF-8 string
	ImageRGBA	**anImageRGBA)		//	output
{
	ErrorText		theErrorMessage		= NULL;
	ImageRGBA		*theImage			= NULL;
	char			*theTextureFolder,
					theFilePath[4096]	= "";
	unsigned int	i;
	float			theAlphaFraction;
	
	static bool		theFractionsAreInitialized	= false;
	static float	theFractions[256];
	
	//	Check the output variable.
	if (anImageRGBA == NULL)
	{
		theErrorMessage = "anImageRGBA == NULL in ReadImageRGBA()";
		goto CleanUpReadImageRGBA;
	}
	if (*anImageRGBA != NULL)
	{
		theErrorMessage = "*anImageRGBA != NULL in ReadImageRGBA()";
		goto CleanUpReadImageRGBA;
	}

	//	Allocate and initialize an empty ImageRGBA.
	theImage = GET_MEMORY(sizeof(ImageRGBA));
	if (theImage == NULL)
	{
		theErrorMessage = "Couldn't allocate theImage in ReadImageRGBA().";
		goto CleanUpReadImageRGBA;
	}
	theImage->itsNumRawBytes	= 0;
	theImage->itsRawBytes		= NULL;
	theImage->itsWidth			= 0;
	theImage->itsHeight			= 0;
	theImage->itsPixels			= NULL;
	
	//	Select a texture collection appropriate for the platform.
#ifdef TORUS_GAMES_FOR_OPENGL_ES
	theTextureFolder = "Textures - Mobile";		//	limited collection of lower-resolution textures
#else
	theTextureFolder = "Textures - Desktop";	//	full collection of high-resolution textures
#endif

	//	Read the texture file's bytes.
	if ((theErrorMessage = GetAbsolutePath(theTextureFolder, aTextureFileName, theFilePath, BUFFER_LENGTH(theFilePath))) != NULL)
		goto CleanUpReadImageRGBA;
	if ((theErrorMessage = GetFileContents(theFilePath, &theImage->itsNumRawBytes, &theImage->itsRawBytes)) != NULL)
		goto CleanUpReadImageRGBA;

	//	Did we get a header?
	if (theImage->itsNumRawBytes < 8)
	{
		theErrorMessage = "Texture file lacks width and/or height specification.";
		goto CleanUpReadImageRGBA;
	}

	//	Parse the image width and height as big-endian 4-byte integers.
	theImage->itsWidth  = ((unsigned int)theImage->itsRawBytes[0] << 24)
						+ ((unsigned int)theImage->itsRawBytes[1] << 16)
						+ ((unsigned int)theImage->itsRawBytes[2] <<  8)
						+ ((unsigned int)theImage->itsRawBytes[3] <<  0);
	theImage->itsHeight = ((unsigned int)theImage->itsRawBytes[4] << 24)
						+ ((unsigned int)theImage->itsRawBytes[5] << 16)
						+ ((unsigned int)theImage->itsRawBytes[6] <<  8)
						+ ((unsigned int)theImage->itsRawBytes[7] <<  0);

	//	OpenGL requires power-of-two dimensions.
	if ( ! IsPowerOfTwo(theImage->itsWidth ) 
	 ||  ! IsPowerOfTwo(theImage->itsHeight) )
	{
		theErrorMessage = "OpenGL requires that each texture's width and height be powers of two.";
		goto CleanUpReadImageRGBA;
	}

	//	Is the total file length correct?
	if (theImage->itsNumRawBytes !=  8  +  4 * theImage->itsWidth * theImage->itsHeight)
	{
		theErrorMessage = "Number of pixels in texture file does not match stated width and height.";
		goto CleanUpReadImageRGBA;
	}

	//	Locate the pixels.
	theImage->itsPixels = (PixelRGBA *) (theImage->itsRawBytes + 8);
	
	//	Premultiply each pixel's red, green and blue components by its alpha component.
	//
	//	Note:  The RGBA files store un-premultiplied values for greater flexibility.
	//	For example, we want to be able to edit an image's alpha mask without losing RGB values.
	if ( ! theFractionsAreInitialized )
	{
		for (i = 0; i < 256; i++)
			theFractions[i] = i / 255.0;
		theFractionsAreInitialized = true;
	}
	for (i = 0; i < theImage->itsWidth * theImage->itsHeight; i++)
	{
		if (theImage->itsPixels[i].a != 0xFF)
		{
			theAlphaFraction = theFractions[theImage->itsPixels[i].a];
			theImage->itsPixels[i].r *= theAlphaFraction;
			theImage->itsPixels[i].g *= theAlphaFraction;
			theImage->itsPixels[i].b *= theAlphaFraction;
		}
	}
	
	//	Assign theImage to the output variable.
	*anImageRGBA = theImage;

CleanUpReadImageRGBA:
	
	if (theErrorMessage != NULL)
	{
		FreeImageRGBA(&theImage);
		if (anImageRGBA != NULL)
			*anImageRGBA = NULL;
	}
	
	return theErrorMessage;
};

void FreeImageRGBA(ImageRGBA **anImage)
{
	if (anImage != NULL && *anImage != NULL)
	{
		FreeFileContents(&(*anImage)->itsNumRawBytes, &(*anImage)->itsRawBytes);

		FREE_MEMORY_SAFELY(*anImage);
	}
}


static void ShutDownTextures(ModelData *md)
{
	//	Release the common textures.
#ifndef TORUS_GAMES_FOR_OPENGL_ES
	ShutDownOneTexture(&md->itsTextureNamesCommon[TextureCommonHandFlat]	);
	ShutDownOneTexture(&md->itsTextureNamesCommon[TextureCommonHandGrab]	);
#endif
	ShutDownOneTexture(&md->itsTextureNamesCommon[TextureCommonClockBase]	);
	ShutDownOneTexture(&md->itsTextureNamesCommon[TextureCommonClockSweep]	);
	ShutDownOneTexture(&md->itsTextureNamesCommon[TextureCommonClockRim]	);
	ShutDownOneTexture(&md->itsTextureNamesCommon[TextureCommonClockArrow]	);
	ShutDownOneTexture(&md->itsTextureNamesCommon[TextureCommonLineDashed]	);
	ShutDownOneTexture(&md->itsTextureNamesCommon[TextureCommonHalo]		);

	//	Release the textures for the current game.
	if (md->itsGameShutDownTextures != NULL)
		(*md->itsGameShutDownTextures)(md);
}

void ShutDownOneTexture(unsigned int *aTextureName)
{
	glDeleteTextures(1, (GLuint *) aTextureName);	//	silently ignores *aTextureName == 0
	*aTextureName = 0;
}


static ErrorText SetUpVAOs(ModelData *md)
{
	//	Release any pre-existing objects.
	ShutDownVAOs(md);
	
	//	Generate new object names.
	if (RUNNING_OPENGL_3)
		;	//	uncomment the following two lines:
//	if (md->itsOpenGLVersion >= VERSION_NUMBER(3, 0))
//		glGenVertexArrays(NumVertexBuffers, (GLuint *) md->itsVertexArrayNames);
	glGenBuffers(NumVertexBuffers, (GLuint *) md->itsVertexBufferNames);
	glGenBuffers(NumVertexBuffers, (GLuint *) md->itsIndexBufferNames );

	//	Set up the individual Vertex Array Objects and Vertex Buffer Objects.
	MakeRectangleVAO(				md->itsVertexArrayNames	[VertexBufferRectangle],
									md->itsVertexBufferNames[VertexBufferRectangle],
									md->itsIndexBufferNames [VertexBufferRectangle],
//									md->itsOpenGLVersion);	RUNNING_OPENGL_3
									0);	//	temporarily pass OpenGL version number 0.0
	MakeRectangleWithEndcapsVAO(	md->itsVertexArrayNames	[VertexBufferRectangleWithEndcaps],
									md->itsVertexBufferNames[VertexBufferRectangleWithEndcaps],
									md->itsIndexBufferNames [VertexBufferRectangleWithEndcaps],
//									md->itsOpenGLVersion);	RUNNING_OPENGL_3
									0);	//	temporarily pass OpenGL version number 0.0
	MakeTriangleVAO(				md->itsVertexArrayNames	[VertexBufferTriangle],
									md->itsVertexBufferNames[VertexBufferTriangle],
									md->itsIndexBufferNames [VertexBufferTriangle],
//									md->itsOpenGLVersion);	RUNNING_OPENGL_3
									0);	//	temporarily pass OpenGL version number 0.0

	//	Did any OpenGL errors occur?
	return GetErrorString();
}

static void ShutDownVAOs(ModelData *md)
{
	unsigned int	i;

	//	The platform-dependent UI #includes StarterCode-Common.h,
	//	but shouldn't know about OpenGL.  To keep things clean,
	//	we've defined itsVertexBufferNames and itsIndexBufferNames
	//	as arrays of unsigned int's rather than arrays of GLuint's.
	//	For MacOS compatibility, though, we need to explicitly cast them
	//	to GLuint's before passing the arrays to glDeleteBuffers().

	//	Delete all Vertex Array Objects and Vertex Buffer Objects.
	//	glDeleteVertexArrays() and glDeleteBuffers() will 
	//	silently ignore zeros and invalid names.
	if (RUNNING_OPENGL_3)
		;	//	uncomment the following two lines:
//	if (md->itsOpenGLVersion >= VERSION_NUMBER(3, 0))
//		glDeleteVertexArrays(NumVertexBuffers, (GLuint *) md->itsVertexArrayNames);
	glDeleteBuffers(NumVertexBuffers, (GLuint *) md->itsVertexBufferNames);
	glDeleteBuffers(NumVertexBuffers, (GLuint *) md->itsIndexBufferNames );

	//	Clear all object names.
	for (i = 0; i < NumVertexBuffers; i++)
	{
		md->itsVertexArrayNames [i] = 0;
		md->itsVertexBufferNames[i] = 0;
		md->itsIndexBufferNames [i] = 0;
	}
}
