//	TorusGamesIntro.c
//
//	Code specific to the Intro "game".
//	This code knows about OpenGL, but does not
//	know about Windows or any other operating system.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <math.h>


//	How many copies of the sand texture to tile
//	in each row of the background?
#define SAND_TILES					2

//	When I first wrote this code, I had hoped that the collision detection
//	would provide a more realistic simulation of a 2-dimensional universe.
//	Alas the collision detection is not very good (it relies on approximating
//	each object as a circle), and it turns out that poor collision detection
//	is worse than none at all.  So I've disabled it.
//#define DETECT_COLLISIONS

//	Size of each flounder or seashell.
//	(I changed from all flounders to one flounder and three seashells.
//	The hit testing and collision detection is still a bit of a hack.)
#define FLOUNDER_SIZE			0.50
#define CRAB_SIZE				0.30
#define CLAM_SIZE				0.30
#define SCALLOP_SIZE			0.20
#define OBJECT_HIT_RADIUS		0.20	//	was (0.80 * FLOUNDER_SIZE/2)
#ifdef DETECT_COLLISIONS
#define OBJECT_COLLISION_RADIUS	0.08	//	was (0.50 * FLOUNDER_SIZE/2)
#endif


//	Public functions
static void	IntroShutDown(ModelData *md);
static void	IntroSetUpTextures(ModelData *md, ErrorText *aFirstError);
static void	IntroShutDownTextures(ModelData *md);
static void	IntroReset(ModelData *md);
static bool	IntroDragBegin(ModelData *md, bool aShiftOrControlKeyIsDown);
static void	IntroDragObject(ModelData *md, double aHandLocalDeltaH, double aHandLocalDeltaV);
static void	IntroDraw(ModelData *md);

//	Private functions
//	<no private functions>


void IntroSetUp(ModelData *md)
{
	unsigned int	i;

	//	Initialize function pointers.
	md->itsGameShutDown				= &IntroShutDown;
	md->itsGameSetUpTextures		= &IntroSetUpTextures;
	md->itsGameShutDownTextures		= &IntroShutDownTextures;
	md->itsGameReset				= &IntroReset;
	md->itsGameHandMoved			= NULL;
	md->itsGameDragBegin			= &IntroDragBegin;
	md->itsGameDragObject			= &IntroDragObject;
	md->itsGameDragEnd				= NULL;
	md->itsGameCharacterInput		= NULL;
	md->itsGameSpecialCommand		= NULL;
	md->itsGameSimulationUpdate		= NULL;
	md->itsGameDraw					= &IntroDraw;
	
	//	Initialize game-specific texture names.
	md->itsGameOf.Intro.itsTextureNameBackground = 0;
	for (i = 0; i < 4; i++)
		md->itsGameOf.Intro.itsTextureNameSprites[i] = 0;

	//	Reset the scene.
	IntroReset(md);
}


static void IntroShutDown(ModelData *md)
{
	//	The caller must ensure that the OpenGL context is set
	//	so textures may be released.
	IntroShutDownTextures(md);
}


static void IntroSetUpTextures(
	ModelData	*md,
	ErrorText	*aFirstError)
{
	SetUpOneTexture(md, &md->itsGameOf.Intro.itsTextureNameBackground,	"Intro/Sand.rgba",			TextureWrap,	TextureRGBA, aFirstError);

	SetUpOneTexture(md, &md->itsGameOf.Intro.itsTextureNameSprites[0],	"Intro/Flounder.rgba",		TextureClamp,	TextureRGBA, aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Intro.itsTextureNameSprites[1],	"Intro/MittenCrab.rgba",	TextureClamp,	TextureRGBA, aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Intro.itsTextureNameSprites[2],	"Intro/FlutedClam.rgba",	TextureClamp,	TextureRGBA, aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Intro.itsTextureNameSprites[3],	"Intro/CalicoScallop.rgba",	TextureClamp,	TextureRGBA, aFirstError);
}


static void IntroShutDownTextures(ModelData *md)
{
	unsigned int	i;

	ShutDownOneTexture(&md->itsGameOf.Intro.itsTextureNameBackground);
	
	for (i = 0; i < 4; i++)
		ShutDownOneTexture(&md->itsGameOf.Intro.itsTextureNameSprites[i]);
}



static void IntroReset(ModelData *md)
{
	//	Place the flounder in the center.

	md->itsGameOf.Intro.itsSpritePlacement[0].itsH		= 0.0;
	md->itsGameOf.Intro.itsSpritePlacement[0].itsV		= 0.0;
	md->itsGameOf.Intro.itsSpritePlacement[0].itsFlip	= false;
	md->itsGameOf.Intro.itsSpritePlacement[0].itsAngle	= 0.0;
	md->itsGameOf.Intro.itsSpritePlacement[0].itsSizeH	= FLOUNDER_SIZE;
	md->itsGameOf.Intro.itsSpritePlacement[0].itsSizeV	= FLOUNDER_SIZE;

	//	Place the crab and the shells arbitrarily, but avoiding the flounder.

	md->itsGameOf.Intro.itsSpritePlacement[1].itsH		= -0.45;
	md->itsGameOf.Intro.itsSpritePlacement[1].itsV		= -0.30;
	md->itsGameOf.Intro.itsSpritePlacement[1].itsFlip	= (rand() % 2 != 0);
	md->itsGameOf.Intro.itsSpritePlacement[1].itsAngle	= TWOPI * ((double) rand() / (double) RAND_MAX);
	md->itsGameOf.Intro.itsSpritePlacement[1].itsSizeH	= CRAB_SIZE;
	md->itsGameOf.Intro.itsSpritePlacement[1].itsSizeV	= CRAB_SIZE;

	md->itsGameOf.Intro.itsSpritePlacement[2].itsH		= -0.3;
	md->itsGameOf.Intro.itsSpritePlacement[2].itsV		= +0.4;
	md->itsGameOf.Intro.itsSpritePlacement[2].itsFlip	= (rand() % 2 != 0);
	md->itsGameOf.Intro.itsSpritePlacement[2].itsAngle	= TWOPI * ((double) rand() / (double) RAND_MAX);
	md->itsGameOf.Intro.itsSpritePlacement[2].itsSizeH	= CLAM_SIZE;
	md->itsGameOf.Intro.itsSpritePlacement[2].itsSizeV	= CLAM_SIZE;

	md->itsGameOf.Intro.itsSpritePlacement[3].itsH		= +0.3;
	md->itsGameOf.Intro.itsSpritePlacement[3].itsV		= +0.3;
	md->itsGameOf.Intro.itsSpritePlacement[3].itsFlip	= (rand() % 2 != 0);
	md->itsGameOf.Intro.itsSpritePlacement[3].itsAngle	= TWOPI * ((double) rand() / (double) RAND_MAX);
	md->itsGameOf.Intro.itsSpritePlacement[3].itsSizeH	= SCALLOP_SIZE;
	md->itsGameOf.Intro.itsSpritePlacement[3].itsSizeV	= SCALLOP_SIZE;

	//	Abort any pending simulation.
	SimulationEnd(md);			//	unused in Intro game

	//	Just for good form...
	md->itsGameOf.Intro.itsDragSprite	= 0;		//	ignored until drag begins
	md->itsGameIsOver					= false;	//	unused in Intro game
}


static bool IntroDragBegin(
	ModelData	*md,
	bool		aShiftOrControlKeyIsDown)
{
	unsigned int	i,
					j,
					k;

	UNUSED_PARAMETER(aShiftOrControlKeyIsDown);

	for (i = 0; i < 4; i++)
	{
		md->itsDragIsReversed = (	md->itsHandPlacement.itsFlip
								 != md->itsGameOf.Intro.itsSpritePlacement[i].itsFlip);

		if (ShortestDistance(
					md->itsHandPlacement.itsH,
					md->itsHandPlacement.itsV,
					md->itsGameOf.Intro.itsSpritePlacement[i].itsH,
					md->itsGameOf.Intro.itsSpritePlacement[i].itsV,
					md->itsTopology,
					&md->itsDragIsReversed)
				< OBJECT_HIT_RADIUS)
		{
			//	Record which sprite was hit.
			md->itsGameOf.Intro.itsDragSprite = i;
			
			//	Clear the history.
			for (j = 0; j < DRAG_HISTORY_DEPTH; j++)
				for (k = 0; k < 2; k++)
					md->itsDragHistory[j][k] = 0.0;

			//	Report a hit.
			return true;
		}
	}

	return false;
}


static void IntroDragObject(
	ModelData	*md,
	double		aHandLocalDeltaH,
	double		aHandLocalDeltaV)
{
	double			theFlounderLocalDeltaH,
					theFlounderLocalDeltaV,
					theHandAmbientDeltaH,
					theHandAmbientDeltaV,
					theFlounderAmbientDeltaH,
					theFlounderAmbientDeltaV,
					theNewPositionH,
					theNewPositionV;
#ifdef DETECT_COLLISIONS
	unsigned int	theSpriteIndex;
#endif
	double			theTotalFlounderLocalDeltaH,
					theTotalFlounderLocalDeltaV;
	unsigned int	i,
					j;

	if (md->itsGameOf.Intro.itsDragSprite >= 4)	//	should never occur
		return;

	theFlounderLocalDeltaH		= md->itsDragIsReversed
								? -aHandLocalDeltaH : aHandLocalDeltaH;
	theFlounderLocalDeltaV		= aHandLocalDeltaV;

	theHandAmbientDeltaH		= md->itsHandPlacement.itsFlip
								? -aHandLocalDeltaH : aHandLocalDeltaH;
	theHandAmbientDeltaV		= aHandLocalDeltaV;

	theFlounderAmbientDeltaH	= md->itsGameOf.Intro.itsSpritePlacement[md->itsGameOf.Intro.itsDragSprite].itsFlip
								? -theFlounderLocalDeltaH : theFlounderLocalDeltaH;
	theFlounderAmbientDeltaV	= theFlounderLocalDeltaV;

	theNewPositionH = md->itsGameOf.Intro.itsSpritePlacement[md->itsGameOf.Intro.itsDragSprite].itsH + theFlounderAmbientDeltaH;
	theNewPositionV = md->itsGameOf.Intro.itsSpritePlacement[md->itsGameOf.Intro.itsDragSprite].itsV + theFlounderAmbientDeltaV;

#ifdef DETECT_COLLISIONS
	for (theSpriteIndex = 0; theSpriteIndex < 4; theSpriteIndex++)
	{
		if (theSpriteIndex == md->itsGameOf.Intro.itsDragSprite)
			continue;

		if (ShortestDistance(
				md->itsGameOf.Intro.itsSpritePlacement[theSpriteIndex].itsH,
				md->itsGameOf.Intro.itsSpritePlacement[theSpriteIndex].itsV,
				theNewPositionH,
				theNewPositionV,
				md->itsTopology,
				NULL)
			< 2.0 * OBJECT_COLLISION_RADIUS)

			return;	//	Collision!
					//	Return without doing the drag.
	}
#endif

	//	No collision.

	//	Move the flounder.
	md->itsGameOf.Intro.itsSpritePlacement[md->itsGameOf.Intro.itsDragSprite].itsH = theNewPositionH;
	md->itsGameOf.Intro.itsSpritePlacement[md->itsGameOf.Intro.itsDragSprite].itsV = theNewPositionV;
	NormalizePlacement(&md->itsGameOf.Intro.itsSpritePlacement[md->itsGameOf.Intro.itsDragSprite], md->itsTopology);
	//	Note:  The flounder's angle will get set below.

	//	Move the hand.
	md->itsHandPlacement.itsH += theHandAmbientDeltaH;
	md->itsHandPlacement.itsV += theHandAmbientDeltaV;
	NormalizePlacement(&md->itsHandPlacement, md->itsTopology);

	//	Only the flounder and the crab change direction when dragged.
	if (md->itsGameOf.Intro.itsDragSprite == 0	//	flounder
	 || md->itsGameOf.Intro.itsDragSprite == 1)	//	crab
	{
		//	Update the drag history.
		for (i = DRAG_HISTORY_DEPTH - 1; i-- > 0; )
			for (j = 0; j < 2; j++)
				md->itsDragHistory[i+1][j] = md->itsDragHistory[i][j];
		md->itsDragHistory[0][0] = theFlounderLocalDeltaH;
		md->itsDragHistory[0][1] = theFlounderLocalDeltaV;

		//	Adjust the flounder's heading.
		//	To avoid processor-speed dependencies, sum the flounder's local motions
		//	over a fixed distance rather than a fixed number of frames.
		theTotalFlounderLocalDeltaH = 0.0;
		theTotalFlounderLocalDeltaV = 0.0;
		for (i = 0; i < DRAG_HISTORY_DEPTH; i++)
		{
			theTotalFlounderLocalDeltaH += md->itsDragHistory[i][0];
			theTotalFlounderLocalDeltaV += md->itsDragHistory[i][1];

			if (fabs(theTotalFlounderLocalDeltaH) > 0.1
			 || fabs(theTotalFlounderLocalDeltaV) > 0.1)
				break;
		}
		if (theTotalFlounderLocalDeltaH != 0.0
		 || theTotalFlounderLocalDeltaV != 0.0)
			md->itsGameOf.Intro.itsSpritePlacement[md->itsGameOf.Intro.itsDragSprite].itsAngle = atan2(theTotalFlounderLocalDeltaV, theTotalFlounderLocalDeltaH);
	}
}


static void IntroDraw(ModelData *md)
{
	unsigned int	i;

	//	Draw the background.
	//
	//	Technical Note:  Just by dumb luck the sand texture
	//	I'm using matches up well enough with itself even
	//	in Klein bottle mode.  If I ever switch to a less friendly
	//	texture, I'll have to tile it by reflection.
	//	Unfortunately I don't think OpenGL can do that automatically.
	DrawFundamentalDomain(md, md->itsGameOf.Intro.itsTextureNameBackground, SAND_TILES);

	//	Draw glide reflection axes if appropriate.
	DrawKleinAxes(	md,
					PREMULTIPLY_RGBA(1.0, 0.0, 0.0, 1.0),
					PREMULTIPLY_RGBA(0.0, 0.0, 1.0, 1.0) );

	//	Draw the shells, then the crab, then finally the flounder.
	for (i = 4; i-- > 0; )
		DrawSprite(	md,
					md->itsGameOf.Intro.itsTextureNameSprites[i],
					&md->itsGameOf.Intro.itsSpritePlacement[i]);
}
