//	TorusGamesLocalization.c
//
//	Manages a dictionary for translating all 
//	user-visible phrases to the current (human) language.
//
//	If necessary the dictionary could be implemented 
//	as a binary tree, but in practice the dictionary 
//	will be small (~100 entries) and lookups will occur 
//	only rarely (for example, when changing the active language), 
//	so a simple linear list should be fine.  
//	More precisely, we'll use a NULL-terminated singly-linked list.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include <stdio.h>	//	for sprintf()

#define MAX_KEY_VALUE_LENGTH	2048	//	includes terminating zero

//	TorusGames-Common.h provided only an opaque typedef
//	for a LocalizedPhrase.  Here is its concrete realization.
struct LocalizedPhrase
{
	char					*itsKey,		//	zero-terminated UTF-8 string
							*itsValue;		//	zero-terminated UTF-8 string
	struct LocalizedPhrase	*itsNextPhrase;	//	next entry in the NULL-terminated singly-linked list
};

//	We'll parse dictionary files with the following format:
//
//		//	This is a comment.
//		"sample key #1" = "sample value #1"
//		"sample key #2" = "sample value #2"
//
//	Within a quoted string, we'll support the escape sequences
//	\", \t, \r and \n.  Otherwise \<character> is is interpreted
//	as <character>, including the special case \\ which gives
//	a single \ character.  Tabs, return and newlines may also
//	be entered directly.
typedef enum
{
	ParseExpectingKey,
	ParseExpectingKeyWithCommentBeginning,	//	one '/' received
	ParseExpectingKeyWithCommentBegun,		//	second '/' received
	ParseInKey,
	ParseInKeyWithEscapePending,	//	Lets parser read "key with\"internal quotes\"" correctly.
	ParseExpectingEqualSign,
	ParseExpectingEqualSignWithCommentBeginning,	//	one '/' received
	ParseExpectingEqualSignWithCommentBegun,		//	second '/' received
	ParseExpectingValue,
	ParseExpectingValueWithCommentBeginning,	//	one '/' received
	ParseExpectingValueWithCommentBegun,		//	second '/' received
	ParseInValue,
	ParseInValueWithEscapePending	//	Lets parser read "value with\"internal quotes\"" correctly.
} ParseStatus;


static void			SetUpLocalizedDictionary(const char aTwoLetterLanguageCode[3]);
static void			ShutDownLocalizedDictionary(void);
static inline bool	SameString(const char *aStringA, const char *aStringB);


void RefreshLocalization(
	const char	aTwoLetterLanguageCode[3])	//	for example "en" or "ja"
											//		or pass "--" to clear the language
											//		and release gDictionary's memory
{
	gLanguageCode[0] = aTwoLetterLanguageCode[0];
	gLanguageCode[1] = aTwoLetterLanguageCode[1];

	if (gDictionary != NULL)
		ShutDownLocalizedDictionary();
	
	if (aTwoLetterLanguageCode[0] != '-'
	 && aTwoLetterLanguageCode[1] != '-')
		SetUpLocalizedDictionary(aTwoLetterLanguageCode);
}


static void SetUpLocalizedDictionary(
	const char aTwoLetterLanguageCode[3])
{
	ErrorText		theErrorMessage		= NULL;
	char			theFileName[32],
					theFilePath[2048];
	Byte			*theFileContents	= NULL;
	unsigned int	theFileSize			= 0,
					theFileStart		= 0;
	ParseStatus		theParseStatus;
	char			theKey[MAX_KEY_VALUE_LENGTH]	= "",	//	zero-terminated UTF-8 string
					theValue[MAX_KEY_VALUE_LENGTH]	= "",	//	zero-terminated UTF-8 string
					theChar				= 0;	//	to cleanly convert a Byte to a char
	unsigned int	theKeyLength		= 0,	//	number of bytes (including terminating zero), not number of Unicode characters
					theValueLength		= 0;	//	number of bytes (including terminating zero), not number of Unicode characters
	LocalizedPhrase	*theNewPhrase		= NULL;
	unsigned int	i,
					j;

	sprintf(theFileName, "TorusGames-%c%c.txt", aTwoLetterLanguageCode[0], aTwoLetterLanguageCode[1]);

	if ((theErrorMessage = GetAbsolutePath("Languages", theFileName, theFilePath, BUFFER_LENGTH(theFilePath))) != NULL)
		goto CleanUpSetUpLocalizedDictionary;

	if ((theErrorMessage = GetFileContents(theFilePath, &theFileSize, &theFileContents)) != NULL)
		goto CleanUpSetUpLocalizedDictionary;
	
	//	We don't expect a byte-order mark (BOM) but if the end user
	//	edits the language file and saves it with a BOM, we want
	//	to handle it gracefully.
	if (theFileSize >= 3
	 && theFileContents[0] == 0xEF	//	UTF-8 encoding of BOM = 0xFEFF
	 && theFileContents[1] == 0xBB
	 && theFileContents[2] == 0xBF)
	{
		theFileStart = 3;
	}
	else
		theFileStart = 0;

	//	theFileContents contains UTF-8 text.
	//	Because the characters defining the dictionary's
	//	structure are 7-bit ASCII characters (namely 
	//	the double quotation marks " and the equal sign = )
	//	we may parse the file one byte at a time,
	//	in effect ignoring the fact that keys and values
	//	may contain multibyte UTF-8 sequences.
	theParseStatus = ParseExpectingKey;
	for (i = theFileStart; i < theFileSize; i++)
	{
		theChar = (char) theFileContents[i];

		switch (theParseStatus)
		{
			case ParseExpectingKey:
				switch (theChar)
				{
					case '"':
						theParseStatus	= ParseInKey;
						theKeyLength	= 0;
						break;

					case '/':
						theParseStatus = ParseExpectingKeyWithCommentBeginning;
						break;

					case '=':
						theErrorMessage = "Encountered '=' while expecting a key.";
						goto CleanUpSetUpLocalizedDictionary;

					case '\n':
					case '\r':
					case ' ':
					case '\t':
						//	White space is legal.  Ignore it.
						break;

					case ';':
						theErrorMessage = "Encountered an unnecessary semicolon ';' .  This program's dictionary format does not require them (unlike Macintosh .strings files).";
						goto CleanUpSetUpLocalizedDictionary;

					default:
						theErrorMessage = "Encountered extraneous non-commented character while expecting a key.";
						goto CleanUpSetUpLocalizedDictionary;
				}
				break;

			case ParseExpectingKeyWithCommentBeginning:
				switch (theChar)
				{
					case '/':
						theParseStatus = ParseExpectingKeyWithCommentBegun;
						break;

					default:
						theErrorMessage = "Found first '/' beginning a comment, but not second '/'.";
						goto CleanUpSetUpLocalizedDictionary;
				}
				break;

			case ParseExpectingKeyWithCommentBegun:
				switch (theChar)
				{
					case '\n':
					case '\r':
						theParseStatus = ParseExpectingKey;
						break;

					default:
						//	Ignore contents of comment.
						break;
				}
				break;

			case ParseInKey:
				switch (theChar)
				{
					case '\\':	//	single backslash character
						theParseStatus = ParseInKeyWithEscapePending;
						break;

					case '"':
						//	Set the new ParseStatus.
						theParseStatus = ParseExpectingEqualSign;
						
						//	Prepare the terminating zero.
						theChar = 0;

						//	Fall through to the generic case
						//	to add the terminating zero to theKey.
					default:
						//	Add theChar (or terminating zero from fall-through case)
						//	if there's room for it.
						if (theKeyLength < MAX_KEY_VALUE_LENGTH)
							theKey[theKeyLength++] = theChar;
						else
						{
							theErrorMessage = "One of the dictionary's keys is too long.";
							goto CleanUpSetUpLocalizedDictionary;
						}
						break;
				}
				break;

			case ParseInKeyWithEscapePending:

				switch (theChar)
				{
					//	Interpret the escape sequences \t, \n and \r.
					case 't':	theChar = '\t';	break;
					case 'n':	theChar = '\n';	break;
					case 'r':	theChar = '\r';	break;
					
					//	Leave all other characters unmodified.
					//	In particular, \" gives a " and
					//	\\ gives a \.
					default:	break;
				}
					
				//	Add theChar if there's room for it.
				if (theKeyLength < MAX_KEY_VALUE_LENGTH)
					theKey[theKeyLength++] = theChar;
				else
				{
					theErrorMessage = "One of the dictionary's keys is too long.";
					goto CleanUpSetUpLocalizedDictionary;
				}
				
				//	Clear the pending escape.
				theParseStatus = ParseInKey;

				break;

			case ParseExpectingEqualSign:
				switch (theChar)
				{
					case '/':
						theParseStatus = ParseExpectingEqualSignWithCommentBeginning;
						break;

					case '=':
						theParseStatus = ParseExpectingValue;
						break;

					case '\n':
					case '\r':
					case ' ':
					case '\t':
						//	White space is legal.  Ignore it.
						break;

					default:
						theErrorMessage = "Encountered extraneous non-commented character while expecting an equals sign '=' .";
						goto CleanUpSetUpLocalizedDictionary;
				}
				break;

			case ParseExpectingEqualSignWithCommentBeginning:
				switch (theChar)
				{
					case '/':
						theParseStatus = ParseExpectingEqualSignWithCommentBegun;
						break;

					default:
						theErrorMessage = "Found first '/' beginning a comment, but not second '/'.";
						goto CleanUpSetUpLocalizedDictionary;
				}
				break;

			case ParseExpectingEqualSignWithCommentBegun:
				switch (theChar)
				{
					case '\n':
					case '\r':
						theParseStatus = ParseExpectingEqualSign;
						break;

					default:
						//	Ignore contents of comment.
						break;
				}
				break;

			case ParseExpectingValue:
				switch (theChar)
				{
					case '"':
						theParseStatus	= ParseInValue;
						theValueLength	= 0;
						break;

					case '/':
						theParseStatus = ParseExpectingValueWithCommentBeginning;
						break;

					case '=':
						theErrorMessage = "Encountered '=' while expecting a value.";
						goto CleanUpSetUpLocalizedDictionary;

					case '\n':
					case '\r':
					case ' ':
					case '\t':
						//	White space is legal.  Ignore it.
						break;

					default:
						theErrorMessage = "Encountered extraneous non-commented character while expecting a value.";
						goto CleanUpSetUpLocalizedDictionary;
				}
				break;

			case ParseExpectingValueWithCommentBeginning:
				switch (theChar)
				{
					case '/':
						theParseStatus = ParseExpectingValueWithCommentBegun;
						break;

					default:
						theErrorMessage = "Found first '/' beginning a comment, but not second '/'.";
						goto CleanUpSetUpLocalizedDictionary;
				}
				break;

			case ParseExpectingValueWithCommentBegun:
				switch (theChar)
				{
					case '\n':
					case '\r':
						theParseStatus = ParseExpectingValue;
						break;

					default:
						//	Ignore contents of comment.
						break;
				}
				break;

			case ParseInValue:
				switch (theChar)
				{
					case '\\':	//	single backslash character
						theParseStatus = ParseInValueWithEscapePending;
						break;

					case '"':
						//	Set the new ParseStatus.
						theParseStatus = ParseExpectingKey;
						
						//	Prepare the terminating zero.
						theChar = 0;

						//	Fall through to the generic case
						//	to add the terminating zero to theValue.
					default:
						//	Add theChar (or terminating zero from fall-through case)
						//	if there's room for it.
						if (theValueLength < MAX_KEY_VALUE_LENGTH)
							theValue[theValueLength++] = theChar;
						else
						{
							theErrorMessage = "One of the dictionary's values is too long.";
							goto CleanUpSetUpLocalizedDictionary;
						}
						
						//	If we just received the terminating '"',
						//	then both theKey and theValue are complete.
						//	Create a new dictionary entry.
						if (theParseStatus == ParseExpectingKey)
						{
							//	Allocate theNewPhrase itself.
							theNewPhrase = (LocalizedPhrase *) GET_MEMORY(sizeof(LocalizedPhrase));
							if (theNewPhrase == NULL)
							{
								theErrorMessage = "Memory allocation failed (theNewPhrase).";
								goto CleanUpSetUpLocalizedDictionary;
							}

							//	Set pointers to NULL for robust error handling.
							theNewPhrase->itsKey		= NULL;
							theNewPhrase->itsValue		= NULL;
							theNewPhrase->itsNextPhrase	= NULL;

							//	Allocate and initialize itsKey.
							theNewPhrase->itsKey	= (char *) GET_MEMORY( theKeyLength  * sizeof(char));
							if (theNewPhrase->itsKey == NULL)
							{
								theErrorMessage = "Memory allocation failed (theNewPhrase->itsKey).";
								goto CleanUpSetUpLocalizedDictionary;
							}
							for (j = 0; j < theKeyLength; j++)	//	includes terminating zero
								theNewPhrase->itsKey[j] = theKey[j];

							//	Allocate and initialize itsValue.
							theNewPhrase->itsValue	= (char *) GET_MEMORY(theValueLength * sizeof(char));
							if (theNewPhrase->itsValue == NULL)
							{
								theErrorMessage = "Memory allocation failed (theNewPhrase->itsValue).";
								goto CleanUpSetUpLocalizedDictionary;
							}
							for (j = 0; j < theValueLength; j++)	//	includes terminating zero
								theNewPhrase->itsValue[j] = theValue[j];

							//	theNewPhrase is complete.
							//	Install it into gDictionary.
							theNewPhrase->itsNextPhrase	= gDictionary;
							gDictionary					= theNewPhrase;
							theNewPhrase				= NULL;
						}
		
						break;
				}
				break;

			case ParseInValueWithEscapePending:

				switch (theChar)
				{
					//	Interpret the escape sequences \t, \n and \r.
					case 't':	theChar = '\t';	break;
					case 'n':	theChar = '\n';	break;
					case 'r':	theChar = '\r';	break;
					
					//	Leave all other characters unmodified.
					//	In particular, \" gives a " and
					//	\\ gives a \.
					default:	break;
				}
					
				//	Add theChar if there's room for it.
				if (theValueLength < MAX_KEY_VALUE_LENGTH)
					theValue[theValueLength++] = theChar;
				else
				{
					theErrorMessage = "One of the dictionary's values is too long.";
					goto CleanUpSetUpLocalizedDictionary;
				}
				
				//	Clear the pending escape.
				theParseStatus = ParseInValue;

				break;
			
			default:
				theErrorMessage = "Internal Error:  theParseStatus has an illegal value.";
				goto CleanUpSetUpLocalizedDictionary;
		}
	}
	
	//	Did the file end at a reasonable place?
	if (theParseStatus != ParseExpectingKey
	 && theParseStatus != ParseExpectingKeyWithCommentBegun)
	{
		theErrorMessage = "Syntax error:  unexpected end-of-file.";
		goto CleanUpSetUpLocalizedDictionary;
	}

CleanUpSetUpLocalizedDictionary:
	
	if (theNewPhrase != NULL)
	{
		FREE_MEMORY_SAFELY(theNewPhrase->itsKey);
		FREE_MEMORY_SAFELY(theNewPhrase->itsValue);
		
		FREE_MEMORY_SAFELY(theNewPhrase);
	}

	FreeFileContents(&theFileSize, &theFileContents);
	
	if (theErrorMessage != NULL)
		ErrorMessage(theErrorMessage, "Error in SetUpLocalizedDictionary()", false);
}


static void ShutDownLocalizedDictionary(void)
{
	LocalizedPhrase	*theDeadPhrase;

	while (gDictionary != NULL)
	{
		theDeadPhrase	= gDictionary;
		gDictionary		= gDictionary->itsNextPhrase;

		FREE_MEMORY_SAFELY(theDeadPhrase->itsKey);
		FREE_MEMORY_SAFELY(theDeadPhrase->itsValue);
		
		FREE_MEMORY_SAFELY(theDeadPhrase);
	}
}


//	Caution:  GetLocalizedText() returns a pointer
//	to one of the values in gDictionary.
//	The caller should finish using that pointer
//	and discard it before calling RefreshLocalization(),
//	because ShutDownLocalizedDictionary() will invalidate
//	all such pointers.
const char *GetLocalizedText(
	const char *aKey)
{
	LocalizedPhrase	*thePhrase;

	if (aKey == NULL)
		return "<missing key>";

	for (	thePhrase = gDictionary;
			thePhrase != NULL;
			thePhrase = thePhrase->itsNextPhrase)
	{
		if (SameString(aKey, thePhrase->itsKey))
			return thePhrase->itsValue;
	}
	
	return "<text not found>";
}


static inline bool SameString(
	const char	*aStringA,
	const char	*aStringB)
{
	while (true)
	{
		if (*aStringA != *aStringB)
			return false;
		
		if (*aStringA == 0)
			return true;
		
		aStringA++;
		aStringB++;
	}
}
