//	TorusGamesMaze.c
//
//	Code specific to the Maze game.
//	This code knows about OpenGL, but does not
//	know about Windows or any other operating system.
//
//	Maze coordinate conventions
//
//	Integer coordinates (h,v) label the cells, for h,v ∈ {0, 1, … , n-1}.
//	The first coordinate runs left-to-right, while the second runs bottom-to-top.
//	Cell centers align with the natural n×n grid, so for example
//	cells straddle the edges of the fundamental domain
//	and a single cell sits centered at the fundamental domain's corners.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <math.h>


//	Compensate for the rotation in the raw graphics (value in radians, counterclockwise).
#define MOUSE_ROTATION	0.3


typedef struct
{
	unsigned int	itsH,
					itsV;
	MazeWallIndex	itsSide;
} WallSpec;

typedef struct
{
	unsigned int	itsH,
					itsV;
} CellSpec;

typedef struct
{
	CellSpec		itsSelf,
					itsParent;
} MazeTreeNode;


//	Public functions
static void			MazeShutDown(ModelData *md);
static void			MazeSetUpTextures(ModelData *md, ErrorText *aFirstError);
static void			MazeShutDownTextures(ModelData *md);
static void			MazeReset(ModelData *md);
static bool			MazeDragBegin(ModelData *md, bool aShiftOrControlKeyIsDown);
static void			MazeDragObject(ModelData *md, double aHandLocalDeltaH, double aHandLocalDeltaV);
static void			MazeDragEnd(ModelData *md);
static void			MazeSimulationUpdate(ModelData *md);
static void			MazeDraw(ModelData *md);

//	Private functions
static bool			AllocMazeCells(ModelData *md);
static void			FreeMazeCells(ModelData *md);
static bool			CreateNewMaze(ModelData *md);
static void			SetAllWallsTrue(ModelData *md);
static void			InitCellIndices(ModelData *md);
static void			InitWallList(ModelData *md, WallSpec *aWallList);
static void			RandomizeWallList(ModelData *md, WallSpec *aWallList);
static void			RemoveUnneededWalls(ModelData *md, WallSpec *aWallList);
static WallSpec		GetOtherSide(WallSpec aNearSide, unsigned int aMazeSize, TopologyType aTopology);
static ErrorText	DrawMazeAsTexture(ModelData *md);
static void			PositionMouseAndCheese(ModelData *md);
static void			FindFurthestPoint(ModelData *md, CellSpec aNearCell, CellSpec *aFarCell);


void MazeSetUp(ModelData *md)
{
	//	Initialize function pointers.
	md->itsGameShutDown			= &MazeShutDown;
	md->itsGameSetUpTextures	= &MazeSetUpTextures;
	md->itsGameShutDownTextures	= &MazeShutDownTextures;
	md->itsGameReset			= &MazeReset;
	md->itsGameHandMoved		= NULL;
	md->itsGameDragBegin		= &MazeDragBegin;
	md->itsGameDragObject		= &MazeDragObject;
	md->itsGameDragEnd			= &MazeDragEnd;
	md->itsGameCharacterInput	= NULL;
	md->itsGameSpecialCommand	= NULL;
	md->itsGameSimulationUpdate	= &MazeSimulationUpdate;
	md->itsGameDraw				= &MazeDraw;

	//	Initialize variables.
	md->itsGameOf.Maze.itsSize			= 0;
	md->itsGameOf.Maze.itsMazeIsValid	= false;
	md->itsGameOf.Maze.itsCells			= NULL;
	
	//	Initialize game-specific texture names.
	md->itsGameOf.Maze.itsTextureNameGrass	= 0;
	md->itsGameOf.Maze.itsTextureNameMouse	= 0;
	md->itsGameOf.Maze.itsTextureNameCheese	= 0;
	md->itsGameOf.Maze.itsTextureNameMaze	= 0;

	//	Create a maze.
	MazeReset(md);
}


static void MazeShutDown(ModelData *md)
{
	//	The caller must ensure that the OpenGL context is set
	//	so textures may be released.
	MazeShutDownTextures(md);

	//	Free the maze cells.
	FreeMazeCells(md);
}


static void MazeSetUpTextures(
	ModelData	*md,
	ErrorText	*aFirstError)
{
	SetUpOneTexture(md, &md->itsGameOf.Maze.itsTextureNameGrass,	"Maze/Grass.rgba",	TextureWrap,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Maze.itsTextureNameMouse,	"Maze/Mouse.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Maze.itsTextureNameCheese,	"Maze/Cheese.rgba",	TextureClamp,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.Maze.itsTextureNameMaze,		NULL,				TextureClamp,	TextureAlpha,	aFirstError);

	//	If a valid maze is present draw it into the maze texture.
	//	Otherwise push onward with the default empty texture.
	if (md->itsGameOf.Maze.itsMazeIsValid)
		(void) DrawMazeAsTexture(md);
}


static void MazeShutDownTextures(ModelData *md)
{
	ShutDownOneTexture(&md->itsGameOf.Maze.itsTextureNameGrass	);
	ShutDownOneTexture(&md->itsGameOf.Maze.itsTextureNameMouse	);
	ShutDownOneTexture(&md->itsGameOf.Maze.itsTextureNameCheese	);
	ShutDownOneTexture(&md->itsGameOf.Maze.itsTextureNameMaze	);
}


static void MazeReset(ModelData *md)
{
	ErrorText	theErrorMessage	= NULL;

	FreeMazeCells(md);

	switch (md->itsDifficultyLevel)
	{
		case 0:  md->itsGameOf.Maze.itsSize =  2;  break;
		case 1:  md->itsGameOf.Maze.itsSize =  4;  break;
		case 2:  md->itsGameOf.Maze.itsSize =  8;  break;
		case 3:  md->itsGameOf.Maze.itsSize = 16;  break;
		case 4:  md->itsGameOf.Maze.itsSize = 32;  break;
		default: md->itsGameOf.Maze.itsSize =  2;  break;	//	should never occur
	}

	if ( ! AllocMazeCells(md) )
	{
		theErrorMessage = "Couldn't allocate memory for maze walls.";
		goto CleanUpMazeReset;
	}

	if ( ! CreateNewMaze(md) )
	{
		theErrorMessage = "Couldn't allocate memory to create new maze.";
		goto CleanUpMazeReset;
	}

	//	Place the mouse and the cheese as far from each other as possible.
	PositionMouseAndCheese(md);
	
	//	Request that the new maze get drawn into the texture.
	md->itsPreparedTextures = false;

	//	Abort any pending simulation.
	SimulationEnd(md);

	//	Ready to go.
	md->itsGameIsOver = false;

CleanUpMazeReset:

	if (theErrorMessage != NULL)
	{
		md->itsGameOf.Maze.itsMazeIsValid = false;
		ErrorMessage(theErrorMessage, "MazeReset() Error", false);
	}
	else
		md->itsGameOf.Maze.itsMazeIsValid = true;
}


static bool AllocMazeCells(ModelData *md)
{
	unsigned int	i;

	md->itsGameOf.Maze.itsCells = (MazeCell **) GET_MEMORY(md->itsGameOf.Maze.itsSize * sizeof(MazeCell *));
	if (md->itsGameOf.Maze.itsCells == NULL)
		return false;

	for (i = 0; i < md->itsGameOf.Maze.itsSize; i++)
	{
		md->itsGameOf.Maze.itsCells[i] = (MazeCell *) GET_MEMORY(md->itsGameOf.Maze.itsSize * sizeof(MazeCell));
		if (md->itsGameOf.Maze.itsCells[i] == NULL)
		{
			while (i-- > 0)
				FREE_MEMORY(md->itsGameOf.Maze.itsCells[i]);
			FREE_MEMORY(md->itsGameOf.Maze.itsCells);
			md->itsGameOf.Maze.itsCells = NULL;
			return false;
		}
	}

	return true;
}


static void FreeMazeCells(ModelData *md)
{
	unsigned int	i;

	if (md->itsGameOf.Maze.itsCells != NULL)
	{
		for (i = 0; i < md->itsGameOf.Maze.itsSize; i++)
			FREE_MEMORY(md->itsGameOf.Maze.itsCells[i]);

		FREE_MEMORY(md->itsGameOf.Maze.itsCells);

		md->itsGameOf.Maze.itsCells = NULL;
	}
}


static bool CreateNewMaze(ModelData *md)
{
	//	This algorithm is from the Maze program written
	//	by Jeff Weeks for Adam Weeks Marano, Christmas 1992.

	//	The plan is to give each cell its own index.  We then put
	//	all the possible Walls (WallWest and WallSouth for each cell)
	//	on a list, randomize the order of the list, and then go down it
	//	one item at a time.  Whenever a wall separates two cells
	//	with different indices, erase it and merge the neighboring indices
	//	(that is, set all occurrences of one index to equal the value
	//	of the other index).  When a wall separates two cells
	//	of the same index, leave it there.  It's easy to see that
	//	this will yield a maze with a unique path between any two points.

	//	We assume memory for md->itsGameOf.Maze.itsCells has already been allocated.

	WallSpec	*theWallList	= NULL;

	SetAllWallsTrue(md);
	InitCellIndices(md);
	theWallList = (WallSpec *) GET_MEMORY(2 * md->itsGameOf.Maze.itsSize * md->itsGameOf.Maze.itsSize * sizeof(WallSpec));
	if (theWallList == NULL)
		return false;
	InitWallList(md, theWallList);
	RandomizeWallList(md, theWallList);
	RemoveUnneededWalls(md, theWallList);
	FREE_MEMORY_SAFELY(theWallList);

	return true;
}


static void SetAllWallsTrue(ModelData *md)
{
	unsigned int	i,
					j,
					k;

	for (i = 0; i < md->itsGameOf.Maze.itsSize; i++)
		for (j = 0; j < md->itsGameOf.Maze.itsSize; j++)
			for (k = 0; k < 4; k++)
				md->itsGameOf.Maze.itsCells[i][j].itsWalls[k] = true;
}


static void InitCellIndices(ModelData *md)
{
	unsigned int	theCount,
					i,
					j;

	theCount = 0;

	for (i = 0; i < md->itsGameOf.Maze.itsSize; i++)
		for (j = 0; j < md->itsGameOf.Maze.itsSize; j++)
			md->itsGameOf.Maze.itsCells[i][j].itsIndex = theCount++;
}


static void InitWallList(
	ModelData	*md,
	WallSpec	*aWallList)
{
	WallSpec		theWallSpec;
	unsigned int	theCount;

	theCount = 0;

	for (theWallSpec.itsH = 0; theWallSpec.itsH < md->itsGameOf.Maze.itsSize; theWallSpec.itsH++)
		for (theWallSpec.itsV = 0; theWallSpec.itsV < md->itsGameOf.Maze.itsSize; theWallSpec.itsV++)
			for (theWallSpec.itsSide = 0; theWallSpec.itsSide < 2; theWallSpec.itsSide++)	//	WallWest and WallSouth
				aWallList[theCount++] = theWallSpec;
}


static void RandomizeWallList(
	ModelData	*md,
	WallSpec	*aWallList)
{
	unsigned int	theLastIndex,
					theRandomIndex;
	WallSpec		theTempWallSpec;

	theLastIndex = 2 * md->itsGameOf.Maze.itsSize * md->itsGameOf.Maze.itsSize;

	while (--theLastIndex > 0)
	{
		theRandomIndex = rand() % (theLastIndex + 1);

		theTempWallSpec				= aWallList[theLastIndex];
		aWallList[theLastIndex]		= aWallList[theRandomIndex];
		aWallList[theRandomIndex]	= theTempWallSpec;
	}
}


static void RemoveUnneededWalls(
	ModelData	*md,
	WallSpec	*aWallList)
{
	WallSpec		theNearSide,
					theFarSide;
	unsigned int	theNumWalls,
					theNearIndex,
					theFarIndex,
					i,
					h,
					v;

	theNumWalls = 2 * md->itsGameOf.Maze.itsSize * md->itsGameOf.Maze.itsSize;

	for (i = 0; i < theNumWalls; i++)
	{
		theNearSide	= aWallList[i];
		theFarSide	= GetOtherSide(theNearSide, md->itsGameOf.Maze.itsSize, md->itsTopology);

		theNearIndex = md->itsGameOf.Maze.itsCells[theNearSide.itsH][theNearSide.itsV].itsIndex;
		theFarIndex  = md->itsGameOf.Maze.itsCells[ theFarSide.itsH][ theFarSide.itsV].itsIndex;

		if (theNearIndex != theFarIndex)
		{
			for (h = 0; h < md->itsGameOf.Maze.itsSize; h++)
				for (v = 0; v < md->itsGameOf.Maze.itsSize; v++)
					if (md->itsGameOf.Maze.itsCells[h][v].itsIndex == theNearIndex)
						md->itsGameOf.Maze.itsCells[h][v].itsIndex = theFarIndex;

			md->itsGameOf.Maze.itsCells[theNearSide.itsH][theNearSide.itsV].itsWalls[theNearSide.itsSide] = false;
			md->itsGameOf.Maze.itsCells[ theFarSide.itsH][ theFarSide.itsV].itsWalls[ theFarSide.itsSide] = false;
		}
	}
}


static WallSpec GetOtherSide(
	WallSpec		aNearSide,
	unsigned int	aMazeSize,
	TopologyType	aTopology)
{
	WallSpec	theFarSide;

	switch (aNearSide.itsSide)
	{
		case WallWest:
			if (aNearSide.itsH > 0)
				theFarSide.itsH	= aNearSide.itsH - 1;
			else
				theFarSide.itsH	= aMazeSize - 1;
			theFarSide.itsV		= aNearSide.itsV;
			theFarSide.itsSide	= WallEast;
			break;

		case WallSouth:
			theFarSide.itsH		= aNearSide.itsH;
			if (aNearSide.itsV > 0)
				theFarSide.itsV	= aNearSide.itsV - 1;
			else
			{
				theFarSide.itsV = aMazeSize - 1;
				if (aTopology == TopologyKlein)
					theFarSide.itsH = (theFarSide.itsH > 0) ? aMazeSize - theFarSide.itsH : 0;
			}
			theFarSide.itsSide	= WallNorth;
			break;

		case WallEast:
			if (aNearSide.itsH < aMazeSize - 1)
				theFarSide.itsH	= aNearSide.itsH + 1;
			else
				theFarSide.itsH	= 0;
			theFarSide.itsV		= aNearSide.itsV;
			theFarSide.itsSide	= WallWest;
			break;

		case WallNorth:
			theFarSide.itsH		= aNearSide.itsH;
			if (aNearSide.itsV < aMazeSize - 1)
				theFarSide.itsV	= aNearSide.itsV + 1;
			else
			{
				theFarSide.itsV = 0;
				if (aTopology == TopologyKlein)
					theFarSide.itsH = (theFarSide.itsH > 0) ? aMazeSize - theFarSide.itsH : 0;
			}
			theFarSide.itsSide	= WallSouth;
			break;
		
		default:	//	unused, suppresses compiler warnings
			theFarSide.itsH		= 0;
			theFarSide.itsV		= 0;
			theFarSide.itsSide	= WallWest;
			break;
	}

	return theFarSide;
}


static ErrorText DrawMazeAsTexture(ModelData *md)
{
	Byte			*theImage			= NULL,
					*thePixel			= NULL;
	unsigned int	thePixelsPerCell,
					theLineWidth,
					thePixelH,
					thePixelV,
					theCellH,
					theCellV,
					theRelativeH,
					theRelativeV;

	//	Draw an alpha-only maze texture to conserve memory.
	//	It will modulate whatever the current color is at render time.

	//	We'll need to
	//
	//		draw small mazes at high resolution to avoid blurriness, but
	//		draw large mazes at low  resolution to conserve memory.
	//
	//	We can meet both those demands by always using a 512×512 texture.
	//
	//		maze size (cells per row)		  2	  4	  8	 16	 32
	//		cell size (pixels per cell)		256	128	 64	 32	 16
	//		total size (pixels per row)		512	512	512	512	512
	//
	//	Within each 16n×16n cell, the outer n-pixel boundary
	//	will accommodate the cell's walls.

	//	Don't exceed 32×32.
	if (md->itsGameOf.Maze.itsSize > 32)
		return "Maze size cannot exceed 32×32.";

	//	The image dimensions must be powers of two.
	//	For simplicity insist that the maze be square.
	//	(If that requirement proves burdensome, the code could be
	//	modified to use a larger-than-needed bitmap, but for now
	//	I won't bother.)
	if ( ! IsPowerOfTwo(md->itsGameOf.Maze.itsSize) )
		return "Maze size must be a power of two.";

	//	Get some memory.
	//	For alpha-only we'll need one byte per pixel.
	theImage = (Byte *) GET_MEMORY(512*512*1);
	if (theImage == NULL)
		return "Couldn't get memory to draw maze image.";

	//	Note the dimensions.
	thePixelsPerCell	= 512 / md->itsGameOf.Maze.itsSize;
	theLineWidth		= thePixelsPerCell / 16;

	//	Draw the walls.
	//
	//	theImage contains the rows in bottom-to-top order,
	//	so the two outermost loops scan the image in pixel order.
	thePixel = theImage;
	for (thePixelV = 0; thePixelV < 512; thePixelV++)
		for (thePixelH = 0; thePixelH < 512; thePixelH++)
		{
			//	Note the cell to which the pixel belongs
			//	and also its relative coordinates within that cell.
			//	(Because thePixelsPerCell is a power of two,
			//	we could replace the division by a bitshift
			//	and the modulus by a mask, if performance is an issue.)
			theCellH		= (thePixelH + thePixelsPerCell/2) / thePixelsPerCell;
			theCellV		= (thePixelV + thePixelsPerCell/2) / thePixelsPerCell;
			theRelativeH	= (thePixelH + thePixelsPerCell/2) % thePixelsPerCell;
			theRelativeV	= (thePixelV + thePixelsPerCell/2) % thePixelsPerCell;
			if (theCellV == md->itsGameOf.Maze.itsSize)
			{
				theCellV = 0;
				if (md->itsTopology == TopologyKlein)
				{
					theCellH		= md->itsGameOf.Maze.itsSize - theCellH;
					theRelativeH	= (thePixelsPerCell - 1) - theRelativeH;
				}
			}
			if (theCellH == md->itsGameOf.Maze.itsSize)
			{
				theCellH = 0;
			}
			
			//	Initialize the pixel to be fully transparent.
			*thePixel = 0x00;
			
			//	Draw the corners no matter what.
			if ((	theRelativeH <  theLineWidth
				 || theRelativeH >= thePixelsPerCell - theLineWidth)
			 && (	theRelativeV <  theLineWidth
				 || theRelativeV >= thePixelsPerCell - theLineWidth))
			{
				*thePixel = 0xFF;
			}
			
			//	Test the pixel for inclusion in each wall.
			if ((	theRelativeH < theLineWidth
				 && md->itsGameOf.Maze.itsCells[theCellH][theCellV].itsWalls[WallWest])
			 || (	theRelativeV < theLineWidth
				 && md->itsGameOf.Maze.itsCells[theCellH][theCellV].itsWalls[WallSouth])
			 || (	theRelativeH >= thePixelsPerCell - theLineWidth
				 && md->itsGameOf.Maze.itsCells[theCellH][theCellV].itsWalls[WallEast])
			 || (	theRelativeV >= thePixelsPerCell - theLineWidth
				 && md->itsGameOf.Maze.itsCells[theCellH][theCellV].itsWalls[WallNorth]))
			{
				*thePixel = 0xFF;
			}

			//	Advance to the next pixel.
			thePixel++;
		}

	//	Set the alpha-only texture.
	SetTextureImageFancy(	md->itsGameOf.Maze.itsTextureNameMaze,
							512,
							512,
							1,
							theImage,
							md->itsShadersAvailableFlag && md->itsShadersEnabledFlag);

	//	Release the temporary memory.
	FREE_MEMORY_SAFELY(theImage);

	//	Return no error.
	return NULL;
}


static void PositionMouseAndCheese(ModelData *md)
{
	//	Proposition
	//
	//	Let cell A be an arbitrary cell in the maze.
	//	Let cell B be a cell maximally far from cell A
	//		(following the maze, of course).
	//	Let cell C be a cell maximally far from cell B
	//		(following the maze, of course).
	//	Then the distance from cell B to cell C is
	//		greater than or equal to the distance between
	//		any other pair of cells.
	//
	//	The proof is easy, but not quite so brief that
	//	I feel like writing it down here.  Presumably this
	//	result appears somewhere in the literature, even
	//	though I couldn't find it on the web (and Patti Lock
	//	was unaware of it).

	CellSpec		theCellA = {0, 0},
					theCellB,
					theCellC;
	unsigned int	theSide;

	//	Find a pair of maximally distant points in the maze.
	FindFurthestPoint(md, theCellA, &theCellB);
	FindFurthestPoint(md, theCellB, &theCellC);

	//	Place the mouse, converting from integer "maze coordinates" in [0, size)
	//	to floating-point "fundamental domain coordinates" in [-0.5, +0.5).
	//	Scale the mouse by 7/8 = 0.875 to avoid overlapping the maze walls.
	md->itsGameOf.Maze.itsMousePlacement.itsH		= (double)( (signed int)(2 * theCellB.itsH) - (signed int)(md->itsGameOf.Maze.itsSize) )
													/ (double)( 2 * md->itsGameOf.Maze.itsSize );
	md->itsGameOf.Maze.itsMousePlacement.itsV		= (double)( (signed int)(2 * theCellB.itsV) - (signed int)(md->itsGameOf.Maze.itsSize) )
													/ (double)( 2 * md->itsGameOf.Maze.itsSize );
	md->itsGameOf.Maze.itsMousePlacement.itsFlip	= false;
	md->itsGameOf.Maze.itsMousePlacement.itsAngle	= 0.0;	//	to be overwritten immediately below, except in 1×1 case
	md->itsGameOf.Maze.itsMousePlacement.itsSizeH	= 0.875 * (1.0 / md->itsGameOf.Maze.itsSize);
	md->itsGameOf.Maze.itsMousePlacement.itsSizeV	= 0.875 * (1.0 / md->itsGameOf.Maze.itsSize);

	//	Set the mouse's heading, measured in radians counterclockwise from east.
	for (theSide = 0; theSide < 4; theSide++)	//	theSide = WallWest, WallSouth, WallEast, WallNorth
		if ( ! md->itsGameOf.Maze.itsCells[theCellB.itsH][theCellB.itsV].itsWalls[theSide] )
			md->itsGameOf.Maze.itsMousePlacement.itsAngle = -PI + (0.5*PI)*theSide + MOUSE_ROTATION;

	//	Place the cheese, converting from integer "maze coordinates" in [0, size)
	//	to floating-point "fundamental domain coordinates" in [-0.5, +0.5).
	//	Scale the cheese by a little less than 7/8 = 0.875 to avoid touching the maze walls.
	md->itsGameOf.Maze.itsCheesePlacement.itsH		= (double)( (signed int)(2 * theCellC.itsH) - (signed int)(md->itsGameOf.Maze.itsSize) )
													/ (double)( 2 * md->itsGameOf.Maze.itsSize );
	md->itsGameOf.Maze.itsCheesePlacement.itsV		= (double)( (signed int)(2 * theCellC.itsV) - (signed int)(md->itsGameOf.Maze.itsSize) )
													/ (double)( 2 * md->itsGameOf.Maze.itsSize );
	md->itsGameOf.Maze.itsCheesePlacement.itsFlip	= false;
	md->itsGameOf.Maze.itsCheesePlacement.itsAngle	= 0.0;
	md->itsGameOf.Maze.itsCheesePlacement.itsSizeH	= 0.8 * (1.0 / md->itsGameOf.Maze.itsSize);
	md->itsGameOf.Maze.itsCheesePlacement.itsSizeV	= 0.8 * (1.0 / md->itsGameOf.Maze.itsSize);
}


static void FindFurthestPoint(
	ModelData	*md,
	CellSpec	aNearCell,
	CellSpec	*aFarCell)
{
	MazeTreeNode	*theQueue		= NULL;
	CellSpec		theInvalidCell	= {-1, -1};
	unsigned int	theQueueStart,
					theQueueEnd;
	CellSpec		theSelf,
					theParent;
	unsigned int	theSide;
	WallSpec		theSelfWallSpec,
					theNeighborWallSpec;

	theQueue = (MazeTreeNode *) GET_MEMORY(md->itsGameOf.Maze.itsSize * md->itsGameOf.Maze.itsSize * sizeof(MazeTreeNode));
	if (theQueue == NULL)
	{
		ErrorMessage(	"Couldn't get memory to properly position mouse and cheese in maze.",
						"FindFurthestPoint() Error",
						false);
		aFarCell->itsH = (md->itsGameOf.Maze.itsSize > 1) ? ( ! aNearCell.itsH) : 0;
		aFarCell->itsV = 0;
		return;
	}

	theQueueStart			= 0;
	theQueueEnd				= 0;
	theQueue[0].itsSelf		= aNearCell;
	theQueue[0].itsParent	= theInvalidCell;

	while (theQueueStart <= theQueueEnd)
	{
		//	Consider the next node on the queue.
		theSelf		= theQueue[theQueueStart].itsSelf;
		theParent	= theQueue[theQueueStart].itsParent;

		//	Each of theNode's accessible neighbors,
		//	excluding its own parent, gets added to the queue.
		for (theSide = 0; theSide < 4; theSide++)	//	theSide = WallWest, WallSouth, WallEast, WallNorth
			if ( ! md->itsGameOf.Maze.itsCells[theSelf.itsH][theSelf.itsV].itsWalls[theSide] )
			{
				theSelfWallSpec.itsH	= theSelf.itsH;
				theSelfWallSpec.itsV	= theSelf.itsV;
				theSelfWallSpec.itsSide	= theSide;

				theNeighborWallSpec = GetOtherSide(theSelfWallSpec, md->itsGameOf.Maze.itsSize, md->itsTopology);

				if (theNeighborWallSpec.itsH == theParent.itsH
				 && theNeighborWallSpec.itsV == theParent.itsV)
					continue;

				theQueueEnd++;
				theQueue[theQueueEnd].itsSelf.itsH	= theNeighborWallSpec.itsH;
				theQueue[theQueueEnd].itsSelf.itsV	= theNeighborWallSpec.itsV;
				theQueue[theQueueEnd].itsParent		= theSelf;
			}

		theQueueStart++;
	}

	if (theQueueStart != md->itsGameOf.Maze.itsSize * md->itsGameOf.Maze.itsSize)
	{
		ErrorMessage(	"Impossible condition in FindFurthestPoint().",
						"Internal Error",
						false);
		exit(1);
	}

	aFarCell->itsH = theSelf.itsH;
	aFarCell->itsV = theSelf.itsV;

	FREE_MEMORY_SAFELY(theQueue);
}


static bool MazeDragBegin(
	ModelData	*md,
	bool		aShiftOrControlKeyIsDown)
{
	unsigned int	i,
					j;

	UNUSED_PARAMETER(aShiftOrControlKeyIsDown);

	if ( ! md->itsGameOf.Maze.itsMazeIsValid )
		return false;

	//	Did we hit the (field)mouse?
	md->itsDragIsReversed = (	md->itsHandPlacement.itsFlip
							 != md->itsGameOf.Maze.itsMousePlacement.itsFlip);
	if (ShortestDistance(
							md->itsHandPlacement.itsH,
							md->itsHandPlacement.itsV,
							md->itsGameOf.Maze.itsMousePlacement.itsH,
							md->itsGameOf.Maze.itsMousePlacement.itsV,
							md->itsTopology,
							&md->itsDragIsReversed)
						< 0.5 / md->itsGameOf.Maze.itsSize)
	{
		//	Clear the history.
		for (i = 0; i < DRAG_HISTORY_DEPTH; i++)
			for (j = 0; j < 2; j++)
				md->itsDragHistory[i][j] = 0.0;

		//	Report a hit.
		return true;
	}
	else
		return false;
}


static void MazeDragObject(
	ModelData	*md,
	double		aHandLocalDeltaH,
	double		aHandLocalDeltaV)
{
	double			theMouseLocalDeltaH,
					theMouseLocalDeltaV,
					theHandAmbientDeltaH,
					theHandAmbientDeltaV,
					theMouseAmbientDeltaH,
					theMouseAmbientDeltaV;
	unsigned int	theCellH,
					theCellV,
					theNormalizedCellH,
					theNormalizedCellV;
	double			theCellCenterH,
					theCellCenterV,
					theRelativePositionH,
					theRelativePositionV,
					theNormalizedRelativePositionH,
					theNormalizedRelativePositionV,
					theCellWidth;
	signed int		theSignH,
					theSignV;
	double			theSizeH,
					theSizeV,
					theRadius;
	unsigned int	i,
					j;
	double			theTotalMouseLocalDeltaH,
					theTotalMouseLocalDeltaV;

	if ( ! md->itsGameOf.Maze.itsMazeIsValid )
		return;

	//	Compute the requested motion relative to the (field)mouse.

	theMouseLocalDeltaH		= md->itsDragIsReversed
							? -aHandLocalDeltaH : aHandLocalDeltaH;
	theMouseLocalDeltaV		= aHandLocalDeltaV;

	theMouseAmbientDeltaH	= md->itsGameOf.Maze.itsMousePlacement.itsFlip
							? -theMouseLocalDeltaH : theMouseLocalDeltaH;
	theMouseAmbientDeltaV	= theMouseLocalDeltaV;

	//	Which cell are we in?
	//	Allow theCellH == md->itsGameOf.Maze.itsSize even though
	//	that's equivalent to theCellH == 0, and similarly for theCellV.
	//	We'll compute the normalized indices immediately below.
	theCellH = (unsigned int) floor((md->itsGameOf.Maze.itsMousePlacement.itsH + 0.5)*md->itsGameOf.Maze.itsSize + 0.5);
	theCellV = (unsigned int) floor((md->itsGameOf.Maze.itsMousePlacement.itsV + 0.5)*md->itsGameOf.Maze.itsSize + 0.5);

	//	Normalize theCellH and theCellV.
	theNormalizedCellH = theCellH;
	theNormalizedCellV = theCellV;
	if (theNormalizedCellV >= md->itsGameOf.Maze.itsSize)	//	special case for  top  edge of fundamental domain (plus protection from invalid input)
	{
		theNormalizedCellV = 0;
		if (md->itsTopology == TopologyKlein)
			theNormalizedCellH = md->itsGameOf.Maze.itsSize - theNormalizedCellH;
	}
	if (theNormalizedCellH >= md->itsGameOf.Maze.itsSize)	//	special case for right edge of fundamental domain (plus protection from invalid input)
		theNormalizedCellH = 0;

	//	Where is the cell's center?
	//	(We're back to the non-normalized form here.)
	theCellCenterH = (double)theCellH/(double)md->itsGameOf.Maze.itsSize - 0.5;
	theCellCenterV = (double)theCellV/(double)md->itsGameOf.Maze.itsSize - 0.5;

	//	Express the new (field)mouse position relative to the cell's center.
	theRelativePositionH = md->itsGameOf.Maze.itsMousePlacement.itsH + theMouseAmbientDeltaH - theCellCenterH;
	theRelativePositionV = md->itsGameOf.Maze.itsMousePlacement.itsV + theMouseAmbientDeltaV - theCellCenterV;

	//	theRelativePositionV never depends on the normalization,
	//	but in a Klein bottle theRelativePositionH does.
	theNormalizedRelativePositionH =
		(md->itsTopology == TopologyKlein && theCellV == md->itsGameOf.Maze.itsSize) ? 
		-theRelativePositionH : theRelativePositionH;
	theNormalizedRelativePositionV = theRelativePositionV;

	//	Don't pass through any walls, or even go near them.
	if (md->itsGameOf.Maze.itsCells[theNormalizedCellH][theNormalizedCellV].itsWalls[WallWest ]
	 && theNormalizedRelativePositionH < 0)
		theNormalizedRelativePositionH = 0;
	if (md->itsGameOf.Maze.itsCells[theNormalizedCellH][theNormalizedCellV].itsWalls[WallSouth]
	 && theNormalizedRelativePositionV < 0)
		theNormalizedRelativePositionV = 0;
	if (md->itsGameOf.Maze.itsCells[theNormalizedCellH][theNormalizedCellV].itsWalls[WallEast ]
	 && theNormalizedRelativePositionH > 0)
		theNormalizedRelativePositionH = 0;
	if (md->itsGameOf.Maze.itsCells[theNormalizedCellH][theNormalizedCellV].itsWalls[WallNorth]
	 && theNormalizedRelativePositionV > 0)
		theNormalizedRelativePositionV = 0;
	
	//	Recover the unnormalized relative position.
	theRelativePositionH =
		(md->itsTopology == TopologyKlein && theCellV == md->itsGameOf.Maze.itsSize) ? 
		-theNormalizedRelativePositionH : theNormalizedRelativePositionH;
	theRelativePositionV = theNormalizedRelativePositionV;

	//	For convenience, note the cell size.
	theCellWidth = 1.0 / md->itsGameOf.Maze.itsSize;

	//	If the (field)mouse is moving outside the current cell...
	if (fabs(theRelativePositionH) >= 0.5*theCellWidth
	 || fabs(theRelativePositionV) >= 0.5*theCellWidth)
	{
		//	...snap to the nearest passageway's centerline
		if (fabs(theRelativePositionH) < fabs(theRelativePositionV))
			theRelativePositionH = 0;
		else
			theRelativePositionV = 0;

		//	..and don't travel any further than the center
		//	of the neighboring cell.
		if (theRelativePositionH >  theCellWidth)
			theRelativePositionH =  theCellWidth;
		if (theRelativePositionH < -theCellWidth)
			theRelativePositionH = -theCellWidth;
		if (theRelativePositionV >  theCellWidth)
			theRelativePositionV =  theCellWidth;
		if (theRelativePositionV < -theCellWidth)
			theRelativePositionV = -theCellWidth;
	}
	else
	{
		//	Otherwise the (field)mouse is staying within the current cell,
		//	and it suffices to stay a half cell-width away from
		//	corners (keeping in mind that we're already a half
		//	cell-width away from all walls -- see above).
		//	At this point you might want to make yourself
		//	a little sketch showing the cell, with a line
		//	through the center parallel to each wall,
		//	and a quarter-circle of radius one-half cell width
		//	centered at each vertex (or, more to the point,
		//	at each vertex not adjacent to a cell wall).

		//	To avoid special cases, split theRelativePosition
		//	into signs and magnitudes.
		theSignH = (theRelativePositionH >= 0) ? +1 : -1;
		theSignV = (theRelativePositionV >= 0) ? +1 : -1;
		theSizeH = fabs(theRelativePositionH);
		theSizeV = fabs(theRelativePositionV);

		//	Re-express the point relative to the nearest corner.
		theSizeH = 0.5*theCellWidth - theSizeH;
		theSizeV = 0.5*theCellWidth - theSizeV;

		//	Are we closer to the corner than a half a cell width?
		theRadius = sqrt(theSizeH*theSizeH + theSizeV*theSizeV);
		if (theRadius < 0.0001*theCellWidth)	//	Guard against unexpected input.
		{
			theRelativePositionH = 0.0;
			theRelativePositionV = 0.0;
		}
		else
		if (theRadius < 0.5*theCellWidth)	//	If we're too close to the corner...
		{
			//	...dilate radially to lie exactly
			//	a half cell width from the corner
			theSizeH *= 0.5*theCellWidth / theRadius;
			theSizeV *= 0.5*theCellWidth / theRadius;

			//	...switch back to cell-center coordinates
			theSizeH = 0.5*theCellWidth - theSizeH;
			theSizeV = 0.5*theCellWidth - theSizeV;

			//	...and multiply the signs back in
			//	to recover the adjusted RelativePosition.
			theRelativePositionH = theSignH * theSizeH;
			theRelativePositionV = theSignV * theSizeV;
		}
	}

	//	Now that we've clipped theRelativePosition to avoid
	//	the maze walls and corners, (re)compute the deltas.

	theMouseAmbientDeltaH	= theCellCenterH + theRelativePositionH - md->itsGameOf.Maze.itsMousePlacement.itsH;
	theMouseAmbientDeltaV	= theCellCenterV + theRelativePositionV - md->itsGameOf.Maze.itsMousePlacement.itsV;

	theMouseLocalDeltaH		= md->itsGameOf.Maze.itsMousePlacement.itsFlip
							? -theMouseAmbientDeltaH : theMouseAmbientDeltaH;
	theMouseLocalDeltaV		= theMouseAmbientDeltaV;

	aHandLocalDeltaH		= md->itsDragIsReversed
							? -theMouseLocalDeltaH : theMouseLocalDeltaH;
	aHandLocalDeltaV		= theMouseLocalDeltaV;

	theHandAmbientDeltaH	= md->itsHandPlacement.itsFlip
							? -aHandLocalDeltaH : aHandLocalDeltaH;
	theHandAmbientDeltaV	= aHandLocalDeltaV;

	//	Move the (field)mouse.
	md->itsGameOf.Maze.itsMousePlacement.itsH += theMouseAmbientDeltaH;
	md->itsGameOf.Maze.itsMousePlacement.itsV += theMouseAmbientDeltaV;
	NormalizePlacement(&md->itsGameOf.Maze.itsMousePlacement, md->itsTopology);

	//	Move the hand.
	md->itsHandPlacement.itsH += theHandAmbientDeltaH;
	md->itsHandPlacement.itsV += theHandAmbientDeltaV;
	NormalizePlacement(&md->itsHandPlacement, md->itsTopology);

	//	Update the drag history.
	for (i = DRAG_HISTORY_DEPTH - 1; i-- > 0; )
		for (j = 0; j < 2; j++)
			md->itsDragHistory[i+1][j] = md->itsDragHistory[i][j];
	md->itsDragHistory[0][0] = theMouseLocalDeltaH;
	md->itsDragHistory[0][1] = theMouseLocalDeltaV;

	//	Adjust the (field)mouse's heading.
	//	To avoid processor-speed dependencies, sum the (field)mouse's local motions
	//	over a fixed distance rather than a fixed number of frames.
	theTotalMouseLocalDeltaH = 0.0;
	theTotalMouseLocalDeltaV = 0.0;
	for (i = 0; i < DRAG_HISTORY_DEPTH; i++)
	{
		theTotalMouseLocalDeltaH += md->itsDragHistory[i][0];
		theTotalMouseLocalDeltaV += md->itsDragHistory[i][1];

		if (fabs(theTotalMouseLocalDeltaH) > 0.5 * theCellWidth
		 || fabs(theTotalMouseLocalDeltaV) > 0.5 * theCellWidth)
			break;
	}
	if (theTotalMouseLocalDeltaH != 0.0
	 || theTotalMouseLocalDeltaV != 0.0)
		md->itsGameOf.Maze.itsMousePlacement.itsAngle = atan2(theTotalMouseLocalDeltaV, theTotalMouseLocalDeltaH) + MOUSE_ROTATION;
}


static void MazeDragEnd(ModelData *md)
{
	//	Did the mouse reach the cheese?

	if ( ! md->itsGameIsOver
	 && ShortestDistance(	md->itsGameOf.Maze.itsMousePlacement.itsH,
							md->itsGameOf.Maze.itsMousePlacement.itsV,
							md->itsGameOf.Maze.itsCheesePlacement.itsH,
							md->itsGameOf.Maze.itsCheesePlacement.itsV,
							md->itsTopology,
							NULL)
		< 0.25 / md->itsGameOf.Maze.itsSize )
	{
		md->itsGameIsOver = true;
		PlayTheSound(SoundMazeSolved);
		SimulationBegin(md, SimulationMazeFlash);
	}
}


static void MazeSimulationUpdate(ModelData *md)
{
	switch (md->itsSimulationStatus)
	{
		case SimulationMazeFlash:
			md->itsFlashFlag = (((unsigned int) floor(10.0 * md->itsSimulationElapsedTime)) % 2) ? true : false;
			if (md->itsSimulationElapsedTime >= 1.4)
			{
				SimulationEnd(md);
				md->itsFlashFlag = false;
			}
			break;

		default:
			break;
	}
}


static void MazeDraw(ModelData *md)
{
	static const unsigned int	theNumTextureReps[5] = {1, 2, 3, 3, 3};	//	How many repetitions of the grass texture?

	static Placement	theMazePlacement	= {0.0, 0.0, false, 0.0, 1.0, 1.0};
	static float		theMazeColor[4]		= PREMULTIPLY_RGBA(1.00, 1.00, 0.50, 0.75);

	//	Draw the background.
	//
	//	Technical Note:  Just by dumb luck the grass texture
	//	I'm using matches up well enough with itself even
	//	in Klein bottle mode.
	DrawFundamentalDomain(	md,
							md->itsGameOf.Maze.itsTextureNameGrass,
							theNumTextureReps[md->itsDifficultyLevel]);

	//	Draw glide reflection axes if appropriate.
	DrawKleinAxes(	md,
					PREMULTIPLY_RGBA(1.0, 0.0, 0.0, 1.0),
					PREMULTIPLY_RGBA(0.5, 1.0, 1.0, 1.0) );

	//	If the maze is valid...
	if (md->itsGameOf.Maze.itsMazeIsValid)
	{
		//	Draw the maze.
		DrawColorWithAlphaMask(	md,
								md->itsGameOf.Maze.itsTextureNameMaze,
								&theMazePlacement,
								theMazeColor);

		//	Draw the cheese before the mouse, so the mouse stacks
		//	on top of it at the end of the game.

		if ( ! md->itsGameIsOver
		 || md->itsFlashFlag)
			DrawSprite(	md,
						md->itsGameOf.Maze.itsTextureNameCheese,
						&md->itsGameOf.Maze.itsCheesePlacement);

		DrawSprite(	md,
					md->itsGameOf.Maze.itsTextureNameMouse,
					&md->itsGameOf.Maze.itsMousePlacement);
	}
}
