//	TorusGamesMouse.c
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include <math.h>


//	On iPhone OS, a scroll coasts to a gradual stop
//	after the user has lifted his/her finger.
#define COASTING_DECELERATION	2.0	//	distance / sec²
#define MAX_COASTING_DISTANCE	0.5	//	distance
#define MAX_COASTING_SPEED		( sqrt(2.0 * COASTING_DECELERATION * MAX_COASTING_DISTANCE) )	//	distance / sec

//	How close must two clicks be to count as a double-click in the Apples game?
#define DOUBLE_CLICK_RADIUS		0.01


static void		MoveTheTouchPoint(ModelData *md, double aHandLocalDeltaH, double aHandLocalDeltaV);
static void		ScrollTheBoard(ModelData *md, double aDeltaH, double aDeltaV);


void MouseDown(
	ModelData	*md,
	double		aMouseH,	//	horizontal coordinate (intrinsic units rightward from center, in range [-0.5, +0.5])
	double		aMouseV,	//	 vertical  coordinate (intrinsic units   upward  from center, in range [-0.5, +0.5])
	bool		aShiftOrControlKeyIsDown,
	bool		aTemporalDoubleClick)
{
	//	In ViewSmall or ViewTiling, rescale mouse coordinates
	//	from [-0.5, +0.5] to [-1.5, +1.5].
	switch (md->itsView)
	{
		case ViewDomain:
			//	no change
			break;
		
		case ViewSmall:
		case ViewTiling:
			aMouseH *= 3.0;
			aMouseV *= 3.0;
			break;
	}
	
	//	Convert mouse from window coordinates to fundamental domain coordinates.
	aMouseH -= md->itsOffset.itsH;
	aMouseV -= md->itsOffset.itsV;
	if (md->itsOffset.itsFlip)
		aMouseH = -aMouseH;

	//	Handle the mouse-down according to itsHandStatus.
	switch (md->itsHandStatus)
	{
		case HandNone:
			//	Place our internal cursor where the system cursor was.
			md->itsHandPlacement.itsH		= aMouseH;
			md->itsHandPlacement.itsV		= aMouseV;
			md->itsHandPlacement.itsFlip	= md->itsOffset.itsFlip;

			//	Normalize itsHandPlacement to the fundamental domain.
			NormalizePlacement(&md->itsHandPlacement, md->itsTopology);

			//	Note that we now have the cursor.
			md->itsHandStatus = HandFree;

			//	For Tic-Tac-Toe, Pool or Apples let's break here,
			//	to avoid accidently making a tic-tac-toe move,
			//	shooting the cue ball or uncovering an unintended apple.
			if (md->itsGame == GameTicTacToe
			 || md->itsGame == GamePool
			 || md->itsGame == GameApples)
				break;
			//	Otherwise fall through to next case,
			//	in case the user is trying to click on something...
		case HandFree:

			//	Begin a drag or scroll.

			//	Normally the game decides whether the mouse hit point
			//	determines a drag (if some object is hit) or a scroll (otherwise).
			//	Exception:  If an animation is in progress, all mouse-down events
			//	get treated as scrolls, to avoid interfering with whatever
			//	action the game has in progress.
			if (md->itsSimulationStatus == SimulationNone
			 && md->itsGameDragBegin != NULL
			 && (*md->itsGameDragBegin)(md, aShiftOrControlKeyIsDown))
			{
				md->itsHandStatus = HandDrag;
			}
			else
			{
				md->itsHandStatus = HandScroll;
			}

			break;

		default:	//	unused (I think)
			break;
	}
	
	//	The operating system (Mac OS or Win32) judges a click to be
	//	a double-click based solely on the elapsed time since the preceding click.
	//	In the case of GameApples we insist that the two clicks be spatially close
	//	as well, so that rapid clicking (to eat apples) does not accidently
	//	exit the user from the torus or Klein bottle.
	md->itsDoubleClickFlag =
	(
		aTemporalDoubleClick
	 && 
		(	
			md->itsGame != GameApples
		 ||	
			ShortestDistance(	md->itsHandPlacement.itsH,
								md->itsHandPlacement.itsV,
								md->itsPrevClickPlacement.itsH,
								md->itsPrevClickPlacement.itsV,
								md->itsTopology,
								NULL) < DOUBLE_CLICK_RADIUS
		)
	);
	md->itsPrevClickPlacement = md->itsHandPlacement;
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


void MouseMove(
	ModelData	*md,
	double		aMouseDeltaH,	//	incremental motion, with same coordinates as in MouseDown()
	double		aMouseDeltaV)	//	incremental motion, with same coordinates as in MouseDown()
{
	double		theHandLocalDeltaH,
				theHandLocalDeltaV;

	//	Interpret the mouse motion as a motion of the hand
	//	relative to its own coordinate system.
	theHandLocalDeltaH = aMouseDeltaH;
	theHandLocalDeltaV = aMouseDeltaV;

	//	Handle the MouseMove according to the HandStatus.
	switch (md->itsHandStatus)
	{
		case HandNone:
			break;

		case HandFree:

			//	Scroll the hand along, taking into account
			//	the hand's orientation within the fundamental domain.
			md->itsHandPlacement.itsH += (md->itsHandPlacement.itsFlip ? -theHandLocalDeltaH : theHandLocalDeltaH);
			md->itsHandPlacement.itsV += theHandLocalDeltaV;
#ifdef CYLINDER_HACK
			if (md->itsHandPlacement.itsV > +0.45)
				md->itsHandPlacement.itsV = +0.45;
			if (md->itsHandPlacement.itsV < -0.45)
				md->itsHandPlacement.itsV = -0.45;
#endif
			NormalizePlacement(&md->itsHandPlacement, md->itsTopology);

			//	Some games don't care about the position of a free hand, but some do.
			if (md->itsGameHandMoved != NULL)
				(*md->itsGameHandMoved)(md);

			break;

		case HandDrag:
			//	The game will take responsibility for modifying
			//	the motion to avoid collisions and then moving
			//	both the hand and the object as necessary.
			if (md->itsGameDragObject != NULL)
				(*md->itsGameDragObject)(md, theHandLocalDeltaH, theHandLocalDeltaV);
			break;

		case HandScroll:
			ScrollTheBoard(md, theHandLocalDeltaH, theHandLocalDeltaV);
			break;
	}
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


void MouseUp(ModelData *md)
{
	switch (md->itsHandStatus)
	{
		case HandNone:
		case HandFree:
			break;

		case HandScroll:
			md->itsHandStatus = HandFree;
			break;

		case HandDrag:

			//	Let the game do what it needs to do.
			if (md->itsGameDragEnd != NULL)
				(*md->itsGameDragEnd)(md);

			//	Free the hand.
			md->itsHandStatus = HandFree;

			break;
	}
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


void MouseGone(ModelData *md)
{
	//	Typically we expect the user to complete any scrolling or dragging
	//	before relinquishing the torus cursor, but just in case a drag is still
	//	in progress when the user hits the ESC key, or in case some other program
	//	unexpectedly grabs the cursor, let's simulate a MouseUp().
	MouseUp(md);

	//	We're done with the cursor.
	md->itsHandStatus = HandNone;
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


void MouseWheel(
	ModelData	*md,
	double		aDeltaH,
	double		aDeltaV)
{
	ScrollTheBoard(md, aDeltaH, aDeltaV);

	md->itsRedrawRequestFlag = true;
}


void TouchBegan(
	ModelData	*md,
	double		aTouchH,	//	horizontal coordinate (intrinsic units rightward from center, in range [-0.5, +0.5])
	double		aTouchV,	//	 vertical  coordinate (intrinsic units   upward  from center, in range [-0.5, +0.5])
	bool		aFlickFlag)	//	Did the touch begin with a flick?
{
	//	In ViewSmall or ViewTiling, rescale touch coordinates
	//	from [-0.5, +0.5] to [-1.5, +1.5].  NOT CURRENTLY USED.
	switch (md->itsView)
	{
		case ViewDomain:
			//	no change
			break;
		
		case ViewSmall:
		case ViewTiling:
			aTouchH *= 3.0;
			aTouchV *= 3.0;
			break;
	}

	//	(aTouchH, aTouchV) tells us where our "cursor" (the user's finger) is.
	//	In absolute coordinates:
	md->itsHandPlacement.itsH		= aTouchH;
	md->itsHandPlacement.itsV		= aTouchV;
	md->itsHandPlacement.itsFlip	= false;

	//	Convert itsHandPlacement from window coordinates to fundamental domain coordinates.
	md->itsHandPlacement.itsH -= md->itsOffset.itsH;
	md->itsHandPlacement.itsV -= md->itsOffset.itsV;
	if (md->itsOffset.itsFlip)
	{
		md->itsHandPlacement.itsH		= - md->itsHandPlacement.itsH;
		md->itsHandPlacement.itsFlip	= ! md->itsHandPlacement.itsFlip;
	}

	//	Normalize itsHandPlacement to the fundamental domain.
	NormalizePlacement(&md->itsHandPlacement, md->itsTopology);

	//	If a recent touch is still coasting...
	if (md->itsTouchIsCoasting)
	{
		//	...keep that touch alive.

		//	Clear the itsTouchIsCoasting flag.
		md->itsTouchIsCoasting	= false;
		md->itsTouchVelocity[0] = 0.0;
		md->itsTouchVelocity[1] = 0.0;
		
		//	Sit back and wait for TouchMoved() events.
	}
	else
	{
		//	Begin a new scroll or drag.

		//	Handle the touch according to itsHandStatus.
		switch (md->itsHandStatus)
		{
			case HandNone:

				//	Unlike the desktop Torus Games, which makes a distinction between 
				//	HandNone (the system's arrow cursor is active) and 
				//	HandFree (the Torus Games' internal hand cursor is active), 
				//	the iPhone/iTouch Torus Games treat the user's finger as the cursor.
				//	The user's finger hovers above the torus or Klein bottle,
				//	so even though we seldom know where it is, we always treat it
				//	as HandFree, never as HandNone.
				//
				//	Fall through to process the TouchBegin...
			case HandFree:
				
				//	Begin a drag or scroll.

				//	Normally the game decides whether the touch point
				//	determines a drag (if some object is hit) or a scroll (otherwise).
				//
				//	Exception #1.  If the use flicked the board,
				//	treat the TouchBegan as a scroll, no matter what point
				//	the user touched.
				//
				//	Exception #2.  If an animation is in progress, all TouchBegan events
				//	get treated as scrolls, to avoid interfering with whatever
				//	action the game has in progress.
				if ( ! aFlickFlag
				 && md->itsSimulationStatus == SimulationNone
				 && md->itsGameDragBegin != NULL
				 && (*md->itsGameDragBegin)(md, false))
				{
					md->itsHandStatus = HandDrag;
				}
				else
				{
					md->itsHandStatus = HandScroll;
				}

				break;
			
			case HandDrag:	//	unused (I think)
			default:		//	unused (I think)
				break;
		}
	}
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


void TouchMoved(
	ModelData	*md,
	double		aTouchDeltaH,	//	incremental motion, with same coordinates as in MouseDown()
	double		aTouchDeltaV,	//	incremental motion, with same coordinates as in MouseDown()
	double		aTouchDeltaT)	//	incremental time
{
	double		theHandLocalDeltaH,
				theHandLocalDeltaV;

	//	Just to be safe...
	if (md->itsHandStatus != HandScroll && md->itsHandStatus != HandDrag)
		return;	//	should never occur

	//	Interpret the TouchMoved as a motion of the hand
	//	relative to its own coordinate system.
	theHandLocalDeltaH = aTouchDeltaH;
	theHandLocalDeltaV = aTouchDeltaV;

	//	Handle the TouchMoved according to the HandStatus.
	MoveTheTouchPoint(md, theHandLocalDeltaH, theHandLocalDeltaV);
	
	//	Compute and remember itsTouchVelocity.

	if (aTouchDeltaT <= 0.0 || aTouchDeltaT > 1.0)	//	just to be safe
		aTouchDeltaT = 1.0;

	md->itsTouchVelocity[0] = aTouchDeltaH / aTouchDeltaT;
	md->itsTouchVelocity[1] = aTouchDeltaV / aTouchDeltaT;
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


void TouchEnded(ModelData *md)
{
	double	theSpeed,
			theFactor;

	switch (md->itsHandStatus)
	{
		case HandScroll:

			//	Let a scroll gradually coast to a stop.
			//	SimulationUpdate() will call TouchMomentum() once per frame
			//	until TouchMomentum() clears the itsTouchIsCoasting flag.
			md->itsTouchIsCoasting = true;
			
			//	Don't let the board coast further than MAX_COASTING_DISTANCE.
			//	A quick Physics 101 calculation shows that a board with initial velocity
			//	MAX_COASTING_SPEED = sqrt(2.0 * COASTING_DECELERATION * MAX_COASTING_DISTANCE)
			//	travels a distance MAX_COASTING_DISTANCE before coming to rest.

			theSpeed = sqrt(md->itsTouchVelocity[0] * md->itsTouchVelocity[0]
						  + md->itsTouchVelocity[1] * md->itsTouchVelocity[1]);
			
			if (theSpeed > MAX_COASTING_SPEED)
			{
				theFactor = MAX_COASTING_SPEED / theSpeed;
				md->itsTouchVelocity[0] *= theFactor;
				md->itsTouchVelocity[1] *= theFactor;
			}

			break;
		
		case HandDrag:

			//	Finalize the drag.
			
			//	Let the game do what it needs to do.
			if (md->itsGameDragEnd != NULL)
				(*md->itsGameDragEnd)(md);

			//	Free the hand.
			md->itsHandStatus = HandFree;

			break;
		
		default:	//	should never occur
			break;
	}
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


void TouchCancelled(ModelData *md)
{
	//	TouchCancelled should get called only under exceptional circumstances,
	//	for example if a phone call interrupts a program on an iPhone.
	//	Treat TouchCancelled the same as TouchEnded.

	TouchEnded(md);
}


void TouchMomentum(
	ModelData	*md,
	double		aTimeInterval)	//	in seconds
{
	double	theSpeedDecrease,
			theSpeed,
			theFactor,
			theDeltaH,
			theDeltaV;

	//	Just to be safe...
	if (md->itsHandStatus != HandScroll)
		return;	//	should never occur
	
	//	The user has lifted his/her finger,
	//	but for a brief period we let the scroll or drag
	//	continue under its own momentum.

	theSpeedDecrease = aTimeInterval * COASTING_DECELERATION;
	
	theSpeed = sqrt(md->itsTouchVelocity[0] * md->itsTouchVelocity[0]
				  + md->itsTouchVelocity[1] * md->itsTouchVelocity[1]);
	
	if (theSpeed > theSpeedDecrease)
	{
		theFactor = (theSpeed - theSpeedDecrease) / theSpeed;
		md->itsTouchVelocity[0] *= theFactor;
		md->itsTouchVelocity[1] *= theFactor;
		
		theDeltaH = md->itsTouchVelocity[0] * aTimeInterval;
		theDeltaV = md->itsTouchVelocity[1] * aTimeInterval;

		MoveTheTouchPoint(md, theDeltaH, theDeltaV);
	}
	else
	{
		//	The "coasting" period is over.

		md->itsTouchIsCoasting	= false;
		md->itsTouchVelocity[0]	= 0.0;
		md->itsTouchVelocity[1]	= 0.0;

		md->itsHandStatus = HandFree;
	}
	
	//	Ask the idle-time routine to redraw the scene.
	md->itsRedrawRequestFlag = true;
}


static void MoveTheTouchPoint(
	ModelData	*md,
	double		aHandLocalDeltaH,
	double		aHandLocalDeltaV)
{
	switch (md->itsHandStatus)
	{
		case HandNone:
		case HandFree:
			//	should never occur
			break;

		case HandDrag:
			//	The game will take responsibility for modifying
			//	the motion to avoid collisions and then moving
			//	both the (invisible) hand and the object as necessary.
			if (md->itsGameDragObject != NULL)
				(*md->itsGameDragObject)(md, aHandLocalDeltaH, aHandLocalDeltaV);
			break;

		case HandScroll:
			ScrollTheBoard(md, aHandLocalDeltaH, aHandLocalDeltaV);
			break;
	}
}


static void ScrollTheBoard(
	ModelData	*md,
	double		aDeltaH,
	double		aDeltaV)
{
#ifdef CYLINDER_HACK
	aDeltaV = 0.0;
#endif

	//	Different images of the hand may have different chiralities
	//	in the Klein bottle, especially in Tiling view.
	//	The scroll can't possibly agree with all of them.
	//	So scroll relative to an unreflected hand
	//	to keep the interface as intuitive as possible.
	md->itsOffset.itsH += aDeltaH;
	md->itsOffset.itsV += aDeltaV;
	NormalizeOffset(&md->itsOffset, md->itsTopology);
}


void CharacterInput(
	ModelData	*md,
	Char16		aCharacter)
{
	if (md->itsGameCharacterInput != NULL)
		(*md->itsGameCharacterInput)(md, aCharacter);

	md->itsRedrawRequestFlag = true;
}
