//	TorusGamesSimulation.c
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"


//	How much can the simulation advance in one frame?
#define MAX_FRAME_PERIOD	0.1


void ResetGame(ModelData *md)
{
	if (md->itsGameReset != NULL)
		(*md->itsGameReset)(md);

	md->itsRedrawRequestFlag = true;
}


void SetUpGame(ModelData *md)
{
	//	Let the new game set up its internal data, allocate any memory
	//	it may need, and initialize function pointers.

	//	We assume the caller has already set up all platform-dependent items,
	//	such as fonts and a possible side window.

	switch (md->itsGame)
	{
		case GameNone:									break;
		case GameIntro:			     IntroSetUp(md);	break;
		case GameTicTacToe:		 TicTacToeSetUp(md);	break;
		case GameMaze:			      MazeSetUp(md);	break;
		case GameCrossword:		 CrosswordSetUp(md);	break;
		case GameWordSearch:	WordSearchSetUp(md);	break;
		case GameJigsaw:		    JigsawSetUp(md);	break;
		case GameChess:			     ChessSetUp(md);	break;
		case GamePool:			      PoolSetUp(md);	break;
		case GameApples:		    ApplesSetUp(md);	break;
		default:										return;	//	should never occur
	}

	//	Reset scrolling.
	md->itsOffset.itsH		= 0.0;
	md->itsOffset.itsV		= 0.0;
	md->itsOffset.itsFlip	= false;
#ifdef TORUS_GAMES_FOR_TALK
	if (md->itsGame == GameTicTacToe)
	{
		md->itsOffset.itsH = 0.5;
		md->itsOffset.itsV = 0.5;
	}
#endif
	
	//	Request new textures.
	md->itsPreparedTextures = false;

	//	Request a redraw.
	md->itsRedrawRequestFlag = true;
}


void ShutDownGame(ModelData *md)	//	Assumes OpenGL context is current,
									//	so itsGameShutDown() can free textures.
{
	//	Let the old game shut down its internal data
	//	and free any memory it may have allocated.

	//	We assume all platform-dependent items, such as fonts and 
	//	a possible side window, are still present, and that
	//	the caller will release them after ShutDownGame() returns.
	//	(In practice, no current game uses fonts or a side window
	//	during shutdown, but if one ever needs them, they're there.)

	//	Cancel any pending simulation.
	SimulationEnd(md);
	
	//	Cancel any coasting.
	md->itsTouchIsCoasting	= false;
	md->itsTouchVelocity[0]	= 0.0;
	md->itsTouchVelocity[1]	= 0.0;
	
	//	Clear cached sounds.
	ClearSoundCache();

	//	Let the game free its private memory,
	//	terminate any private threads,
	//	release game-specific OpenGL textures,
	//	and generally clean up as it sees fit.
	//
	//	Note:  The caller must ensure that the OpenGL context
	//	is set so textures may be released.
	if (md->itsGameShutDown != NULL)
		(*md->itsGameShutDown)(md);

	//	Clear function pointers.
	ClearFunctionPointers(md);
	
	//	This game is no longer available.
	md->itsGame = GameNone;
}


void SpecialMenuCommand(ModelData *md)
{
	if (md->itsGameSpecialCommand != NULL)
		(*md->itsGameSpecialCommand)(md);

	md->itsRedrawRequestFlag = true;
}


void ChangeHumanVsComputer(
	ModelData	*md,
	bool		aNewHumanVsComputer)
{
	md->itsHumanVsComputer = aNewHumanVsComputer;

	//	Update the Pool status window.
	//	Other games require no special notification.
	if (md->itsGame == GamePool)
		PoolSetStatusMessage(md);

	md->itsRedrawRequestFlag = true;
}


void ChangeDifficultyLevel(
	ModelData		*md,
	unsigned int	aNewDifficultyLevel)
{
	//	Just to be safe...
	if (aNewDifficultyLevel > 4)
		return;

	//	Ignore redundant changes to prevent unneeded resets.
	if (aNewDifficultyLevel == md->itsDifficultyLevel)
		return;

	//	Set the new difficulty level.
	md->itsDifficultyLevel = aNewDifficultyLevel;

	//	The maze, jigsaw puzzle and apples games need to reset,
	//	but chess does not.
	if (md->itsGame == GameMaze
	 || md->itsGame == GameJigsaw
	 || md->itsGame == GameApples)
		ResetGame(md);
}


void ChangeTopology(
	ModelData		*md,
	TopologyType	aNewTopology)
{
	//	Just to be safe...
	if (aNewTopology != TopologyTorus
	 && aNewTopology != TopologyKlein)
		return;

	//	Set the new topology.
	md->itsTopology = aNewTopology;

	//	Reset the game as necessary.
	ResetGame(md);

	//	Reset scrolling.
	md->itsOffset.itsH		= 0.0;
	md->itsOffset.itsV		= 0.0;
	md->itsOffset.itsFlip	= false;

	//	Request a redraw.
	md->itsRedrawRequestFlag = true;
}


void ChangeView(
	ModelData	*md,
	ViewType	aNewView)
{
	//	Just to be safe...
	if (aNewView != ViewDomain
	 && aNewView != ViewSmall
	 && aNewView != ViewTiling)
		return;

	//	Set the new view.
	md->itsView = aNewView;

	//	Request a redraw.
	md->itsRedrawRequestFlag = true;
}


void SimulationBegin(
	ModelData		*md,
	SimulationType	aSimulationStatus)
{
	if (md->itsSimulationStatus != SimulationNone)
		ErrorMessage(	"A new simulation is trying to begin while an old one is still active.",
						"Internal Error in SimulationBegin()",
						false);

	md->itsSimulationStatus			= aSimulationStatus;
	md->itsSimulationElapsedTime	= 0.0;
	md->itsSimulationDeltaTime		= 0.0;
}


void SimulationEnd(ModelData *md)
{
	md->itsSimulationStatus			= SimulationNone;
	md->itsSimulationElapsedTime	= 0.0;
	md->itsSimulationDeltaTime		= 0.0;
}


bool SimulationWantsUpdates(
	ModelData	*md)
{
	return md->itsSimulationStatus != SimulationNone
		|| md->itsTouchIsCoasting
		|| md->itsRedrawRequestFlag;
}


void SimulationUpdate(
	ModelData	*md,
	double		aFramePeriod)	//	in seconds
{
	//	Keep track of total elapsed time.
	md->itsSimulationElapsedTime += aFramePeriod;

	//	If some external delay suspends the animation for a few seconds
	//	(for example if the user holds down a menu) we'll receive
	//	a huge frame period.  To avoid a discontinuous jump,
	//	limit the frame period to MAX_FRAME_PERIOD.
	//	This limit should also have the desirable effect of slowing the animation
	//	on systems with humble graphics cards and very slow frame rates.
	if (aFramePeriod > MAX_FRAME_PERIOD)
		aFramePeriod = MAX_FRAME_PERIOD;
	
	//	Record the incremental time.
	md->itsSimulationDeltaTime = aFramePeriod;

	//	Let the current game respond as it wishes.
	if (md->itsGameSimulationUpdate != NULL
	 && md->itsSimulationStatus != SimulationNone)
		(*md->itsGameSimulationUpdate)(md);
	
	//	Let the touch-handing code advance the scroll or drag
	//	as it wishes (iPhone OS only).
	if (md->itsTouchIsCoasting)
		TouchMomentum(md, aFramePeriod);
		
	//	We may assume that the UI-specific code will redraw
	//	the scene after this call to SimulationUpdate() returns,
	//	so this is a convenient spot to clear itsRedrawRequestFlag.
	md->itsRedrawRequestFlag = false;
}
