//	TorusGamesTicTacToe.c
//
//	Code specific to the TicTacToe game.
//	This code knows about OpenGL, but does not
//	know about Windows or any other operating system.
//
//	The board's nine cells are indexed as
//
//		(0,2)	(1,2)	(2,2)
//		(0,1)	(1,1)	(2,1)
//		(0,0)	(1,0)	(2,0)
//
//	to agree with the standard OpenGL coordinate system,
//	at the expense of violating the usual (row, column)
//	convention for arrays.  Cell (0,0) straddles all four corners.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include "TorusGames-OpenGL.h"
#include <math.h>


#define ROOT_TWO	1.41421356237309504880


//	Public functions
static void				TicTacToeShutDown(ModelData *md);
static void				TicTacToeSetUpTextures(ModelData *md, ErrorText *aFirstError);
static void				TicTacToeShutDownTextures(ModelData *md);
static void				TicTacToeReset(ModelData *md);
static bool				TicTacToeDragBegin(ModelData *md, bool aShiftOrControlKeyIsDown);
static void				TicTacToeDragObject(ModelData *md, double aHandLocalDeltaH, double aHandLocalDeltaV);
static void				TicTacToeDragEnd(ModelData *md);
static void				TicTacToeSimulationUpdate(ModelData *md);
static void				TicTacToeDraw(ModelData *md);

//	Private functions
static void				ComputerMoves(ModelData *md);
static bool				ComputerMovesToWin(ModelData *md, TicTacToePlayer aPlayer);
static bool				ComputerMovesToBlock(ModelData *md,TicTacToePlayer aPlayer);
static bool				ComputerMovesRandomly(ModelData *md, TicTacToePlayer aPlayer);
static TicTacToePlayer	CheckForWin(ModelData *md, Placement *aThreeInARow);
static void				NormalizeCell(signed int *h, signed int *v, TopologyType aTopology);


void TicTacToeSetUp(ModelData *md)
{
	unsigned int	h,
					v;

	//	Initialize function pointers.
	md->itsGameShutDown			= &TicTacToeShutDown;
	md->itsGameSetUpTextures	= &TicTacToeSetUpTextures;
	md->itsGameShutDownTextures	= &TicTacToeShutDownTextures;
	md->itsGameReset			= &TicTacToeReset;
	md->itsGameHandMoved		= NULL;
	md->itsGameDragBegin		= &TicTacToeDragBegin;
	md->itsGameDragObject		= &TicTacToeDragObject;
	md->itsGameDragEnd			= &TicTacToeDragEnd;
	md->itsGameCharacterInput	= NULL;
	md->itsGameSpecialCommand	= NULL;
	md->itsGameSimulationUpdate	= &TicTacToeSimulationUpdate;
	md->itsGameDraw				= &TicTacToeDraw;

	//	Initialize variables.

	for (h = 0; h < 3; h++)
		for (v = 0; v < 3; v++)
			md->itsGameOf.TicTacToe.itsCellFlip[h][v] = false;

	md->itsGameOf.TicTacToe.itsWinLine.itsH		= 0.0;
	md->itsGameOf.TicTacToe.itsWinLine.itsV		= 0.0;
	md->itsGameOf.TicTacToe.itsWinLine.itsFlip	= false;
	md->itsGameOf.TicTacToe.itsWinLine.itsAngle	= 0.0;
	md->itsGameOf.TicTacToe.itsWinLine.itsSizeH	= 0.0;
	md->itsGameOf.TicTacToe.itsWinLine.itsSizeV	= 0.0;
	
	//	Initialize game-specific texture names.
	md->itsGameOf.TicTacToe.itsTextureNameBackground	= 0;
	md->itsGameOf.TicTacToe.itsTextureNameX				= 0;
	md->itsGameOf.TicTacToe.itsTextureNameO				= 0;
	md->itsGameOf.TicTacToe.itsTextureNameWin			= 0;

	//	Set up the board.
	TicTacToeReset(md);
}


static void TicTacToeShutDown(ModelData *md)
{
	//	The caller must ensure that the OpenGL context is set
	//	so textures may be released.
	TicTacToeShutDownTextures(md);
}


static void TicTacToeSetUpTextures(
	ModelData	*md,
	ErrorText	*aFirstError)
{
	SetUpOneTexture(md, &md->itsGameOf.TicTacToe.itsTextureNameBackground,	"Tic-Tac-Toe/SquarePlainOffset.rgba",	TextureWrap,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.TicTacToe.itsTextureNameX,			"Tic-Tac-Toe/SquareX.rgba",				TextureWrap,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.TicTacToe.itsTextureNameO,			"Tic-Tac-Toe/SquareO.rgba",				TextureWrap,	TextureRGBA,	aFirstError);
	SetUpOneTexture(md, &md->itsGameOf.TicTacToe.itsTextureNameWin,			"Tic-Tac-Toe/Win.rgba",					TextureWrap,	TextureAlpha,	aFirstError);
}


static void TicTacToeShutDownTextures(ModelData *md)
{
	ShutDownOneTexture(&md->itsGameOf.TicTacToe.itsTextureNameBackground);
	ShutDownOneTexture(&md->itsGameOf.TicTacToe.itsTextureNameX			);
	ShutDownOneTexture(&md->itsGameOf.TicTacToe.itsTextureNameO			);
	ShutDownOneTexture(&md->itsGameOf.TicTacToe.itsTextureNameWin		);
}


static void TicTacToeReset(ModelData *md)
{
	unsigned int	h,
					v;

	for (h = 0; h < 3; h++)
		for (v = 0; v < 3; v++)
			md->itsGameOf.TicTacToe.itsBoard[h][v] = PlayerNone;

	//	Abort any pending simulation.
	SimulationEnd(md);

	md->itsGameIsOver						= false;
	md->itsGameOf.TicTacToe.itsWhoseTurn	= PlayerX;
}


static bool TicTacToeDragBegin(
	ModelData	*md,
	bool		aShiftOrControlKeyIsDown)
{
	double			theIntegerPart;	//	value is ignored
	unsigned int	h,
					v;
	bool			theFlip;

	UNUSED_PARAMETER(aShiftOrControlKeyIsDown);

	//	If the game is already over, ignore all hits.
	if (md->itsGameIsOver)
		return false;

	//	Are we near the center of a cell?
	//	If not, treat this hit as a scroll.
	if (modf(3.0*(md->itsHandPlacement.itsH + 0.5) + 0.25, &theIntegerPart) > 0.5
	 || modf(3.0*(md->itsHandPlacement.itsV + 0.5) + 0.25, &theIntegerPart) > 0.5)
	{
		return false;
	}

	//	Find the integer coordinates (h,v) of the hit cell.
	//	While we're at it, note the chirality of the hand.
	//	When recording chirality, cells with coordinates h = 0 or/and v = 0
	//	are considered to lie along the left and/or bottom edges
	//	of the fundamental domain (rather than the right and/or top edges).
	h		= (unsigned int) floor(3.0*(md->itsHandPlacement.itsH + 2.0/3.0));
	v		= (unsigned int) floor(3.0*(md->itsHandPlacement.itsV + 2.0/3.0));
	theFlip	= md->itsHandPlacement.itsFlip;
	if (v > 2)	//	Test for v > 2 rather than v == 3 to protect against invalid input.
	{
		v = 0;
		if (md->itsTopology == TopologyKlein)
		{
			h = 3 - h;
			theFlip = ! theFlip;
		}
	}
	if (h > 2)
	{
		h = 0;
	}

	//	If the cell is already occupied, treat this hit as a scroll.
	if (md->itsGameOf.TicTacToe.itsBoard[h][v] != PlayerNone)
		return false;

	//	Mark O or X in the hit cell, according to whose turn it is.
	md->itsGameOf.TicTacToe.itsBoard[h][v] = md->itsGameOf.TicTacToe.itsWhoseTurn;

	//	The orientation of the symbol relative to the fundamental domain 
	//	should match the orientation of the nearest image of the hand
	//	relative to the fundamental domain.
	md->itsGameOf.TicTacToe.itsCellFlip[h][v] = theFlip;

	//	Toggle itsGameOf.TicTacToe.itsWhoseTurn.
	md->itsGameOf.TicTacToe.itsWhoseTurn = (md->itsGameOf.TicTacToe.itsWhoseTurn == PlayerX) ? PlayerO : PlayerX;

	//	Play the move sound.
	PlayTheSound(SoundTicTacToeMove);

	//	Wait until TicTacToeDragEnd() to let the computer respond.
	//	This will make the game a little more civilized for the user.

	return true;
}


static void TicTacToeDragObject(
	ModelData	*md,
	double		aHandLocalDeltaH,
	double		aHandLocalDeltaV)
{
	md->itsHandPlacement.itsH += (md->itsHandPlacement.itsFlip ? -aHandLocalDeltaH : aHandLocalDeltaH);
	md->itsHandPlacement.itsV += aHandLocalDeltaV;
	NormalizePlacement(&md->itsHandPlacement, md->itsTopology);
}


static void TicTacToeDragEnd(ModelData *md)
{
	//	Has somebody won?
	if (CheckForWin(md, &md->itsGameOf.TicTacToe.itsWinLine) != PlayerNone)
	{
		SimulationBegin(md, SimulationTicTacToeWaitToDisplayWin);
	}
	else
	{
		if (md->itsHumanVsComputer)
		{
			//	Wait half a second before letting the computer respond,
			//	for the user's psychological benefit.
			SimulationBegin(md, SimulationTicTacToeWaitToMove);
		}
	}
}


static void ComputerMoves(ModelData *md)
{
	//	Start the move sound playing.
	PlayTheSound(SoundTicTacToeMove);

	if (ComputerMovesToWin(md, md->itsGameOf.TicTacToe.itsWhoseTurn))
	{
		//	Record the win and flash the winning line.
		SimulationBegin(md, SimulationTicTacToeWaitToDisplayWin);
	}
	else
	{
		if ( ! ComputerMovesToBlock(md, md->itsGameOf.TicTacToe.itsWhoseTurn) )
			(void) ComputerMovesRandomly(md, md->itsGameOf.TicTacToe.itsWhoseTurn);
	}

	md->itsGameOf.TicTacToe.itsWhoseTurn = (md->itsGameOf.TicTacToe.itsWhoseTurn == PlayerX) ? PlayerO : PlayerX;
}


static bool ComputerMovesToWin(
	ModelData		*md,
	TicTacToePlayer	aPlayer)
{
	unsigned int	h,
					v;

	for (h = 0; h < 3; h++)
		for (v = 0; v < 3; v++)
			if (md->itsGameOf.TicTacToe.itsBoard[h][v] == PlayerNone)
			{
				md->itsGameOf.TicTacToe.itsBoard[h][v] = aPlayer;
				if (CheckForWin(md, &md->itsGameOf.TicTacToe.itsWinLine) == aPlayer)
					return true;
				else
					md->itsGameOf.TicTacToe.itsBoard[h][v] = PlayerNone;
			}

	return false;
}


static bool ComputerMovesToBlock(
	ModelData		*md,
	TicTacToePlayer	aPlayer)
{
	TicTacToePlayer	theOpponent;
	unsigned int	h,
					v;

	theOpponent = (aPlayer == PlayerX ? PlayerO : PlayerX);

	for (h = 0; h < 3; h++)
		for (v = 0; v < 3; v++)
			if (md->itsGameOf.TicTacToe.itsBoard[h][v] == PlayerNone)
			{
				md->itsGameOf.TicTacToe.itsBoard[h][v] = theOpponent;
				if (CheckForWin(md, NULL) == theOpponent)
				{
					md->itsGameOf.TicTacToe.itsBoard[h][v] = aPlayer;
					return true;
				}
				else
					md->itsGameOf.TicTacToe.itsBoard[h][v] = PlayerNone;
			}

	return false;
}


static bool ComputerMovesRandomly(
	ModelData		*md,
	TicTacToePlayer	aPlayer)
{
	unsigned int	theNumEmptyCells,
					h,
					v,
					theCount;

	//	Count the number of empty cells.
	theNumEmptyCells = 0;
	for (h = 0; h < 3; h++)
		for (v = 0; v < 3; v++)
			if (md->itsGameOf.TicTacToe.itsBoard[h][v] == PlayerNone)
				theNumEmptyCells++;

	if (theNumEmptyCells == 0)
		return false;	//	should never occur

	//	Move randomly.
	theCount = rand() % theNumEmptyCells;

	for (h = 0; h < 3; h++)
		for (v = 0; v < 3; v++)
			if (md->itsGameOf.TicTacToe.itsBoard[h][v] == PlayerNone)
				if (theCount-- == 0)
				{
					md->itsGameOf.TicTacToe.itsBoard[h][v] = aPlayer;
					return true;
				}

	return false;	//	should never occur
}


static TicTacToePlayer CheckForWin(
	ModelData	*md,
	Placement	*aThreeInARow)	//	may be NULL
{
	signed int		h,	//	Signed integers allow for ±1 offsets.
					v,
					h1,
					v1,
					h2,
					v2;
	unsigned int	d;

	//	The winning three-in-a-row may go in any of four
	//	possible directions: E-W, SE-NW, S-N, SW-NE.
	static const signed int	theWinDirections[4][2]	= { {-1, 0}, {-1, 1}, { 0, 1}, { 1, 1} };
	static const double		theWinAngles[4]			= { 0.0,  -0.25*PI, 0.5*PI, 0.25*PI  },
							theWinLengths[4]		= { 1.0,  ROOT_TWO,  1.0,   ROOT_TWO };

	//	Consider each possible center (h,v) for the three-in-a-row.
	for (h = 0; h < 3; h++)
		for (v = 0; v < 3; v++)
		{
			//	Ignore empty cells.
			if (md->itsGameOf.TicTacToe.itsBoard[h][v] == PlayerNone)
				continue;

			//	Consider each possible direction.
			for (d = 0; d < 4; d++)
			{
#ifdef CYLINDER_HACK
Need to think some more to bring cylindrical tic-tac-toe up to date.
Clearly I need to somehow suppress the half-square offset.
				//	Center all vertical or diagonal wins
				//	in the middle row, so the three-in-a-row marker
				//	respects the cylinder's boundaries.
				if (theWinDirections[d][1] != 0 && v != 1)
					continue;
#endif
				//	Let (h1,v1) be the square in front of us
				//	in the given direction, and (h2,v2) be
				//	the square behind us in that direction.
				//	If both agree with (h,v), we've got a win.

				h1 = h + theWinDirections[d][0];
				v1 = v + theWinDirections[d][1];
				NormalizeCell(&h1, &v1, md->itsTopology);
				if (md->itsGameOf.TicTacToe.itsBoard[h1][v1] != md->itsGameOf.TicTacToe.itsBoard[h][v])
					continue;

				h2 = h - theWinDirections[d][0];
				v2 = v - theWinDirections[d][1];
				NormalizeCell(&h2, &v2, md->itsTopology);
				if (md->itsGameOf.TicTacToe.itsBoard[h2][v2] != md->itsGameOf.TicTacToe.itsBoard[h][v])
					continue;

				//	We've got a win!
				if (aThreeInARow != NULL)
				{
					aThreeInARow->itsH		= -0.5 + (1.0/3.0)*h;
					aThreeInARow->itsV		= -0.5 + (1.0/3.0)*v;
					aThreeInARow->itsFlip	= false;	//	ignored
					aThreeInARow->itsAngle	= theWinAngles[d];
					aThreeInARow->itsSizeH	= theWinLengths[d];	//	line length
					aThreeInARow->itsSizeV	= 0.08;				//	line thickness
				}
				return md->itsGameOf.TicTacToe.itsBoard[h][v];
			}
		}

	//	Nobody's won yet.
	return PlayerNone;
}


static void NormalizeCell(
	signed int		*h,
	signed int		*v,
	TopologyType	aTopology)
{
	if (*v == 3)
	{
		*v = 0;
		if (aTopology == TopologyKlein)
			*h = 3 - *h;
	}
	if (*v == -1)
	{
		*v = 2;
		if (aTopology == TopologyKlein)
			*h = 3 - *h;
	}

	if (*h > 2)
		*h -= 3;
	if (*h < 0)
		*h += 3;
}


static void TicTacToeSimulationUpdate(ModelData *md)
{
	switch (md->itsSimulationStatus)
	{
		case SimulationTicTacToeWaitToMove:
			//	Wait half a second before letting the computer move.
			if (md->itsSimulationElapsedTime >= 0.5)
			{
				SimulationEnd(md);
				ComputerMoves(md);
			}
			break;

		case SimulationTicTacToeWaitToDisplayWin:
			//	Pause briefly so the user may appreciate
			//	the move before the winning line appears.
			if (md->itsSimulationElapsedTime >= 0.5)
			{
				SimulationEnd(md);
				md->itsGameIsOver = true;						//	Record the win.
				PlayTheSound(SoundTicTacToeWin);				//	Start a victory sound playing.
				SimulationBegin(md, SimulationTicTacToeFlash);	//	Flash the three-in-a-row marker.
			}
			break;

		case SimulationTicTacToeFlash:
			md->itsFlashFlag = (((unsigned int) floor(10.0 * md->itsSimulationElapsedTime)) % 2) ? true : false;
			if (md->itsSimulationElapsedTime >= 1.4)
			{
				SimulationEnd(md);
				md->itsFlashFlag = false;
			}
			break;

		default:
			break;
	}
}


static void TicTacToeDraw(ModelData *md)
{
	unsigned int	h,
					v;
	Placement		theCellPlacement;

	//	Draw the background.
	DrawFundamentalDomain(md, md->itsGameOf.TicTacToe.itsTextureNameBackground, 3);

	//	Draw glide reflection axes if appropriate.
	DrawKleinAxes(	md,
					PREMULTIPLY_RGBA(0.75, 1.00, 0.50, 1.00),
					PREMULTIPLY_RGBA(0.25, 0.00, 0.50, 1.00) );

	//	Draw the nine cells.
	for (h = 0; h < 3; h++)
	{
		for (v = 0; v < 3; v++)
		{
			if (md->itsGameOf.TicTacToe.itsBoard[h][v] != PlayerNone)
			{
				theCellPlacement.itsH		= -0.5 + (1.0/3.0)*h;
				theCellPlacement.itsV		= -0.5 + (1.0/3.0)*v;
				theCellPlacement.itsFlip	= md->itsGameOf.TicTacToe.itsCellFlip[h][v];
				theCellPlacement.itsAngle	= 0.0;
				theCellPlacement.itsSizeH	= 1.0/3.0;
				theCellPlacement.itsSizeV	= 1.0/3.0;

				DrawSprite(	md,
							md->itsGameOf.TicTacToe.itsBoard[h][v] == PlayerX ?
								md->itsGameOf.TicTacToe.itsTextureNameX :
								md->itsGameOf.TicTacToe.itsTextureNameO,
							&theCellPlacement);
			}
		}
	}

	//	If the game is over, draw the winning three-in-a-row.
	if (md->itsGameIsOver
	 && ! md->itsFlashFlag )
		DrawSegment(md,
					PREMULTIPLY_RGBA(0.00, 0.75, 0.00, 1.00),
					md->itsGameOf.TicTacToe.itsTextureNameWin,
					md->itsGameOf.TicTacToe.itsWinLine.itsH,
					md->itsGameOf.TicTacToe.itsWinLine.itsV,
#ifdef CYLINDER_HACK
					0.8 * md->itsGameOf.TicTacToe.itsWinLine.itsSizeH,	//	line length
#else
					md->itsGameOf.TicTacToe.itsWinLine.itsSizeH,		//	line length
#endif
					md->itsGameOf.TicTacToe.itsWinLine.itsSizeV,		//	line thickness
					md->itsGameOf.TicTacToe.itsWinLine.itsAngle,
					fabs(fabs(md->itsGameOf.TicTacToe.itsWinLine.itsAngle) - 0.25*PI) < 0.01 ? 3 : 2);	//	angle is ±π/4 ?
}
