//	TorusGamesUtilities.c
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Common.h"
#include <math.h>
#include <wctype.h>	//	for towlower()


//	IS_PRECOMPOSED_HANGUL() tests for a precomposed Hangul syllable
//	(unicode range 0xAC00 - 0xD7A3).
#define IS_PRECOMPOSED_HANGUL(c)	((c) >= L'가' && (c) <= L'힣')

//	IS_COMPATIBILITY_JAMO tests for a Hangul compatibility jamo
//	(unicode range 0x3131 - 0x3163).
#define IS_COMPATIBILITY_JAMO(c)	((c) >= L'ㄱ' && (c) <= L'ㅣ')

//	IS_JAMO_CONSONANT() tests for a Hangul compatibility jamo consonant
//	(unicode range range 0x3131 - 0x314E).
#define IS_JAMO_CONSONANT(c)	((c) >= L'ㄱ' && (c) <= L'ㅎ')

//	Unicode's precomposed Hangul syllables are packed systematically
//	into a 3-dimensional array, indexed by the leading consonant (L),
//	vowel (V) and trailing consonant (T).  The trailing consonant may be
//	empty (with index value 0), but the leading consonant and the vowel
//	must be present.

//	The innermost blocks range over T, for fixed L and V.
#define BLOCK_SIZE_T	28
#define BLOCK_SIZE_VT	(21*28)


typedef enum
{
	FormatMIDI,
	FormatWAV
} SoundFormat;

typedef struct
{
	char		*itsFileName;	//	zero-terminated UTF-8 string
	SoundFormat	itsFormat;	
} SoundFileInfo;



static Char16	MakeDipthongJamo(Char16 aVowelA, Char16 aVowelB);
static Char16	ComposeHangulSyllable(HangulSyllable *aHangulSyllable);
static Char16	MakeLeadingJamo(Char16 aConsonant);
static Char16	MakeVowelJamo(Char16 aConsonant);
static Char16	MakeTrailingJamo(Char16 aConsonant);

	
void NormalizePlacement(
	Placement		*aPlacement,
	TopologyType	aTopology)
{
	//	Normalize aPlacement to lie within the fundamental domain,
	//	with coordinates in the range -0.5 to +0.5.

	//	Note that the rotation through itsAngle occurs before itsFlip,
	//	so toggling the latter has no effect on the former.

	while (aPlacement->itsV > +0.5)
	{
		aPlacement->itsV -= 1.0;
		if (aTopology == TopologyKlein)
		{
			aPlacement->itsH	= - aPlacement->itsH;
			aPlacement->itsFlip	= ! aPlacement->itsFlip;
		}
	}
	while (aPlacement->itsV < -0.5)
	{
		aPlacement->itsV += 1.0;
		if (aTopology == TopologyKlein)
		{
			aPlacement->itsH	= - aPlacement->itsH;
			aPlacement->itsFlip	= ! aPlacement->itsFlip;
		}
	}

	while (aPlacement->itsH > +0.5)
		aPlacement->itsH -= 1.0;
	while (aPlacement->itsH < -0.5)
		aPlacement->itsH += 1.0;
}


void NormalizeOffset(
	Offset			*anOffset,
	TopologyType	aTopology)
{
	//	Normalize anOffset to lie within the fundamental domain,
	//	with coordinates in the range -0.5 to +0.5.

	//	NormalizeOffset() differs from NormalizePlacement()
	//	not only in setting fewer fields, but also in never
	//	negating itsH in a Klein bottle.

	while (anOffset->itsV > +0.5)
	{
		anOffset->itsV -= 1.0;
		if (aTopology == TopologyKlein)
			anOffset->itsFlip = ! anOffset->itsFlip;
	}
	while (anOffset->itsV < -0.5)
	{
		anOffset->itsV += 1.0;
		if (aTopology == TopologyKlein)
			anOffset->itsFlip = ! anOffset->itsFlip;
	}

	while (anOffset->itsH > +0.5)
		anOffset->itsH -= 1.0;
	while (anOffset->itsH < -0.5)
		anOffset->itsH += 1.0;
}


double ShortestDistance(
	double			h0,
	double			v0,
	double			h1,
	double			v1,
	TopologyType	aTopology,
	bool			*aMirrorFlag)	//	may be NULL
{
	//	Compute the distance between two points in the torus
	//	or Klein bottle.  The coordinates (h0, v0) and (h1, v1)
	//	may lie anywhere in the universal cover.   Conceptually
	//	each point projects down to a point within the fundamental domain,
	//	i.e. with coordinates in the range from -0.5 to +0.5.

	//	Note:  This algorithm finds nearest images iff the objects
	//	are fairly close.  For more distant pairs of objects
	//	the closest images might involve |v1 - v0| > 1/2.

	double	dh,
			dv;

	while (v1 - v0 > +0.5)
	{
		v1 -= 1.0;
		if (aTopology == TopologyKlein)
		{
			h1 = -h1;
			if (aMirrorFlag != NULL)
				*aMirrorFlag = ! *aMirrorFlag;
		}
	}
	while (v1 - v0 < -0.5)
	{
		v1 += 1.0;
		if (aTopology == TopologyKlein)
		{
			h1 = -h1;
			if (aMirrorFlag != NULL)
				*aMirrorFlag = ! *aMirrorFlag;
		}
	}

	while (h1 - h0 > +0.5)
		h1 -= 1.0;
	while (h1 - h0 < -0.5)
		h1 += 1.0;

	dh = h1 - h0;
	dv = v1 - v0;

	return sqrt(dh*dh + dv*dv);
}


void PlayTheSound(SoundType aSound)
{
	static SoundFileInfo	theSoundFiles[NumSounds] =
							{
								{"TicTacToeMove.mid",		FormatMIDI},
								{"TicTacToeWin.mid",		FormatMIDI},
								{"MazeMunch.wav",			FormatWAV },
								{"CrosswordKeypress.mid",	FormatMIDI},	//	touch-based UI only
								{"CrosswordComplete.mid",	FormatMIDI},
								{"WordSearchComplete.mid",	FormatMIDI},
								{"JigsawJoin.mid",			FormatMIDI},
								{"JigsawComplete.mid",		FormatMIDI},
								{"ChessWrongColor.mid",		FormatMIDI},
								{"ChessCheck.mid",			FormatMIDI},
								{"ChessCheckmate.mid",		FormatMIDI},
								{"ChessStalemate.mid",		FormatMIDI},
								{"PoolCue.wav",				FormatWAV },
								{"PoolClick.wav",			FormatWAV },
								{"PoolSink.wav",			FormatWAV },
								{"PoolLoss.mid",			FormatMIDI},
								{"PoolWin.mid",				FormatMIDI},
								{"ApplesLoss.mid",			FormatMIDI},
								{"ApplesWin.mid",			FormatMIDI}
							};

	if (gPlaySounds)
	{
		switch (theSoundFiles[aSound].itsFormat)
		{
			case FormatMIDI:
				PlayMIDI(theSoundFiles[aSound].itsFileName, aSound);
				break;

			case FormatWAV:
				PlayWAV(theSoundFiles[aSound].itsFileName, aSound);
				break;
		}
	}
}


Char16 ToLowerCase(Char16 aLetter)
{
	//	The built-in towlower() function misses Cyrillic.
	//	Handle the basic letters А-Я here,
	//	then handle Ё → ё in the switch statement below.
	//	If I ever need to accommodate additional Cyrillic letters,
	//	say Ђ or Љ, I'll need to add them to the switch as well.
	if (aLetter >= L'А' && aLetter <= L'Я')
		return aLetter + (L'а' - L'А');

	//	The built-in towlower() function misses certain accented letters.
	//	Handle known flaws manually.
	switch (aLetter)
	{
		case L'Đ':	return L'đ';
		case L'Ý':	return L'ý';
		case L'Á':	return L'á';
		case L'Ú':	return L'ú';
		case L'Ĳ':	return L'ĳ';
		
		case L'Ё':	return L'ё';	//	Cyrillic
		
		default:	return towlower(aLetter);
	}
}


Char16 BaseLetter(Char16 aLetter)
{
	//	Remove any accents from aLetter and return the "base letter".
	//	For example, 'é', 'è' and 'e' map to 'e'.
	
	//	For comparing kana in Japanese crossword puzzles,
	//	let the hiragana form be the "base letter".
	//	We may assume that small characters have already been converted
	//	to large ones, and that half-width katakana have already been
	//	converted to full-width.
	//	In other words, only full-width large kana should arrive here,
	//	so mapping katakana into hiragana will make the base letter unique.
	
	if (aLetter >= L'ぁ' && aLetter <= L'ん')	//	hiragana
		return aLetter;
	
	if (aLetter >= L'ァ' && aLetter <= L'ン')	//	katakana
		return aLetter - (L'ァ' - L'ぁ');		//		shift to parallel hiragana

	switch (aLetter)
	{
		case L'à':
		case L'á':
		case L'â':
		case L'ã':
		case L'ä':
		case L'ă':
		case L'ạ':
		case L'ả':
		case L'ấ':
		case L'ầ':
		case L'ẩ':
		case L'ẫ':
		case L'ậ':
		case L'ắ':
		case L'ằ':
		case L'ẳ':
		case L'ẵ':
		case L'ặ':
			return L'a';

		case L'è':
		case L'é':
		case L'ê':
		case L'ë':
		case L'ẹ':
		case L'ẻ':
		case L'ẽ':
		case L'ế':
		case L'ề':
		case L'ể':
		case L'ễ':
		case L'ệ':
			return L'e';

		case L'ё':			//	Cyrillic 'yo'
			return L'е';	//	Cyrillic 'ye'

		case L'ì':
		case L'í':
		case L'î':
		case L'ï':
		case L'ĩ':
		case L'ỉ':
		case L'ị':
			return L'i';

		case L'ò':
		case L'ó':
		case L'ô':
		case L'õ':
		case L'ö':
		case L'ő':
		case L'ơ':
		case L'ọ':
		case L'ỏ':
		case L'ố':
		case L'ồ':
		case L'ổ':
		case L'ỗ':
		case L'ộ':
		case L'ớ':
		case L'ờ':
		case L'ở':
		case L'ỡ':
		case L'ợ':
			return L'o';

		case L'ù':
		case L'ú':
		case L'û':
		case L'ü':
		case L'ũ':
		case L'ű':
		case L'ư':
		case L'ụ':
		case L'ủ':
		case L'ứ':
		case L'ừ':
		case L'ử':
		case L'ữ':
		case L'ự':
			return L'u';

		case L'ý':
		case L'ỳ':
		case L'ỵ':
		case L'ỷ':
		case L'ỹ':
			return L'y';

		case L'ç':
			return L'c';

		case L'đ':
			return L'd';

		case L'ñ':
			return L'n';

		default:
			return aLetter;
	}
}


Char16 PromoteHalfWidthKatakana(Char16 aCharacter)
{
	//	The following half-width characters occupy the Unicode range
	//	from 0xFF61 though 0xFF9F
	//
	//		｡｢｣､･ｦｧｨｩｪｫｬｭｮｯｰｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐﾑﾒﾓﾔﾕﾖﾗﾘﾙﾚﾛﾜﾝﾞﾟ
	//
	//	Map them to their full-width equivalents
	//
	//		。「」、・ヲァィゥェォャュョッーアイウエオカキクケコサシスセソタチツテトナニヌネノハヒフヘホマミムメモヤユヨラリルレロワン゛゜
	//
	
	//	Technical note:  L"…" gives a string of wchar_t's, which will be
	//	2-byte characters on Windows but 4-byte characters on Mac OS.
	static wchar_t	theFullWidthEquivalents[] = L"。「」、・ヲァィゥェォャュョッーアイウエオカキクケコサシスセソタチツテトナニヌネノハヒフヘホマミムメモヤユヨラリルレロワン゛゜";

	if (aCharacter >= L'｡'
	 && aCharacter <= L'ﾟ')
	{
		aCharacter = (Char16) theFullWidthEquivalents[aCharacter - L'｡'];
	}
	
	return aCharacter;
}


Char16 PromoteSmallKana(Char16 aCharacter)
{
	switch (aCharacter)
	{
		//	hiragana
		case L'ぁ':	return L'あ';
		case L'ぃ':	return L'い';
		case L'ぅ':	return L'う';
		case L'ぇ':	return L'え';
		case L'ぉ':	return L'お';
		case L'っ':	return L'つ';
		case L'ゃ':	return L'や';
		case L'ゅ':	return L'ゆ';
		case L'ょ':	return L'よ';
		case L'ゎ':	return L'わ';
		
		//	katakana
		case L'ァ':	return L'ア';
		case L'ィ':	return L'イ';
		case L'ゥ':	return L'ウ';
		case L'ェ':	return L'エ';
		case L'ォ':	return L'オ';
		case L'ッ':	return L'ツ';
		case L'ャ':	return L'ヤ';
		case L'ュ':	return L'ユ';
		case L'ョ':	return L'ヨ';
		case L'ヮ':	return L'ワ';
		case L'ヵ':	return L'カ';	//	not sure what small カ is for
		case L'ヶ':	return L'ケ';	//	not sure what small ケ is for

		//	half-width katakana (seldom used?)
		case L'ｧ':	return L'ｱ';  
		case L'ｨ':	return L'ｲ';
		case L'ｩ':	return L'ｳ';
		case L'ｪ':	return L'ｴ';
		case L'ｫ':	return L'ｵ';
		case L'ｯ':	return L'ﾂ';
		case L'ｬ':	return L'ﾔ';
		case L'ｭ':	return L'ﾕ';
		case L'ｮ':	return L'ﾖ';
		
		default:	return aCharacter;
	}
}


Char16 ComposeHiragana(
	Char16		aCharacter,
	Char16		*aPendingCharacter)
{
	Char16	thePreviousPendingCharacter;
	
	//	Read *aPendingCharacter then clear it immediately, 
	//	so we don't have to worry about clearing it
	//	in the various cases of the switch statement.
	thePreviousPendingCharacter = *aPendingCharacter;
	*aPendingCharacter = NO_PENDING_CHARACTER;

	switch (thePreviousPendingCharacter)
	{
		case NO_PENDING_CHARACTER:
			switch (aCharacter)
			{
				case L'a':	return L'あ';
				case L'i':	return L'い';
				case L'u':	return L'う';
				case L'e':	return L'え';
				case L'o':	return L'お';
				
				case L'b':
				case L'c':
				case L'd':
				case L'f':
				case L'g':
				case L'h':
				case L'j':
				case L'k':
				case L'm':
				case L'n':
				case L'p':
				case L'r':
				case L's':
				case L't':
				case L'v':
				case L'w':
				case L'y':
				case L'z':
					*aPendingCharacter = aCharacter;
					return L' ';
				
				default:
					return aCharacter;
			}
		
		case L'b':
			switch (aCharacter)
			{
				case L'a':	return L'ば';
				case L'i':	return L'び';
				case L'u':	return L'ぶ';
				case L'e':	return L'べ';
				case L'o':	return L'ぼ';
				default:	return aCharacter;
			}
		
		case L'c':
			switch (aCharacter)
			{
				case L'h':
					*aPendingCharacter = L'č';	//	facilitates "chi" → 'ち'
					return L' ';

				case L'i':	return L'ち';
				default:	return aCharacter;
			}
		
		case L'č':	//	facilitates "chi" → 'ち'
			switch (aCharacter)
			{
				case L'i':	return L'ち';
				default:	return aCharacter;
			}
		
		case L'd':
			switch (aCharacter)
			{
				case L'a':	return L'だ';
				case L'i':	return L'ぢ';
				case L'u':	return L'づ';
				case L'e':	return L'で';
				case L'o':	return L'ど';
				default:	return aCharacter;
			}
		
		case L'f':
			switch (aCharacter)
			{
				case L'u':	return L'ふ';
				default:	return aCharacter;
			}
		
		case L'g':
			switch (aCharacter)
			{
				case L'a':	return L'が';
				case L'i':	return L'ぎ';
				case L'u':	return L'ぐ';
				case L'e':	return L'げ';
				case L'o':	return L'ご';
				default:	return aCharacter;
			}
		
		case L'h':
			switch (aCharacter)
			{
				case L'a':	return L'は';
				case L'i':	return L'ひ';
				case L'u':	return L'ふ';
				case L'e':	return L'へ';
				case L'o':	return L'ほ';
				default:	return aCharacter;
			}
		
		case L'j':
			switch (aCharacter)
			{
				case L'i':	return L'じ';
				default:	return aCharacter;
			}
		
		case L'k':
			switch (aCharacter)
			{
				case L'a':	return L'か';
				case L'i':	return L'き';
				case L'u':	return L'く';
				case L'e':	return L'け';
				case L'o':	return L'こ';
				default:	return aCharacter;
			}
		
		case L'm':
			switch (aCharacter)
			{
				case L'a':	return L'ま';
				case L'i':	return L'み';
				case L'u':	return L'む';
				case L'e':	return L'め';
				case L'o':	return L'も';
				default:	return aCharacter;
			}
		
		case L'n':
			switch (aCharacter)
			{
				case L'a':	return L'な';
				case L'i':	return L'に';
				case L'u':	return L'ぬ';
				case L'e':	return L'ね';
				case L'o':	return L'の';
				case L'n':	return L'ん';
				case L' ':	return L'ん';
				default:	return aCharacter;
			}
		
		case L'p':
			switch (aCharacter)
			{
				case L'a':	return L'ぱ';
				case L'i':	return L'ぴ';
				case L'u':	return L'ぷ';
				case L'e':	return L'ぺ';
				case L'o':	return L'ぽ';
				default:	return aCharacter;
			}
		
		case L'r':
			switch (aCharacter)
			{
				case L'a':	return L'ら';
				case L'i':	return L'り';
				case L'u':	return L'る';
				case L'e':	return L'れ';
				case L'o':	return L'ろ';
				default:	return aCharacter;
			}
		
		case L's':
			switch (aCharacter)
			{
				case L'a':	return L'さ';
				case L'i':	return L'し';
				case L'u':	return L'す';
				case L'e':	return L'せ';
				case L'o':	return L'そ';

				case L'h':
					*aPendingCharacter = L'š';	//	facilitates "shi" → 'し'
					return L' ';

				default:	return aCharacter;
			}
		
		case L'š':	//	facilitates "shi" → 'し'
			switch (aCharacter)
			{
				case L'i':	return L'し';
				default:	return aCharacter;
			}
		
		case L't':
			switch (aCharacter)
			{
				case L'a':	return L'た';
				case L'i':	return L'ち';
				case L'u':	return L'つ';
				case L'e':	return L'て';
				case L'o':	return L'と';
				case L's':	return L'つ';
				default:	return aCharacter;
			}
		
		case L'w':
			switch (aCharacter)
			{
				case L'a':	return L'わ';
				case L'o':	return L'を';
				default:	return aCharacter;
			}
		
		case L'y':
			switch (aCharacter)
			{
				case L'a':	return L'や';
				case L'u':	return L'ゆ';
				case L'o':	return L'よ';
				default:	return aCharacter;
			}
		
		case L'z':
			switch (aCharacter)
			{
				case L'a':	return L'ざ';
				case L'i':	return L'じ';
				case L'u':	return L'ず';
				case L'e':	return L'ぜ';
				case L'o':	return L'ぞ';
				default:	return aCharacter;
			}
		
		default:
			return aCharacter;
	}
}


Char16 ComposeHangul(
	Char16			aCharacter,
	Char16			*aPendingCharacter,
	HangulSyllable	*aPendingHangulSyllable)
{
	Char16	theComposedCharacter,
			theReinterpretedJamo;

	if (IS_PRECOMPOSED_HANGUL(aCharacter))	//	Is aCharacter a complete, precomposed Hangul syllable?
	{
		//	The Windows Korean IME passes fully formed Hangul syllables,
		//	unlike the Mac OS Korean Input Method which passes jamo.
		*aPendingHangulSyllable	= EMPTY_HANGUL_SYLLABLE;
		theComposedCharacter	= aCharacter;
	}
	else
	if (aCharacter == ' ' || aCharacter == '\t' || aCharacter == '\r' || aCharacter == '\n')
	{
		//	Treat a space, tab, return or newline
		//	as a request to complete the pending syllable.
		theComposedCharacter	= ComposeHangulSyllable(aPendingHangulSyllable);
		*aPendingHangulSyllable	= EMPTY_HANGUL_SYLLABLE;

		//	The Win32 Korean Input Method Editor (IME) passes 
		//	precomposed Hangul syllables, which is fine, but then it also 
		//	passes along the character (typically a return, newline, tab or space)
		//	that the user typed to terminate the word's last syllable.
		//	To avoid overwriting the next crossword cell with an unwanted space,
		//	return 0 in cases where a return or newline arrived when no jamo were pending.
		if ((aCharacter == '\r' || aCharacter == '\n') && theComposedCharacter == ' ')
			theComposedCharacter = 0;
	}
	else
	if ( ! IS_COMPATIBILITY_JAMO(aCharacter))
	{
		//	Aside from precomposed Hangul syllables and white space, 
		//	the only legal input is unicode's "Hangul compatibility jamo".
		//	Flag illegal input with a '?'.
		*aPendingHangulSyllable	= EMPTY_HANGUL_SYLLABLE;
		theComposedCharacter	= '?';
	}
	else	//	Break into cases according to whether we're expecting
			//	a leading consonant, a vowel or a trailing consonant.
	if (aPendingHangulSyllable->itsLeadingConsonant == 0)
	{
		//	aPendingHangulSyllable is empty.
		
		if (IS_JAMO_CONSONANT(aCharacter))
		{
			aPendingHangulSyllable->itsLeadingConsonant = aCharacter;
			theComposedCharacter = ' ';	//	syllable remains incomplete
		}
		else	//	aCharacter is a vowel jamo
		{
			aPendingHangulSyllable->itsLeadingConsonant	= L'ᄋ';
			aPendingHangulSyllable->itsVowel			= aCharacter;
			theComposedCharacter = ' ';	//	syllable remains (tentatively) incomplete
		}
	}
	else
	if (aPendingHangulSyllable->itsVowel == 0)
	{
		//	aPendingHangulSyllable already contains a leading consonant,
		//	but no vowel.
		
		if (IS_JAMO_CONSONANT(aCharacter))
		{
			//	Something is wrong, because a Hangul syllable should
			//	have a simple (non-compound) leading consonant.
			//	Overwrite the old leading consonant and hope for the best.
			aPendingHangulSyllable->itsLeadingConsonant = aCharacter;
			theComposedCharacter = ' ';	//	syllable remains incomplete
		}
		else	//	aCharacter is a vowel jamo
		{
			aPendingHangulSyllable->itsVowel = aCharacter;
			theComposedCharacter = ' ';	//	syllable remains (tentatively) incomplete
		}
	}
	else
	if (aPendingHangulSyllable->itsTrailingConsonant == 0)
	{
		//	aPendingHangulSyllable already contains a leading consonant
		//	and a vowel, but no trailing consonant.
		
		if (IS_JAMO_CONSONANT(aCharacter))
		{
			aPendingHangulSyllable->itsTrailingConsonant = aCharacter;

			//	The syllable remains nominally incomplete,
			//	because we're not sure whether the trailing consonant
			//	really belongs here, or whether it's intended
			//	as the leading consonant for the next syllable.
			theComposedCharacter = ' ';
		}
		else	//	aCharacter is a vowel jamo
		{
			aPendingHangulSyllable->itsVowel = MakeDipthongJamo(aPendingHangulSyllable->itsVowel, aCharacter);
			theComposedCharacter = ' ';	//	syllable remains (tentatively) incomplete
		}
	}
	else
	{
		//	aPendingHangulSyllable already contains a leading consonant,
		//	a vowel and a trailing consonant.
		
		if (IS_JAMO_CONSONANT(aCharacter))
		{
			//	Here we go against the usual Hangul input conventions.
			//	Conventionally, two trailing jamo, for example ㄹ and ㄱ,
			//	get merged to a single trailing jamo ㄺ.  However, such combinations
			//	seem to be rare, and in particular do not occur in the Korean
			//	Torus Games crossword puzzles.  So rather than forcing the user
			//	to type a vowel to rescue the ㄱ, it seems friendlier to simply
			//	let the ㄱ begin a new syllable immedately.

			theComposedCharacter = ComposeHangulSyllable(aPendingHangulSyllable);

			aPendingHangulSyllable->itsLeadingConsonant		= aCharacter;
			aPendingHangulSyllable->itsVowel				= 0x0000;
			aPendingHangulSyllable->itsTrailingConsonant	= 0x0000;
		}
		else	//	aCharacter is a vowel jamo
		{
			//	Re-interpret the trailing jamo of aPendingHangulSyllable
			//	as the leading jamo in a new syllable.

			theReinterpretedJamo = aPendingHangulSyllable->itsTrailingConsonant;

			aPendingHangulSyllable->itsTrailingConsonant = 0x0000;

			theComposedCharacter = ComposeHangulSyllable(aPendingHangulSyllable);

			aPendingHangulSyllable->itsLeadingConsonant		= theReinterpretedJamo;
			aPendingHangulSyllable->itsVowel				= aCharacter;
			aPendingHangulSyllable->itsTrailingConsonant	= 0x0000;
		}
	}

	*aPendingCharacter = ComposeHangulSyllable(aPendingHangulSyllable);

	return theComposedCharacter;
}

static Char16 MakeDipthongJamo(	//	output in unicode's Hangul Compatibility Jamo range 0x3131 - 0x3163
	Char16	aVowelA,			//	 input in unicode's Hangul Compatibility Jamo range 0x3131 - 0x3163
	Char16	aVowelB)
{
	switch (aVowelA)
	{
		case L'ㅗ':
			switch (aVowelB)
			{
				case L'ㅏ': return L'ㅘ';
				case L'ㅐ': return L'ㅙ';
				case L'ㅣ': return L'ㅚ';
			}
			break;

		case L'ㅜ':
			switch (aVowelB)
			{
				case L'ㅓ': return L'ㅝ';
				case L'ㅔ': return L'ㅞ';
				case L'ㅣ': return L'ㅟ';
			}
			break;

		case L'ㅡ':
			switch (aVowelB)
			{
				case L'ㅣ': return L'ㅢ';
			}
			break;
	}
	
	//	If the program gets to this point,
	//	it means aVowelA and aVowelB don't normally combine.
	//	Return aVowelB and hope for the best.
	return aVowelB;
}

static Char16 ComposeHangulSyllable(	//	output typically in unicode's precomposed Hangul range 0xAC00 - 0xD7A3
	HangulSyllable	*aHangulSyllable)
{
	Char16			theLeadingJamo,		//	in unicode's Hangul Jamo  leading subrange 0x1100 - 0x1112
					theVowelJamo,		//	in unicode's Hangul Jamo    vowel subrange 0x1161 - 0x1175
					theTrailingJamo;	//	in unicode's Hangul Jamo trailing subrange 0x11A8 - 0x11C2
	unsigned int	theLeadingIndex,
					theVowelIndex,
					theTrailingIndex;

	if (aHangulSyllable->itsLeadingConsonant == 0)
	{
		//	aHangulSyllable is empty.
		return ' ';	//	in unicode's ASCII range
	}
	else
	if (aHangulSyllable->itsVowel == 0)
	{
		//	aHangulSyllable contains a leading consonant, but no vowel.
		return aHangulSyllable->itsLeadingConsonant;	//	in unicode's Hangul Compatibility Jamo range
	}
	else
	{
		//	aHangulSyllable already contains a leading consonant and a vowel,
		//	but may or may not contain a trailing consonant.

		theLeadingJamo	= MakeLeadingJamo(aHangulSyllable->itsLeadingConsonant);
		theVowelJamo	= MakeVowelJamo(aHangulSyllable->itsVowel);
		theTrailingJamo	= MakeTrailingJamo(aHangulSyllable->itsTrailingConsonant);	//	handles 0 (to trailing jamo) correctly

		theLeadingIndex		= theLeadingJamo  - 0x1100;
		theVowelIndex		= theVowelJamo    - 0x1161;
		theTrailingIndex	= theTrailingJamo - 0x11A7;
		
		return 0xAC00 + theLeadingIndex*BLOCK_SIZE_VT + theVowelIndex*BLOCK_SIZE_T + theTrailingIndex;
	}
}

static Char16 MakeLeadingJamo(	//	output in unicode's Hangul Jamo leading subrange 0x1100 - 0x1112
	Char16	aConsonant)			//	 input in unicode's Hangul Compatibility Jamo range 0x3131 - 0x3163
{
	switch (aConsonant)
	{
		case L'ㄱ':	return 0x1100;	//	return a leading ㄱ
		case L'ㄲ':	return 0x1101;	//	return a leading ㄲ
		case L'ㄴ':	return 0x1102;	//	return a leading ㄴ
		case L'ㄷ':	return 0x1103;	//	return a leading ㄷ
		case L'ㄸ':	return 0x1104;	//	return a leading ㄸ
		case L'ㄹ':	return 0x1105;	//	return a leading ㄹ
		case L'ㅁ':	return 0x1106;	//	return a leading ㅁ
		case L'ㅂ':	return 0x1107;	//	return a leading ㅂ
		case L'ㅃ':	return 0x1108;	//	return a leading ㅃ
		case L'ㅅ':	return 0x1109;	//	return a leading ㅅ
		case L'ㅆ':	return 0x110A;	//	return a leading ㅆ
		case L'ㅇ':	return 0x110B;	//	return a leading ㅇ
		case L'ㅈ':	return 0x110C;	//	return a leading ㅈ
		case L'ㅉ':	return 0x110D;	//	return a leading ㅉ
		case L'ㅊ':	return 0x110E;	//	return a leading ㅊ
		case L'ㅋ':	return 0x110F;	//	return a leading ㅋ
		case L'ㅌ':	return 0x1110;	//	return a leading ㅌ
		case L'ㅍ':	return 0x1111;	//	return a leading ㅍ
		case L'ㅎ':	return 0x1112;	//	return a leading ㅎ

		default:	return 0x1100;	//	Input is not valid leading consonant --
									//	  return a leading ㄱ and hope for the best.
	}
}

static Char16 MakeVowelJamo(	//	output in unicode's Hangul Jamo vowel subrange 0x1161 - 0x1175
	Char16	aConsonant)			//	 input in unicode's Hangul Compatibility Jamo range 0x3131 - 0x3163
{
	if (aConsonant >= L'ㅏ' && aConsonant <= L'ㅣ')	//	aConsonant is a valid vowel, so...
	{
		return 0x1161 + (aConsonant -  L'ㅏ');		//	...shift it to the Hangul Jamo range.
	}
	else				//	aConsonant is not a valid vowel, so...
	{
		return 0x1161;	//	...return an ㅏ and hope for the best.
	}
}

static Char16 MakeTrailingJamo(	//	output in unicode's Hangul Jamo trailing subrange 0x11A8 - 0x11C2
	Char16	aConsonant)			//	 input in unicode's Hangul Compatibility Jamo range 0x3131 - 0x3163
{
	switch (aConsonant)
	{
		case 0x0000:	return 0x11A7;	//	No trailing consonant is present.
										//	The unicode precomposed Hangul code tables
										//	will handle this case correctly.

		case L'ㄱ':	return 0x11A8;	//	return a trailing ㄱ
		case L'ㄲ':	return 0x11A9;	//	return a trailing ㄲ
		case L'ㄳ':	return 0x11AA;	//	return a trailing ㄳ
		case L'ㄴ':	return 0x11AB;	//	return a trailing ㄴ
		case L'ㄵ':	return 0x11AC;	//	return a trailing ㄵ
		case L'ㄶ':	return 0x11AD;	//	return a trailing ㄶ
		case L'ㄷ':	return 0x11AE;	//	return a trailing ㄷ
		case L'ㄹ':	return 0x11AF;	//	return a trailing ㄹ
		case L'ㄺ':	return 0x11B0;	//	return a trailing ㄺ
		case L'ㄻ':	return 0x11B1;	//	return a trailing ㄻ
		case L'ㄼ':	return 0x11B2;	//	return a trailing ㄼ
		case L'ㄽ':	return 0x11B3;	//	return a trailing ㄽ
		case L'ㄾ':	return 0x11B4;	//	return a trailing ㄾ
		case L'ㄿ':	return 0x11B5;	//	return a trailing ㄿ
		case L'ㅀ':	return 0x11B6;	//	return a trailing ㅀ
		case L'ㅁ':	return 0x11B7;	//	return a trailing ㅁ
		case L'ㅂ':	return 0x11B8;	//	return a trailing ㅂ
		case L'ㅄ':	return 0x11B9;	//	return a trailing ㅄ
		case L'ㅅ':	return 0x11BA;	//	return a trailing ㅅ
		case L'ㅆ':	return 0x11BB;	//	return a trailing ㅆ
		case L'ㅇ':	return 0x11BC;	//	return a trailing ㅇ
		case L'ㅈ':	return 0x11BD;	//	return a trailing ㅈ
		case L'ㅊ':	return 0x11BE;	//	return a trailing ㅊ
		case L'ㅋ':	return 0x11BF;	//	return a trailing ㅋ
		case L'ㅌ':	return 0x11C0;	//	return a trailing ㅌ
		case L'ㅍ':	return 0x11C1;	//	return a trailing ㅍ
		case L'ㅎ':	return 0x11C2;	//	return a trailing ㅎ

		default:	return 0x1100;	//	Input is not valid trailing consonant --
									//	  return a trailing ㄱ and hope for the best.
	}
}

Char16 QWERTYtoDubeolshik(
	Char16 aCharacter)
{
	//	Interpret a keystroke on a QWERTY keyboard layout as if 
	//	the same key had been pressed on a Dubeolshik (두벌식) layout.
	switch (aCharacter)
	{
		case 'q':	return L'ㅂ';
		case 'w':	return L'ㅈ';
		case 'e':	return L'ㄷ';
		case 'r':	return L'ㄱ';
		case 't':	return L'ㅅ';
		case 'y':	return L'ㅛ';
		case 'u':	return L'ㅕ';
		case 'i':	return L'ㅑ';
		case 'o':	return L'ㅐ';
		case 'p':	return L'ㅔ';

		case 'a':	return L'ㅁ';
		case 's':	return L'ㄴ';
		case 'd':	return L'ㅇ';
		case 'f':	return L'ㄹ';
		case 'g':	return L'ㅎ';
		case 'h':	return L'ㅗ';
		case 'j':	return L'ㅓ';
		case 'k':	return L'ㅏ';
		case 'l':	return L'ㅣ';

		case 'z':	return L'ㅋ';
		case 'x':	return L'ㅌ';
		case 'c':	return L'ㅊ';
		case 'v':	return L'ㅍ';
		case 'b':	return L'ㅠ';
		case 'n':	return L'ㅜ';
		case 'm':	return L'ㅡ';

		case 'Q':	return L'ㅃ';
		case 'W':	return L'ㅉ';
		case 'E':	return L'ㄸ';
		case 'R':	return L'ㄲ';
		case 'T':	return L'ㅆ';
		case 'O':	return L'ㅒ';
		case 'P':	return L'ㅖ';

		default:	return aCharacter;
	}
}


Char16 ComposeCyrillic(
	Char16	anExistingCharacter,	//	already in the puzzle cell
	Char16	aNewCharacter)			//	newly arrived
{
	switch (aNewCharacter)
	{
		case L'a':
			switch (anExistingCharacter)
			{
				case L'ы':	return L'я';
				default:	return L'а';
			}		
		case L'b':	return L'б';
		case L'c':
			switch (anExistingCharacter)
			{
				case L'ш':	return L'щ';
				default:	return L'ч';
			}		
		case L'd':	return L'д';
		case L'e':
			switch (anExistingCharacter)
			{
				case L'е':	return L'э';
				case L'э':	return L'е';
				default:	return L'е';
			}		
		case L'f':	return L'ф';
		case L'g':	return L'г';
		case L'h':
			switch (anExistingCharacter)
			{
				case L'ч':	return L'ч';
				case L'с':	return L'ш';
				case L'щ':	return L'щ';
				case L'з':	return L'ж';
				case L'к':	return L'х';
				default:	return L'?';
			}		
		case L'i':	return L'и';
		case L'j':	return L'й';
		case L'k':	return L'к';
		case L'l':	return L'л';
		case L'm':	return L'м';
		case L'n':	return L'н';
		case L'o':
			switch (anExistingCharacter)
			{
				case L'ы':	return L'ё';
				default:	return L'о';
			}		
		case L'p':	return L'п';
		case L'r':	return L'р';
		case L's':
			switch (anExistingCharacter)
			{
				case L'т':	return L'ц';
				default:	return L'с';
			}		
		case L't':	return L'т';
		case L'u':
			switch (anExistingCharacter)
			{
				case L'ы':	return L'ю';
				default:	return L'у';
			}		
		case L'v':	return L'в';
		case L'w':	return L'ш';
		case L'x':	return L'х';
		case L'y':	return L'ы';
		case L'z':	return L'з';
		case L':':	return L'ё';
		case L'\'':	return L'ь';
		case L'\"':	return L'ъ';

		//	All other characters -- including
		//	directly-typed Cyrillic -- pass through unchanged.
		default:	return aNewCharacter;
	}
}


Char16 ComposeLatin(
	Char16		aCharacter,
	Char16		*aPendingCharacter)
{
	Char16	thePreviousPendingCharacter;
	
	//	Read *aPendingCharacter then clear it immediately, 
	//	so we don't have to worry about clearing it
	//	in the various cases of the switch statement.
	thePreviousPendingCharacter = *aPendingCharacter;
	*aPendingCharacter = NO_PENDING_CHARACTER;

	switch (thePreviousPendingCharacter)
	{
		case NO_PENDING_CHARACTER:
			switch (aCharacter)
			{
				case L'\'':	*aPendingCharacter = L'´';	return L' ';
				case L'`':	*aPendingCharacter = L'`';	return L' ';
				case L'^':	*aPendingCharacter = L'ˆ';	return L' ';
				case L':':	*aPendingCharacter = L'¨';	return L' ';
				case L'"':	*aPendingCharacter = L'˝';	return L' ';
				case L',':	*aPendingCharacter = L'¸';	return L' ';
				case L'~':	*aPendingCharacter = L'˜';	return L' ';
				
				default:
					return aCharacter;
			}
		
		case L'´':
			switch (aCharacter)
			{
				case L'a':	return L'á';
				case L'e':	return L'é';
				case L'i':	return L'í';
				case L'o':	return L'ó';
				case L'u':	return L'ú';
				default:	return aCharacter;
			}
		
		case L'`':
			switch (aCharacter)
			{
				case L'a':	return L'à';
				case L'e':	return L'è';
				case L'i':	return L'ì';
				case L'o':	return L'ò';
				case L'u':	return L'ù';
				default:	return aCharacter;
			}
		
		case L'ˆ':
			switch (aCharacter)
			{
				case L'a':	return L'â';
				case L'e':	return L'ê';
				case L'i':	return L'î';
				case L'o':	return L'ô';
				case L'u':	return L'û';
				default:	return aCharacter;
			}
		
		case L'¨':
			switch (aCharacter)
			{
				case L'a':	return L'ä';
				case L'e':	return L'ë';
				case L'i':	return L'ï';
				case L'o':	return L'ö';
				case L'u':	return L'ü';
				default:	return aCharacter;
			}
		
		case L'˝':
			switch (aCharacter)
			{
				case L'o':	return L'ő';
				case L'u':	return L'ű';
				default:	return aCharacter;
			}
		
		case L'¸':
			switch (aCharacter)
			{
				case L'c':	return L'ç';
				default:	return aCharacter;
			}
		
		case L'˜':
			switch (aCharacter)
			{
				case L'a':	return L'ã';
				case L'n':	return L'ñ';
				case L'o':	return L'õ';
				default:	return aCharacter;
			}
		
		default:
			return aCharacter;
	}
}


ErrorText GetAbsolutePath(
	const char		*aDirectory,	//	input,  zero-terminated UTF-8 string, may be NULL
	const char		*aFileName,		//	input,  zero-terminated UTF-8 string, may be NULL
	char			*aPath,			//	output, zero-terminated UTF-8 string
	unsigned int	aPathSize)		//	output buffer size, including space for terminating zero
{
	//	Assemble an absolute path of the form
	//
	//		<base path>/<directory name>/<file name>
	//
	//	where
	//
	//		<base path> says where the application lives,
	//			and won't be known until runtime,
	//
	//		<directory name> specifies the directory of interest,
	//			for example "Languages", "Sounds" or "Textures".
	//
	//		<file name> specifies the particular file,
	//			for example "StarterCode-ja.txt" or "Clouds.rgba".

	char	*theTerminator,
			theChar;

	//	Get the base path.
	if ( ! GetBasePath(aPath, aPathSize) )
	{
		aPath[0] = 0;
		return "Couldn't get base path in GetAbsolutePath()";
	}
	
	//	Locate the terminating zero (which GetBasePath() guarantees).
	theTerminator = aPath;
	while (*theTerminator != 0)
		theTerminator++;

	//	Append aDirectory.
	if (aDirectory != NULL)
	{
		while ((theChar = *aDirectory++) != 0)
		{
			*theTerminator = theChar;
			if (++theTerminator < aPath + aPathSize)
				*theTerminator = 0;
			else
			{
				aPath[0] = 0;
				return "Path name too long (1) in  in GetAbsolutePath()";
			}
		}
		
		//	Append a path separator.
		*theTerminator = '/';
		if (++theTerminator < aPath + aPathSize)
			*theTerminator = 0;
		else
		{
			aPath[0] = 0;
			return "Path name too long (2) in  in GetAbsolutePath()";
		}
	}

	//	Append aFileName.
	if (aFileName != NULL)
	{
		while ((theChar = *aFileName++) != 0)
		{
			*theTerminator = theChar;
			if (++theTerminator < aPath + aPathSize)
				*theTerminator = 0;
			else
			{
				aPath[0] = 0;
				return "Path name too long (3) in  in GetAbsolutePath()";
			}
		}
	}

	return NULL;
}
