//	TorusGamesMacOSView.m
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#import "TorusGamesMacOSView.h"
#import "TorusGamesMacOSUtilities.h"

//	When displaying the frame rate in the window's title bar,
//	how often should we update it?
#define FRAME_RATE_UPDATE_PERIOD	1000000		//	time interval in microseconds

//	How far should the board scroll in response to one notch (0.1 units)
//	of mouse wheel rotation?
#define MOUSEWHEEL_SPEED_MAC	(0.001 / 0.1)

//	Virtual key codes
#define       ENTER_KEY		 36
#define         TAB_KEY		 48
#define      DELETE_KEY		 51
#define	     ESCAPE_KEY		 53
#define  LEFT_ARROW_KEY		123
#define	RIGHT_ARROW_KEY		124
#define  DOWN_ARROW_KEY		125
#define    UP_ARROW_KEY		126

//	Allow for a possible toolbar along the top of the window.
#define TOOLBAR_HEIGHT	24

//	The Change Game panel uses each button's "tag" to identify the game
//	that the button selects.  The only exception is the Cancel button,
//	which carries the CHANGE_GAME_CANCEL_BUTTON_TAG.  Make sure
//	the value defined here agrees with the value set in Interface Builder.
//
//	NOTE:  THE CANCEL BUTTON IS NOT CURRENTLY BEING USED.
//	I'VE REMOVED IT FROM THE NIB FILE, BUT LEFT THE CODE HERE
//	IN CASE I EVER WANT TO RESTORE THIS FEATURE.
//	IF I DO RESTORE THE CANCEL BUTTON HERE, I'LL NEED
//	TO IMPLEMENT IT IN THE WIN32 VERSION AS WELL.
#define CHANGE_GAME_CANCEL_BUTTON_TAG	128



@implementation TorusGamesMacOSView


+ (NSOpenGLPixelFormat*)pixelFormatSupportingMultisampling:(bool)aMultisampleFlag
{
	NSOpenGLPixelFormatAttribute	thePlainPixelFormatAttributes[] =
									{
										NSOpenGLPFADoubleBuffer,

										NSOpenGLPFADepthSize,			0,
										NSOpenGLPFAAlphaSize,			8,

										(NSOpenGLPixelFormatAttribute)	0
									},
									theMultisamplePixelFormatAttributes[] =
									{
										NSOpenGLPFADoubleBuffer,

										NSOpenGLPFADepthSize,			0,
										NSOpenGLPFAAlphaSize,			8,

										//	Request multisampling support.
										NSOpenGLPFASampleBuffers,		1,
										NSOpenGLPFASamples,				8,	//	Just guess a large number and let the driver reduce it?

										(NSOpenGLPixelFormatAttribute)	0
									};
	
	return [[[NSOpenGLPixelFormat alloc] initWithAttributes:(aMultisampleFlag ? theMultisamplePixelFormatAttributes : thePlainPixelFormatAttributes)] autorelease];
}

- (id)initWithFrame:(NSRect)frameRect
{
	//	Set harmless defaults for the ModelData,
	//	without assuming the presence of OpenGL or anything else.
	//	In particular, set all pointers to NULL so that
	//	windowShouldClose() can call ShutDownModelData() safely.
	SetUpModelData(&md);

	//	Initialize the superclass.
	//	Request a multisample buffer iff InitWindowData() set md.itsMultisamplingEnabledFlag to true.
	//	The graphics card may or may not support multisampling, but it's always safe to ask.
	self = [super	initWithFrame:	frameRect
					pixelFormat:	[TorusGamesMacOSView pixelFormatSupportingMultisampling:md.itsMultisamplingEnabledFlag]];
	if (self != nil)
	{
		//	What time are we starting at?
		Microseconds( (UnsignedWide *) &itsLastRenderTime    );
		Microseconds( (UnsignedWide *) &itsLastFrameRateTime );
		
		//	Allocate and initialize the TIFF file directory name.
		itsLastImageDirectory	= [[NSMutableString stringWithCapacity:512] retain];
		[itsLastImageDirectory setString:NSHomeDirectory()];
		
		//	Initialize itsTimer to nil, then...
		itsTimer = nil;
		
		//	...set the swap parameter and the timer.
		//	SetUpModelData() has already set a default value for itsSyncFlag.
		[self setRedrawWithSync:md.itsSyncFlag];
		
		//	We don't have the cursor.
		itsTorusCursorFlag = false;
		
		//	No special menu item is present.
		itsMenuGameSpecial = NULL;
	}

	return self;
}

- (void)dealloc
{
	[itsTimer invalidate];
	[itsTimer release];
	
	[itsLastImageDirectory	release];

	[super dealloc];
}

- (void)awakeFromNib
{		
	NSFont	*theSideWindowFont;

	//	Set the pointer linking the ModelData's abstract side window
	//	to the MacOS-specific information.
	md.itsSideWindow = &itsMacOSSideWindow;
	itsMacOSSideWindow.itsCrosswordCluesDrawer	= itsCrosswordCluesDrawer;
	itsMacOSSideWindow.itsCrosswordClueAcross	= itsCrosswordClueAcross;
	itsMacOSSideWindow.itsCrosswordClueDown		= itsCrosswordClueDown;
	itsMacOSSideWindow.itsWordSearchListDrawer	= itsWordSearchListDrawer;
	itsMacOSSideWindow.itsWordSearchList		= itsWordSearchList;
	itsMacOSSideWindow.itsPoolStatusDrawer		= itsPoolStatusDrawer;
	itsMacOSSideWindow.itsPoolStatus			= itsPoolStatus;

	//	Interface Builder seems not to offer the option
	//	of suppressing a TextView's background, so let's set that here.
	//	It seems we must make the call for the enclosing NSScrollView
	//	as well the NSTextView itself.
	[itsWordSearchList setDrawsBackground:NO];
	[[itsWordSearchList enclosingScrollView] setDrawsBackground:NO];
	
	//	We could easily assign different font sizes to the different
	//	text field and views, but they all look good 
	//	at the same fairly large size (20pt), so let's go with that.
	//	By the way, none of these fields respond to awakeFromNib,
	//	so we needn't worry that they will undo what we do here.
	theSideWindowFont = [NSFont systemFontOfSize:20.0];
	[itsCrosswordClueAcross	setFont:theSideWindowFont];
	[itsCrosswordClueDown	setFont:theSideWindowFont];
	[itsWordSearchList		setFont:theSideWindowFont];
	[itsPoolStatus			setFont:theSideWindowFont];
	
	//	itsGameChoicePanel shouldn't prevent the user from quitting.
//UNCOMMENT WHEN COMPILING FOR MAC OS 10.6
//	if ([itsGameChoicePanel respondsToSelector:@selector(setPreventsApplicationTerminationWhenModal:)])
//		[itsGameChoicePanel setPreventsApplicationTerminationWhenModal:NO];
	
	//	I don't see how to set the window's background color
	//	in Interface Builder, so let's set it here.
	//	The background color is visible only in the toolbar,
	//	surrounding the Reset and Change Games buttons.
	[[self window] setBackgroundColor:[NSColor grayColor]];

	//	Zoom to the largest window that fits on the screen.
	//	The method windowWillResize: (see below) ensures that
	//	the window's content region will remain square,
	//	with width and height that are multiples of three. 
	[[self window] zoom:self];
}

- (NSSize)windowWillResize:(NSWindow *)sender toSize:(NSSize)frameSize
{
	NSSize	theContentSize;

	//	Convert the frameSize to theContentRect.
	theContentSize = [[self window] contentRectForFrameRect:(NSRect){{0.0, 0.0}, frameSize}].size;

	//	Subtract off the space for the toolbar.
	theContentSize.height -= TOOLBAR_HEIGHT;

	//	Constrain the width and height to multiples of three,
	//	so the ViewSmall image may exactly fill the center 1/9
	//	of the ViewTiling image.
	theContentSize.width  -= (CGFloat)(((unsigned int)theContentSize.width ) % 3);
	theContentSize.height -= (CGFloat)(((unsigned int)theContentSize.height) % 3);
	
	//	Constrain to a square.
	if (theContentSize.width  > theContentSize.height)
		theContentSize.width  = theContentSize.height;
	if (theContentSize.height > theContentSize.width )
		theContentSize.height = theContentSize.width;
	
	//	Add back the space for the toolbar.
	theContentSize.height += TOOLBAR_HEIGHT;

	//	Convert theContentSize back to the frameSize.
	frameSize = [[self window] frameRectForContentRect:(NSRect){{0.0, 0.0}, theContentSize}].size;
	
	return frameSize;
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
	NSArray		*thePreferredLanguages;
	NSString	*thePreferredLanguage;

	//	Uncomment the following line in test versions.
//	ErrorMessage("...", "Pre-release copy for ...");

	//	Initialize the language code and dictionary to the user's default language.
	//
	//	Note #1.  preferredLocalizations determines language availability
	//	based on what localizations it finds in our program's application bundle.
	//	Thus, as a practical matter, localizing InfoPList.strings to a new language
	//	will bring that language to the system's attention.
	//	
	//	Note #2.  One would expect preferredLocalizations to return an NSArray containing
	//	all localizations supported by our mainBundle and acceptable to the user,
	//	listed in order of the user's preferences.  In practice, however,
	//	preferredLocalizations always returns a 1-element array, even when several
	//	acceptable languages are available.  Apple's documentation is evasive
	//	on this question, saying only that preferredLocalizations
	//
	//		Returns one or more localizations contained in the receiver’s bundle 
	//		that the receiver uses to locate resources based on the user’s preferences.
	//
	//	Fortunately preferredLocalizations reliably returns the user's most preferred localization,
	//	from among those supported.
	thePreferredLanguages	= [[NSBundle mainBundle] preferredLocalizations];
	thePreferredLanguage	= ([thePreferredLanguages count] > 0 ?			//	If a preferred language exists...
								[thePreferredLanguages objectAtIndex:0] :	//	...use it.
								@"en");										//	Otherwise fall back to English.
	[self changeLanguage:thePreferredLanguage];

	//	Set up the MIDI player.
	SetUpMIDIPlayer();
	
	//	Let the user select a game.
	[self runGameChoicePanel:self];
}

- (IBAction)runGameChoicePanel:(id)sender;
{
	//	Display the Game Choice panel so the user may select a game.
	//	The Game Choice panel's buttons call the gameChoicePanelSelection: method below.
	[NSApp beginSheet:itsGameChoicePanel modalForWindow:[self window] modalDelegate:nil didEndSelector:nil contextInfo:nil];
}

- (IBAction)gameChoicePanelSelection:(id)sender
{
	[NSApp endSheet:itsGameChoicePanel];
	[itsGameChoicePanel orderOut:self];
	
	if ([sender tag] != CHANGE_GAME_CANCEL_BUTTON_TAG)
	{
		[self changeGameMacOS:(GameType)[sender tag]];

#ifndef TORUS_GAMES_FOR_TALK
		//	If the user asked for GameIntro,
		//	show the How to Play and Wraparound Universe pages.
		if (md.itsGame == GameIntro)
		{
			//	Load Wraparound Universe first,
			//	so How to Play ends up on top.
			[self commandHelpWraparoundUniverse:nil];
			[self commandHelpHowToPlay:nil];
		}
#endif
	}
}

- (void)localizeGameChoicePanel
{
	NSView			*theContentView;
	NSArray			*theSubviews;
	NSEnumerator	*theEnumerator;
	NSView			*theSubview;
	NSButton		*theButton;
	char			*theKey;
	
	//	Translate itsGameChoicePanel's window title into the current language.
	//	In principle the key could be anything.  In practice it's the English title
	//	"Choose a Game".  At the moment this question is moot, because we're 
	//	displaying the panel as a "sheet" and so its title doesn't get shown.

	//	OLD SCHEME:
	//		itsGameChoicePanel gets translated and displayed only once, at launch,
	//		so it's OK to overwrite the key with the localized title.
//	[itsGameChoicePanel setTitle:GetLocalizedTextAsNSString([[itsGameChoicePanel title] UTF8String])];

	//	NEW SCHEME:
	//		itsGameChoicePanel get translated each time the user presses
	//		the Change Game button, so we used a fixed key.
	[itsGameChoicePanel setTitle:GetLocalizedTextAsNSString("Choose a Game")];	//	currently irrelevant
	
	//	Translate the button labels.

	theContentView	= [itsGameChoicePanel contentView];
	theSubviews		= [[theContentView subviews] copy];	//	NSView documentation says to copy before using.

//	In Mac OS 10.5 and later, use
//
//		for (theSubview in theSubviews)
//
//	instead of the NSEnumerator.
	theEnumerator = [theSubviews objectEnumerator];
	while ((theSubview = [theEnumerator nextObject]) != nil)
	{
		if ([theSubview isKindOfClass:[NSButton class]])	//	should always be true
		{
			theButton = (NSButton *)theSubview;
			
			//	OLD SCHEME:
			//		theButton get translated only once, at launch.
			//		Use the button's initial label as a key.
//			[theButton setTitle:GetLocalizedTextAsNSString([[theButton title] UTF8String])];

			//	NEW SCHEME:
			//		theButton gets translated each time the user
			//		presses Change Game, so let theButton's tag
			//		determine the key.
			switch ([theButton tag])
			{
				//	Game buttons
				case GameNone:			theKey = "?";				break;
				case GameIntro:			theKey = "Help";			break;
				case GameTicTacToe:		theKey = "Tic-Tac-Toe";		break;
				case GameMaze:			theKey = "Maze";			break;
				case GameCrossword:		theKey = "Crossword";		break;
				case GameWordSearch:	theKey = "Word Search";		break;
				case GameJigsaw:		theKey = "Jigsaw Puzzle";	break;
				case GameChess:			theKey = "Chess";			break;
				case GamePool:			theKey = "Pool";			break;
				case GameApples:		theKey = "Apples";			break;
				
				//	Cancel button
				case CHANGE_GAME_CANCEL_BUTTON_TAG:
										theKey = "Cancel";			break;

				//	Just to be safe...
				default:				theKey = "?";				break;
			}
			[theButton setTitle:GetLocalizedTextAsNSString(theKey)];
		}
	}
	
	[theSubviews release];
}

- (BOOL)acceptsFirstResponder
{
	return YES;
}

- (void)setRedrawWithSync:(bool)aSyncFlag
{
	GLint	theSwapValue;

	//	Record the new sync value.
	md.itsSyncFlag = aSyncFlag;

	//	Set the swap interval.
	theSwapValue = (aSyncFlag ? 1 : 0);
	[[self openGLContext] setValues:&theSwapValue forParameter:NSOpenGLCPSwapInterval];

	//	Set a timer for continuous redraw.
	//	The scene may get redrawn less often than the timer interval
	//	(especially if we're synchronizing to the monitor)
	//	but it won't get drawn more often.

	//		Out with the old...
	if (itsTimer != nil)
	{
		[itsTimer invalidate];
		[itsTimer release];
		itsTimer = nil;
	}

	//		...in with the new.
	itsTimer = [NSTimer
			scheduledTimerWithTimeInterval:	(md.itsSyncFlag ? 0.01 : 0.001)
			target:							self
			selector:						@selector( pleaseRedraw )
			userInfo:						nil
			repeats:						YES];
	[itsTimer retain];
}

- (void)pleaseRedraw
{
	long long	theCurrentTime;
	double		theElapsedTime;

	//	Compute the time since the last redraw.
	Microseconds( (UnsignedWide *) &theCurrentTime );
	theElapsedTime = 0.000001 * (theCurrentTime - itsLastRenderTime);
	itsLastRenderTime = theCurrentTime;

	//	If the window is miniaturized, proceed no further.
	if ([[self window] isMiniaturized])
		return;
	
	//	Need to redraw the image?
	if (SimulationWantsUpdates(&md))
	{
		//	Update the simulation.
		SimulationUpdate(&md, theElapsedTime);

		//	Request a redraw.
		[self setNeedsDisplay:YES];
	}
	
	//	If desired, update the frame rate display about once per second.
	if (gShowFrameRate
	 && theCurrentTime >= itsLastFrameRateTime + FRAME_RATE_UPDATE_PERIOD)
	{
		[[self window] setTitle:
			(SimulationWantsUpdates(&md) ?
				(theElapsedTime > 0.0001 ?
					[NSString localizedStringWithFormat:
						GetLocalizedTextAsNSString("fps"),
						(unsigned int)(1.0 / theElapsedTime)]
				:
					GetLocalizedTextAsNSString("fps2")
				)
			:
				[NSString localizedStringWithFormat:
					GetLocalizedTextAsNSString("fps"), 0]
			)
		];
		itsLastFrameRateTime = theCurrentTime;
	}
}

- (void)keyDown:(NSEvent *)anEvent
{
	switch ([anEvent keyCode])
	{
		case ESCAPE_KEY:
			[self exitTorusCursorMode];
			break;

		case ENTER_KEY:			CharacterInput(&md, L'\n');	break;
		case TAB_KEY:			CharacterInput(&md, L'\t');	break;
		case DELETE_KEY:		CharacterInput(&md, L'\b');	break;
		
		case LEFT_ARROW_KEY:	CharacterInput(&md, L'←');	break;
		case RIGHT_ARROW_KEY:	CharacterInput(&md, L'→');	break;
		case DOWN_ARROW_KEY:	CharacterInput(&md, L'↓');	break;
		case UP_ARROW_KEY:		CharacterInput(&md, L'↑');	break;

		default:
			//	Process the keystroke normally.
			[self interpretKeyEvents:[NSArray arrayWithObject:anEvent]];
			break;
	}
}

- (void)insertText:(id)aString
{
	NSString	*theNewText;
	unichar		theCharacter;

	//	The documentation promises only that aString will be either
	//	an NSString or an NSAttributedString. In practice it's an NSString
	//	for both the US English server and Hiragana server.
	if ([[aString class] isSubclassOfClass:[NSAttributedString class]])
	{
		theNewText = [aString string];
	}
	else
	if ([[aString class] isSubclassOfClass:[NSString class]])
	{
		theNewText = aString;
	}
	else
	{
		theNewText = @"?";
	}
	
	if ([theNewText length] > 0)
	{
		theCharacter = [theNewText characterAtIndex:0];

		if (md.itsGame == GameCrossword)
		{
			//	GameCrossword will want character input.
			CharacterInput(&md, theCharacter);
		}
		else
		if (theCharacter == 'f')	//	"secret" keyboard command to show frame rate
		{
			gShowFrameRate = ! gShowFrameRate;
			if ( ! gShowFrameRate )
				[[self window] setTitle:GetLocalizedTextAsNSString("WindowTitle")];
		}
		else
		if (theCharacter == 'm')	//	"secret" keyboard command to sync to the monitor
		{
			[self setRedrawWithSync:( ! md.itsSyncFlag )];
		}
		else
		if (theCharacter == 's')	//	"secret" keyboard command to toggle shaders
		{
			SetShadersEnabledFlag(&md, ! md.itsShadersEnabledFlag );
		}
		else
		{
			//	Otherwise assume the user is flailing.
			//	Perhaps s/he doesn't know how to exit torus cursor mode?
			if (itsTorusCursorFlag)
			{
				[self exitTorusCursorMode];
				ErrorMessage(GetLocalizedText("TorusExitMessage"), GetLocalizedText("TorusExitTitle"), false);
			}
		}
	}
}

- (BOOL)performKeyEquivalent:(NSEvent *)anEvent
{
	if ([anEvent modifierFlags] & NSCommandKeyMask)
	{
		//	The user just pressed a key combination such as ⌘S, ⌘C, ⌘? or ⌘, .
		//	Exit torus cursor mode to restore the system cursor.
		[self exitTorusCursorMode];
	}
	
	//	Return NO to let the caller handle the key equivalent in the usual way.
	return NO;
}

- (void)mouseDown:(NSEvent *)anEvent
{
	//	Not all Mac's have 2-button mice, so treat a shift- or control-click
	//	as a right-click.  Right clicks are important in the Apples game
	//	for marking wormy apples.

	bool	theShiftOrControlKeyIsDown;

	theShiftOrControlKeyIsDown	= (([anEvent modifierFlags] & (NSShiftKeyMask | NSControlKeyMask)) != 0);

	[self mouseDown:anEvent withRightButton:theShiftOrControlKeyIsDown];
}

- (void)mouseDown:(NSEvent *)anEvent withRightButton:(bool)aRightClickFlag
{
	NSPoint	theMouseLocation;

	//	A mouseDown takes us into torus cursor mode if we're not already there.
	if ( ! itsTorusCursorFlag )
		[self enterTorusCursorMode];
	
	//	Pass the mouse click to the game.
	theMouseLocation = [self mouseLocation:anEvent];
	MouseDown(&md, theMouseLocation.x, theMouseLocation.y, aRightClickFlag, [anEvent clickCount] > 1);
}

- (void)mouseDragged:(NSEvent *)anEvent
{
	if (itsTorusCursorFlag)
		[self mouseMoved:anEvent];
}

- (void)mouseMoved:(NSEvent *)anEvent
{
	NSPoint	theMouseDelta;

	theMouseDelta.x =   [anEvent deltaX] / [self bounds].size.width;
	theMouseDelta.y = - [anEvent deltaY] / [self bounds].size.height;

	MouseMove(&md, theMouseDelta.x, theMouseDelta.y);
}

- (void)mouseUp:(NSEvent *)anEvent
{
	MouseUp(&md);

	if (md.itsDoubleClickFlag)
		[self exitTorusCursorMode];
}

- (void)rightMouseDown:(NSEvent *)anEvent
{
	[self mouseDown:anEvent withRightButton:true];
}

- (void)rightMouseDragged:(NSEvent *)anEvent
{
	[self mouseDragged:anEvent];
}

- (void)rightMouseUp:(NSEvent *)anEvent;
{
	[self mouseUp:anEvent];
}

- (void)enterTorusCursorMode
{
	//	Hide the cursor and disassociate it from mouse motion
	//	so the (invisible) cursor stays parked over our window.
	//	Mouse motions will nevertheless arrive in mouseDragged calls.

	if ( ! itsTorusCursorFlag )
	{
		CGDisplayHideCursor(kCGDirectMainDisplay);	//	parameter is ignored
		CGAssociateMouseAndMouseCursorPosition(NO);
		[[self window] setAcceptsMouseMovedEvents:YES];
		itsTorusCursorFlag = true;

		md.itsRedrawRequestFlag = true;
	}
}

- (void)exitTorusCursorMode
{
	//	Restore the usual cursor.

	if (itsTorusCursorFlag)
	{
		CGDisplayShowCursor(kCGDirectMainDisplay);	//	parameter is ignored
		CGAssociateMouseAndMouseCursorPosition(YES);
		[[self window] setAcceptsMouseMovedEvents:NO];
		itsTorusCursorFlag = false;

		MouseGone(&md);
	}
}

- (NSPoint)mouseLocation:(NSEvent *)anEvent
{
	NSPoint	theMouseLocation;

	//	mouse location in window coordinates
	theMouseLocation = [anEvent locationInWindow];

	//	mouse location in view coordinates
	theMouseLocation = [self convertPoint:theMouseLocation fromView:nil];

	//	mouse location in [-0.5, +0.5] coordinates
	theMouseLocation.x = (theMouseLocation.x / [self bounds].size.width ) - 0.5;
	theMouseLocation.y = (theMouseLocation.y / [self bounds].size.height) - 0.5;
	
	return theMouseLocation;
}

- (void)scrollWheel:(NSEvent *)anEvent
{
	MouseWheel(&md, MOUSEWHEEL_SPEED_MAC * [anEvent deltaX], -MOUSEWHEEL_SPEED_MAC * [anEvent deltaY]);
}

- (void)drawRect:(NSRect)bounds
{
	ErrorText	theError;

	if ((theError = SetUpGraphicsAsNeeded(&md)) != NULL)
		ErrorMessage(theError, "OpenGL Setup Error", true);

	if ((theError = Render(&md)) != NULL)
		ErrorMessage(theError, "OpenGL Rendering Error", true);

	[[self openGLContext] flushBuffer];
}

- (BOOL)validateMenuItem:(NSMenuItem *)aMenuItem
{
	SEL	theAction;
	
	//	Disable menus when the Game Choice panel is active.
	if ([NSApp modalWindow] != nil)
		return NO;

	theAction = [aMenuItem action];

	//	game menu

	if (theAction == @selector(commandGameReset:))
	{
		//	Set appropriate text for the Reset menu item.
		[aMenuItem setTitle:[self gameSpecificResetText]];

		return YES;
	}
	if (theAction == @selector(commandGameSpecial:))
		return YES;
		
	if (theAction == @selector(commandGameTicTacToe:))
	{
		[aMenuItem setState:(md.itsGame == GameTicTacToe	? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandGameMaze:))
	{
		[aMenuItem setState:(md.itsGame == GameMaze			? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandGameCrossword:))
	{
		[aMenuItem setState:(md.itsGame == GameCrossword	? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandGameWordSearch:))
	{
		[aMenuItem setState:(md.itsGame == GameWordSearch	? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandGameJigsaw:))
	{
		[aMenuItem setState:(md.itsGame == GameJigsaw		? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandGameChess:))
	{
		[aMenuItem setState:(md.itsGame == GameChess		? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandGamePool:))
	{
		[aMenuItem setState:(md.itsGame == GamePool			? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandGameApples:))
	{
		[aMenuItem setState:(md.itsGame == GameApples		? NSOnState : NSOffState)];
		return YES;
	}

	//	options menu

	if (md.itsGame == GameTicTacToe
	 || md.itsGame == GameChess
	 || md.itsGame == GamePool)
	{
		if (theAction == @selector(commandOptionsHumanVsHuman:))
		{
			[aMenuItem setState:( ! md.itsHumanVsComputer ? NSOnState : NSOffState)];
			return YES;
		}
		if (theAction == @selector(commandOptionsHumanVsComputer:))
		{
			[aMenuItem setState:(   md.itsHumanVsComputer ? NSOnState : NSOffState)];
			return YES;
		}
	}
	else
	{
		if (theAction == @selector(commandOptionsHumanVsHuman:)
		 || theAction == @selector(commandOptionsHumanVsComputer:))
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}

	if (md.itsGame == GameMaze
	 || md.itsGame == GameJigsaw
	 || (	md.itsGame == GameChess
		 && md.itsHumanVsComputer)
	 || (	md.itsGame == GamePool
		 && md.itsHumanVsComputer)
	 || md.itsGame == GameApples)
	{
		if (theAction == @selector(commandOptionsExtraEasy:))
		{
			[aMenuItem setState:(md.itsDifficultyLevel == 0	? NSOnState : NSOffState)];
			return YES;
		}
		if (theAction == @selector(commandOptionsEasy:))
		{
			[aMenuItem setState:(md.itsDifficultyLevel == 1	? NSOnState : NSOffState)];
			return YES;
		}
		if (theAction == @selector(commandOptionsMedium:))
		{
			[aMenuItem setState:(md.itsDifficultyLevel == 2	? NSOnState : NSOffState)];
			return YES;
		}
		if (theAction == @selector(commandOptionsHard:))
		{
			[aMenuItem setState:(md.itsDifficultyLevel == 3	? NSOnState : NSOffState)];
			return YES;
		}
		if (theAction == @selector(commandOptionsExtraHard:))
		{
			[aMenuItem setState:(md.itsDifficultyLevel == 4	? NSOnState : NSOffState)];
			return (md.itsGame != GameChess);	//	Suppress Extra Hard chess because it's too slow.
		}
	}
	else
	{
		if (theAction == @selector(commandOptionsExtraEasy:)
		 || theAction == @selector(commandOptionsEasy:     )
		 || theAction == @selector(commandOptionsMedium:   )
		 || theAction == @selector(commandOptionsHard:     )
		 || theAction == @selector(commandOptionsExtraHard:))
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}

	if (md.itsTopology == TopologyKlein)
	{
		if (theAction == @selector(commandOptionsGlideReflections:))
		{
			[aMenuItem setState:(md.itsShowGlideAxes ? NSOnState : NSOffState)];
			return YES;
		}
	}
	else
	{
		if (theAction == @selector(commandOptionsGlideReflections:))
		{
			[aMenuItem setState:NSOffState];
			return NO;
		}
	}

	if (theAction == @selector(commandOptionsSound:))
	{
		[aMenuItem setState:(gPlaySounds ? NSOnState : NSOffState)];
		return YES;
	}

	if (theAction == @selector(commandOptionsExportCopy:)
	 || theAction == @selector(commandOptionsExportSave:))
	{
		return YES;
	}

	//	topology menu

	if (theAction == @selector(commandTopologyTorus:))
	{
		[aMenuItem setState:(md.itsTopology == TopologyTorus	? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandTopologyKlein:))
	{
		[aMenuItem setState:(md.itsTopology == TopologyKlein	? NSOnState : NSOffState)];
		return YES;
	}
	
	//	view menu

	if (theAction == @selector(commandViewDomainLarge:))
	{
		[aMenuItem setState:(md.itsView == ViewDomain	? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewDomainSmall:))
	{
		[aMenuItem setState:(md.itsView == ViewSmall	? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandViewTiling:))
	{
		[aMenuItem setState:(md.itsView == ViewTiling	? NSOnState : NSOffState)];
		return YES;
	}
	
	//	language menu

	if (theAction == @selector(commandLanguageDe:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "de") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageEn:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "en") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageEs:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "es") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageFr:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "fr") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageHu:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "hu") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageIt:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "it") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageJa:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "ja") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageKo:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "ko") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageNl:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "nl") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguagePt:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "pt") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageRu:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "ru") ? NSOnState : NSOffState)];
		return YES;
	}
	if (theAction == @selector(commandLanguageVi:))
	{
		[aMenuItem setState:(SAME_LANGUAGE_CODE(gLanguageCode, "vi") ? NSOnState : NSOffState)];
		return YES;
	}
	
	//	help menu

	if (theAction == @selector(commandHelpPracticeBoard:))
	{
		[aMenuItem setState:(md.itsGame == GameIntro ? NSOnState : NSOffState)];
		return YES;
	}

	return YES;
}

- (IBAction)commandGameReset:(id)sender
{
	ResetGame(&md);
}

- (IBAction)commandGameSpecial:(id)sender
{
	SpecialMenuCommand(&md);
}

- (IBAction)commandGameTicTacToe:(id)sender
{
	[self changeGameMacOS:GameTicTacToe];
}

- (IBAction)commandGameMaze:(id)sender
{
	[self changeGameMacOS:GameMaze];
}

- (IBAction)commandGameCrossword:(id)sender
{
	[self changeGameMacOS:GameCrossword];
}

- (IBAction)commandGameWordSearch:(id)sender
{
	[self changeGameMacOS:GameWordSearch];
}

- (IBAction)commandGameJigsaw:(id)sender
{
	[self changeGameMacOS:GameJigsaw];
}

- (IBAction)commandGameChess:(id)sender
{
	[self changeGameMacOS:GameChess];
}

- (IBAction)commandGamePool:(id)sender
{
	[self changeGameMacOS:GamePool];
}

- (IBAction)commandGameApples:(id)sender
{
	[self changeGameMacOS:GameApples];
}

- (IBAction)commandOptionsHumanVsHuman:(id)sender
{
	ChangeHumanVsComputer(&md, false);
}

- (IBAction)commandOptionsHumanVsComputer:(id)sender
{
	ChangeHumanVsComputer(&md, true);
}

- (IBAction)commandOptionsExtraEasy:(id)sender
{
	ChangeDifficultyLevel(&md, 0);
}

- (IBAction)commandOptionsEasy:(id)sender
{
	ChangeDifficultyLevel(&md, 1);
}

- (IBAction)commandOptionsMedium:(id)sender
{
	ChangeDifficultyLevel(&md, 2);
}

- (IBAction)commandOptionsHard:(id)sender
{
	ChangeDifficultyLevel(&md, 3);
}

- (IBAction)commandOptionsExtraHard:(id)sender
{
	ChangeDifficultyLevel(&md, 4);
}

- (IBAction)commandOptionsGlideReflections:(id)sender
{
	SetShowGlideAxes(&md, ! md.itsShowGlideAxes );
}

- (IBAction)commandOptionsSound:(id)sender
{
	gPlaySounds = ! gPlaySounds;
}

- (IBAction)commandOptionsExportCopy:(id)sender
{
	[self copyImage];
}

- (IBAction)commandOptionsExportSave:(id)sender
{
	[self saveImage];
}

- (IBAction)commandTopologyTorus:(id)sender
{
	ChangeTopology(&md, TopologyTorus);
}

- (IBAction)commandTopologyKlein:(id)sender
{
	ChangeTopology(&md, TopologyKlein);
}

- (IBAction)commandViewDomainLarge:(id)sender
{
	ChangeView(&md, ViewDomain);
}

- (IBAction)commandViewDomainSmall:(id)sender
{
	ChangeView(&md, ViewSmall);
}

- (IBAction)commandViewTiling:(id)sender
{
	ChangeView(&md, ViewTiling);
}

- (IBAction)commandLanguageDe:(id)sender
{
	[self changeLanguage:@"de"];
}

- (IBAction)commandLanguageEn:(id)sender
{
	[self changeLanguage:@"en"];
}

- (IBAction)commandLanguageEs:(id)sender
{
	[self changeLanguage:@"es"];
}

- (IBAction)commandLanguageFr:(id)sender
{
	[self changeLanguage:@"fr"];
}

- (IBAction)commandLanguageHu:(id)sender
{
	[self changeLanguage:@"hu"];
}

- (IBAction)commandLanguageIt:(id)sender
{
	[self changeLanguage:@"it"];
}

- (IBAction)commandLanguageJa:(id)sender
{
	[self changeLanguage:@"ja"];
}

- (IBAction)commandLanguageKo:(id)sender
{
	[self changeLanguage:@"ko"];
}

- (IBAction)commandLanguageNl:(id)sender
{
	[self changeLanguage:@"nl"];
}

- (IBAction)commandLanguagePt:(id)sender
{
	[self changeLanguage:@"pt"];
}

- (IBAction)commandLanguageRu:(id)sender
{
	[self changeLanguage:@"ru"];
}

- (IBAction)commandLanguageVi:(id)sender
{
	[self changeLanguage:@"vi"];
}

- (IBAction)commandHelpHowToPlay:(id)sender
{
	//	The Mac OS Help Book system is terrible:
	//	it's very difficult to get it accept different pages,
	//	won't accept multiple named pages in the same directory,
	//	tends to fail or crash at the least excuse,
	//	and caches so agressively that you can't get
	//	old versions of a Help Book to go away.
	//
	//	By contrast, sending a page to the default web browser
	//	is simple and effective.  A real breath of fresh air!
	[[NSWorkspace sharedWorkspace] openFile:[[[NSBundle mainBundle] resourcePath]
		stringByAppendingString:[NSString stringWithFormat:@"/Help/HowToPlay-%c%c.html", gLanguageCode[0], gLanguageCode[1]]]];
}

- (IBAction)commandHelpWraparoundUniverse:(id)sender
{
	[[NSWorkspace sharedWorkspace] openFile:[[[NSBundle mainBundle] resourcePath]
		stringByAppendingString:[NSString stringWithFormat:@"/Help/WraparoundUniverse-%c%c.html", gLanguageCode[0], gLanguageCode[1]]]];
}

- (IBAction)commandHelpPracticeBoard:(id)sender
{
	[self changeGameMacOS:GameIntro];
}

- (IBAction)commandHelpContact:(id)sender
{
	[[NSWorkspace sharedWorkspace] openFile:[[[NSBundle mainBundle] resourcePath]
		stringByAppendingString:[NSString stringWithFormat:@"/Help/Contact-%c%c.html", gLanguageCode[0], gLanguageCode[1]]]];
}

- (void)changeGameMacOS:(GameType)aNewGame
{
	//	Just to be safe...
	if (aNewGame >= NumGameTypes)
		aNewGame = GameNone;
	
	//	If the user re-selects the game that's already running,
	//	leave that game untouched.  In other words, don't reset it.
	if (aNewGame == md.itsGame)
		return;

	//	Shut down the old game.
	//	Make sure the OpenGL context is current
	//	so the old game can release its textures,
	//	whose OpenGL "names" only it knows.
	[[self openGLContext] makeCurrentContext];
	ShutDownGame(&md);
	[NSOpenGLContext clearCurrentContext];

	//	Switch to the new game.
	md.itsGame = aNewGame;

	//	Delete fonts (if any) from the old game and 
	//	create fonts (if any) for the new game.
	[self prepareFonts];
	
	//	Delete the side window (if any) from the old game and
	//	create the side window (if any) for the new game,
	//	setting the font for the latter.
	[self prepareSideWindow];
	
	//	Delete the special menu (if any) from the old game and
	//	create the special menu (if any) for the new game.
	//	While we're at it, set the correct text for the Reset menu item.
	[self prepareGameMenu];

	//	Update the toolbar to reflect the new game.
	[itsResetButton setTitle:[self gameSpecificResetText]];

	//	Set up the new game.
	SetUpGame(&md);
}

- (void)prepareFonts
{
}

- (void)prepareSideWindow
{
	//	Notes on drawer management:
	//
	//	In Interface Builder, setting the Crossword Clues View's
	//	topmost autosizing line to "springy" and its bottommost
	//	autosizing line to "straight" seems to have the desired effect
	//	of keeping the view pushed to the top of the drawer, even
	//	though this is exactly the opposite of what I would have expected.
	//	Each NSText field within the NSView has its topmost line "straight"
	//	and its bottommost line "springy" to keep it anchored relative
	//	to the top of the view, as one would expect.

	//	I tried setting the drawer's height to the height of the actual content,
	//	but it didn't work out well.  The problem is that NSDrawer lets us 
	//	set the leading and trailing offsets, not the height of the drawer itself.
	//	So even if we calculate the offsets needed to make it look good initially,
	//	once the user starts resizing the window it looks terrible because
	//	it preserves the offsets, not the drawer height.  So better to leave
	//	well enough alone and put up with a too-tall drawer.

	//	Close any open drawer.
	[itsCrosswordCluesDrawer close];
	[itsWordSearchListDrawer close];
	[itsPoolStatusDrawer     close];
	
	//	Open a new drawer if required.
	switch (md.itsGame)
	{
		case GameCrossword:

			//	Set the labels to the current language.
			[itsCrosswordLabelAcross setStringValue:
				[GetLocalizedTextAsNSString("Across") stringByAppendingString:@":  "]];
			[itsCrosswordLabelDown   setStringValue:
				[GetLocalizedTextAsNSString( "Down" ) stringByAppendingString:@":  "]];

			//	Open the drawer.
			[itsCrosswordCluesDrawer open];

			break;

		case GameWordSearch:
		
			//	The NSTextView remembers characteristics of its previous font.
			//	For example, the Russian word list looks fine if you go straight to it,
			//	but if you first view a Japanese word list (which gets drawn correctly
			//	in a mono-spaced font) and then pass to Russian, the Russian comes
			//	out mono-spaced, with the same spacing as the previous Japanese text
			//	(which looks good for Japanese, but is much too wide for Russian).
			//	Passing from Japanese to English contaminates the vertical spacing
			//	but not the horizontal spacing.
			//	The workaround for these problems is to clear the old text
			//	before installing the new.
			//	Note that the Crossword and Pool text avoids this issue,
			//	presumably because they are NSTextFields rather than NSTextViews.
			[itsWordSearchList setString:@""];
			
			//	Open the drawer.
			[itsWordSearchListDrawer open];

			break;

		case GamePool:

			//	Open the drawer.
			[itsPoolStatusDrawer open];

			break;

		default:
			//	Most games don't need a drawer.
			break;
	}
}

-(void)prepareGameMenu
{
	char	*theSpecialKey;
	NSMenu	*theGameMenu;
	
	//	Assume theGameMenu is the first menu following the application menu.
	theGameMenu = [[[NSApp mainMenu] itemAtIndex:1] submenu];

	//	Create a Special menu item for those games that require it.
	switch (md.itsGame)
	{
		case GameJigsaw:	theSpecialKey = "Send to Back";			break;
		case GameChess:		theSpecialKey = "Computer Goes First";	break;
		default:			theSpecialKey = NULL;					break;
	}
	if (theSpecialKey != NULL)
	{
		//	The game needs a Special menu item.

		//	If none exists, create it.
		if (itsMenuGameSpecial == nil)
		{
			//	theGameMenu will own itsMenuGameSpecial.
			itsMenuGameSpecial = [theGameMenu
				insertItemWithTitle:	@""	//	to be overridden below
				action:					@selector( commandGameSpecial: )
				keyEquivalent:			@""
				atIndex:				1];
		}
		
		//	Set an appropriate menu title.
		[itsMenuGameSpecial setTitle:GetLocalizedTextAsNSString(theSpecialKey)];
	}
	else
	{
		//	The game needs no Special menu item.
		//	If one exists, delete it.
		if (itsMenuGameSpecial != nil)
		{
			//	theGameMenu owns itsMenuGameSpecial,
			//	so removing it automatically releases it.
			[theGameMenu removeItem:itsMenuGameSpecial];
			itsMenuGameSpecial = nil;
		}
	}
}


- (void)changeLanguage:(NSString *)aLanguageCode
{
	//	Record the new language code and initialize a dictionary.
	RefreshLocalization([aLanguageCode UTF8String]);

	//	Reconstruct the menu bar in the new language.
	[NSApp setMainMenu:[self buildLocalizedMenuBar]];

	//	The old menu bar owned the special menu item, if one was present,
	//	so replacing the menu bar automatically deleted the old special
	//	menu item along with it.
	itsMenuGameSpecial = nil;
	
	//	Create a new special menu item, if necessary.
	[self prepareGameMenu];

	//	Set the Crossword labels to the current language.
	[itsCrosswordLabelAcross setStringValue:
		[GetLocalizedTextAsNSString("Across") stringByAppendingString:@":  "]];
	[itsCrosswordLabelDown   setStringValue:
		[GetLocalizedTextAsNSString( "Down" ) stringByAppendingString:@":  "]];
	
	//	Update the window title.
	//	The title will be "Torus Games" in all languages
	//	except Japanese where it's 「トーラス・ゲームズ」
	//	and Russian where it's "Игры Торус".
	[[self window] setTitle:GetLocalizedTextAsNSString("WindowTitle")];
	
	//	Update the toolbar to reflect the new language.
	[itsResetButton			setTitle:[self gameSpecificResetText]];
	[itsChangeGameButton	setTitle:GetLocalizedTextAsNSString("Change Game")];

	//	Localize itsGameChoicePanel, so it will be ready when needed.
	[self localizeGameChoicePanel];
	
	//	Refresh any game-specific text.
	switch (md.itsGame)
	{
		case GameCrossword:
		case GameWordSearch:
			//	Reset Crossword and Word Search to accommodate the new language.
			//	<game>Reset() will notice that the language has changed and 
			//	set the puzzle indices to 0, so the user will get the easiest
			//	puzzle in the new language.
			ResetGame(&md);
			break;
		
		case GamePool:
			//	No need to disturb the Pool game in progress.
			//	Just refresh the current status message in the new language.
			PoolSetStatusMessage(&md);
			break;
		
		default:
			//	Do nothing.
			break;
	}
}

- (NSMenu *)buildLocalizedMenuBar
{
	NSMenu		*theMenuBar,
				*theMenu,
				*theSubmenu;
	NSMenuItem	*theItemWithKeyModifier;

	//	Construct all menus using the current language.

	//	Design Note:  A more traditional approach would be to maintain 
	//	a separate .nib file for each language, but it's a pain
	//	to maintain numerous parallel .nib files as the software evolves --
	//	one must either maintain them by hand or take on the added complexity
	//	of dealing with AppleGlot to automate the process.
	//	Personally I prefer to keep a program's functionality separated 
	//	from its language-specific strings.  Thus the menus (including all
	//	selectors and key equivalents) get defined here here in this
	//	language-independent function, while all language-specific strings 
	//	(including the text of each menu item) get stored in the language files.
	//	This not only makes program maintenance easier (because a change 
	//	to the menu bar happens in one place only) but also simplifies localization
	//	(because the translator works with a simple text-only file containing 
	//	the English phrases alongside their local language equivalents).
	
	//	Main menu bar
	theMenuBar = [[[NSMenu alloc] initWithTitle:@"main menu"] autorelease];
	
	//	Application menu
	//	(On Mac OS 10.6, title will be overridden by localized CFBundleName.
	//	But on 10.4 or 10.5, the title given here may persist.)
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Torus Games"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("About Torus Games")
			action:@selector(orderFrontStandardAboutPanel:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Hide Torus Games")
			action:@selector(hide:) keyEquivalent:@"h"];
		theItemWithKeyModifier = [[NSMenuItem alloc]
					initWithTitle:GetLocalizedTextAsNSString("Hide Others")
					action:@selector(hideOtherApplications:) keyEquivalent:@"h"];
			[theItemWithKeyModifier setKeyEquivalentModifierMask:(NSCommandKeyMask | NSAlternateKeyMask)];
			[theMenu addItem:theItemWithKeyModifier];
			[theItemWithKeyModifier release];
			theItemWithKeyModifier = nil;
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Show All")
			action:@selector(unhideAllApplications:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Quit Torus Games")
			action:@selector(terminate:) keyEquivalent:@"q"];
	theMenu = nil;

	//	Game menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Game"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Reset")
			action:@selector(commandGameReset:) keyEquivalent:@"n"];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Tic-Tac-Toe")
			action:@selector(commandGameTicTacToe:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Maze")
			action:@selector(commandGameMaze:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Crossword")
			action:@selector(commandGameCrossword:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Word Search")
			action:@selector(commandGameWordSearch:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Jigsaw Puzzle")
			action:@selector(commandGameJigsaw:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Chess")
			action:@selector(commandGameChess:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Pool")
			action:@selector(commandGamePool:) keyEquivalent:@""];
#ifndef TORUS_GAMES_FOR_TALK
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Apples")
			action:@selector(commandGameApples:) keyEquivalent:@""];
#endif
	theMenu = nil;

	//	Options menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Options"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Human vs. Human")
			action:@selector(commandOptionsHumanVsHuman:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Human vs. Computer")
			action:@selector(commandOptionsHumanVsComputer:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Extra Easy")
			action:@selector(commandOptionsExtraEasy:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Easy")
			action:@selector(commandOptionsEasy:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Medium")
			action:@selector(commandOptionsMedium:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Hard")
			action:@selector(commandOptionsHard:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Extra Hard")
			action:@selector(commandOptionsExtraHard:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Show Glide Reflections")
			action:@selector(commandOptionsGlideReflections:) keyEquivalent:@""];
		[theMenu addItem:[NSMenuItem separatorItem]];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Sound")
			action:@selector(commandOptionsSound:) keyEquivalent:@""];
		theSubmenu = [self addToMenu:theMenu submenuWithTitle:"Export"];;
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Copy Image")
				action:@selector(commandOptionsExportCopy:) keyEquivalent:@"c"];
			[theSubmenu addItemWithTitle:GetLocalizedTextAsNSString("Save Image…")
				action:@selector(commandOptionsExportSave:) keyEquivalent:@"s"];
		theSubmenu = nil;
	theMenu = nil;

	//	Topology menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Topology"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Torus")
			action:@selector(commandTopologyTorus:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Klein Bottle")
			action:@selector(commandTopologyKlein:) keyEquivalent:@""];
	theMenu = nil;

	//	View menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"View"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Fundamental Domain - Large")
			action:@selector(commandViewDomainLarge:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Fundamental Domain - Small")
			action:@selector(commandViewDomainSmall:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Tiling")
			action:@selector(commandViewTiling:) keyEquivalent:@""];
	theMenu = nil;

	//	Language menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Language"];
		//	To add a new language, please see the instructions
		//	in the file "To add a new language.txt".
		[theMenu addItemWithTitle:@"Deutsch"
			action:@selector(commandLanguageDe:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"English"
			action:@selector(commandLanguageEn:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Español"
			action:@selector(commandLanguageEs:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Français"
			action:@selector(commandLanguageFr:) keyEquivalent:@""];
//Don't forget to restore the Hungarian localization of InfoPlist.strings.
//		[theMenu addItemWithTitle:@"Magyar"
//			action:@selector(commandLanguageHu:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Italiano"
			action:@selector(commandLanguageIt:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"日本語"
			action:@selector(commandLanguageJa:) keyEquivalent:@""];
//Don't forget to restore the Korean localization of InfoPlist.strings.
//		[theMenu addItemWithTitle:@"한글"
//			action:@selector(commandLanguageKo:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Nederlands"
			action:@selector(commandLanguageNl:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Português"
			action:@selector(commandLanguagePt:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Русский"
			action:@selector(commandLanguageRu:) keyEquivalent:@""];
		[theMenu addItemWithTitle:@"Tiếng Việt"
			action:@selector(commandLanguageVi:) keyEquivalent:@""];
	theMenu = nil;
	
	//	Help menu
	theMenu = [self addToMenu:theMenuBar submenuWithTitle:"Help"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("How to Play")
			action:@selector(commandHelpHowToPlay:) keyEquivalent:@"?"];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Wraparound Universe")
			action:@selector(commandHelpWraparoundUniverse:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Practice Board")
			action:@selector(commandHelpPracticeBoard:) keyEquivalent:@""];
		[theMenu addItemWithTitle:GetLocalizedTextAsNSString("Contact")
			action:@selector(commandHelpContact:) keyEquivalent:@""];
	theMenu = nil;
	
	return theMenuBar;
}

- (NSMenu *)addToMenu:(NSMenu *)aParentMenu submenuWithTitle:(char *)aSubmenuTitle
{
	NSString	*theLocalizedTitle;
	NSMenuItem	*theNewParentMenuItem;
	NSMenu		*theNewMenu;

	//	Technical Note:  For top-level menus (i.e. those attached to the menu bar),
	//	theNewMenu's title gets displayed while theNewParentMenuItem's title gets ignored.
	//	For submenus, the reverse is true.  To be safe, pass the same title to both.

	//	Localize aSubmenuTitle to the current language.
	theLocalizedTitle = GetLocalizedTextAsNSString(aSubmenuTitle);

	//	Create a new item for aParentMenu, and a new menu to attach to it.
	theNewParentMenuItem	= [[NSMenuItem alloc] initWithTitle:theLocalizedTitle action:NULL keyEquivalent:@""];
	theNewMenu				= [[NSMenu alloc] initWithTitle:theLocalizedTitle];
	
	//	Attach theNewMenu to theNewParentMenuItem.
	[theNewParentMenuItem setSubmenu:theNewMenu];
	
	//	Add theNewParentMenuItem to aParentMenu.
	[aParentMenu addItem:theNewParentMenuItem];

	//	aParentMenu retains theNewParentMenuItem,
	//	so we can release theNewParentMenuItem.
	[theNewParentMenuItem release];
	
	//	aParentMenu retains theNewParentMenuItem,
	//	and theNewParentMenuItem retains theNewMenu,
	//	so we can release theNewMenu.
	[theNewMenu release];
	
	//	Pass the caller a weak reference to theNewMenu.
	return theNewMenu;
}

- (NSString *)gameSpecificResetText
{
	char	*theResetKey;
	
	//	The Reset command's text depends on both the current game
	//	and the current language.  This same text appears both
	//	as a toolbar button and as an item on the Game menu.

	switch (md.itsGame)
	{
		case GameNone:			theResetKey = "Reset";						break;
		case GameIntro:			theResetKey = "Reset Practice Board";		break;
		case GameTicTacToe:		theResetKey = "Erase Tic-Tac-Toe Board";	break;
		case GameMaze:			theResetKey = "New Maze";					break;
		case GameCrossword:		theResetKey = "New Crossword Puzzle";		break;
		case GameWordSearch:	theResetKey = "New Word Search";			break;
		case GameJigsaw:		theResetKey = "New Jigsaw Puzzle";			break;
		case GameChess:			theResetKey = "Reset Chess Board";			break;
		case GamePool:			theResetKey = "Reset Pool Balls";			break;
		case GameApples:		theResetKey = "New Apples";					break;
		default:				theResetKey = NULL;							break;
	}

	return GetLocalizedTextAsNSString(theResetKey);
}


- (void)copyImage
{
	NSData			*theTiff				= nil;
	NSPasteboard	*theGeneralPasteboard	= nil;

	theTiff = [self contentsAsTIFF];

	theGeneralPasteboard = [NSPasteboard generalPasteboard];
	
	[theGeneralPasteboard declareTypes:[NSArray arrayWithObject:NSTIFFPboardType] owner:nil];
	[theGeneralPasteboard setData:theTiff forType:NSTIFFPboardType];
}


- (void)saveImage
{
	NSSavePanel	*theSavePanel	= nil;
	NSData		*theTiff		= nil;

	//	Create a save-a-file panel.
	theSavePanel = [NSSavePanel savePanel];
	[theSavePanel setAllowedFileTypes:[NSArray arrayWithObject:@"tiff"]];
	[theSavePanel setTitle:@"Save image as TIFF file"];

	//	Ask the user to choose a file name.
	if ([theSavePanel runModalForDirectory:itsLastImageDirectory file:nil] == NSFileHandlingPanelCancelButton)
		return;	//	user cancelled
	
	//	Remember the directory for next time.
	[itsLastImageDirectory setString:[theSavePanel directory]];

	//	Read the OpenGL image.
	theTiff = [self contentsAsTIFF];
	if (theTiff == nil)
		return;
	
	//	Save the data under the requested filename.
	if ([theTiff writeToFile:[theSavePanel filename] atomically:YES] == NO)
		ErrorMessage("Couldn't save image under requested filename.", "Save failed", false);
}


- (NSImage *)contentsAsImage
{
	NSBitmapImageRep	*theBitmap	= nil;
	NSImage				*theImage	= nil;
	
	//	Get a bitmap...
	theBitmap = [self contentsAsBitmapImageRep];
	
	//	...and wrap it as an NSImage.
	//
	//	Note:  The documentation for addRepresentation makes the vague comment that
	//	"After invoking this method, you may need to explicitly set features
	//	of the new image representation, such as the size, number of colors, and so on."
	//	but in practice it seems to work OK.
	theImage = [[[NSImage alloc] initWithSize:(NSSize){0.0, 0.0}] autorelease];
	[theImage addRepresentation:theBitmap];

	return theImage;
}


- (NSData *)contentsAsTIFF
{
	NSBitmapImageRep	*theBitmap	= nil;
	NSData				*theTIFF	= nil;
	
	//	Get a bitmap...
	theBitmap = [self contentsAsBitmapImageRep];

	//	...and convert it to TIFF format.
	theTIFF = [theBitmap TIFFRepresentationUsingCompression:NSTIFFCompressionLZW factor:0.0f];
	if (theTIFF == nil)
		ErrorMessage("Couldn't create TIFF representation.", "Image Copying Error", false);

	return theTIFF;
}


- (NSBitmapImageRep *)contentsAsBitmapImageRep
{
	ErrorText			theError	= NULL;
	NSRect				theBounds	= {{0.0f, 0.0f}, {0.0f, 0.0f}};
	NSBitmapImageRep	*theBitmap	= nil;
	unsigned char		*thePixels	= NULL;
	
	//	Note the view's bounds.
	theBounds = [self bounds];
	if (NSMinX(theBounds) != 0 || NSMinY(theBounds) != 0)
	{
		theError = "View does not have origin (0,0).";
		goto CleanUpContentsAsBitmapImageRep;
	}

	//	Create an empty bitmap of the correct dimensions.
	//	By default the NSBitmapImageRep assumes premultiplied alpha.
	theBitmap = [[[NSBitmapImageRep alloc]
		initWithBitmapDataPlanes:	NULL
		pixelsWide:					NSWidth(theBounds)
		pixelsHigh:					NSHeight(theBounds)
		bitsPerSample:				8
		samplesPerPixel:			4
		hasAlpha:					YES
		isPlanar:					NO
		colorSpaceName:				NSDeviceRGBColorSpace
		bytesPerRow:				NSWidth(theBounds) * 4
		bitsPerPixel:				32
		] autorelease];
	if (theBitmap == nil)
	{
		theError = "Couldn't allocate theBitmap.";
		goto CleanUpContentsAsBitmapImageRep;
	}
	
	//	Read the pixels into theBitmap.
	thePixels = [theBitmap bitmapData];
	[self lockFocus];
		theError = ReadRGBAPixels(NSWidth(theBounds), NSHeight(theBounds), thePixels);
	[self unlockFocus];
	if (theError != NULL)
		goto CleanUpContentsAsBitmapImageRep;
	
	//	OpenGL reads the image bottom-to-top, but Cocoa seems to be expecting
	//	it top-to-bottom.  So we must reverse the rows.
	[self invertRawImage:(unsigned int *)thePixels pixelsWide:NSWidth(theBounds) pixelsHigh:NSHeight(theBounds)];
	
CleanUpContentsAsBitmapImageRep:

	if (theError != NULL)
		ErrorMessage(theError, "Error Copying OpenGL Image from Window", false);

	return theBitmap;
}


- (void)invertRawImage:(unsigned int *)aPixelBuffer pixelsWide:(unsigned int)aWidth pixelsHigh:(unsigned int)aHeight
{
	unsigned int	theLowerRow,
					theUpperRow,
					theCount;
	
	//	Rather than thinking of a pixel as four 1-byte components {R,G,B,A},
	//	think of it as a single 4-byte integer.  The byte order,
	//	0xRRGGBBAA or 0xAABBGGRR, makes no difference here.
	unsigned int	*theLowerPixel,
					*theUpperPixel,
					theTempPixel;
	
	theLowerRow	= 0;
	theUpperRow	= aHeight - 1;

	while (theUpperRow > theLowerRow)
	{
		theLowerPixel = aPixelBuffer + theLowerRow*aWidth;
		theUpperPixel = aPixelBuffer + theUpperRow*aWidth;

		for (theCount = aWidth; theCount-- > 0; )
		{
			theTempPixel	= *theLowerPixel;
			*theLowerPixel	= *theUpperPixel;
			*theUpperPixel	= theTempPixel;
			theLowerPixel++;
			theUpperPixel++;
		}

		theLowerRow++;
		theUpperRow--;
	}
}


- (void)reshape
{
	ResizeGLView(	(unsigned int) [self bounds].size.width,
					(unsigned int) [self bounds].size.height);

	//	Refresh the window.  You need to call update first to have it 
	//	come to terms with its context.  If not bizzare things happen.
	//	Adam found this solution at:
	//		http://developer.apple.com/qa/qa2001/qa1209.html
	//
	[self update];
	[self setNeedsDisplay:YES];
}

- (void)windowWillMiniaturize:(NSNotification *)aNotification
{
	//	When the user minimizes a window, Mac OS creates a dock icon to represent it.
	//	Normally the icon is a low-resolution image of the window's contents,
	//	which Mac OS creates automatically with no help from the programmer.
	//	Alas that automated process fails for OpenGL windows, so let's read
	//	the OpenGL image buffer ourselves and report its contents via setMiniwindowImage.
	[[self window] setMiniwindowImage:[self contentsAsImage]];
}

- (BOOL)windowShouldClose:(id)sender
{
	//	Caution:  Cocoa calls this method only when the user explictly clicks
	//	the window's close button.  When the user chooses Quit Torus Games
	//	we must call this method ourselves.  In the former case this method
	//	will get called twice, but that's OK.

	//	Switch to the null game to free any allocated memory.
	[self changeGameMacOS:GameNone];

	//	Free the ModelData here, rather than in dealloc(),
	//	so applicationWillTerminate() can reliably test for memory leaks.
	ShutDownModelData(&md);
	
	return YES;
}

- (BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication *)theApplication
{
	return YES;
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
	//	If the user chooses Quit from the menu,
	//	Cocoa will *not* ask windows whether they should close or not,
	//	so we must do it ourselves.
	if ([self windowShouldClose:self] == YES)
		return NSTerminateNow;
	else
		return NSTerminateCancel;
}

- (void)applicationWillTerminate:(NSNotification *)aNotification
{
	//	Shut down the MIDI player.
	ShutDownMIDIPlayer();

	//	Free memory used for localized dictionary.
	RefreshLocalization("--");

	//	We should be clean.
	if (gMemCount != 0)
		ErrorMessage("Memory allocated does not equal memory freed.\r\nPlease report this error to\r\n\twww.geometrygames.org/contact.html", "Memory Leak", false);
}

@end
