//	TorusGames-Win32-WndProc.c
//
//	Provides MainWndProc(), DrawingWndProc() and supporting functions
//	for the Win32 version of Torus Games.
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#include "TorusGames-Win32.h"
#include "TorusGamesIDs.h"
#include <commdlg.h>	//	for Save dialog
#include <stdio.h>		//	for sprintf()
#include <shellapi.h>	//	for ShellExecute()
#include <commctrl.h>	//	for tab control in preferences panel
//#include <imm.h>		//	for Input Method Editor support
						//		[IME not needed -- we compose characters manually]
#include "GeometryGames-Win32-LoadEntryPoints.h"

//	What is the minimum acceptable size for the window's client area?
//	Make this a multiple of 3 so that the ViewSmall image
//	will sit cleanly within ViewTiling image.
#define MIN_CLIENT_SIZE				300
	
//	The CROSSWORD_CHARACTER_INSET prevents a "descender",
//	for example the tail of a 'g', from touching the square's border.
//	It's needed only in the Win32 version, not the Mac OS or iPhone OS version.
#define CROSSWORD_CHARACTER_INSET	(CROSSWORD_CHARACTER_TEXTURE_SIZE_DESKTOP / 32)

//	How should the Change Game panel's buttons be layed out?
//	Be sure to coordinate these with the dialog's dimensions
//	as given in TorusGames.rc.
#define CHANGE_GAME_STRIDE_H		64		//	in "dialog layout units", center-to-center
#define CHANGE_GAME_STRIDE_V		56		//	in "dialog layout units", center-to-center
#define CHANGE_GAME_MARGIN_TOP		 8		//	in "dialog layout units"
#define CHANGE_GAME_BUTTON_SIZE		32		//	in "dialog layout units"
#define CHANGE_GAME_MARGIN_MIDDLE	 4		//	in "dialog layout units"
#define CHANGE_GAME_LABEL_HEIGHT	20		//	in "dialog layout units"


//	Command IDs -- main window

#define IDC_FILE_NEW					0x0000
#define IDC_FILE_EXIT					0x0001

#define IDC_GAME_RESET					0x0100
#define IDC_GAME_SPECIAL				0x0101
#define IDC_GAME_TICTACTOE				0x0102
#define IDC_GAME_MAZE					0x0103
#define IDC_GAME_CROSSWORD				0x0104
#define IDC_GAME_WORDSEARCH				0x0105
#define IDC_GAME_JIGSAW					0x0106
#define IDC_GAME_CHESS					0x0107
#define IDC_GAME_POOL					0x0108
#define IDC_GAME_APPLES					0x0109

#define IDC_OPTIONS_HUMAN				0x0200
#define IDC_OPTIONS_COMPUTER			0x0201
#define IDC_OPTIONS_EXTRA_EASY			0x0202
#define IDC_OPTIONS_EASY				0x0203
#define IDC_OPTIONS_MEDIUM				0x0204
#define IDC_OPTIONS_HARD				0x0205
#define IDC_OPTIONS_EXTRA_HARD			0x0206
#define IDC_OPTIONS_GLIDE_AXES			0x0207
#define IDC_OPTIONS_SOUND				0x0208
#define IDC_OPTIONS_EXPORT_COPY			0x0220
#define IDC_OPTIONS_EXPORT_SAVE			0x0221

#define IDC_TOPOLOGY_TORUS				0x0300
#define IDC_TOPOLOGY_KLEIN				0x0301

#define IDC_VIEW_DOMAIN					0x0400
#define IDC_VIEW_SMALL					0x0401
#define IDC_VIEW_TILING					0x0402

#define IDC_LANGUAGE_DE					0x0510
#define IDC_LANGUAGE_EN					0x0511
#define IDC_LANGUAGE_ES					0x0512
#define IDC_LANGUAGE_FR					0x0513
#define IDC_LANGUAGE_HU					0x0514
#define IDC_LANGUAGE_IT					0x0515
#define IDC_LANGUAGE_JA					0x0516
#define IDC_LANGUAGE_KO					0x0517
#define IDC_LANGUAGE_NL					0x0518
#define IDC_LANGUAGE_PT					0x0519
#define IDC_LANGUAGE_RU					0x051A
#define IDC_LANGUAGE_VI					0x051B

#define IDC_HELP_HOW_TO_PLAY			0x0600
#define IDC_HELP_WRAPAROUND_UNIVERSE	0x0601
#define IDC_HELP_PRACTICE_BOARD			0x0602
#define IDC_HELP_CONTACT				0x0603
#define IDC_HELP_ABOUT					0x0604

//	Command IDs -- Change Game window
//
//	Technical Note:  The dialog box's close box sends IDCANCEL = 0x0002,
//	so let's start our own commands at 0x0100 to avoid conflicts.
#define IDC_CHANGE_GAME_COMMAND_BASE	0x0100


//	How far should the board scroll in response to one notch
//	of mouse wheel rotation?
//
//	Note:  One notch on a standard mouse wheel generates
//	a rotation distance of 120 ( = WHEEL_DELTA);
//	more subtle mouse wheels may generate half notches (60),
//	quarter notches (30), and so on.
#define MOUSEWHEEL_SPEED_WIN	(0.05 / ((double)WHEEL_DELTA))


//	Support for WGL extensions.

#define WGL_DRAW_TO_WINDOW_ARB         0x2001
#define WGL_ACCELERATION_ARB           0x2003
#define WGL_SUPPORT_OPENGL_ARB         0x2010
#define WGL_DOUBLE_BUFFER_ARB          0x2011
#define WGL_COLOR_BITS_ARB             0x2014
#define WGL_ALPHA_BITS_ARB             0x201B
#define WGL_DEPTH_BITS_ARB             0x2022
#define WGL_STENCIL_BITS_ARB           0x2023
#define WGL_FULL_ACCELERATION_ARB      0x2027
#define WGL_SAMPLE_BUFFERS_ARB         0x2041
#define WGL_SAMPLES_ARB                0x2042

typedef const char	*(WINAPI * PFNWGLGETEXTENSIONSSTRINGARBPROC)(HDC hdc);
typedef bool		(WINAPI * PFNWGLCHOOSEPIXELFORMATARBPROC)(HDC hdc, const int *piAttribIList, const FLOAT *pfAttribFList, UINT nMaxFormats, int *piFormats, UINT *nNumFormats);
typedef bool		(WINAPI * PFNWGLSWAPINTERVALEXTPROC )(int interval);

static  PFNWGLSWAPINTERVALEXTPROC	wglSwapInterval	= NULL;


static INT_PTR CALLBACK	ChangeGameDialogProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL CALLBACK	DeleteChooseGameBitmapsAndClearFonts(HWND aChildWindow, LPARAM aParameter);
static bool				SetUpWindowData(HWND aWindow);
static void				ShutDownWindowData(WindowData *wd);
static bool				SetUpOpenGL(WindowData *wd);
static void				ShutDownOpenGL(WindowData *wd);
static bool				SetUpDrawingPanel(WindowData *wd);
static bool				SetUpDrawingPanelWithPixelFormat(WindowData *wd, unsigned int aPixelFormat);
static unsigned int		ChooseMultisamplePixelFormat(WindowData *wd);
static void				ShutDownDrawingPanel(WindowData *wd);
static bool				WholeWordSubstring(const char *aString, const char *aPotentialSubstring);
static void				CharacterInputWin32(WindowData *wd, WCHAR aCharacter);
static void				MenuPrepare(WindowData *wd, HMENU aMenu);
static WCHAR			*GameSpecificResetText(GameType aGame, WCHAR *aValueBuffer, unsigned int aValueBufferLength);
static void				MenuCommand(WindowData *wd, WORD aCommand);
static BOOL CALLBACK	PostCloseMessage(HWND aWindow, LPARAM anUnusedParameter);
static void				ChangeGameWin32(WindowData *wd, GameType aNewGame);
static void				PrepareFonts(WindowData *wd);
static void				PrepareSideWindow(WindowData *wd);
static void				CreateSideWindow(WindowData *wd, char *aWindowTitleKey, unsigned int aWidth, unsigned int aHeight1, unsigned int aHeight2, DWORD aStyleSet, FontIndex aFont);
static unsigned int		StringToUnsignedInt(WCHAR *aString);
static void				CopyTheImage(WindowData *wd);
static void				SaveTheImage(WindowData *wd);
static bool				CopyDirectoryName(WCHAR *aPathName, DWORD aPathNameLength, WCHAR *aDirectoryName, DWORD aDirectoryNameLength);
static HANDLE			FetchTheImage(WindowData *wd);
static void				ChangeLanguage(const char aLanguageCode[3]);
static void				RefreshMirroring(HWND aWindow);
static HMENU			BuildLocalizedMenuBar(void);
static void				OpenHelpPage(DWORD aHelpCommand);
static void				SetSync(WindowData *wd);
static void				SetMinWindowSize(HWND aWindow, POINT *aMinSize);
static void				SetMaxWindowSize(HWND aWindow, POINT *aMaxPosition, POINT *aMaxSize);
static void				ResizeDrawingPanel(WindowData *wd, unsigned int aWidth, unsigned int aHeight);
static void				DisplayChange(WindowData *wd);
static void				GetAndResetMouseLocation(HWND aDrawingWindow, LPARAM lParam, double *theMouseH, double *theMouseV);
static void				ReleaseTorusCursor(WindowData *wd);
static void				SetConvenientMouseLocation(HWND aDrawingWindow);
static void				MakeContentSquare3(HWND hWnd, WPARAM aSizingMode, RECT *aWindowRect);
static void				PaintWindow(WindowData *wd);


//	Callback function to process messages in the main window.
LRESULT CALLBACK MainWndProc(
	HWND	hWnd,
	UINT	uMsg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData	*wd = NULL;

	//	For all messages except WM_CREATE and WM_DESTROY, locate the WindowData.
	//	If we confirm once and for all that wd != NULL, then
	//	the routines we call in the message loop can safely rely on it.
	if (hWnd != NULL)
	{
		if (uMsg != WM_CREATE
		 && uMsg != WM_DESTROY)
		{
			wd = (WindowData *) GetWindowLongPtr(hWnd, GWLP_USERDATA);
			if (wd == NULL)
				return DefWindowProc(hWnd, uMsg, wParam, lParam);
		}
		else
			wd = NULL;
	}
	else
		return 0;	//	should never occur

	//	Process the message.
	switch (uMsg)
	{
		case WM_CREATE:
			if (SetUpWindowData(hWnd))
				return 0;	//	success -- continue normally
			else
				return -1;	//	failure -- abort window creation

		case WM_CLOSE:
			//	Shut down our window data, including itsRenderingContext,
			//	itsDeviceContext and itsDrawingPanel.  Otherwise Windows will destroy
			//	child windows prematurely, leaving itsRenderingContext invalid.
			ShutDownWindowData(wd);
			DestroyWindow(hWnd);
			return 0;

//		case WM_DESTROY:
//			return 0;

		case WM_KEYDOWN:
			//	Note:  Return, Escape, Space, Tab and Backspace
			//	can be handled in WM_CHAR (as 0x0D, 0x1B, ' ',
			//	'\t' and '\b', respectively), but the arrow keys 
			//	don't generate characters so they must be handled here.
			switch (wParam)	//	virtual key code
			{
				case VK_LEFT:	CharacterInput(&wd->md, L'←');	break;
				case VK_RIGHT:	CharacterInput(&wd->md, L'→');	break;
				case VK_DOWN:	CharacterInput(&wd->md, L'↓');	break;
				case VK_UP:		CharacterInput(&wd->md, L'↑');	break;
				
				default:
					return DefWindowProc(hWnd, uMsg, wParam, lParam);
			}
			return 0;
		
		case WM_CHAR:
			
			if ((WCHAR) wParam == 0x1B)	//	ASCII escape character (ESC)
			{
				ReleaseTorusCursor(wd);
			}
			else
			if (wd->md.itsGame == GameCrossword)	//	GameCrossword will want character input.
			{
				CharacterInputWin32(wd, (WCHAR) wParam);
			}
			else
			if ((WCHAR) wParam == 'f')	//	"secret" keyboard command to show frame rate
			{
				gShowFrameRate = ! gShowFrameRate;
				EnumThreadWindows(	GetCurrentThreadId(),
									SetWindowTitle,
									(LPARAM) (gShowFrameRate ? L"…" : NULL));
			}
			else
			if ((WCHAR) wParam == 's')	//	"secret" keyboard command to sync to the monitor
			{
				wd->md.itsSyncFlag = ! wd->md.itsSyncFlag;
				SetSync(wd);
			}
			else
			{
				//	Otherwise assume the user is flailing.
				//	Perhaps s/he doesn't know how to exit torus cursor mode?
				if (GetCapture() != NULL)
				{
					ReleaseTorusCursor(wd);
					ErrorMessage(GetLocalizedText("TorusExitMessage"), GetLocalizedText("TorusExitTitle"), false);
				}
			}
			
			return 0;

		case WM_MOUSEWHEEL:
			MouseWheel(&wd->md, 0, - MOUSEWHEEL_SPEED_WIN * ((double)(signed short int)HIWORD(wParam)));
			return 0;

		case WM_INITMENUPOPUP:
			MenuPrepare(wd, (HMENU) wParam);
			return 0;

		case WM_COMMAND:
			if (lParam == 0)	//	command comes from menu
				MenuCommand(wd, LOWORD(wParam));
			else				//	command comes from toolbar
			{
				if ((HWND)lParam == wd->itsResetButton)
					ResetGame(&wd->md);
				if ((HWND)lParam == wd->itsChangeGameButton)
					ShowChangeGameDialog(wd);
					
				//	Don't let itsResetButton or itsChangeGameButton steal the keyboard focus.
				SetFocus(hWnd);
			}
			return 0;

//		case WM_EXITMENULOOP:
//		case WM_EXITSIZEMOVE:
//		case WM_ACTIVATE:
//			//	Warning:  wd will be NULL when a WM_ACTIVATE event
//			//	arrives during window destruction.
//			return 0;

		case WM_GETMINMAXINFO:
			SetMinWindowSize(	hWnd,
								&((MINMAXINFO*)lParam)->ptMinTrackSize);
			SetMaxWindowSize(	hWnd,
								&((MINMAXINFO*)lParam)->ptMaxPosition,
								&((MINMAXINFO*)lParam)->ptMaxSize);
			return 0;

		case WM_SIZING:
			MakeContentSquare3(hWnd, wParam, (RECT *)lParam);
			return TRUE;

		case WM_SIZE:
			MoveWindow(	wd->itsResetButton,
						0,
						0,
						LOWORD(lParam) / 2,			//	discards fractional part
						TORUS_GAMES_WIN32_TOOLBAR_HEIGHT,
						true);
			MoveWindow(	wd->itsChangeGameButton,
						LOWORD(lParam) / 2,			//	discards fractional part
						0,
						(LOWORD(lParam) + 1) / 2,	//	recovers fractional part
						TORUS_GAMES_WIN32_TOOLBAR_HEIGHT,
						true);
			MoveWindow(	wd->itsDrawingPanel,
						0,
						TORUS_GAMES_WIN32_TOOLBAR_HEIGHT,
						LOWORD(lParam),
						HIWORD(lParam) - TORUS_GAMES_WIN32_TOOLBAR_HEIGHT,
						false);
			return 0;

		case WM_DISPLAYCHANGE:
			DisplayChange(wd);
			return 0;

		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
}


void ShowChangeGameDialog(
	WindowData	*wd)
{
	DialogBoxParam(	GetModuleHandle(NULL),
					(void *) IDD_CHANGE_GAME_DIALOG,
					wd->itsWindow,
					(DLGPROC)ChangeGameDialogProc,
					(LPARAM) wd);
}

static INT_PTR CALLBACK ChangeGameDialogProc(
	HWND	hDlg,
	UINT	msg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData		*wd = NULL;
	HWND			theParentWindow;
	RECT			theParentRect,
					theDialogRect;
	WCHAR			theText[128];
	HDC				theDeviceContext;
	unsigned int	thePixelsPerInch,
					theFontSizeInPoints,
					theFontSizeInPixels,
					theRow,
					theCol;
	RECT			theChildRect;
	HWND			theButton,
					theLabel;
	GameType		theNewGame;

	LOGFONT			theLogicalFont =
					{
						0,				//	font height gets set below
						0, 0, 0, 0, 0, 0, 0, DEFAULT_CHARSET, 0, 0, 0, 0,
						L"Tahoma"		//	font face name
					};
		
	static const GameType theGame[3][3] =
	{
		{GameTicTacToe,		GameMaze,	GameCrossword	},
		{GameWordSearch,	GameJigsaw,	GameChess		},
		{GamePool,			GameApples,	GameIntro		}
	};
	static const WCHAR *theFileName[3][3] =
	{
		{L"Buttons\\TicTacToe.bmp",		L"Buttons\\Maze.bmp",	L"Buttons\\Crossword.bmp"	},
		{L"Buttons\\WordSearch.bmp",	L"Buttons\\Jigsaw.bmp",	L"Buttons\\Chess.bmp"		},
		{L"Buttons\\Pool.bmp",			L"Buttons\\Apples.bmp",	L"Buttons\\QuestionMark.bmp"}
	};
	static const char *theLabelKey[3][3] =
	{
		{"Tic-Tac-Toe",	"Maze",				"Crossword"	},
		{"Word Search",	"Jigsaw Puzzle",	"Chess"		},
		{"Pool",		"Apples",			"Help"		}
	};

	switch (msg)
	{
		case WM_INITDIALOG:

			//	lParam contains a pointer to the parent window's WindowData.
			//	Stash a copy in this dialog box as well.
			wd = (WindowData *) lParam;
			SetWindowLongPtr(hDlg, GWLP_USERDATA, (LONG_PTR) wd);
			if (wd == NULL)	//	should never occur
				return true;	//	set focus normally
			
			//	itsChangeGamePanelFont should not be in use.
			if (wd->itsChangeGamePanelFont != NULL)
				return true;	//	set focus normally
			
			//	Position the dialog over its parent window.
			//	Position it dead center left-to-right,
			//	but allocate extra vertical space in a 1:3 ratio,
			//	with the dialog sitting above the parent's center.
			theParentWindow = GetParent(hDlg);
			if (theParentWindow == NULL)
				theParentWindow = GetDesktopWindow();
			GetWindowRect(theParentWindow, &theParentRect);
			GetWindowRect(hDlg, &theDialogRect);
			SetWindowPos(	hDlg,
							HWND_TOP,
							theParentRect.left + ((theParentRect.right - theParentRect.left) - (theDialogRect.right - theDialogRect.left))/2,
							theParentRect.top  + ((theParentRect.bottom - theParentRect.top) - (theDialogRect.bottom - theDialogRect.top))/4,
							0,	//	ignored with SWP_NOSIZE
							0,	//	ignored with SWP_NOSIZE
							SWP_NOSIZE);

			//	Set the window title.
			SetWindowText(hDlg, GetLocalizedTextAsWCHAR("Choose a Game", theText, BUFFER_LENGTH(theText)));

			//	Create the font for the nine button labels.
			theDeviceContext = GetDC(NULL);
			thePixelsPerInch = GetDeviceCaps(theDeviceContext, LOGPIXELSY);
			ReleaseDC(NULL, theDeviceContext);
			theFontSizeInPoints			= 12;
			theFontSizeInPixels			= -MulDiv(theFontSizeInPoints, thePixelsPerInch, 72);
			theLogicalFont.lfHeight		= theFontSizeInPixels;
			wd->itsChangeGamePanelFont	= CreateFontIndirect(&theLogicalFont);

			//	Create the nine buttons.
			for (theRow = 0; theRow < 3; theRow++)
			{
				for (theCol = 0; theCol < 3; theCol++)
				{
					//	Create the image button.

					//	The conversion from dialog layout units (DLU) to pixels
					//	is anisotropic (!), at least with the Arial font.
					//	We want the button to be square no matter what,
					//	so we'll convert its height according to Microsoft's
					//	official DLU-to-pixels conversion, and then set
					//	its width (in pixels) to equal its height (in pixels).
					//	The reason we fudge the width and not the height is
					//	that the button as plenty of slack horizontally,
					//	but vertically it needs to respect the layout.
					
					//	Create a rectangle with zero width (and all dimensions in DLU).
					theChildRect.left	= theCol * CHANGE_GAME_STRIDE_H
										+ CHANGE_GAME_STRIDE_H/2;
					theChildRect.top	= theRow * CHANGE_GAME_STRIDE_V
										+ CHANGE_GAME_MARGIN_TOP;
					theChildRect.right	= theChildRect.left
										+ 0;
					theChildRect.bottom	= theChildRect.top
										+ CHANGE_GAME_BUTTON_SIZE;
					
					//	Convert from DLU to pixels.
					MapDialogRect(hDlg, &theChildRect);
					
					//	Adjust the width (in pixels) to equal the height (in pixels).
					theChildRect.left	-= (theChildRect.bottom - theChildRect.top)/2;
					theChildRect.right	+= (theChildRect.bottom - theChildRect.top)/2;

					//	Create the button.
					theButton = CreateWindow(
						L"button",	//	built-in window class name
						L"n/a",
						WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON | BS_BITMAP,
						theChildRect.left, theChildRect.top,
						theChildRect.right - theChildRect.left, theChildRect.bottom - theChildRect.top,
						hDlg,
						(HMENU)(IDC_CHANGE_GAME_COMMAND_BASE + theGame[theRow][theCol]),
						GetModuleHandle(NULL), NULL);

					//	Load the bitmap at the desired size,
					//	and pass it to the button.
					//	When we process WM_DESTROY we'll free the bitmaps.
					SendMessage(theButton,
								BM_SETIMAGE,
								(WPARAM)IMAGE_BITMAP,
								(LPARAM)LoadImage(NULL,
									theFileName[theRow][theCol],
									IMAGE_BITMAP,
									theChildRect.right - theChildRect.left,
									theChildRect.bottom - theChildRect.top,
									LR_LOADFROMFILE));

					//	Create the text label.

					//	Lay out the text area in DLU.
					theChildRect.left	= theCol * CHANGE_GAME_STRIDE_H
										+ 0;
					theChildRect.top	= theRow * CHANGE_GAME_STRIDE_V
										+ CHANGE_GAME_MARGIN_TOP
										+ CHANGE_GAME_BUTTON_SIZE
										+ CHANGE_GAME_MARGIN_MIDDLE;
					theChildRect.right	= theChildRect.left
										+ CHANGE_GAME_STRIDE_H;
					theChildRect.bottom	= theChildRect.top
										+ CHANGE_GAME_LABEL_HEIGHT;
					
					//	Convert DLU to pixels.
					MapDialogRect(hDlg, &theChildRect);

					//	Create the label.
					theLabel = CreateWindow(
						L"static",	//	built-in window class name
						GetLocalizedTextAsWCHAR(theLabelKey[theRow][theCol], theText, BUFFER_LENGTH(theText)),
						WS_CHILD | WS_VISIBLE | SS_CENTER,
						theChildRect.left, theChildRect.top,
						theChildRect.right - theChildRect.left, theChildRect.bottom - theChildRect.top,
						hDlg,
						(HMENU) 0,	//	ignored
						GetModuleHandle(NULL), NULL);
					
					//	Set the font.
					SendMessage(theLabel, WM_SETFONT, (WPARAM) wd->itsChangeGamePanelFont, true);
				}
			}
			
			//	Set the keyboard focus to NULL to ignore keystrokes.
			SetFocus(NULL);

			return false;	//	Don't let the caller set the keyboard focus.

		case WM_DESTROY:

			EnumChildWindows(hDlg, DeleteChooseGameBitmapsAndClearFonts, 0);

			wd = (WindowData *) GetWindowLongPtr(hDlg, GWLP_USERDATA);
			if (wd != NULL
			 && wd->itsChangeGamePanelFont != NULL)
			{
				DeleteObject(wd->itsChangeGamePanelFont);
				wd->itsChangeGamePanelFont = NULL;
			}

			return 0;

		case WM_COMMAND:
			wd = (WindowData *) GetWindowLongPtr(hDlg, GWLP_USERDATA);
			if (wd != NULL)
			{
				//	The window's close box sends IDCANCEL = 0x0002.
				//	The only other commands should be offsets
				//	from IDC_CHANGE_GAME_COMMAND_BASE, which represent games.
				if (msg >= IDC_CHANGE_GAME_COMMAND_BASE)
				{
					//	Which game did the user request?
					theNewGame = (GameType)(LOWORD(wParam) - IDC_CHANGE_GAME_COMMAND_BASE);

					//	Load the requested game.
					ChangeGameWin32(wd, theNewGame);

					//	If the user asked for GameIntro,
					//	show the How to Play and Wraparound Universe pages.
					if (theNewGame == GameIntro)
					{
						//	Load Wraparound Universe first,
						//	so How to Play ends up on top.
						OpenHelpPage(IDC_HELP_WRAPAROUND_UNIVERSE);
						OpenHelpPage(IDC_HELP_HOW_TO_PLAY);
					}
				}
				
				EndDialog(hDlg, true);
			}
			SetWindowLong(hDlg, DWL_MSGRESULT, 0);	//	return value
			return true;	//	message processed

		default:
			return false;	//	message ignored
	}
}

static BOOL CALLBACK DeleteChooseGameBitmapsAndClearFonts(
	HWND	aChildWindow,
	LPARAM	aParameter)
{
	HBITMAP	theBitmap;

	UNUSED_PARAMETER(aParameter);

	//	Pass a NULL image to aChildWindow.
	//	If aChildWindow is a static text control, 
	//		it will pass back NULL.
	//	If aChildWindow is a button with an image,
	//		it will relinquish that image
	//		and pass back the handle to it.
	theBitmap = (HBITMAP)SendMessage(
							aChildWindow,
							BM_SETIMAGE,
							(WPARAM)IMAGE_BITMAP,
							(LPARAM)NULL);
	if (theBitmap != NULL)
	{
		DeleteObject(theBitmap);
		theBitmap = NULL;
	}
	
	//	Set the font to NULL, in preparation for releasing the shared font.
	SendMessage(aChildWindow, WM_SETFONT, (WPARAM)NULL, false);
	
	return true;
}


//	Callback function to process messages in the drawing panel.
LRESULT CALLBACK DrawingWndProc(
	HWND	aDrawingWindow,
	UINT	uMsg,
	WPARAM	wParam,
	LPARAM	lParam)
{
	WindowData	*wd;
	double		theMouseH,
				theMouseV,
				theMouseDeltaH,
				theMouseDeltaV;

	//	If we confirm once and for all that wd != NULL, then the routines
	//	we call in the message switch can safely rely on it.
	//	See MainWndProc() for an explanation of why the WindowData
	//	isn't available at WM_DESTROY time.
	if (aDrawingWindow != NULL)
	{
		if (uMsg != WM_CREATE
		 && uMsg != WM_DESTROY)
		{
			wd = (WindowData *) GetWindowLongPtr(GetParent(aDrawingWindow), GWLP_USERDATA);
			if (wd == NULL)
				return DefWindowProc(aDrawingWindow, uMsg, wParam, lParam);
		}
		else
			wd = NULL;
	}
	else
		return 0;	//	should never occur

	switch (uMsg)
	{
		case WM_CREATE:
			return 0;

		case WM_LBUTTONDOWN:
		case WM_LBUTTONDBLCLK:
			SetCapture(aDrawingWindow);	//	usually redundant
			SetCursor(NULL);			//	usually redundant
			GetAndResetMouseLocation(aDrawingWindow, lParam, &theMouseH, &theMouseV);
			MouseDown(&wd->md, theMouseH, theMouseV, (wParam & (MK_CONTROL | MK_SHIFT)) != 0, uMsg == WM_LBUTTONDBLCLK);
			return 0;

		case WM_RBUTTONDOWN:
			return DrawingWndProc(aDrawingWindow, WM_LBUTTONDOWN, wParam | MK_SHIFT, lParam);

		case WM_MOUSEMOVE:
			if (GetCapture() == aDrawingWindow)
			{
				GetAndResetMouseLocation(aDrawingWindow, lParam, &theMouseDeltaH, &theMouseDeltaV);
				if (theMouseDeltaH != 0.0 || theMouseDeltaV != 0.0)	//	Ignore the secondary WM_MOUSEMOVE event that SetCursorPos() triggers.
					MouseMove(&wd->md, theMouseDeltaH, theMouseDeltaV);
			}
			return 0;

		case WM_LBUTTONUP:
		case WM_RBUTTONUP:
			MouseUp(&wd->md);
			if (wd->md.itsDoubleClickFlag)
				ReleaseTorusCursor(wd);
			return 0;

		case WM_PAINT:
			PaintWindow(wd);
			return 0;

		case WM_SIZE:
			ResizeDrawingPanel(wd, LOWORD(lParam), HIWORD(lParam));
			return 0;

		default:
			return DefWindowProc(aDrawingWindow, uMsg, wParam, lParam);
	}
}


static bool SetUpWindowData(HWND aWindow)
{
	WindowData		*wd;
	unsigned int	i;
	WCHAR			theButtonText[128];

	//	Antialiasing doesn't work for Japanese text (under Windows XP at least).
	//	After struggling with the fonts for a couple hours,
	//	I finally Googled "WindowsXP doesn't antialias Japanese text"
	//	and found that this is a well-known problem.
	//	So, no need to provide a separate font for Japanese.
//	LOGFONT			theLogicalFontTahoma =
//					{
//						18,				//	font height
//						0, 0, 0, 0, 0, 0, 0, DEFAULT_CHARSET, 0, 0, 0, 0,
//						L"Tahoma"		//	font face name
//					},
//					theLogicalFontGothic =
//					{
//						18,				//	font height
//						0, 0, 0, 0, 0, 0, 0, DEFAULT_CHARSET, 0, 0, 0, 0,
//						L"MS UI Gothic"	//	font face name
//					};
	LOGFONT			theLogicalFont =
					{
						18,				//	font height
						0, 0, 0, 0, 0, 0, 0, DEFAULT_CHARSET, 0, 0, 0, 0,
						L"Tahoma"		//	font face name
					};

	//	Attempt to allocate the WindowData and -- whether or not the call
	//	succeeds -- attach the resulting pointer to the window.
	wd = GET_MEMORY(sizeof(WindowData));
	SetWindowLongPtr(aWindow, GWLP_USERDATA, (LONG_PTR) wd);

	//	If the memory allocation failed, return false
	//	to tell Windows to abort window construction.
	if (wd == NULL)
	{
		ErrorMessage("Couldn't get memory for WindowData.", "SetUpWindowData() Error", false);
		return false;
	}

	//	Record the window.
	wd->itsWindow = aWindow;

	//	Set pointers to NULL for robust error handling.
	wd->itsButtonFont						= NULL;
	wd->itsResetButton						= NULL;
	wd->itsChangeGameButton					= NULL;
	wd->itsChangeGamePanelFont				= NULL;
	wd->itsDrawingPanel						= NULL;
	wd->itsDeviceContext					= NULL;
	wd->itsRenderingContext					= NULL;
	wd->itsPreferencesPane					= NULL;
	for (i = 0; i < NumberOfFonts; i++)
		wd->itsWin32Fonts[i].itsHFONT = NULL;
	wd->itsWin32SideWindow.itsHWND			= NULL;
	wd->itsWin32SideWindow.itsTextField1	= NULL;
	wd->itsWin32SideWindow.itsTextField2	= NULL;

	//	Set platform-independent pointers to point to platform-dependent fonts.
	for (i = 0; i < NumberOfFonts; i++)
		wd->md.itsFonts[i] = &wd->itsWin32Fonts[i];
	
	//	Set platform-independent pointer to point to platform-dependent side window.
	wd->md.itsSideWindow = &wd->itsWin32SideWindow;

	//	Initialize the model's internal data.
	//	The game will initially be GameNone.
	SetUpModelData(&wd->md);

	//	Create itsDrawingPanel and set OpenGL up within it.
	if ( ! SetUpOpenGL(wd) )
		goto FailedToSetUpData;

	//	Create the toolbar.
	wd->itsResetButton = CreateWindow(
		L"button",	//	built-in window class name
		GameSpecificResetText(wd->md.itsGame, theButtonText, BUFFER_LENGTH(theButtonText)),
		WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
		0, 0,	//	Final placement gets set in response to WM_SIZE.
		128, TORUS_GAMES_WIN32_TOOLBAR_HEIGHT,
		wd->itsWindow, NULL, GetModuleHandle(NULL), NULL);
	wd->itsChangeGameButton	= CreateWindow(
		L"button",	//	built-in window class name
		GetLocalizedTextAsWCHAR("Change Game", theButtonText, BUFFER_LENGTH(theButtonText)),
		WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
		128, 0,	//	Final placement gets set in response to WM_SIZE.
		128, TORUS_GAMES_WIN32_TOOLBAR_HEIGHT,
		wd->itsWindow, NULL, GetModuleHandle(NULL), NULL);
	if (wd->itsResetButton == NULL || wd->itsChangeGameButton == NULL)
		goto FailedToSetUpData;
	
	//	Set the fonts in the toolbar buttons.
//	wd->itsButtonFont = CreateFontIndirect(SAME_LANGUAGE_CODE(gLanguageCode, "ja") ?
//								&theLogicalFontGothic : &theLogicalFontTahoma);
	wd->itsButtonFont = CreateFontIndirect(&theLogicalFont);
	SendMessage(wd->itsResetButton,      WM_SETFONT, (WPARAM) wd->itsButtonFont, true);
	SendMessage(wd->itsChangeGameButton, WM_SETFONT, (WPARAM) wd->itsButtonFont, true);

#ifdef CYLINDER_HACK
	//	Set window title.
	SetWindowText(wd->itsWindow, L"Cylindrical Tic-Tac-Toe");
#endif

	//	Report success.
	return true;

FailedToSetUpData:

	//	Clean up.
	ShutDownWindowData(wd);	//	wd is known to be non-NULL here

	//	Report failure.
	ErrorMessage("Failed to set up window.", "SetUpWindowData() Error", false);
	return false;
}


static void ShutDownWindowData(WindowData *wd)
{
	//	Switch to the null game to free any allocated memory.
	ChangeGameWin32(wd, GameNone);
	
	//	Detach itsButtonFont from the buttons.
	//	The buttons should get freed automatically,
	//	as children of the main window.
	if (wd->itsResetButton != NULL)
		SendMessage(wd->itsResetButton, WM_SETFONT, (WPARAM)NULL, false);
	if (wd->itsChangeGameButton != NULL)
		SendMessage(wd->itsChangeGameButton, WM_SETFONT, (WPARAM)NULL, false);
	
	//	Free itsButtonFont.
	if (wd->itsButtonFont != NULL)
	{
		DeleteObject(wd->itsButtonFont);
		wd->itsButtonFont = NULL;
	}

	//	Free any memory that the null game might hold.
	ShutDownModelData(&wd->md);

	//	Shut down the OpenGL rendering context.
	ShutDownOpenGL(wd);

	//	Once all dependent data have been cleaned up,
	//	then free the WindowData itself.
	SetWindowLongPtr(wd->itsWindow, GWLP_USERDATA, 0);
	FREE_MEMORY(wd);
}


static bool SetUpOpenGL(WindowData *wd)
{
	bool	theLoadResult;

	//	Create a drawing panel, get a device context and
	//	attach an OpenGL rendering context.
	if ( ! SetUpDrawingPanel(wd) )
		return false;

	//	On Win32 we must explicitly load functions from OpenGL 1.2 or later.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	theLoadResult = LoadOpenGLFunctions();
	wglMakeCurrent(NULL, NULL);	//	Deactivate context before displaying error message.
	if ( ! theLoadResult )
		ErrorMessage("This program cannot continue because your computer's graphics library does not support \"Vertex Buffer Objects\".", "OpenGL 1.5 Required", true);
	
	//	Load an entry point for wglSwapInterval, if it's present.
	//	Caution:  My mother's Dell with Intel Extreme Graphics 
	//	does not support WGL_EXT_swap_control but nevertheless returns a non-NULL
	//	(but ultimately useless) entry point for wglGetProcAddress("wglSwapIntervalEXT").
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	wglSwapInterval = (PFNWGLSWAPINTERVALEXTPROC) wglGetProcAddress("wglSwapIntervalEXT");
	wglMakeCurrent(NULL, NULL);

	//	Set sync-to-monitor as desired.
	SetSync(wd);

	//	The following flags should already be false
	//	(either from the original window initialization
	//	or from a call to ShutDownOpenGL()), but reset them
	//	here just to be safe.
	wd->md.itsPreparedExtensions	= false;
	wd->md.itsPreparedShaders		= false;
	wd->md.itsPreparedTextures		= false;

	return true;
}


static void ShutDownOpenGL(WindowData *wd)
{
	//	Shut down application-specific OpenGL shaders and textures.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	ShutDownGraphicsAsNeeded(&wd->md);
	wglMakeCurrent(NULL, NULL);

	//	Shut down the drawing panel along with its OpenGL context.
	ShutDownDrawingPanel(wd);
}


static bool SetUpDrawingPanel(WindowData *wd)
{
	unsigned int	thePixelFormat;

	//	The multisampling extension presents a chicken-and-egg problem:
	//	we need to use it to set up the OpenGL rendering context,
	//	but we can't find out whether it's available -- or even get
	//	a pixel format -- until after the rendering context has been created
	//	and made current.

	//	Set up an OpenGL drawing panel with a plain (non-multisampling) buffer.
	if (SetUpDrawingPanelWithPixelFormat(wd, 0) != true)
		return false;

	//	Try for a multi-sample pixel format if the user wants one.
	//	Note:  For sure we'll have at least the default value of
	//	itsMultisamplingEnabledFlag, but at start-up we won't yet
	//	have tested or set itsMultisamplingAvailableFlag.
	//	We hope it's available, but if it's not ChooseMultisamplePixelFormat()
	//	will fail harmlessly.
	if (wd->md.itsMultisamplingEnabledFlag)
	{
		//	Use the plain context to access the WGL_ARB_multisample extension
		//	and select a multisample-capable pixel format.
		thePixelFormat = ChooseMultisamplePixelFormat(wd);

		//	If a multisample-capable pixel format is available,
		//	destroy the plain drawing panel and replace it
		//	with a multisample-capable panel.
		if (thePixelFormat != 0)
		{
			ShutDownDrawingPanel(wd);

			if (SetUpDrawingPanelWithPixelFormat(wd, thePixelFormat) != true)
			{
				ErrorMessage("Failed to create multisample-capable OpenGL context.  Falling back to plain OpenGL context.", "Full-scene antialiasing not working", false);
				wd->md.itsMultisamplingEnabledFlag = false;
				if (SetUpDrawingPanelWithPixelFormat(wd, 0) != true)
					return false;
			}
		}
	}

	return true;
}


static bool SetUpDrawingPanelWithPixelFormat(
	WindowData		*wd,
	unsigned int	aPixelFormat)	//	needed only for multisampling;  otherwise pass 0
{
	ErrorText				theErrorMessage	= NULL,
							theErrorTitle	= "SetUpDrawingPanel() failed";	//	default title, may be overridden
	RECT					theClientRect;
	PIXELFORMATDESCRIPTOR	thePixelFormatDescriptor =
	{
		sizeof(PIXELFORMATDESCRIPTOR),	//	size
		1,								//	version number
		  PFD_DRAW_TO_WINDOW			//	format
		| PFD_SUPPORT_OPENGL
	//	| PFD_STEREO	//	could be useful in the future, but not needed for anaglyphic stereo 3D
		| PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,					//	preferred format is rgba (not palette)
		32, 							//	preferred color buffer depth (including alpha bits)
		0, 0, 0, 0, 0, 0,				//	color bits ignored
		8,								//	alpha buffer depth
		0,								//	shift bit ignored
		0,								//	no accumulation buffer
		0, 0, 0, 0, 					//	accumulation bits ignored
		0, 								//	no Z-buffer
		0,								//	no stencil buffer
		0,								//	no auxiliary buffer
		0,								//	ignored
		PFD_MAIN_PLANE, 				//	underlay/overlay planes
		0,								//	ignored
		0,								//	underlay color
		0								//	ignored
	};
	
	if (wd->itsDrawingPanel != NULL)
	{
		theErrorMessage = "Window has pre-existing drawing panel.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}
	if (wd->itsDeviceContext != NULL)
	{
		theErrorMessage = "Window has pre-existing device context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}
	if (wd->itsRenderingContext != NULL)
	{
		theErrorMessage = "Window has pre-existing rendering context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	//	You'd think you could create itsDrawingPanel once and for all,
	//	and then replace its OpenGL rendering context as needed,
	//	but in practice that doesn't work.  So instead we must
	//	re-create itsDrawingPanel whenever we need to change
	//	the OpenGL rendering context (for example to change
	//	to a multisample-capable pixel format).
	//
	//	Note:  Calling CreateWindow() with the WS_VISIBLE style
	//	triggers a WM_SIZE message which leads to a ResizeDrawingPanel()
	//	call which adjusts the window size variables set earlier.
	GetClientRect(wd->itsWindow, &theClientRect);
	wd->itsDrawingPanel = CreateWindow(
		DRAWING_WINDOW_CLASS_NAME,
		NULL,
		WS_CHILD | WS_VISIBLE,
		theClientRect.left,										//	theClientRect.left should be 0
		theClientRect.top + TORUS_GAMES_WIN32_TOOLBAR_HEIGHT,	//	theClientRect.top  should be 0
		theClientRect.right - theClientRect.left,
		(theClientRect.bottom - theClientRect.top) - TORUS_GAMES_WIN32_TOOLBAR_HEIGHT,
		wd->itsWindow, 0, GetModuleHandle(NULL), NULL);
	if (wd->itsDrawingPanel == NULL)
	{
		theErrorMessage = "Can't create a window for OpenGL graphics.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	wd->itsDeviceContext = GetDC(wd->itsDrawingPanel);
	if (wd->itsDeviceContext == NULL)
	{
		theErrorMessage = "Can't get a device context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	if (aPixelFormat == 0)
	{
		aPixelFormat = ChoosePixelFormat(wd->itsDeviceContext, &thePixelFormatDescriptor);
		if (aPixelFormat == 0)
		{
			theErrorMessage = "Can't get a pixel format.";
			goto CleanUpSetUpDrawingPanelWithPixelFormat;
		}
	}

	if (DescribePixelFormat(wd->itsDeviceContext, aPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &thePixelFormatDescriptor))
	{
		if (thePixelFormatDescriptor.iPixelType	!= PFD_TYPE_RGBA
		 || (	thePixelFormatDescriptor.cColorBits	!= 24	//	Microsoft's documentation claims that cColorBits
		 	 && thePixelFormatDescriptor.cColorBits	!= 32))	//	excludes the alpha bits, but in reality it includes them.
		{
			theErrorMessage	= GetLocalizedText("ErrorColorMessage");
			theErrorTitle	= GetLocalizedText("ErrorColorTitle");
			goto CleanUpSetUpDrawingPanelWithPixelFormat;
		}

		if (thePixelFormatDescriptor.cAlphaBits	!= 8)
		{
			theErrorMessage = GetLocalizedText("ErrorAlphaMessage");
			theErrorTitle	= GetLocalizedText("ErrorAlphaTitle");
			goto CleanUpSetUpDrawingPanelWithPixelFormat;
		}
	}

	if ( ! SetPixelFormat(wd->itsDeviceContext, aPixelFormat, &thePixelFormatDescriptor) )
	{
		theErrorMessage = "Can't set the pixel format.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	wd->itsRenderingContext = wglCreateContext(wd->itsDeviceContext);
	if (wd->itsRenderingContext == NULL)
	{
		theErrorMessage = "Can't get an OpenGL rendering context.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	//	Let's do a "dry run" to test itsRenderingContext right here and now.
	//	That way we won't need to clutter up the rest of the code with
	//	error checking on wglMakeCurrent().  In the highly unlikely event
	//	that wglMakeCurrent() succeeds now and fails later on, nothing
	//	too terrible should happen because when wglMakeCurrent() fails
	//	to set the requested context it instead sets wglMakeCurrent(NULL, NULL),
	//	meaning that all subsequent OpenGL calls will be ignored.
	if ( ! wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext) )
	{
		theErrorMessage = "Can't make OpenGL rendering context current.";
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

	//	The generic Windows OpenGL driver ignores 3D graphics hardware.
	//	In practice this means that a computer from, say, Dell will run
	//	OpenGL fine because Dell provides a good driver, but a standard
	//	installation from a Windows XP CD will run OpenGL in software emulation
	//	(no doubt because Microsoft wants Microsoft-only DirectX graphics
	//	to run well while platform-independent OpenGL graphics runs poorly).
	//	I don't know any direct way to query whether hardware acceleration
	//	is active, but while our rendering context is current we can at least
	//	check the renderer name, and if it's "GDI Generic" notify the user.
	if (strcmp(GetRendererName(), "GDI Generic") == 0)
	{
		theErrorMessage = GetLocalizedText("ErrorDriverMessage");
		theErrorTitle	= GetLocalizedText("ErrorDriverTitle");
		goto CleanUpSetUpDrawingPanelWithPixelFormat;
	}

CleanUpSetUpDrawingPanelWithPixelFormat:

	//	Clear the OpenGL context before calling ErrorMessage().
	wglMakeCurrent(NULL, NULL);

	if (theErrorMessage != NULL)
	{
		ErrorMessage(theErrorMessage, theErrorTitle, false);
		ShutDownDrawingPanel(wd);
		return false;
	}
	else
		return true;
}


static unsigned int ChooseMultisamplePixelFormat(WindowData *wd)
{
	signed int							thePixelFormat;
	PFNWGLGETEXTENSIONSSTRINGARBPROC	wglGetExtensionsStringARB;
	const char							*theWGLExtensionsList;	//	always ASCII, never Unicode
	PFNWGLCHOOSEPIXELFORMATARBPROC		wglChoosePixelFormatARB;
	signed int							thePixelFormatIntAttributes[] =
	{
		WGL_DRAW_TO_WINDOW_ARB,	true,
		WGL_SUPPORT_OPENGL_ARB,	true,
		WGL_DOUBLE_BUFFER_ARB,	true,
		WGL_ACCELERATION_ARB,	WGL_FULL_ACCELERATION_ARB,
		WGL_COLOR_BITS_ARB,		32,	//	including alpha bits (contrary to some online documentation)
		WGL_ALPHA_BITS_ARB,		8,
		WGL_DEPTH_BITS_ARB,		0,
		WGL_STENCIL_BITS_ARB,	0,
		WGL_SAMPLE_BUFFERS_ARB,	true,
		WGL_SAMPLES_ARB,		8,	//	the code below assumes the number of samples is at array index 19
		0,						0
	};
	float								thePixelFormatFloatAttributes[] = {0.0f, 0.0f};
	bool								theFormatIsValid;
	unsigned int						theNumPixelFormats;

	//	If we encounter an error before finding a pixel format, we'll return 0.
	thePixelFormat = 0;

	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);

	wglGetExtensionsStringARB = (PFNWGLGETEXTENSIONSSTRINGARBPROC) wglGetProcAddress("wglGetExtensionsStringARB");
	if (wglGetExtensionsStringARB == NULL)
		goto CleanUpChooseMultisamplePixelFormat;

	theWGLExtensionsList = wglGetExtensionsStringARB(wd->itsDeviceContext);
	if (theWGLExtensionsList == NULL)
		goto CleanUpChooseMultisamplePixelFormat;

	if ( ! WholeWordSubstring(theWGLExtensionsList, "WGL_ARB_multisample" )
	 ||  ! WholeWordSubstring(theWGLExtensionsList, "WGL_ARB_pixel_format") )
		goto CleanUpChooseMultisamplePixelFormat;

	wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATARBPROC) wglGetProcAddress("wglChoosePixelFormatARB");
	if (wglChoosePixelFormatARB == NULL)
		goto CleanUpChooseMultisamplePixelFormat;

	//	Try for 8-sample multisampling, then 6 samples, then 4, then 2, then give up.
	//	Note:  Some applications might want to voluntarily use fewer than 8 samples,
	//	to keep memory usage down.  Conversely, someday we might want to be more
	//	ambitious and try for more than 8 samples.
	do
	{
		theFormatIsValid = wglChoosePixelFormatARB(
								wd->itsDeviceContext,
								thePixelFormatIntAttributes,
								thePixelFormatFloatAttributes,
								1,
								&thePixelFormat,
								&theNumPixelFormats)
						&&
							theNumPixelFormats >= 1;

		//	Make sure the WGL_SAMPLES_ARB parameter sits where we think it sits,
		//	then decrement its value by 2.
		if (thePixelFormatIntAttributes[18] != WGL_SAMPLES_ARB)
			goto CleanUpChooseMultisamplePixelFormat;
		thePixelFormatIntAttributes[19] -= 2;

	} while ( ! theFormatIsValid && thePixelFormatIntAttributes[19] > 0);

	if ( ! theFormatIsValid )
	{
		thePixelFormat = 0;
		goto CleanUpChooseMultisamplePixelFormat;
	}

CleanUpChooseMultisamplePixelFormat:

	wglMakeCurrent(NULL, NULL);

	return thePixelFormat;
}


static void ShutDownDrawingPanel(WindowData *wd)
{
	wglMakeCurrent(NULL, NULL);	//	redundant, but safe

	if (wd->itsRenderingContext != NULL)
	{
		wglDeleteContext(wd->itsRenderingContext);
		wd->itsRenderingContext = NULL;
	}

	if (wd->itsDeviceContext != NULL)
	{
		ReleaseDC(wd->itsDrawingPanel, wd->itsDeviceContext);
		wd->itsDeviceContext = NULL;
	}

	if (wd->itsDrawingPanel != NULL)
	{
		DestroyWindow(wd->itsDrawingPanel);
		wd->itsDrawingPanel = NULL;
	}
}


static bool WholeWordSubstring(
	const char	*aString,				//	null-terminated
	const char	*aPotentialSubstring)	//	null-terminated
{
	const char	*thePotentialMatch,
				*s,
				*p;
	bool		theMatchFlag;

	//	Does aPotentialSubstring occur as a complete space-delimited word in aString?

	//	Irrelevant technical detail:  an empty aPotentialSubstring
	//	(consisting of a terminating zero alone) counts as a substring
	//	of any string, even an empty one.

	//	Consider the first character in each space-delimited word of aString
	//	as the starting point for a potential match.
	thePotentialMatch = aString;
	while (*thePotentialMatch == ' ')
		thePotentialMatch++;
	do
	{
		//	Compare each character in aPotentialSubstring
		//	(excluding the terminating zero) to thePotentialMatch.
		s = thePotentialMatch;
		p = aPotentialSubstring;
		theMatchFlag = true;
		while (theMatchFlag && *p != 0)
		{
			if (*p++ != *s++)
				theMatchFlag = false;
		}

		//	If a space or terminating 0 follows a match, return true.
		if (theMatchFlag
		 && (*s == ' ' || *s == 0))
			return true;

		//	Skip to the beginning of the next whole word in thePotentialMatch.
		//		Skip by non-spaces.
		while (*thePotentialMatch != ' ' && *thePotentialMatch != 0)
			thePotentialMatch++;
		//		Skip by intervening spaces.
		while (*thePotentialMatch == ' ')
			thePotentialMatch++;

		//	Continue with the loop to examine the next whole word (if any).

	} while (*thePotentialMatch != 0);

	return false;
}


static void CharacterInputWin32(
	WindowData	*wd,
	WCHAR		aCharacter)
{
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	CharacterInput(&wd->md, aCharacter);	//	OK because WCHAR == Char16
	wglMakeCurrent(NULL, NULL);
}


static void MenuPrepare(
	WindowData	*wd,
	HMENU		aMenu)
{
	HMENU	theMainMenu		= 0;
	char	*theSpecialKey	= NULL;
	WCHAR	theMenuText[128];
	
	//	In theory the C code shouldn't care about the menu bar's physical layout,
	//	but popup menus carry no ID numbers, so alas I see no way to recognize
	//	them except by their position.  (The alternative would be to ignore aMenu
	//	and prepare all popup menus each time, even though only one is needed.)

	theMainMenu = GetMenu(wd->itsWindow);

	//	File menu
	if (aMenu == GetSubMenu(theMainMenu, 0))
	{
	}

	//	Game menu
	if (aMenu == GetSubMenu(theMainMenu, 1))
	{
		ModifyMenu(aMenu, IDC_GAME_RESET, MF_STRING, IDC_GAME_RESET,
			GameSpecificResetText(wd->md.itsGame, theMenuText, BUFFER_LENGTH(theMenuText)));

		//	Remove any "special" menu item that may be lingering
		//	from a previous MenuPrepare() call, then
		//	insert the special menu item associated with the current game,
		//	if it needs one.
		RemoveMenu(aMenu, IDC_GAME_SPECIAL, MF_BYCOMMAND);
		if (wd->md.itsGame == GameJigsaw)
			theSpecialKey = "Send to Back";
		else if (wd->md.itsGame == GameChess && wd->md.itsHumanVsComputer)
			theSpecialKey = "Computer Goes First";
		else
			theSpecialKey = NULL;
		if (theSpecialKey != NULL)
			InsertMenu(aMenu, 1, MF_BYPOSITION, IDC_GAME_SPECIAL,
				GetLocalizedTextAsWCHAR(theSpecialKey, theMenuText, BUFFER_LENGTH(theMenuText)));

		CheckMenuItem(aMenu, IDC_GAME_TICTACTOE,
			wd->md.itsGame == GameTicTacToe		? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_GAME_MAZE,
			wd->md.itsGame == GameMaze			? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_GAME_CROSSWORD,
			wd->md.itsGame == GameCrossword		? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_GAME_WORDSEARCH,
			wd->md.itsGame == GameWordSearch	? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_GAME_JIGSAW,
			wd->md.itsGame == GameJigsaw		? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_GAME_CHESS,
			wd->md.itsGame == GameChess      	? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_GAME_POOL,
			wd->md.itsGame == GamePool      	? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_GAME_APPLES,
			wd->md.itsGame == GameApples      	? MF_CHECKED : MF_UNCHECKED);
	}

	//	Options menu
	if (aMenu == GetSubMenu(theMainMenu, 2))
	{
		if (wd->md.itsGame == GameTicTacToe
		 || wd->md.itsGame == GameChess
		 || wd->md.itsGame == GamePool)
		{
			EnableMenuItem(aMenu, IDC_OPTIONS_HUMAN,	MF_ENABLED);
			EnableMenuItem(aMenu, IDC_OPTIONS_COMPUTER,	MF_ENABLED);

			CheckMenuItem (aMenu, IDC_OPTIONS_HUMAN,
				wd->md.itsHumanVsComputer ? MF_UNCHECKED : MF_CHECKED);
			CheckMenuItem (aMenu, IDC_OPTIONS_COMPUTER,
				wd->md.itsHumanVsComputer ? MF_CHECKED : MF_UNCHECKED);
		}
		else
		{
			EnableMenuItem(aMenu, IDC_OPTIONS_HUMAN,	MF_GRAYED);
			EnableMenuItem(aMenu, IDC_OPTIONS_COMPUTER,	MF_GRAYED);

			CheckMenuItem (aMenu, IDC_OPTIONS_HUMAN,	MF_UNCHECKED);
			CheckMenuItem (aMenu, IDC_OPTIONS_COMPUTER,	MF_UNCHECKED);
		}

		if (wd->md.itsGame == GameMaze
		 || wd->md.itsGame == GameJigsaw
		 || (	wd->md.itsGame == GameChess
			 && wd->md.itsHumanVsComputer)
		 || (	wd->md.itsGame == GamePool
			 && wd->md.itsHumanVsComputer)
		 || wd->md.itsGame == GameApples)
		{
			EnableMenuItem(aMenu, IDC_OPTIONS_EXTRA_EASY,	MF_ENABLED);
			EnableMenuItem(aMenu, IDC_OPTIONS_EASY,			MF_ENABLED);
			EnableMenuItem(aMenu, IDC_OPTIONS_MEDIUM,		MF_ENABLED);
			EnableMenuItem(aMenu, IDC_OPTIONS_HARD,			MF_ENABLED);
			EnableMenuItem(aMenu, IDC_OPTIONS_EXTRA_HARD,	wd->md.itsGame != GameChess ? MF_ENABLED : MF_GRAYED);
															//	Suppress Extra Hard chess because it's too slow.

			CheckMenuItem(aMenu, IDC_OPTIONS_EXTRA_EASY,
				wd->md.itsDifficultyLevel == 0 ? MF_CHECKED : MF_UNCHECKED);
			CheckMenuItem(aMenu, IDC_OPTIONS_EASY,
				wd->md.itsDifficultyLevel == 1 ? MF_CHECKED : MF_UNCHECKED);
			CheckMenuItem(aMenu, IDC_OPTIONS_MEDIUM,
				wd->md.itsDifficultyLevel == 2 ? MF_CHECKED : MF_UNCHECKED);
			CheckMenuItem(aMenu, IDC_OPTIONS_HARD,
				wd->md.itsDifficultyLevel == 3 ? MF_CHECKED : MF_UNCHECKED);
			CheckMenuItem(aMenu, IDC_OPTIONS_EXTRA_HARD,
				wd->md.itsDifficultyLevel == 4 ? MF_CHECKED : MF_UNCHECKED);
		}
		else
		{
			EnableMenuItem(aMenu, IDC_OPTIONS_EXTRA_EASY,	MF_GRAYED);
			EnableMenuItem(aMenu, IDC_OPTIONS_EASY,			MF_GRAYED);
			EnableMenuItem(aMenu, IDC_OPTIONS_MEDIUM,		MF_GRAYED);
			EnableMenuItem(aMenu, IDC_OPTIONS_HARD,			MF_GRAYED);
			EnableMenuItem(aMenu, IDC_OPTIONS_EXTRA_HARD,	MF_GRAYED);

			CheckMenuItem(aMenu, IDC_OPTIONS_EXTRA_EASY,	MF_UNCHECKED);
			CheckMenuItem(aMenu, IDC_OPTIONS_EASY,			MF_UNCHECKED);
			CheckMenuItem(aMenu, IDC_OPTIONS_MEDIUM,		MF_UNCHECKED);
			CheckMenuItem(aMenu, IDC_OPTIONS_HARD,			MF_UNCHECKED);
			CheckMenuItem(aMenu, IDC_OPTIONS_EXTRA_HARD,	MF_UNCHECKED);
		}

		if (wd->md.itsTopology == TopologyKlein)
		{
			EnableMenuItem(aMenu, IDC_OPTIONS_GLIDE_AXES, MF_ENABLED);
			CheckMenuItem (aMenu, IDC_OPTIONS_GLIDE_AXES,
				wd->md.itsShowGlideAxes ? MF_CHECKED : MF_UNCHECKED);
		}
		else
		{
			EnableMenuItem(aMenu, IDC_OPTIONS_GLIDE_AXES, MF_GRAYED);
			CheckMenuItem (aMenu, IDC_OPTIONS_GLIDE_AXES, MF_UNCHECKED);
		}

		CheckMenuItem (aMenu, IDC_OPTIONS_SOUND,
			gPlaySounds ? MF_CHECKED : MF_UNCHECKED);
	}

	//	Topology menu
	if (aMenu == GetSubMenu(theMainMenu, 3))
	{
		CheckMenuItem(aMenu, IDC_TOPOLOGY_TORUS,
			wd->md.itsTopology == TopologyTorus ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_TOPOLOGY_KLEIN,
			wd->md.itsTopology == TopologyKlein ? MF_CHECKED : MF_UNCHECKED);
	}

	//	View menu
	if (aMenu == GetSubMenu(theMainMenu, 4))
	{
		CheckMenuItem(aMenu, IDC_VIEW_DOMAIN,
			wd->md.itsView == ViewDomain ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_VIEW_SMALL,
			wd->md.itsView == ViewSmall  ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_VIEW_TILING,
			wd->md.itsView == ViewTiling ? MF_CHECKED : MF_UNCHECKED);
	}

	//	Language menu
	if (aMenu == GetSubMenu(theMainMenu, 5))
	{
		EnableMenuItem(aMenu, IDC_LANGUAGE_DE, IsValidLocale(MAKELCID(MAKELANGID(LANG_GERMAN,     SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_EN, IsValidLocale(MAKELCID(MAKELANGID(LANG_ENGLISH,    SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_ES, IsValidLocale(MAKELCID(MAKELANGID(LANG_SPANISH,    SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_FR, IsValidLocale(MAKELCID(MAKELANGID(LANG_FRENCH,     SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_HU, IsValidLocale(MAKELCID(MAKELANGID(LANG_HUNGARIAN,  SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_IT, IsValidLocale(MAKELCID(MAKELANGID(LANG_ITALIAN,    SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_JA, IsValidLocale(MAKELCID(MAKELANGID(LANG_JAPANESE,   SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_KO, IsValidLocale(MAKELCID(MAKELANGID(LANG_KOREAN,     SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_NL, IsValidLocale(MAKELCID(MAKELANGID(LANG_DUTCH,      SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_PT, IsValidLocale(MAKELCID(MAKELANGID(LANG_PORTUGUESE, SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_RU, IsValidLocale(MAKELCID(MAKELANGID(LANG_RUSSIAN,    SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);
		EnableMenuItem(aMenu, IDC_LANGUAGE_VI, IsValidLocale(MAKELCID(MAKELANGID(LANG_VIETNAMESE, SUBLANG_DEFAULT), SORT_DEFAULT), LCID_INSTALLED) ? MF_ENABLED : MF_GRAYED);

		CheckMenuItem(aMenu, IDC_LANGUAGE_DE, SAME_LANGUAGE_CODE(gLanguageCode, "de") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_EN, SAME_LANGUAGE_CODE(gLanguageCode, "en") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_ES, SAME_LANGUAGE_CODE(gLanguageCode, "es") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_FR, SAME_LANGUAGE_CODE(gLanguageCode, "fr") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_HU, SAME_LANGUAGE_CODE(gLanguageCode, "hu") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_IT, SAME_LANGUAGE_CODE(gLanguageCode, "it") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_JA, SAME_LANGUAGE_CODE(gLanguageCode, "ja") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_KO, SAME_LANGUAGE_CODE(gLanguageCode, "ko") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_NL, SAME_LANGUAGE_CODE(gLanguageCode, "nl") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_PT, SAME_LANGUAGE_CODE(gLanguageCode, "pt") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_RU, SAME_LANGUAGE_CODE(gLanguageCode, "ru") ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(aMenu, IDC_LANGUAGE_VI, SAME_LANGUAGE_CODE(gLanguageCode, "vi") ? MF_CHECKED : MF_UNCHECKED);
	}

	//	Help menu
	if (aMenu == GetSubMenu(theMainMenu, 6))
	{
		CheckMenuItem(aMenu, IDC_HELP_PRACTICE_BOARD,
			wd->md.itsGame == GameIntro ? MF_CHECKED : MF_UNCHECKED);
	}
}

static WCHAR *GameSpecificResetText(
	GameType		aGame,				//	input
	WCHAR			*aValueBuffer,		//	output;	zero-terminated WCHAR string
	unsigned int	aValueBufferLength)	//	input;	length of aValueBuffer in WCHARs (not bytes),
										//		including space for the terminating zero
{
	char	*theResetKey = NULL;
	
	//	The Reset command's text depends on both the current game
	//	and the current language.  This same text appears both
	//	as a toolbar button and as an item on the Game menu.

	switch (aGame)
	{
		case GameNone:			theResetKey = "Reset";						break;
		case GameIntro:			theResetKey = "Reset Practice Board";		break;
		case GameTicTacToe:		theResetKey = "Erase Tic-Tac-Toe Board";	break;
		case GameMaze:			theResetKey = "New Maze";					break;
		case GameCrossword:		theResetKey = "New Crossword Puzzle";		break;
		case GameWordSearch:	theResetKey = "New Word Search";			break;
		case GameJigsaw:		theResetKey = "New Jigsaw Puzzle";			break;
		case GameChess:			theResetKey = "Reset Chess Board";			break;
		case GamePool:			theResetKey = "Reset Pool Balls";			break;
		case GameApples:		theResetKey = "New Apples";					break;
		default:				theResetKey = NULL;							break;
	}

	return GetLocalizedTextAsWCHAR(theResetKey, aValueBuffer, aValueBufferLength);
}


static void MenuCommand(
	WindowData	*wd,
	WORD		aCommand)
{
	switch (aCommand)
	{
		//	file menu

		case IDC_FILE_NEW:
			(void) CreateTorusGamesWindow();
			break;

		case IDC_FILE_EXIT:
			//	Send each window a WM_CLOSE message.
			//	If all windows comply, our message loop in WinMain() will terminate.
			EnumThreadWindows(GetCurrentThreadId(), PostCloseMessage, 0);
			break;

		//	game menu

		case IDC_GAME_RESET:		ResetGame(&wd->md);						break;
		case IDC_GAME_SPECIAL:		SpecialMenuCommand(&wd->md);			break;

		case IDC_GAME_TICTACTOE:	ChangeGameWin32(wd, GameTicTacToe);		break;
		case IDC_GAME_MAZE:			ChangeGameWin32(wd, GameMaze);			break;
		case IDC_GAME_CROSSWORD:	ChangeGameWin32(wd, GameCrossword);		break;
		case IDC_GAME_WORDSEARCH:	ChangeGameWin32(wd, GameWordSearch);	break;
		case IDC_GAME_JIGSAW:		ChangeGameWin32(wd, GameJigsaw);		break;
		case IDC_GAME_CHESS:		ChangeGameWin32(wd, GameChess);			break;
		case IDC_GAME_POOL:			ChangeGameWin32(wd, GamePool);			break;
		case IDC_GAME_APPLES:		ChangeGameWin32(wd, GameApples);		break;

		//	options menu

		case IDC_OPTIONS_HUMAN:
			ChangeHumanVsComputer(&wd->md, false);
			break;

		case IDC_OPTIONS_COMPUTER:
			ChangeHumanVsComputer(&wd->md, true);
			break;

		case IDC_OPTIONS_EXTRA_EASY:	ChangeDifficultyLevel(&wd->md, 0);	break;
		case IDC_OPTIONS_EASY:			ChangeDifficultyLevel(&wd->md, 1);	break;
		case IDC_OPTIONS_MEDIUM:		ChangeDifficultyLevel(&wd->md, 2);	break;
		case IDC_OPTIONS_HARD:			ChangeDifficultyLevel(&wd->md, 3);	break;
		case IDC_OPTIONS_EXTRA_HARD:	ChangeDifficultyLevel(&wd->md, 4);	break;

		case IDC_OPTIONS_GLIDE_AXES:
			SetShowGlideAxes(&wd->md, ! wd->md.itsShowGlideAxes );
			break;

		case IDC_OPTIONS_SOUND:
			gPlaySounds = ! gPlaySounds;
			break;

		case IDC_OPTIONS_EXPORT_COPY:
			CopyTheImage(wd);
			break;

		case IDC_OPTIONS_EXPORT_SAVE:
			SaveTheImage(wd);
			break;

		//	topology menu

		case IDC_TOPOLOGY_TORUS:	ChangeTopology(&wd->md, TopologyTorus);	break;
		case IDC_TOPOLOGY_KLEIN:	ChangeTopology(&wd->md, TopologyKlein);	break;

		//	view menu

		case IDC_VIEW_DOMAIN:	ChangeView(&wd->md, ViewDomain);	break;
		case IDC_VIEW_SMALL:	ChangeView(&wd->md, ViewSmall);		break;
		case IDC_VIEW_TILING:	ChangeView(&wd->md, ViewTiling);	break;
		
		//	language menu

		case IDC_LANGUAGE_DE:	ChangeLanguage("de");	break;
		case IDC_LANGUAGE_EN:	ChangeLanguage("en");	break;
		case IDC_LANGUAGE_ES:	ChangeLanguage("es");	break;
		case IDC_LANGUAGE_FR:	ChangeLanguage("fr");	break;
		case IDC_LANGUAGE_HU:	ChangeLanguage("hu");	break;
		case IDC_LANGUAGE_IT:	ChangeLanguage("it");	break;
		case IDC_LANGUAGE_JA:	ChangeLanguage("ja");	break;
		case IDC_LANGUAGE_KO:	ChangeLanguage("ko");	break;
		case IDC_LANGUAGE_NL:	ChangeLanguage("nl");	break;
		case IDC_LANGUAGE_PT:	ChangeLanguage("pt");	break;
		case IDC_LANGUAGE_RU:	ChangeLanguage("ru");	break;
		case IDC_LANGUAGE_VI:	ChangeLanguage("vi");	break;

		//	help menu

		case IDC_HELP_HOW_TO_PLAY:
			OpenHelpPage(IDC_HELP_HOW_TO_PLAY);
			break;
		
		case IDC_HELP_WRAPAROUND_UNIVERSE:
			OpenHelpPage(IDC_HELP_WRAPAROUND_UNIVERSE);
			break;

		case IDC_HELP_PRACTICE_BOARD:
			ChangeGameWin32(wd, GameIntro);
			break;
		
		case IDC_HELP_CONTACT:
			OpenHelpPage(IDC_HELP_CONTACT);
			break;

		case IDC_HELP_ABOUT:
			//	Building a fancy dialog box with graphics is tricky
			//	because of the conflict between the "dialog layout units"
			//	and the pixel dimensions of bitmaps.  A simple but humble
			//	solution relays basic information in a MessageBox().
			ErrorMessage(GetLocalizedText("AboutBoxMessage"), GetLocalizedText("AboutBoxTitle"), false);
			break;
	}
}


static BOOL CALLBACK PostCloseMessage(
	HWND	aWindow,
	LPARAM	anUnusedParameter)
{
	UNUSED_PARAMETER(anUnusedParameter);

	//	Skip windows that aren't ours.
	//
	//	Note:  On my system the Input Method Environment (IME) creates
	//	secret windows that shadow the main window.  They are top-level windows,
	//	so EnumThreadWindows() finds them.  Specifically, each main window owns
	//	a secret window of class "IME", which in turn owns another secret window
	//	of class "MSCTFIME UI".
	if (GetClassLong(aWindow, GCW_ATOM) == gMainWindowClass)
		PostMessage(aWindow, WM_CLOSE, 0, 0);

	return true;	//	keep going
}


static void ChangeGameWin32(
	WindowData	*wd,
	GameType	aNewGame)
{
	WCHAR	theButtonText[128];

	//	Just to be safe...
	if (aNewGame >= NumGameTypes)
		aNewGame = GameNone;
	
	//	If the user re-selects the game that's already running,
	//	leave that game untouched.  In other words, don't reset it.
	if (aNewGame == wd->md.itsGame)
		return;

	//	Shut down the old game.
	//	Make sure the OpenGL context is current
	//	so the old game can release its textures,
	//	whose OpenGL "names" only it knows.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	ShutDownGame(&wd->md);
	wglMakeCurrent(NULL, NULL);

	//	Switch to the new game.
	wd->md.itsGame = aNewGame;

	//	Delete fonts (if any) from the old game and 
	//	create fonts (if any) for the new game.
	PrepareFonts(wd);
	
	//	Delete the side window (if any) from the old game and
	//	create the side window (if any) for the new game,
	//	setting the font for the latter.
	PrepareSideWindow(wd);

	//	Update the toolbar to reflect the new game.
	SetWindowText(
		wd->itsResetButton,
		GameSpecificResetText(wd->md.itsGame, theButtonText, BUFFER_LENGTH(theButtonText)));

	//	Set up the new game.
	SetUpGame(&wd->md);
}


static void PrepareFonts(WindowData *wd)
{
	unsigned int	i;
	LOGFONT			theLogicalFont	= {0, 0, 0, 0, 0, 0, 0, 0, DEFAULT_CHARSET, 0, 0, 0, 0, L""};

	//	Delete all pre-existing fonts.
	for (i = 0; i < NumberOfFonts; i++)
		if (wd->itsWin32Fonts[i].itsHFONT != NULL)
		{
			DeleteObject(wd->itsWin32Fonts[i].itsHFONT);
			wd->itsWin32Fonts[i].itsHFONT = NULL;
		}

	//	Allocate whatever new fonts may be needed.

	//	For a given language, all games currently use the same font face.
	GetLocalizedTextAsWCHAR("FontFace", theLogicalFont.lfFaceName, LF_FACESIZE);

	//	The font size depends on the context.
	switch (wd->md.itsGame)
	{
		case GameCrossword:

			//	main puzzle
			theLogicalFont.lfHeight = CROSSWORD_CHARACTER_TEXTURE_SIZE_DESKTOP - 2*CROSSWORD_CHARACTER_INSET;
			wd->itsWin32Fonts[FontCrosswordPuzzle].itsHFONT = CreateFontIndirect(&theLogicalFont);

			//	clues
			theLogicalFont.lfHeight = 28;	//	font size in pixels, not points
			wd->itsWin32Fonts[FontCrosswordClue].itsHFONT = CreateFontIndirect(&theLogicalFont);

			break;

		case GameWordSearch:

			//	main puzzle
			theLogicalFont.lfHeight = WORDSEARCH_CHARACTER_TEXTURE_SIZE_DESKTOP;
			wd->itsWin32Fonts[FontWordSearchPuzzle].itsHFONT = CreateFontIndirect(&theLogicalFont);

			//	word list
			theLogicalFont.lfHeight = 28;	//	font size in pixels, not points
			wd->itsWin32Fonts[FontWordSearchList].itsHFONT = CreateFontIndirect(&theLogicalFont);

			break;

		case GamePool:

			//	status window
			//
			//	For the same nominal font size,
			//	"MS UI Gothic" is bigger than "Tahoma".
			//	MS Gothic apparently contains embedded bitmaps
			//	at 21- and 25-pixel sizes, so 25 pixels looks 
			//	much better than 24.
			if (SAME_LANGUAGE_CODE(gLanguageCode, "ja"))
				theLogicalFont.lfHeight = 25;	//	font size in pixels, not points
			else
				theLogicalFont.lfHeight = 28;	//	font size in pixels, not points
			wd->itsWin32Fonts[FontPoolStatus].itsHFONT = CreateFontIndirect(&theLogicalFont);

			break;

		case GameApples:

			//	numerals 1 through 8
			theLogicalFont.lfHeight = APPLES_CHARACTER_TEXTURE_SIZE;
			wd->itsWin32Fonts[FontApplesDigit].itsHFONT = CreateFontIndirect(&theLogicalFont);
			break;

		default:
			//	The other games don't need fonts.
			break;
	}
}


static void PrepareSideWindow(WindowData *wd)
{
	//	Delete the previous side window, if any.
	if (wd->itsWin32SideWindow.itsHWND != NULL)
	{
		//	Destroy the side window.
		//	This will automatically detroy its children (if any) as well.
		DestroyWindow(wd->itsWin32SideWindow.itsHWND);
		
		//	Set all pointers to NULL.
		wd->itsWin32SideWindow.itsHWND			= NULL;
		wd->itsWin32SideWindow.itsTextField1	= NULL;
		wd->itsWin32SideWindow.itsTextField2	= NULL;
	}

	//	Allocate whatever new side window may be needed
	//	and set its font.
	switch (wd->md.itsGame)
	{
		case GameCrossword:

			//	Create the window.
			CreateSideWindow(wd, "CrosswordCluesWindowTitle", 256, 128, 128, ES_LEFT, FontCrosswordClue);

			//	Pre-load the clue prefixes.
			GetLocalizedTextAsWCHAR("Across", wd->md.itsGameOf.Crossword.itsCluePrefixAcross, MAX_CROSSWORD_PREFIX_LENGTH);
			GetLocalizedTextAsWCHAR("Down",   wd->md.itsGameOf.Crossword.itsCluePrefixDown,   MAX_CROSSWORD_PREFIX_LENGTH);

			break;

		case GameWordSearch:
			CreateSideWindow(wd, "WordSearchListWindowTitle", 256, 512, 0, WS_VSCROLL | ES_LEFT, FontWordSearchList);
			break;

		case GamePool:
			CreateSideWindow(wd, "PoolStatusWindowTitle", 256, 128, 0, ES_CENTER, FontPoolStatus);
			break;

		default:
			//	Most games don't need a side window.
			break;
	}
}


static void CreateSideWindow(
	WindowData		*wd,
	char			*aWindowTitleKey,
	unsigned int	aWidth,		//	width of content area
	unsigned int	aHeight1,	//	height of first text field
	unsigned int	aHeight2,	//	height of second text field (may be zero)
	DWORD			aStyleSet,	//	additional styles for the text fields
	FontIndex		aFont)
{
	WCHAR			theTitle[64];
	RECT			theMainWindowRect,
					theSideWindowRect;
	unsigned int	theHeight;
	DWORD			theWindowStyles;
	signed int		theSideWindowPositionH,
					theSideWindowPositionV,
					theSideWindowWidth,
					theSideWindowHeight,
					theLimit;

	//	Redundant but safe.
	wd->itsWin32SideWindow.itsHWND			= NULL;
	wd->itsWin32SideWindow.itsTextField1	= NULL;
	wd->itsWin32SideWindow.itsTextField2	= NULL;
	
	//	Set the default color to black.
	//	Only the Crossword will change this.
	wd->itsWin32SideWindow.itsTextColor1	= 0x00000000;	//	0x00BBGGRR
	wd->itsWin32SideWindow.itsTextColor2	= 0x00000000;	//	0x00BBGGRR

	//	Fetch the side window's title.
	GetLocalizedTextAsWCHAR(aWindowTitleKey, theTitle, BUFFER_LENGTH(theTitle));

	//	Where is the main window?
	//	We'll want to place the side window just to the right of it.
	GetWindowRect(wd->itsWindow, &theMainWindowRect);
	
	//	Note the total height of the client area.
	theHeight = aHeight1 + aHeight2;

	//	Figure out what overall size to request in order to end up
	//	with a client area of dimensions aWidth × theHeight.
	//	[Note:  WS_CAPTION isn't needed for CreateWindow(),
	//	because WS_OVERLAPPED already implies a caption,
	//	but WS_CAPTION *is* required to get AdjustWindowRect()
	//	to work correctly.  Apparently AdjustWindowRect()
	//	doesn't know that WS_OVERLAPPED implies a caption.]
	theWindowStyles = WS_OVERLAPPED | WS_CAPTION | WS_VISIBLE;
	SetRect(&theSideWindowRect, 0, 0, aWidth, theHeight);
	AdjustWindowRect(&theSideWindowRect, theWindowStyles, false);

	//	Tentatively place the side window to the right of the main window.
	theSideWindowWidth		= theSideWindowRect.right - theSideWindowRect.left;
	theSideWindowHeight		= theSideWindowRect.bottom - theSideWindowRect.top;
	theSideWindowPositionH	= theMainWindowRect.right;
	theSideWindowPositionV	= theMainWindowRect.top;

	//	Allow for multimonitor systems and keep the side window 
	//	fully within the virtual screen (if the user has 
	//	an odd arrangement of monitors this might still leave 
	//	the side window hidden, but in most cases it should work fine).

	theLimit = GetSystemMetrics(SM_XVIRTUALSCREEN);
	if (theSideWindowPositionH < theLimit)
		theSideWindowPositionH = theLimit;

	theLimit = (GetSystemMetrics(SM_XVIRTUALSCREEN) + GetSystemMetrics(SM_CXVIRTUALSCREEN)) - theSideWindowWidth;
	if (theSideWindowPositionH > theLimit)
		theSideWindowPositionH = theLimit;

	theLimit = GetSystemMetrics(SM_YVIRTUALSCREEN);
	if (theSideWindowPositionV < theLimit)
		theSideWindowPositionV = theLimit;

	theLimit = (GetSystemMetrics(SM_YVIRTUALSCREEN) + GetSystemMetrics(SM_CYVIRTUALSCREEN)) - theSideWindowHeight;
	if (theSideWindowPositionV > theLimit)
		theSideWindowPositionV = theLimit;

	//	Create the window.
	wd->itsWin32SideWindow.itsHWND = CreateWindow(
		SIDE_WINDOW_CLASS_NAME,
		theTitle,
		theWindowStyles,
		theSideWindowPositionH,
		theSideWindowPositionV,
		theSideWindowWidth,
		theSideWindowHeight,
		wd->itsWindow,
		NULL, GetModuleHandle(NULL), NULL);
	if (wd->itsWin32SideWindow.itsHWND == NULL)
		return;

	//	Set a pointer to the SideWindow data, 
	//	so our custom WindowProc can read the desired text color
	//	and the overridden WindowProc address.
	SetWindowLongPtr(	wd->itsWin32SideWindow.itsHWND,
						GWLP_USERDATA,
						(LONG_PTR) &wd->itsWin32SideWindow);
	
	//	Supplement the caller's style request with the standard styles.
	aStyleSet |= (WS_CHILD | WS_VISIBLE | ES_MULTILINE | ES_READONLY);

	//	If the caller wants a first text field...
	if (aHeight1 > 0)
	{
		//	Create the text field.
		wd->itsWin32SideWindow.itsTextField1 = CreateWindow(
			L"Edit",
			L"",
			aStyleSet,
			0, 0, aWidth, aHeight1,
			wd->itsWin32SideWindow.itsHWND,
			NULL, GetModuleHandle(NULL), NULL);
		
		//	Set its font.
		if (wd->itsWin32SideWindow.itsTextField1 != NULL
		 && wd->itsWin32Fonts[aFont].itsHFONT != NULL)
			SendMessage(wd->itsWin32SideWindow.itsTextField1, WM_SETFONT, (WPARAM) wd->itsWin32Fonts[aFont].itsHFONT, true);
	}

	//	If the caller wants a second text field...
	if (aHeight2 > 0)
	{
		//	Create the text field.
		wd->itsWin32SideWindow.itsTextField2 = CreateWindow(
			L"Edit",
			L"",
			aStyleSet,
			0, aHeight1, aWidth, aHeight2,
			wd->itsWin32SideWindow.itsHWND,
			NULL, GetModuleHandle(NULL), NULL);
		
		//	Set its font.
		if (wd->itsWin32SideWindow.itsTextField2 != NULL
		 && wd->itsWin32Fonts[aFont].itsHFONT != NULL)
			SendMessage(wd->itsWin32SideWindow.itsTextField2, WM_SETFONT, (WPARAM) wd->itsWin32Fonts[aFont].itsHFONT, true);
	}

	//	Move the main window to the foreground,
	//	so keyboard input goes to it rather than the side window.
	SetForegroundWindow(wd->itsWindow);
}


void CrosswordSetClues(
	ModelData	*md,
	Char16		*aClueAcross,
	Char16		*aClueDown,
	Char16		aHotDirection)
{
	HWND	theAcrossField	= NULL,
			theDownField	= NULL;
	
	const COLORREF	theActiveColor		= 0x00FF0000,	//	0x00bbggrr
					theInactiveColor	= 0x00C0A0A0;

	if (aHotDirection == L'↓')
	{
		md->itsSideWindow->itsTextColor1 = theInactiveColor;
		md->itsSideWindow->itsTextColor2 = theActiveColor;
	}
	else
	{
		md->itsSideWindow->itsTextColor1 = theActiveColor;
		md->itsSideWindow->itsTextColor2 = theInactiveColor;
	}

	if ((theAcrossField = md->itsSideWindow->itsTextField1) != NULL)
	{
		SendMessage(theAcrossField, WM_SETTEXT,    0, (LPARAM) L"");
		SendMessage(theAcrossField, EM_REPLACESEL, 0, (LPARAM) md->itsGameOf.Crossword.itsCluePrefixAcross);
		SendMessage(theAcrossField, EM_REPLACESEL, 0, (LPARAM) L":  ");
		SendMessage(theAcrossField, EM_REPLACESEL, 0, (LPARAM) aClueAcross);
	}

	if ((theDownField = md->itsSideWindow->itsTextField2) != NULL)
	{
		SendMessage(theDownField,   WM_SETTEXT,    0, (LPARAM) L"");
		SendMessage(theDownField,   EM_REPLACESEL, 0, (LPARAM) md->itsGameOf.Crossword.itsCluePrefixDown);
		SendMessage(theDownField,   EM_REPLACESEL, 0, (LPARAM) L":  ");
		SendMessage(theDownField,   EM_REPLACESEL, 0, (LPARAM) aClueDown);
	}
}


void WordSearchSetList(ModelData *md)
{
	HWND			theListField	= NULL;
	unsigned int	i;

	theListField = md->itsSideWindow->itsTextField1;

	if (theListField != NULL)
	{
		SendMessage(theListField, WM_SETTEXT, 0, (LPARAM) L"");

		for (i = 0; i < MAX_WORDSEARCH_NUM_WORDS; i++)
		{
			if (md->itsGameOf.WordSearch.itsWords[i][0] != 0)
			{
				SendMessage(theListField, EM_REPLACESEL, 0, (LPARAM) md->itsGameOf.WordSearch.itsWords[i]);
				SendMessage(theListField, EM_REPLACESEL, 0, (LPARAM) L"\r\n");
			}
		}

		SendMessage(theListField, EM_SETSEL, 0, 0);
		SendMessage(theListField, EM_SCROLLCARET, 0, 0);
	}
}


void PoolSetStatusMessage(ModelData *md)
{
	char	*theSubject,
			*thePredicate,
			theKey[32];
	WCHAR	theMessageText[128]	= L"";
	HWND	theStatusField		= NULL;

	//	Determine the active subject.
	if (md->itsHumanVsComputer)
	{
		if (md->itsGameOf.Pool.itsWhoseTurn == md->itsGameOf.Pool.itsComputerPlayer)
			theSubject = "Computer";
		else
			theSubject = "Human";
	}
	else
	{
		if (md->itsGameOf.Pool.itsWhoseTurn == PoolPlayerA)
			theSubject = "PlayerA";
		else
			theSubject = "PlayerB";
	}
	
	//	Determine the appropriate predicate.
	if (md->itsGameOf.Pool.itsWinFlag)
		thePredicate = "HasWon";
	else
	if (md->itsGameOf.Pool.itsLoseFlag)
		thePredicate = "HasLost";
	else
		switch (md->itsGameOf.Pool.itsWhoseTurn == PoolPlayerA ?
					md->itsGameOf.Pool.itsPlayerAGoal :
					md->itsGameOf.Pool.itsPlayerBGoal)
		{
			case PoolUndecided:	thePredicate = "Shoots";			break;
			case PoolSolids:	thePredicate = "ShootsForSolids";	break;
			case PoolStripes:	thePredicate = "ShootsForStripes";	break;
			case PoolBlack:		thePredicate = "ShootsForFourBall";	break;
			default:			thePredicate = "";					break;	//	unused
		}

	//	Note:  sprintf() does not check for buffer overruns.
	sprintf(theKey, "%s%s", theSubject, thePredicate);
	GetLocalizedTextAsWCHAR(theKey, theMessageText, BUFFER_LENGTH(theMessageText));

	theStatusField = md->itsSideWindow->itsTextField1;
	if (theStatusField != NULL)
	{
		SendMessage(theStatusField, WM_SETTEXT, 0, (LPARAM) theMessageText);
		SendMessage(theStatusField, EM_SETSEL, 0, 0);
		SendMessage(theStatusField, EM_SCROLLCARET, 0, 0);
	}
}


unsigned int GetNumPuzzles(
	GameType		aGame,
	TopologyType	aTopology)
{
	char			*thePrefix				= NULL,
					*theGame				= NULL,
					*theTopology			= NULL,
					theKey[64]				= "";
	WCHAR			theNumPuzzlesString[4]	= L"000";
	unsigned int	theNumPuzzles			= 0;

	switch (aGame)
	{
		case GameCrossword:
			thePrefix	= "Num";
			theGame		= "Crossword";
			break;

		case GameWordSearch:
			thePrefix	= "NumDesktop";
			theGame		= "WordSearch";
			break;

		default:
			ErrorMessage("Invalid GameType", "Internal Error in GetNumPuzzles()", false);
			return 1;	//	safest possible return value
	}

	switch (aTopology)
	{
		case TopologyTorus: theTopology = "Torus"; break;
		case TopologyKlein: theTopology = "Klein"; break;
		default:
			ErrorMessage("Invalid TopologyType", "Internal Error in GetNumPuzzles()", false);
			return 1;	//	safest possible return value
	}

	//	Note:  sprintf() does not check for buffer overruns.
	sprintf(theKey, "%s%s%sPuzzles", thePrefix, theTopology, theGame);
	GetLocalizedTextAsWCHAR(theKey, theNumPuzzlesString, BUFFER_LENGTH(theNumPuzzlesString));
	theNumPuzzles = StringToUnsignedInt(theNumPuzzlesString);

	return theNumPuzzles;
}


static unsigned int StringToUnsignedInt(WCHAR *aString)
{
	//	Assume the string contains only digits 0-9,
	//	is NULL-terminated, and is short enough to avoid overflows.

	unsigned int	theValue;

	theValue = 0;

	while (*aString != 0)
	{
		if (*aString < '0' || *aString > '9')
			return 0;

		theValue *= 10;
		theValue += *aString - '0';

		aString++;
	}

	return theValue;
}


void GetRawPuzzleData(
	GameType		aGame,
	TopologyType	aTopology,
	unsigned int	aPuzzleIndex,
	Char16			*aBuffer,
	unsigned int	aBufferLength)	//	number of Char16's (not bytes!) available in aBuffer, including terminating zero
{
	ErrorText	theError		= NULL;
	char		*theGame		= NULL,
				*theTopology	= NULL,
				theKey[32]		= "";
	
	if (aBuffer == NULL
	 || aBufferLength == 0)
	{
		theError = "No output buffer";
		goto CleanUpGetRawPuzzleData;
	}

	switch (aGame)
	{
		case GameCrossword:  theGame = "Crossword";  break;
		case GameWordSearch: theGame = "WordSearch"; break;
		default:
			theError = "Invalid GameType";
			goto CleanUpGetRawPuzzleData;
	}

	switch (aTopology)
	{
		case TopologyTorus: theTopology = "Torus"; break;
		case TopologyKlein: theTopology = "Klein"; break;
		default:
			theError = "Invalid TopologyType";
			goto CleanUpGetRawPuzzleData;
	}

	//	Note:  sprintf() does not check for buffer overruns.
	sprintf(theKey, "%s%s%d", theTopology, theGame, aPuzzleIndex);
	GetLocalizedTextAsWCHAR(theKey, aBuffer, aBufferLength);

CleanUpGetRawPuzzleData:

	if (theError != NULL)
	{
		ErrorMessage(theError, "Internal Error in GetRawPuzzleData()", false);
		if (aBuffer != NULL
		 && aBufferLength > 0)
		{
			aBuffer[0] = 0;
		}
	}
}


static void CopyTheImage(WindowData *wd)
{
	ErrorText	theErrorMessage		= NULL;
	HANDLE		theDibHandle		= NULL;
	bool		theClipboardIsOpen	= false;

	//	Read the image as a DIB.
	theDibHandle = FetchTheImage(wd);
	if (theDibHandle == NULL)
	{
		theErrorMessage = "Couldn't read image to copy to clipboard.";
		goto CleanUpCopyTheImage;
	}

	//	Transfer the DIB to the clipboard.

	if (OpenClipboard(NULL) == 0)
	{
		theErrorMessage = "Couldn't open clipboard.";
		goto CleanUpCopyTheImage;
	}
	theClipboardIsOpen = true;

	if (EmptyClipboard() == 0)
	{
		theErrorMessage = "Couldn't delete previous clipboard contents.";
		goto CleanUpCopyTheImage;
	}

	if (SetClipboardData(CF_DIB, theDibHandle) == NULL)
	{
		theErrorMessage = "Couldn't transfer image to clipboard.";
		goto CleanUpCopyTheImage;
	}
	//	The clipboard now owns the DIB.
	theDibHandle = NULL;

CleanUpCopyTheImage:

	if (theClipboardIsOpen)
		CloseClipboard();

	if (theDibHandle != NULL)
		GlobalFree(theDibHandle);

	if (theErrorMessage != NULL)
		ErrorMessage(theErrorMessage, "Copy Error", false);
}


static void SaveTheImage(WindowData *wd)
{
	ErrorText			theErrorMessage		= NULL;
	WCHAR				theFileName[1024]	= {0};
	OPENFILENAME		theFileInfo			= {sizeof(OPENFILENAME),
		NULL, NULL, L"32-bit Bitmap (*.bmp)\0*.bmp\0", NULL, 0, 0,
		theFileName, BUFFER_LENGTH(theFileName), NULL, 0,
		gLastImageDirectory, NULL, OFN_OVERWRITEPROMPT, 0, 0,
		L"bmp", 0L, NULL, NULL};
	HANDLE				theDibHandle		= NULL;
	unsigned int		theDIBSize			= 0;
	HANDLE				theFile				= NULL;
	BITMAPFILEHEADER	theFileHeader		= {0x4D42, 0, 0, 0, 0};
	DWORD				theNumBytes			= 0;
	Byte				*theDib				= NULL;

	//	Prompt the user for a file name.
	//	The extension will be .bmp (see theFileInfo above).
	if (GetSaveFileName(&theFileInfo) == 0)
	{
		//	The user cancelled or an error occurred.
		goto CleanUpSaveTheImage;
	}

	//	Remember the directory name for next time.
	CopyDirectoryName(	theFileName,
						BUFFER_LENGTH(theFileName),
						gLastImageDirectory,
						BUFFER_LENGTH(gLastImageDirectory));

	//	Read the image as a DIB.
	theDibHandle = FetchTheImage(wd);
	if (theDibHandle == NULL)
	{
		theErrorMessage = "Couldn't read image to save to file.";
		goto CleanUpSaveTheImage;
	}
	
	//	How big is the DIB?
	theDIBSize = GlobalSize(theDibHandle);
	if (theDIBSize == 0)
	{
		theErrorMessage = "Couldn't get image size.";
		goto CleanUpSaveTheImage;
	}
	
	//	Open the output file.
	theFile = CreateFile(	theFileName,
							GENERIC_WRITE,
							0,
							NULL,
							CREATE_ALWAYS,
							FILE_ATTRIBUTE_NORMAL,
							NULL);
	if (theFile == NULL)
	{
		theErrorMessage = "Couldn't create file to save image.";
		goto CleanUpSaveTheImage;
	}
	
	//	Save the DIB to the file.
	
	//		First the BITMAPFILEHEADER.
	theFileHeader.bfSize	= sizeof(BITMAPFILEHEADER) + theDIBSize;
	theFileHeader.bfOffBits	= sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	if (WriteFile(theFile, &theFileHeader, sizeof(BITMAPFILEHEADER), &theNumBytes, NULL) == 0
	 || theNumBytes != sizeof(BITMAPFILEHEADER))
	{
		theErrorMessage = "Couldn't write bitmap file header.";
		goto CleanUpSaveTheImage;
	}
	
	//		Then the DIB itself.
	theDib = (Byte *) GlobalLock(theDibHandle);
	if (theDib == NULL)
	{
		theErrorMessage = "Couldn't lock the image memory to save to file.";
		goto CleanUpSaveTheImage;
	}
	if (WriteFile(theFile, theDib, theDIBSize, &theNumBytes, NULL) == 0
	 || theNumBytes != theDIBSize)
	{
		theErrorMessage = "Couldn't write DIB to file.";
		goto CleanUpSaveTheImage;
	}
	GlobalUnlock(theDibHandle);
	theDib = NULL;

CleanUpSaveTheImage:

	if (theDib != NULL)
		GlobalUnlock(theDibHandle);

	if (theFile != NULL)
		CloseHandle(theFile);

	if (theDibHandle != NULL)
		GlobalFree(theDibHandle);

	if (theErrorMessage != NULL)
		ErrorMessage(theErrorMessage, "Save Error", false);
}


static bool CopyDirectoryName(
	WCHAR	*aPathName,				//	input, 	NULL-terminated string of WCHARs
	DWORD	aPathNameLength,		//	input, 	size of input  buffer in WCHARs, including terminating zero
	WCHAR	*aDirectoryName,		//	output, NULL-terminated string of WCHARs
	DWORD	aDirectoryNameLength)	//	input,	size of output buffer in WCHARs, including terminating zero
{
	//	Copy aPathName to aDirectoryName, excluding the file name itself.
	
	DWORD	thePosition,
			i;
	
	//	Will aDirectoryName accommodate a null string in case of error?
	if (aDirectoryName == NULL
	 || aDirectoryNameLength == 0)
	{
		//	Return *without* setting an empty output string.
		return false;
	}
	else
	{
		//	Preinitialize an empty output string.
		aDirectoryName[0] = 0;
	}

	//	Find the terminating zero at the end of the input path name.
	for (thePosition = 0; thePosition < aPathNameLength; thePosition++)
		if (aPathName[thePosition] == 0)
			break;
	if (thePosition == aPathNameLength)
		return false;	//	no terminating zero
	
	//	Back up to the last backslash character.
	while (thePosition > 0
	    && aPathName[thePosition] != '\\')
	{
		thePosition--;
	}
	if (thePosition == 0)
		return false;	//	no backslash found
	
	if (aDirectoryNameLength < thePosition + 2)
		return false;	//	output buffer too small
	
	//	Copy the directory name from aPathName to aDirectoryName.
	for (i = 0; i <= thePosition; i++)
		aDirectoryName[i] = aPathName[i];
	
	//	Append a terminating zero.
	aDirectoryName[thePosition + 1] = 0;
	
	//	Success
	return true;
}


static HANDLE FetchTheImage(WindowData *wd)
{
	ErrorText			theErrorMessage			= NULL;
	RECT				theClientRect			= {0,0,0,0};
	unsigned int		theImageWidth			= 0,
						theImageHeight			= 0,
						theImageSize			= 0,
						theHeaderSize			= 0,
						theTotalSize			= 0;
	HANDLE				theDibHandle			= NULL;
	Byte				*theDib					= NULL;
	BITMAPINFOHEADER	*theHeader				= NULL;
	Byte				*theImage				= NULL;
	unsigned int		r						= 0,
						c						= 0;
	Byte				*thePixel				= NULL,
						theSwap					= 0;

	//	How big is the image?
	if (GetClientRect(wd->itsDrawingPanel, &theClientRect) == 0
	 || theClientRect.left != 0
	 || theClientRect.top  != 0
	 || (theImageWidth  = theClientRect.right)  == 0
	 || (theImageHeight = theClientRect.bottom) == 0)
	{
		theErrorMessage = "Couldn't read window size.";
		goto CleanUpFetchTheImage;
	}

	//	How many bytes do we have?
	//	(If we were reading RGB instead of RGBA we'd need to pad the rows
	//	to insure a row-stride that's a multiple of four, but for RGBA
	//	it's automatic.  Unfortunately the alpha channel will be ignored,
	//	but let's include the alpha values anyhow.)
	theImageSize = theImageWidth * theImageHeight * 4;

	//	Create a DIB.

	theHeaderSize	= sizeof(BITMAPINFOHEADER);
	theTotalSize	= theHeaderSize + theImageSize;

	theDibHandle = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, theTotalSize);
	if (theDibHandle == NULL)
	{
		theErrorMessage = "Not enough memory to copy image.";
		goto CleanUpFetchTheImage;
	}

	theDib = (Byte *) GlobalLock(theDibHandle);
	if (theDib == NULL)
	{
		theErrorMessage = "Couldn't lock the image memory to write image data.";
		goto CleanUpFetchTheImage;
	}
	theHeader	= (BITMAPINFOHEADER *) theDib;
	theImage	= theDib + theHeaderSize;

	theHeader->biSize			= sizeof(BITMAPINFOHEADER);
	theHeader->biWidth			= theImageWidth;
	theHeader->biHeight			= theImageHeight;
	theHeader->biPlanes			= 1;
	theHeader->biBitCount		= 32;
	theHeader->biCompression	= BI_RGB;
	theHeader->biSizeImage		= 0;
	theHeader->biXPelsPerMeter	= 0;
	theHeader->biYPelsPerMeter	= 0;
	theHeader->biClrUsed		= 0;
	theHeader->biClrImportant	= 0;
	
	//	The back buffer may or may not contain the current image,
	//	depending on whether the driver swaps pointers to two buffers
	//	or whether it copies the contents of fixed back buffer
	//	to a fixed front buffer.  To play it safe, ReadRGBAPixels()
	//	always reads the front buffer.
	//	The only problem with this is that we end up copying
	//	the Export menu's fade effect.  To work around that,
	//	sleep for 1sec before reading the buffer.
//	PaintWindow(wd);	unnecessary
	Sleep(1000);

	//	Read the pixels as RGBA.
	//
	//	Note:  Microsoft's BITMAP format does not support transparency --
	//	the last byte of each pixel gets ignored!
	//	ReadRGBAPixels() returns pixels with premultiplied alpha,
	//	which the BITMAP format interprets as fading to black.
	//	Given the lack of transparency support, this is the best
	//	outcome we could have hoped for.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	ReadRGBAPixels(theImageWidth, theImageHeight, theImage);
	wglMakeCurrent(NULL, NULL);

	//	Swap the byte order manually to conform to .bmp conventions.
	thePixel = theImage;
	for (r = 0; r < theImageHeight; r++)
		for (c = 0; c < theImageWidth; c++)
		{
			theSwap		= thePixel[0];
			thePixel[0]	= thePixel[2];
			thePixel[2]	= theSwap;
			
			thePixel += 4;
		}

	//	Unlock the handle.
	GlobalUnlock(theDibHandle);
	theDib = NULL;

CleanUpFetchTheImage:

	if (theDib != NULL)
		GlobalUnlock(theDibHandle);

	if (theErrorMessage != NULL)
	{
		if (theDibHandle != NULL)
		{
			GlobalFree(theDibHandle);
			theDibHandle = NULL;
		}

		ErrorMessage(theErrorMessage, "Image Export Error", false);
	}
	
	return theDibHandle;
}


static void ChangeLanguage(
	const char	aLanguageCode[3])
{
	//	Record the new language code and initialize a dictionary.
	RefreshLocalization(aLanguageCode);

	//	Refresh each window.
	EnumThreadWindows(GetCurrentThreadId(), RefreshLanguage, 0);
}


BOOL CALLBACK RefreshLanguage(
	HWND	aWindow,
	LPARAM	anUnusedParameter)
{
	WCHAR		theWindowTitle[128];
	HMENU		theOldMenu	= NULL,
				theNewMenu	= NULL;
	WCHAR		theButtonText[128];
	WindowData	*wd			= NULL;

	UNUSED_PARAMETER(anUnusedParameter);

	//	Skip windows that aren't ours.
	//
	//	Note:  On my system the Input Method Environment (IME) creates
	//	secret windows that shadow the main window.  They are top-level windows,
	//	so EnumThreadWindows() finds them.  Specifically, each main window owns
	//	a secret window of class "IME", which in turn owns another secret window
	//	of class "MSCTFIME UI".
	if (GetClassLong(aWindow, GCW_ATOM) != gMainWindowClass)
		return true;

	//	Reset the window's title.
	SetWindowText(aWindow, GetLocalizedTextAsWCHAR("WindowTitle", theWindowTitle, BUFFER_LENGTH(theWindowTitle)));

	//	Reset the mirroring.
	RefreshMirroring(aWindow);

	//	Replace the window's old menu with a new menu in the new language.
	//	Note:  LoadMenu() creates a private menu for each window,
	//	even though they are all based on the same resource.
	//	Thus we may safely translate the new menu and destroy the old one.
	theOldMenu = GetMenu(aWindow);
	theNewMenu = BuildLocalizedMenuBar();
	SetMenu(aWindow, theNewMenu);
	DestroyMenu(theOldMenu);
	DrawMenuBar(aWindow);
	
	//	Locate the WindowData.
	wd = (WindowData *) GetWindowLongPtr(aWindow, GWLP_USERDATA);
	
	//	Update the toolbar to reflect the new language.
	if (wd != NULL)
	{
		SetWindowText(
			wd->itsResetButton,
			GameSpecificResetText(wd->md.itsGame, theButtonText, BUFFER_LENGTH(theButtonText)));
		SetWindowText(
			wd->itsChangeGameButton,
			GetLocalizedTextAsWCHAR("Change Game", theButtonText, BUFFER_LENGTH(theButtonText)));
	}

	//	Refresh the game as required.
	if (wd != NULL)
	{
		switch (wd->md.itsGame)
		{
			case GameCrossword:

				//	Re-load the clue prefixes in the new language.
				GetLocalizedTextAsWCHAR("Across", wd->md.itsGameOf.Crossword.itsCluePrefixAcross, MAX_CROSSWORD_PREFIX_LENGTH);
				GetLocalizedTextAsWCHAR("Down",   wd->md.itsGameOf.Crossword.itsCluePrefixDown,   MAX_CROSSWORD_PREFIX_LENGTH);

				//	Reset Crossword to accommodate the new language.
				//	CrosswordReset() will notice that the language has changed 
				//	and set the puzzle indices to 0, so the user will get 
				//	the easiest puzzle in the new language.
				ResetGame(&wd->md);

				break;
			
			case GameWordSearch:
				//	Reset Word Search to accommodate the new language.
				//	WordSearchReset() will notice that the language has changed 
				//	and set the puzzle indices to 0, so the user will get 
				//	the easiest puzzle in the new language.
				ResetGame(&wd->md);
				break;
			
			case GamePool:
				//	No need to disturb the Pool game in progress. 
				//	Just...

				//	Recreate the language-specific fonts.
				PrepareFonts(wd);

				//	Recreate the pool status window
				//	so it's title will appear in the new language.
				PrepareSideWindow(wd);
				
				//	Reset the current status message in the new language.
				PoolSetStatusMessage(&wd->md);
				
				break;
			
			case GameApples:
				//	No need to disturb the Apples game in progress. 
				//	Just...
				
				//	Recreate the language-specific fonts.
				PrepareFonts(wd);
				
				//	Refresh the textures for the numbers.
				wd->md.itsPreparedTextures = false;

				break;

			default:
				//	Do nothing.
				break;
		}
	}

	return true;	//	keep going
}


static void RefreshMirroring(HWND aWindow)
{
	//	Reset the window layout, left-to-right (LTR) or right-to-left (RTL).

	if (LanguageReadsRightToLeft())
	{
		//	Mirror the main window.
		//	Mirror child windows only if you want to -- it isn't necessary.
		SetWindowLongPtr(	aWindow,
							GWL_EXSTYLE,
							GetWindowLongPtr(aWindow, GWL_EXSTYLE) |  WS_EX_LAYOUTRTL);
	}
	else
	{
		SetWindowLongPtr(	aWindow,
							GWL_EXSTYLE,
							GetWindowLongPtr(aWindow, GWL_EXSTYLE) & ~WS_EX_LAYOUTRTL);
	}
}


static HMENU BuildLocalizedMenuBar(void)
{
	HMENU	theMainMenu		= NULL,
			theSubMenu		= NULL,
			theSubSubMenu	= NULL;
	WCHAR	theBuffer[2048];

	theMainMenu = CreateMenu();
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("File", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_FILE_NEW,  GetLocalizedTextAsWCHAR("New Window", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,             NULL                                                                      );
	AppendMenu(theSubMenu, MF_STRING,    IDC_FILE_EXIT, GetLocalizedTextAsWCHAR("Exit",       theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("Game", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_GAME_RESET,      GetLocalizedTextAsWCHAR("New Game",      theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                   NULL                                                                         );
	AppendMenu(theSubMenu, MF_STRING,    IDC_GAME_TICTACTOE,  GetLocalizedTextAsWCHAR("Tic-Tac-Toe",   theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_GAME_MAZE,       GetLocalizedTextAsWCHAR("Maze",          theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_GAME_CROSSWORD,  GetLocalizedTextAsWCHAR("Crossword",     theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_GAME_WORDSEARCH, GetLocalizedTextAsWCHAR("Word Search",   theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_GAME_JIGSAW,     GetLocalizedTextAsWCHAR("Jigsaw Puzzle", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_GAME_CHESS,      GetLocalizedTextAsWCHAR("Chess",         theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_GAME_POOL,       GetLocalizedTextAsWCHAR("Pool",          theBuffer, BUFFER_LENGTH(theBuffer)));
#ifndef TORUS_GAMES_FOR_TALK
	AppendMenu(theSubMenu, MF_STRING,    IDC_GAME_APPLES,     GetLocalizedTextAsWCHAR("Apples",        theBuffer, BUFFER_LENGTH(theBuffer)));
#endif
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("Options", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_OPTIONS_HUMAN,      GetLocalizedTextAsWCHAR("Human vs. Human",        theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_OPTIONS_COMPUTER,   GetLocalizedTextAsWCHAR("Human vs. Computer",     theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                      NULL                                                                                  );
	AppendMenu(theSubMenu, MF_STRING,    IDC_OPTIONS_EXTRA_EASY, GetLocalizedTextAsWCHAR("Extra Easy",             theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_OPTIONS_EASY,       GetLocalizedTextAsWCHAR("Easy",                   theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_OPTIONS_MEDIUM,     GetLocalizedTextAsWCHAR("Medium",                 theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_OPTIONS_HARD,       GetLocalizedTextAsWCHAR("Hard",                   theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_OPTIONS_EXTRA_HARD, GetLocalizedTextAsWCHAR("Extra Hard",             theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                      NULL                                                                                  );
	AppendMenu(theSubMenu, MF_STRING,    IDC_OPTIONS_GLIDE_AXES, GetLocalizedTextAsWCHAR("Show Glide Reflections", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                      NULL                                                                                  );
	AppendMenu(theSubMenu, MF_STRING,    IDC_OPTIONS_SOUND,      GetLocalizedTextAsWCHAR("Sound",                  theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_SEPARATOR, 0,                      NULL                                                                                  );
	theSubSubMenu = CreateMenu();
		AppendMenu(theSubMenu, MF_POPUP, (UINT_PTR)theSubSubMenu, GetLocalizedTextAsWCHAR("Export", theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_OPTIONS_EXPORT_COPY, GetLocalizedTextAsWCHAR("Copy Image",  theBuffer, BUFFER_LENGTH(theBuffer)));
		AppendMenu(theSubSubMenu, MF_STRING, IDC_OPTIONS_EXPORT_SAVE, GetLocalizedTextAsWCHAR("Save Image…", theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubSubMenu = NULL;
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("Topology", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_TOPOLOGY_TORUS, GetLocalizedTextAsWCHAR("Torus",        theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_TOPOLOGY_KLEIN, GetLocalizedTextAsWCHAR("Klein Bottle", theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("View", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_DOMAIN, GetLocalizedTextAsWCHAR("Fundamental Domain - Large", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_SMALL,  GetLocalizedTextAsWCHAR("Fundamental Domain - Small", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_VIEW_TILING, GetLocalizedTextAsWCHAR("Tiling",                     theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("Language", theBuffer, BUFFER_LENGTH(theBuffer)));
	//	To add a new language, please see the instructions
	//	in the file "To add a new language.txt".
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_DE, L"Deutsch"   );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_EN, L"English"   );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_ES, L"Español"   );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_FR, L"Français"  );
//	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_HU, L"Magyar"    );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_IT, L"Italiano"  );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_JA, L"日本語"       );
//	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_KO, L"한글"        );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_NL, L"Nederlands");
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_PT, L"Português" );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_RU, L"Русский"   );
	AppendMenu(theSubMenu, MF_STRING,    IDC_LANGUAGE_VI, L"Tiếng Việt");
	theSubMenu = NULL;
	
	theSubMenu = CreateMenu();
	AppendMenu(theMainMenu, MF_POPUP, (UINT_PTR)theSubMenu, GetLocalizedTextAsWCHAR("Help", theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_HOW_TO_PLAY,         GetLocalizedTextAsWCHAR("How to Play",          theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_WRAPAROUND_UNIVERSE, GetLocalizedTextAsWCHAR("Wraparound Universe",  theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_PRACTICE_BOARD,      GetLocalizedTextAsWCHAR("Practice Board",       theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_CONTACT,             GetLocalizedTextAsWCHAR("Contact",              theBuffer, BUFFER_LENGTH(theBuffer)));
	AppendMenu(theSubMenu, MF_STRING,    IDC_HELP_ABOUT,               GetLocalizedTextAsWCHAR("About Torus Games…",   theBuffer, BUFFER_LENGTH(theBuffer)));
	theSubMenu = NULL;
	
	return theMainMenu;
}


static void OpenHelpPage(DWORD aHelpCommand)
{
	char	*theFileNameBase,
			theFileName[64]			= "",
			theFilePath[2048]		= "";
	WCHAR	theFilePathUTF16[2048]	= L"";

	switch (aHelpCommand)
	{
		case IDC_HELP_HOW_TO_PLAY:
			theFileNameBase = "HowToPlay";
			break;
		
		case IDC_HELP_WRAPAROUND_UNIVERSE:
			theFileNameBase = "WraparoundUniverse";
			break;
		
		case IDC_HELP_CONTACT:
			theFileNameBase = "Contact";
			break;
		
		default:
			return;	//	should never occur
	}
	sprintf(theFileName, "%s-%c%c.html", theFileNameBase, gLanguageCode[0], gLanguageCode[1]);

	if (GetAbsolutePath("Help", theFileName, theFilePath, BUFFER_LENGTH(theFilePath)) == NULL
	 && MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, theFilePath, -1, theFilePathUTF16, BUFFER_LENGTH(theFilePathUTF16)) != 0)
	{
		ShellExecute(	NULL,
						L"open",
						theFilePathUTF16,
						NULL,
						NULL,
						SW_SHOWNORMAL);
	}
}


static void SetSync(WindowData *wd)
{
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	if (wglSwapInterval != NULL)	//	wglSwapInterval may be non-NULL but ineffective
		wglSwapInterval(wd->md.itsSyncFlag ? 1 : 0);
	wglMakeCurrent(NULL, NULL);
}


static void SetMinWindowSize(
	HWND	aWindow,
	POINT	*aMinSize)
{
	RECT	theRect;

	// Prevent the client area from going smaller than MIN_CLIENT_SIZE.
	SetRect(&theRect, 0, 0, MIN_CLIENT_SIZE, MIN_CLIENT_SIZE);
	AdjustWindowRect(&theRect, (DWORD) GetWindowLongPtr(aWindow, GWL_STYLE), true);
	aMinSize->x = theRect.right - theRect.left;
	aMinSize->y = theRect.bottom - theRect.top;
}

static void SetMaxWindowSize(
	HWND	aWindow,
	POINT	*aMaxPosition,
	POINT	*aMaxSize)
{
	RECT	theOriginalRect;
	
	//	According to msdn.microsoft.com/en-us/library/ms632605(VS.85).aspx
	//
	//		For systems with multiple monitors, the ptMaxSize and ptMaxPosition
	//		members describe the maximized size and position of the window 
	//		on the primary monitor, even if the window ultimately maximizes 
	//		onto a secondary monitor.  In that case, the window manager adjusts 
	//		these values to compensate for differences between the primary monitor 
	//		and the monitor that displays the window. Thus, if the user leaves 
	//		ptMaxSize untouched, a window on a monitor larger than 
	//		the primary monitor maximizes to the size of the larger monitor.
	//
	//	It sounds hopeless to get a square window on a secondary monitor
	//	(unless its size happens to equal that of the primary monitor).
	//	So let's just restore the original window position and size,
	//	and hope for the best.
	
	GetInitialWindowRect(GetWindowLongPtr(aWindow, GWL_STYLE), &theOriginalRect);

	aMaxPosition->x = theOriginalRect.left;
	aMaxPosition->y = theOriginalRect.top;

	aMaxSize->x = theOriginalRect.right - theOriginalRect.left;
	aMaxSize->y = theOriginalRect.bottom - theOriginalRect.top;
}


static void ResizeDrawingPanel(
	WindowData		*wd,
	unsigned int	aWidth,
	unsigned int	aHeight)
{
	//	Let OpenGL resize the viewport.
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
	ResizeGLView(aWidth, aHeight);
	wglMakeCurrent(NULL, NULL);

	//	Request a refresh.
	InvalidateRect(wd->itsDrawingPanel, NULL, false);
}


static void DisplayChange(WindowData *wd)
{
	//	Maybe the user changed the color depth, screen size or multisampling,
	//	or maybe it's just that the screen saver had been running
	//	and is now passing control back to our program.
	//	Either way, we need to refresh our OpenGL context, shaders
	//	and textures.

	//	Free the old OpenGL context along with its shaders and textures.
	ShutDownOpenGL(wd);

	//	Create a new OpenGL context with new shaders and textures.
	SetUpOpenGL(wd);
}


static void GetAndResetMouseLocation(
	HWND		aDrawingWindow,
	LPARAM		lParam,			//	input location in drawing panel coordinates
	double		*theMouseH,		//	output horizontal coordinate (intrinsic units rightward from center, in range [-0.5, +0.5])
	double		*theMouseV)		//	output  vertical  coordinate (intrinsic units   upward  from center, in range [-0.5, +0.5])
{
	signed short int	hh,
						vv;
	RECT				theRect;
	POINT				theCenter;

	//	Split lParam into its horizontal and vertical parts.
	hh = (signed short int) LOWORD(lParam);
	vv = (signed short int) HIWORD(lParam);

	//	Fetch the client rectangle.
	//	GetClientRect() reports theClientRect in client coordinates.
	//	The upper-left corner is always (0,0).
	//	The lower-right corner tells the width and height.
	GetClientRect(aDrawingWindow, &theRect);

	//	Select a center.  Don't worry about off-by-one issues --
	//	any reasonable center will do just so we're consistent.
	theCenter.x = theRect.right /2;
	theCenter.y = theRect.bottom/2;

	//	Compute coordinates in range [-0.5, +0.5] with origin at the center.
	*theMouseH = (double)(hh - theCenter.x) / (double)theRect.right;
	*theMouseV = (double)(theCenter.y - vv) / (double)theRect.bottom;

	//	Move the (hidden) system cursor to the center of the client area if necessary.
	//	Be careful not to call SetCursorPos() needlessly, though, because
	//	it will trigger another WM_MOUSEMOVE event and thus an infinite loop.
	if (hh != theCenter.x
	 || vv != theCenter.y)
	{
		ClientToScreen(aDrawingWindow, &theCenter);
		SetCursorPos(theCenter.x, theCenter.y);
	}
}


static void ReleaseTorusCursor(WindowData *wd)
{
	if (GetCapture() != NULL)
	{
		SetConvenientMouseLocation(wd->itsDrawingPanel);
		ReleaseCapture();
		MouseGone(&wd->md);
	}
}


static void SetConvenientMouseLocation(HWND aDrawingWindow)
{
	POINT	theCursorRestorePoint;

	//	Move the cursor to an aesthetically sensible location.
	//	Most likely the user wants to use the menu, so put the cursor nearby.
	theCursorRestorePoint.x = 32;
	theCursorRestorePoint.y = 32;
	ClientToScreen(aDrawingWindow, &theCursorRestorePoint);
	SetCursorPos(theCursorRestorePoint.x, theCursorRestorePoint.y);
}


static void MakeContentSquare3(
	HWND	hWnd,
	WPARAM	aSizingMode,
	RECT	*aWindowRect)
{
	RECT		theDummyClientRect,
				theDummyWindowRect,
				theClientRect,
				theGameRect;
	signed int	theExcessWidth,
				theExcessHeight;
	
	//	The function AdjustWindowRectEx() converts a client rect
	//	to a window rect, but... there is no inverse function!
	//	So we must use a dummy rectangle to deduce the insets.
	theDummyClientRect = (RECT){0, 0, 256, 256};
	theDummyWindowRect = theDummyClientRect;
	AdjustWindowRectEx(	&theDummyWindowRect,
						GetWindowLongPtr(hWnd, GWL_STYLE),
						TRUE,
						GetWindowLongPtr(hWnd, GWL_EXSTYLE));
	
	//	Deduce the proposed client rect.
	theClientRect.left		= aWindowRect->left   + (theDummyClientRect.left   - theDummyWindowRect.left  );
	theClientRect.top		= aWindowRect->top    + (theDummyClientRect.top    - theDummyWindowRect.top   );
	theClientRect.right		= aWindowRect->right  + (theDummyClientRect.right  - theDummyWindowRect.right );
	theClientRect.bottom	= aWindowRect->bottom + (theDummyClientRect.bottom - theDummyWindowRect.bottom);
	
	//	Deduce the proposed game rect, excluding the toolbar.
	theGameRect = theClientRect;
	theGameRect.top += TORUS_GAMES_WIN32_TOOLBAR_HEIGHT;

	//	Constrain the game rect to a square.
	theExcessWidth	= (theGameRect.right - theGameRect.left)
					- (theGameRect.bottom - theGameRect.top);
	theExcessHeight	= - theExcessWidth;
	switch (aSizingMode)
	{
		//	As the user drags the left or right side,
		//	keep the opposite side fixed, and adjust
		//	the bottom to keep the game rectangle square.
		case WMSZ_LEFT:
		case WMSZ_RIGHT:
			theGameRect.bottom += theExcessWidth;
			break;

		//	As the user drags the top or bottom side,
		//	keep the opposite side fixed, and adjust
		//	the right side to keep the game rectangle square.
		case WMSZ_TOP:
		case WMSZ_BOTTOM:
			theGameRect.right  += theExcessHeight;
			break;

		//	As the user drags a corner, keep the opposite sides fixed
		//	and clamp to the smaller dimension.

		case WMSZ_BOTTOMRIGHT:
			if (theExcessWidth > 0)
				theGameRect.right  -= theExcessWidth;
			else
				theGameRect.bottom -= theExcessHeight;
			break;

		case WMSZ_TOPRIGHT:
			if (theExcessWidth > 0)
				theGameRect.right  -= theExcessWidth;
			else
				theGameRect.top    += theExcessHeight;
			break;

		case WMSZ_BOTTOMLEFT:
			if (theExcessWidth > 0)
				theGameRect.left   += theExcessWidth;
			else
				theGameRect.bottom -= theExcessHeight;
			break;

		case WMSZ_TOPLEFT:
			if (theExcessWidth > 0)
				theGameRect.left   += theExcessWidth;
			else
				theGameRect.top    += theExcessHeight;
			break;

		default:
			break;
	}

	//	Constrain the width and height to multiples of three,
	//	so the ViewSmall image may exactly fill the center 1/9
	//	of the ViewTiling image.
	switch (aSizingMode)
	{
		case WMSZ_LEFT:
		case WMSZ_TOPLEFT:
		case WMSZ_BOTTOMLEFT:
			theGameRect.left   += (theGameRect.right - theGameRect.left) % 3;
			break;

		case WMSZ_RIGHT:
		case WMSZ_TOPRIGHT:
		case WMSZ_BOTTOMRIGHT:
		case WMSZ_TOP:
		case WMSZ_BOTTOM:
		default:
			theGameRect.right  -= (theGameRect.right - theGameRect.left) % 3;
			break;
	}
	switch (aSizingMode)
	{
		case WMSZ_TOP:
		case WMSZ_TOPLEFT:
		case WMSZ_TOPRIGHT:
			theGameRect.top    += (theGameRect.bottom - theGameRect.top) % 3;
			break;

		case WMSZ_BOTTOM:
		case WMSZ_BOTTOMLEFT:
		case WMSZ_BOTTOMRIGHT:
		case WMSZ_LEFT:
		case WMSZ_RIGHT:
		default:
			theGameRect.bottom -= (theGameRect.bottom - theGameRect.top) % 3;
			break;
	}
	
	//	Convert theGameRect back to theClientRect.
	theClientRect = theGameRect;
	theClientRect.top -= TORUS_GAMES_WIN32_TOOLBAR_HEIGHT;

	//	Convert theClientRect back to a window rect.
	*aWindowRect = theClientRect;
	AdjustWindowRectEx(	aWindowRect,
						GetWindowLongPtr(hWnd, GWL_STYLE),
						TRUE,
						GetWindowLongPtr(hWnd, GWL_EXSTYLE));
}


static void PaintWindow(WindowData *wd)
{
	ErrorText	theSetUpError				= NULL,
				theRenderError				= NULL;
	HDC			theSavedDeviceContext		= NULL;
	HGLRC		theSavedRenderingContext	= NULL;

	//	In the normal flow of events there's no need to save and restore
	//	the rendering context.  However, we may get "unexpected" WM_PAINT
	//	messages -- for example if a MessageBox takes it upon itself
	//	to send WM_PAINT messages to underlying windows -- so play it safe
	//	and leave the rendering context as we found it.

	theSavedDeviceContext		= wglGetCurrentDC();		//	typically NULL
	theSavedRenderingContext	= wglGetCurrentContext();	//	typically NULL
	wglMakeCurrent(wd->itsDeviceContext, wd->itsRenderingContext);
		theSetUpError = SetUpGraphicsAsNeeded(&wd->md);
		if (theSetUpError == NULL)
			theRenderError = Render(&wd->md);
		if (theSetUpError == NULL && theRenderError == NULL)
			SwapBuffers(wd->itsDeviceContext);
	wglMakeCurrent(theSavedDeviceContext, theSavedRenderingContext);

	ValidateRect(wd->itsDrawingPanel, NULL);

	if (theSetUpError != NULL)
		ErrorMessage(theSetUpError, "OpenGL Setup Error", true);
	if (theRenderError != NULL)
		ErrorMessage(theRenderError, "OpenGL Rendering Error", true);
}
