// 	TorusGamesEAGLView.m
//
//	© 2010 by Jeff Weeks
//	See TermsOfUse.txt

#import "TorusGamesEAGLView.h"
#import "TorusGamesMainViewController.h"
#import "TorusGamesModelData.h"
#import "TorusGames-Common.h"
#import "TorusGames-OpenGL.h"
#import <QuartzCore/QuartzCore.h>	//	for CAEAGLLayer


//	Torus Games uses not depth buffer,
//	but I've included code to handling one
//	in case I ever need to copy this code
//	to a project that needs a depth buffer.
//	So keep the depth buffer code well maintained!
#define USE_DEPTH_BUFFER 0


@implementation TorusGamesEAGLView


+ (Class)layerClass
{
	//	Specify the class we want to use to create 
	//	the Core Animation layer for this view.  
	//	Instead of the default CALayer, request a CAEAGLLayer.
	return [CAEAGLLayer class];
}


- (id)initWithCoder:(NSCoder*)coder
{
	CAEAGLLayer	*theEAGLLayer;

	if ((self = [super initWithCoder:coder]) != nil)
	{
		itsContext				= nil;
		itsBackingWidth			= 0;
		itsBackingHeight		= 0;
		itsFramebuffer			= 0;
		itsColorRenderbuffer	= 0;
		itsDepthRenderbuffer	= 0;	//	Torus Games uses no depth buffer.

		theEAGLLayer = (CAEAGLLayer *)[self layer];
		[theEAGLLayer setOpaque:YES];
		[theEAGLLayer setDrawableProperties:[NSDictionary dictionaryWithObjectsAndKeys:
			[NSNumber numberWithBool:NO],	kEAGLDrawablePropertyRetainedBacking,
			kEAGLColorFormatRGBA8,			kEAGLDrawablePropertyColorFormat,
			nil]];

#ifdef SUPPORT_OPENGL_ES_2
		itsContext = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES2];
#endif
#ifdef SUPPORT_OPENGL_ES_1
		if (itsContext == nil)
			itsContext = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
#endif
//		if (itsContext == nil
//		 || ! [EAGLContext setCurrentContext:itsContext])
//		{
//			[self release];
//			return nil;
//		}

		itsLastTouchTime = 0.0;
	}
	return self;
}

- (EAGLContext *)context
{
	return itsContext;
}


- (void)dealloc
{
	//	This dealloc method never gets called under normal operation.
	
	EAGLContext	*theSavedContext;

	if (itsContext != nil)
	{
		if ([EAGLContext currentContext] != itsContext)
			theSavedContext = [EAGLContext currentContext];
		else
			theSavedContext = nil;

		[EAGLContext setCurrentContext:itsContext];

		[self destroyFramebuffer];
		
		[EAGLContext setCurrentContext:theSavedContext];

		[itsContext release];
		itsContext = nil;
	}

	[super dealloc];
}


- (void)layoutSubviews
{
	[EAGLContext setCurrentContext:itsContext];
	[self destroyFramebuffer];
	[self createFramebuffer];
}


- (BOOL)createFramebuffer
{
	//	The following code is identical for OpenGL ES 1 and OpenGL ES 2,
	//	except that OpenGL ES 1 uses the OES suffix.
	switch ([itsContext API])
	{
#ifdef SUPPORT_OPENGL_ES_1
		case kEAGLRenderingAPIOpenGLES1:

			glGenFramebuffersOES(1, &itsFramebuffer);
			glBindFramebufferOES(GL_FRAMEBUFFER_OES, itsFramebuffer);

			glGenRenderbuffersOES(1, &itsColorRenderbuffer);
			glBindRenderbufferOES(GL_RENDERBUFFER_OES, itsColorRenderbuffer);
			[itsContext renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:(CAEAGLLayer*)[self layer]];
			glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, itsColorRenderbuffer);

			glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &itsBackingWidth);
			glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &itsBackingHeight);

			if (USE_DEPTH_BUFFER)
			{
				glGenRenderbuffersOES(1, &itsDepthRenderbuffer);
				glBindRenderbufferOES(GL_RENDERBUFFER_OES, itsDepthRenderbuffer);
				glRenderbufferStorageOES(GL_RENDERBUFFER_OES, GL_DEPTH_COMPONENT16_OES, itsBackingWidth, itsBackingHeight);
				glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_DEPTH_ATTACHMENT_OES, GL_RENDERBUFFER_OES, itsDepthRenderbuffer);
			}
			else
			{
				itsDepthRenderbuffer = 0;
			}

			if (glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES)
			{
				NSLog(@"failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
				return NO;
			}

			break;
#endif
#ifdef SUPPORT_OPENGL_ES_2
		case kEAGLRenderingAPIOpenGLES2:

			glGenFramebuffers(1, &itsFramebuffer);
			glBindFramebuffer(GL_FRAMEBUFFER, itsFramebuffer);

			glGenRenderbuffers(1, &itsColorRenderbuffer);
			glBindRenderbuffer(GL_RENDERBUFFER, itsColorRenderbuffer);
			[itsContext renderbufferStorage:GL_RENDERBUFFER fromDrawable:(CAEAGLLayer*)[self layer]];
			glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_RENDERBUFFER, itsColorRenderbuffer);

			glGetRenderbufferParameteriv(GL_RENDERBUFFER, GL_RENDERBUFFER_WIDTH, &itsBackingWidth);
			glGetRenderbufferParameteriv(GL_RENDERBUFFER, GL_RENDERBUFFER_HEIGHT, &itsBackingHeight);

			if (USE_DEPTH_BUFFER)
			{
				glGenRenderbuffers(1, &itsDepthRenderbuffer);
				glBindRenderbuffer(GL_RENDERBUFFER, itsDepthRenderbuffer);
				glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT16, itsBackingWidth, itsBackingHeight);
				glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, itsDepthRenderbuffer);
			}
			else
			{
				itsDepthRenderbuffer = 0;
			}

			if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
			{
				NSLog(@"failed to make complete framebuffer object %x", glCheckFramebufferStatus(GL_FRAMEBUFFER));
				return NO;
			}

			break;
#endif
	}

	return YES;
}


- (void)destroyFramebuffer
{
	//	The following code is identical for OpenGL ES 1 and OpenGL ES 2,
	//	except that OpenGL ES 1 uses the OES suffix.
	switch ([itsContext API])
	{
#ifdef SUPPORT_OPENGL_ES_1
		case kEAGLRenderingAPIOpenGLES1:
			glDeleteFramebuffersOES(1, &itsFramebuffer);
			itsFramebuffer = 0;
			glDeleteRenderbuffersOES(1, &itsColorRenderbuffer);
			itsColorRenderbuffer = 0;
			if (itsDepthRenderbuffer)
			{
				glDeleteRenderbuffersOES(1, &itsDepthRenderbuffer);
				itsDepthRenderbuffer = 0;
			}
			break;
#endif
#ifdef SUPPORT_OPENGL_ES_2
		case kEAGLRenderingAPIOpenGLES2:
			glDeleteFramebuffers(1, &itsFramebuffer);
			itsFramebuffer = 0;
			glDeleteRenderbuffers(1, &itsColorRenderbuffer);
			itsColorRenderbuffer = 0;
			if (itsDepthRenderbuffer)
			{
				glDeleteRenderbuffers(1, &itsDepthRenderbuffer);
				itsDepthRenderbuffer = 0;
			}
			break;
#endif
	}
}


- (void)drawViewUsing:(TorusGamesModelData *)someModelData
{
	ErrorText	theSetUpError	= NULL,
				theRenderError	= NULL;

	[EAGLContext setCurrentContext:itsContext];
	
	//	The NIB file defines a square view so we expect itsBackingHeight == itsBackingWidth,
	//	but handle the more general case, just to be safe.
	if (itsBackingHeight >= itsBackingWidth)
		glViewport(0, (itsBackingHeight - itsBackingWidth)/2, itsBackingWidth, itsBackingWidth);
	else
		glViewport((itsBackingWidth - itsBackingHeight)/2, 0, itsBackingHeight, itsBackingHeight);

	theSetUpError	= SetUpGraphicsAsNeeded([someModelData modelData]);
	theRenderError	= Render([someModelData modelData]);

	switch ([itsContext API])
	{
#ifdef SUPPORT_OPENGL_ES_1
		case kEAGLRenderingAPIOpenGLES1:
			glBindRenderbufferOES(GL_RENDERBUFFER_OES, itsColorRenderbuffer);
			[itsContext presentRenderbuffer:GL_RENDERBUFFER_OES];
			break;
#endif
#ifdef SUPPORT_OPENGL_ES_2
		case kEAGLRenderingAPIOpenGLES2:
			glBindRenderbuffer(GL_RENDERBUFFER, itsColorRenderbuffer);
			[itsContext presentRenderbuffer:GL_RENDERBUFFER];
			break;
#endif
	}

	if (theSetUpError != NULL)
	{
		ErrorMessage(theSetUpError, "Non-fatal OpenGL Setup Error", false);
	}
	if (theRenderError != NULL)
	{
		ErrorMessage(theRenderError, "OpenGL Rendering Error", true);
	}
}


- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
	UITouch	*theTouch;
	CGPoint	theTouchPoint;

	//	By default UIView processes only single touches,
	//	so we're guaranteed to get information about
	//	the first finger to touch the display.
	//	UIView suppresses information about any other fingers
	//	that touch the display as part of the same gesture.
	theTouch = [touches anyObject];
	
	//	Where did theTouch begin?
	theTouchPoint = [self touchLocation:theTouch];
	
	//	When did theTouch begin?
	//	touchesMoved will use this value to compute an elapsed time.
	itsLastTouchTime = [theTouch timestamp];

	//	Pass the touch-began to the main view controller.
	[itsTorusGamesMainViewController touchBeganAt:theTouchPoint];
}

- (void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
	UITouch	*theTouch;
	CGPoint	theTouchMotion;
	double	theTouchTime,
			theElapasedTime;

	//	By default, a UIView gets events for only the first touch,
	//	so [touches anyObject] always returns the same touch.
	theTouch = [touches anyObject];

	//	How far has the user's finger moved since the previous
	//	call to touchesBegan or touchesMoved?
	theTouchMotion = [self touchMotion:theTouch];
	
	//	How fast is the user's finger moving?
	theTouchTime		= [theTouch timestamp];
	theElapasedTime		= theTouchTime - itsLastTouchTime;
	itsLastTouchTime	= theTouchTime;

	//	Pass the touch-motion to the main view controller.
	[itsTorusGamesMainViewController touchMovedAcross:theTouchMotion during:theElapasedTime];
}

- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{
	//	Pass the touch-end to the main view controller.
	[itsTorusGamesMainViewController touchEnded];
}

- (void)touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event
{
	//	Pass the touch-cancelled to the main view controller.
	[itsTorusGamesMainViewController touchCancelled];
}

- (CGPoint)touchLocation:(UITouch *)aTouch
{
	CGPoint	theTouchPoint;
	
	//	Convert theTouchPoint from view coordinates to intrinsic [-0.5, +0.5] coordinates.
	
	//	We expect itsBackingWidth and itsBackingHeight to be equal.

	//	touch location in view coordinates (0,0) to (itsBackingWidth, itsBackingHeight)
	theTouchPoint = [aTouch locationInView:self];

	//	touch location in coordinates 
	//	(-itsBackingWidth/2, -itsBackingHeight/2) to (+itsBackingWidth/2, +itsBackingHeight/2)
	theTouchPoint.x -= 0.5 * itsBackingWidth;
	theTouchPoint.y -= 0.5 * itsBackingHeight;

	//	touch location in (-0.5, -0.5) to (+0.5, +0.5) coordinates
	theTouchPoint.x /= itsBackingWidth;
	theTouchPoint.y /= itsBackingHeight;
	
	//	Flip from the iPhone OS's top-down coordinates to OpenGL's bottom-up coordinates.
	theTouchPoint.y = - theTouchPoint.y;
	
	return theTouchPoint;
}

- (CGPoint)touchMotion:(UITouch *)aTouch
{
	CGPoint	theTouchPointA,
			theTouchPointB,
			theDifference;
	
	//	Compute how far the touch has moved since the previous touchesMoved event.

	//	touch locations in view coordinates (0,0) to (itsBackingWidth, itsBackingHeight)
	theTouchPointA = [aTouch previousLocationInView:self];
	theTouchPointB = [aTouch locationInView:self];
	
	//	the relative motion, in the same coordinates
	theDifference.x = theTouchPointB.x - theTouchPointA.x;
	theDifference.y = theTouchPointB.y - theTouchPointA.y;

	//	the relative motion, in normalized coordinates
	theDifference.x /= itsBackingWidth;
	theDifference.y /= itsBackingHeight;

	//	Flip from the iPhone OS's top-down coordinates to OpenGL's bottom-up coordinates.
	theDifference.y = - theDifference.y;
	
	return theDifference;
}


@end
